/*
  -------------------------------------------------------------------

  Copyright (C) 2006-2012, Andrew W. Steiner

  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  -------------------------------------------------------------------
*/
/*
  Compare O2scl solvers with Boost
*/

#include <cmath>
#include <o2scl/test_mgr.h>
#include <o2scl/funct.h>
#include <o2scl/cern_mroot_root.h>
#include <o2scl/gsl_root_brent.h>
#include <o2scl/gsl_rnga.h>
#include <o2scl/hist_ev.h>

using namespace std;
using namespace o2scl;

double func(double x) {
  return atan((x-0.01)*4)*(1.0+sin((x-0.01)*50.0)/1.1);
}

int main(void) {

  const size_t N=100;

  gsl_rnga gr;

  // Store timing data in a hist_ev object
  hist_ev hev_grb, hev_cmr;

  // Choose a logarithmic grid with 10 blocks
  hist hist;
  hev_grb.set_grid_blocks(uniform_grid_log_end_width<>
			  (1.0e-16,1.0e-4,10.0),10);
  hev_cmr.set_grid_blocks(uniform_grid_log_end_width<>
			  (1.0e-16,1.0e-4,10.0),10);

  cout.setf(ios::scientific);

  test_mgr t;
  t.set_output_level(1);

  // Time gsl_mroot_hybrids
  gsl_root_brent<> grb;
  funct_fptr ff(func);

  cout << "GRB  ";
  for(size_t j=0;j<60;j++) {
    cout << "." << flush;
    
    grb.tol_abs=pow(10.0,-4-((double)j)/9.0);
    grb.tol_rel=pow(10.0,-4-((double)j)/9.0);
    
    for(size_t k=0;k<N;k++) {
      size_t tmp=clock();
      double x0, x1;

      for(size_t ell=0;ell<N;ell++) {

	// Random number from -1 to 2
	x0=-1.0-gr.random()*0.01;
	x1=2.0+gr.random()*0.01;

	grb.solve_bkt(x0,x1,ff);
      }
      
      double acc=fabs(func(x0));
      if (acc<fabs(1.0e-16)) acc=1.0001e-16;
      
      double t1=(clock()-tmp)/10000.0;
      hev_grb.add(acc,t1);
    }
  }
  cout << endl;
  
  cern_mroot_root<funct> cmr;

  cout << "CMR  ";
  for(size_t j=0;j<60;j++) {
    cout << "." << flush;
    
    cmr.tol_abs=pow(10.0,-4-((double)j)/9.0);
    cmr.tol_rel=pow(10.0,-4-((double)j)/9.0);
    
    for(size_t k=0;k<N;k++) {
      size_t tmp=clock();
      double x;
      
      for(size_t ell=0;ell<N;ell++) {
	// Random number from -1 to 2
	x=-1.0+3.0*gr.random();
	cmr.solve(x,ff);
      }
      
      double acc=fabs(func(x));
      if (acc<fabs(1.0e-16)) acc=1.0001e-16;
      
      double t1=(clock()-tmp)/10000.0;
      hev_cmr.add(acc,t1);
    }
  }
  cout << endl;
  cout << endl;

  cout << "GSL, gsl_root_brent" << endl;
  uvector rep, avg, sd, avge;
  hev_grb.current_avg(rep,avg,sd,avge);
  cout << "Accuracy     Avg. time    Error in avg. time" << endl;
  for(size_t i=0;i<rep.size();i++) {
    cout << rep[i] << " " << avg[i] << " " << avge[i] << endl;
  }
  cout << endl;
  
  cout << "CERNLIB, cern_mroot_root" << endl;
  hev_cmr.current_avg(rep,avg,sd,avge);
  cout << "Accuracy     Avg. time    Error in avg. time" << endl;
  for(size_t i=0;i<rep.size();i++) {
    cout << rep[i] << " " << avg[i] << " " << avge[i] << endl;
  }
  cout << endl;

  t.report();
  return 0;
}
