/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/

/* Example: ex_methast.cpp
   -------------------------------------------------------------------

   An example to demonstrate generation of an arbitrary distribution
   through the creation of a Markov chain using the
   Metropolis-Hastings algorithm.

*/
  
#include <iostream>
#include <cmath>
#include <gsl/gsl_randist.h>
#include <o2scl/test_mgr.h>
#include <o2scl/hist.h>
#include <o2scl/gsl_rnga.h>
#include <o2scl/expect_val.h>
#include <o2scl/gsl_inte_qag.h>

using namespace std;
using namespace o2scl;

/*
  An unusual two-peaked probability distribution. 
*/
double function(double x) {
  return exp(-x*x)*(sin(x-1.4)+1.0);
}

// Same function but now weighted with x
double function2(double x) {
  return x*function(x);
}

// The proposal distribution
double weight(double x) {
  return exp(-(x-1)*(x-1)/0.5)+exp(-(x+1)*(x+1)/0.5);
}

int main(int argc, char *argv[]) {
  test_mgr t;
  t.set_output_level(1);

  cout.setf(ios::scientific);

  // The square root of the number of iterations
  static const size_t N=1000;
  
  // The random number generator
  gsl_rnga gr;
  if (argc>=2) {
    gr.set_seed(stoi(argv[1]));
  } 

  for(size_t ell=0;ell<2;ell++) {

    bool hastings;
    if (ell==0) hastings=false;
    else hastings=true;

    // Construct a histogrmam of the distribution. Initialize the grid 
    // spacing with a unform grid between -5 and 5. 
    hist h;
    h.set_bins(uniform_grid_end_width<>(-5.0,5.0,0.1));
    h.clear_wgts();

    // Compute the average value of the distribution, and use
    // a scalar_ev object to estimate the uncertainty
    static const size_t M=20;
    scalar_ev sev(M,N*N/M);

    // An initial point and the initial weight
    double x_old=0.01;
    double w_old=function(x_old);
    double Q_old=weight(x_old);

    // GSL RNG
    const gsl_rng_type *T=gsl_rng_default;
    gsl_rng *grng=gsl_rng_alloc(T);
    gsl_rng_env_setup();

    for(size_t k=0;k<N;k++) {
      for(size_t i=0;i<N;i++) {

	// Perform a step and compute the new weight
	double x_new;

	if (!hastings) {
	  x_new=gr.random()*10.0-5.0;
	} else {
	  do {
	    if (gr.random()>0.5) {
	      x_new=gsl_ran_gaussian(grng,0.5)-1.0;
	    } else {
	      x_new=gsl_ran_gaussian(grng,0.5)+1.0;
	    }
	  } while (fabs(x_new)>5.0);
	}

	double w_new=function(x_new);
	double Q_new=weight(x_new);

	// The Metrpolis-Hastings algorithm
	double r=gr.random();
	if (hastings) {
	  if (r<w_new*Q_old/w_old/Q_new && x_new>-5.0 && x_new<5.0) {
	    x_old=x_new;
	    w_old=w_new;
	    Q_old=Q_new;
	  }
	} else {
	  if (r<w_new/w_old && x_new>-5.0 && x_new<5.0) {
	    x_old=x_new;
	    w_old=w_new;
	  }
	}

	// Update the histogram
	h.update(x_old);

	// Update the average value
	sev.add(x_old);

      }
    }
  
    // Normalize by the value of the distribution at x=1.05
    double norm=h.get_wgt(1.05)/function(1.05);

    // Output the generated distribution and compare to the real
    // distribution.

    for(size_t i=0;i<h.size();i++) {
    
      double xx=h.get_rep_i(i);

      // Limit output to every third point
      if (i%3==0) {
	cout.width(2);
	cout << i << " ";
	cout.setf(ios::showpos);
	cout << xx << " "; 
	cout.unsetf(ios::showpos);
	cout << h.get_wgt_i(i)/norm << " " 
	     << function(xx) << " " 
	     << fabs(h.get_wgt_i(i)/norm-function(xx))/function(xx) << endl;
      }
    
      // Only test regions where we'll have enough statistics
      if (function(xx)>5.0e-3) {
	t.test_rel(h.get_wgt_i(i)/norm,function(xx),3.0e-1,"dist");
      }
    
    }
    cout << endl;

    // Output the average value and its uncertainty
    double avg, std_dev, avg_err;
    sev.current_avg(avg,std_dev,avg_err);
    cout << avg << " " << avg_err << endl;

    // Compare to result from direct integration
    funct_fptr ff(function);
    funct_fptr ff2(function2);
    gsl_inte_qag<> qag;
    double exact=qag.integ(ff2,-5.0,5.0)/qag.integ(ff,-5.0,5.0);
    cout << exact << " " << fabs(avg-exact)/exact << endl;

    t.test_abs(avg,exact,3.0*avg_err,"Average value");
    cout << endl;

  }
  
  t.report();

  return 0;
}
// End of example
