/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_INTERP_H
#define O2SCL_INTERP_H

#include <iostream>
#include <string>

#include <o2scl/err_hnd.h>
#include <o2scl/ovector_tlate.h>
#include <o2scl/ovector_rev_tlate.h>
#include <o2scl/search_vec.h>
#include <o2scl/tridiag.h>

#ifndef DOXYGENP
namespace o2scl {
#endif
  
  /** \brief Base low-level interpolation class [abstract base]
      
      For any pair of vectors \c x and \c y into which you would like
      to interpolate, you need to call allocate() and init() first,
      and then the interpolation functions, and then free(). If the
      next pair of vectors has the same size, then you need only to
      call init() before the next call to an interpolation function
      (you don't need to reallocate). If the \c x and \c y vectors do
      not change, then you may call the interpolation functions in
      succession.

      These interpolation class work so long as the vector \c x is
      either monotonically increasing or monotonically decreasing. All
      of the descendants give identical results to the GSL
      interpolation routines when the vector is monotonically
      increasing. 
  */
  template<class vec_t> class base_interp {

#ifndef DOXYGEN_INTERNAL

  protected:

    /// To perform binary searches
    search_vec <vec_t> *sv;
    
    /// The size of the allocated \ref search_vec object
    size_t n_sv;

    /// True if memory has been allocated
    bool mem_alloc;

    /** \brief An internal function to assist in computing the 
	integral for both the cspline and Akima types
    */
    double integ_eval(double ai, double bi, double ci, double di, double xi, 
		      double a, double b) {

      const double r1=a-xi;
      const double r2=b-xi;
      const double r12=r1+r2;
      const double bterm=0.5*bi*r12;
      const double cterm=(1.0/3.0)*ci*(r1*r1+r2*r2+r1*r2);
      const double dterm=0.25*di*r12*(r1*r1+r2*r2);
      
      return (b-a)*(ai+bterm+cterm+dterm);
    }

#endif

  public:

    // This empty constructor is required since the 
    // default copy constructor is private.
    base_interp() {
      sv=0;
      n_sv=0;
      mem_alloc=false;
    }

    virtual ~base_interp() {
      if (sv!=0) delete sv;
      if (mem_alloc) free();
    }

    /** \brief The minimum size of the vectors to interpolate between

	This needs to be set in the constructor of the children for
	access by the class user
    */
    size_t min_size;

    /// Allocate memory, assuming x and y have size \c size
    virtual int allocate(size_t size) { return 0; }

    /// Free allocated memory
    virtual int free() { return 0; }

    /// Initialize interpolation routine
    virtual int init(const vec_t &x, const vec_t &y, size_t size) {
      return 0;
    }
    
    /// Give the value of the function \f$ y(x=x_0) \f$ .
    virtual int interp(const vec_t &x, const vec_t &y, size_t size, 
		       double x0, double &y0)=0;
    
    /// Give the value of the derivative \f$ y^{\prime}(x=x_0) \f$ .
    virtual int deriv(const vec_t &x, const vec_t &y, size_t size,
		      double x0, double &dydx)=0;
    
    /** \brief Give the value of the second derivative 
	\f$ y^{\prime \prime}(x=x_0) \f$ .
    */
    virtual int deriv2(const vec_t &x, const vec_t &y, size_t size, 
		       double x0, double &d2ydx2)=0;
    
    /// Give the value of the integral \f$ \int_a^{b}y(x)~dx \f$ .
    virtual int integ(const vec_t &x, const vec_t &y, size_t size, 
		      double a, double b, double &result)=0;

    /// Return the type, \c "base_interp".
    virtual const char *type() { return "base_interp"; }

#ifndef DOXYGEN_INTERNAL

  private:

    base_interp<vec_t>(const base_interp<vec_t> &);
    base_interp<vec_t>& operator=(const base_interp<vec_t>&);

#endif

  };
  
  /** \brief Linear interpolation (GSL) 

      Linear interpolation requires no calls to allocate() or free()
      as there is no internal storage required.
  */
  template<class vec_t> class linear_interp : public base_interp<vec_t> {
    
  public:
    
    linear_interp() {
      this->min_size=2;
    }

    virtual ~linear_interp() {}
    
    /// Initialize interpolation routine
    virtual int init(const vec_t &x, const vec_t &y, size_t size) {
      if (this->sv!=0) delete this->sv;
      this->sv=new search_vec<vec_t>(size,x);
      return 0;
    }
    
    /// Give the value of the function \f$ y(x=x_0) \f$ .
    virtual int interp(const vec_t &x_array, const vec_t &y_array, 
		       size_t size, double x, double &y) {

      double x_lo, x_hi;
      double y_lo, y_hi;
      double dx;
      size_t index;

      index=this->sv->find(x);
      
      /* evaluate */
      x_lo=x_array[index];
      x_hi=x_array[index+1];
      y_lo=y_array[index];
      y_hi=y_array[index+1];
      dx=x_hi-x_lo;

      y=y_lo+(x-x_lo)/dx*(y_hi-y_lo);
      return gsl_success;
    }
    
    /// Give the value of the derivative \f$ y^{\prime}(x=x_0) \f$ .
    virtual int deriv(const vec_t &x_array, const vec_t &y_array, 
		      size_t size, double x, double &dydx) {
      
      double x_lo, x_hi;
      double y_lo, y_hi;
      double dx;
      double dy;
      size_t index;
      
      index=this->sv->find(x);
      
      /* evaluate */
      x_lo=x_array[index];
      x_hi=x_array[index+1];
      y_lo=y_array[index];
      y_hi=y_array[index+1];
      dx=x_hi-x_lo;
      dy=y_hi-y_lo;

      dydx=dy/dx;
      return gsl_success;
    }

    /** \brief Give the value of the second derivative  
	\f$ y^{\prime \prime}(x=x_0) \f$ (always zero)
    */
    virtual int deriv2(const vec_t &x, const vec_t &y, size_t size, 
		       double x0, double &d2ydx2) {
      d2ydx2=0.0;
      return gsl_success;
    }

    /// Give the value of the integral \f$ \int_a^{b}y(x)~dx \f$ .
    virtual int integ(const vec_t &x_array, const vec_t &y_array, 
		      size_t size, double a, double b, 
		      double &result) {

      size_t i, index_a, index_b;
      
      double tmp;
      bool flip=false;
      if ((x_array[0]<x_array[size-1] && a>b) ||
	  (x_array[0]>x_array[size-1] && a<b)) {
	tmp=a;
	a=b;
	b=tmp;
	flip=true;
      }

      index_a=this->sv->find(a);
      index_b=this->sv->find(b);
      
      /* endpoints span more than one interval */
      
      result=0.0;
      
      /* interior intervals */
      for(i=index_a; i<=index_b; i++) {
	const double x_hi=x_array[i+1];
	const double x_lo=x_array[i];
	const double y_lo=y_array[i];
	const double y_hi=y_array[i+1];
	const double dx=x_hi-x_lo;
	
	if(dx != 0.0) {
	  if (i == index_a || i == index_b) {
	    double x1=(i == index_a) ? a : x_lo;
	    double x2=(i == index_b) ? b : x_hi;
	    const double D=(y_hi-y_lo)/dx;
	    result += (x2-x1)*(y_lo+0.5*D*((x2-x_lo)+(x1-x_lo)));
	  } else  {
	    result += 0.5*dx*(y_lo+y_hi);
	  }
	}
      }
      
      if (flip) result*=-1.0;
      return gsl_success;
    }

    /// Return the type, \c "linear_interp".
    virtual const char *type() { return "linear_interp"; }

#ifndef DOXYGEN_INTERNAL

  private:

    linear_interp<vec_t>(const linear_interp<vec_t> &);
    linear_interp<vec_t>& operator=(const linear_interp<vec_t>&);

#endif

  };
  
  /** \brief Cubic spline interpolation (GSL)
      
      By default, this class uses natural boundary conditions, where
      the second derivative vanishes at each end point. Extrapolation
      effectively assumes that the second derivative is linear outside
      of the endpoints. 
  */
  template<class vec_t> class cspline_interp : public base_interp<vec_t> {
    
#ifndef DOXYGEN_INTERNAL
    
  protected:

    /// \name Storage for cubic spline interpolation
    //@{
    double *c;
    double *g;
    double *diag;
    double *offdiag;
    //@}

    /// Memory for the tridiagonalization
    o2scl_linalg::uvector_4_mem p4m;

    /// Compute coefficients for cubic spline interpolation
    void coeff_calc(const double c_array[], double dy, double dx, 
		    size_t index, double *b, double *c2, double *d) {

      const double c_i=c_array[index];
      const double c_ip1=c_array[index+1];
      *b=(dy/dx)-dx*(c_ip1+2.0*c_i)/3.0;
      *c2=c_i;
      *d=(c_ip1-c_i)/(3.0*dx);
    }

#endif
    
  public:

    /** \brief Create a base interpolation object with natural or
	periodic boundary conditions
    */
    cspline_interp() {
      this->min_size=3;
    }

    virtual ~cspline_interp() {}

    /// Allocate memory, assuming x and y have size \c size
    virtual int allocate(size_t size) {
      
      c=(double *)malloc(size*sizeof(double));
      if (c == 0) {
	O2SCL_ERR2_RET("Failed to allocate space for c",
		       " in cspline_interp::allocate().",gsl_enomem);
      }
      
      g=(double *)malloc(size*sizeof(double));
      if (g == 0) {
	std::free(c);
	O2SCL_ERR2_RET("Failed to allocate space for g",
		       " in cspline_interp::allocate().",gsl_enomem);
      }
      
      diag=(double *)malloc(size*sizeof(double));
      if (diag == 0) {
	std::free(g);
	std::free(c);
	O2SCL_ERR2_RET("Failed to allocate space for diag",
		       " in cspline_interp::allocate().",gsl_enomem);
      }
      
      offdiag=(double *)malloc(size*sizeof(double));
      if (offdiag == 0) {
	std::free(diag);
	std::free(g);
	std::free(c);
	O2SCL_ERR2_RET("Failed to allocate space for offdiag",
		       " in cspline_interp::allocate().",gsl_enomem);
      }

      p4m.allocate(size);

      this->mem_alloc=true;
      
      return 0;
    }
    
    /** \brief Initialize interpolation routine
     */
    virtual int init(const vec_t &xa, const vec_t &ya, size_t size) {

      if (this->mem_alloc==false) {
	O2SCL_ERR2_RET("Memory allocation not performed in ",
		       "cspline_interp::init().",gsl_efailed);
      }

      if (this->sv!=0) delete this->sv;
      this->sv=new search_vec<vec_t>(size,xa);

      /// Natural boundary conditions

      size_t i;
      size_t num_points=size;
      size_t max_index=num_points-1;
      size_t sys_size=max_index-1;

      c[0]=0.0;
      c[max_index]=0.0;

      for (i=0; i < sys_size; i++) {
	const double h_i=xa[i+1]-xa[i];
	const double h_ip1=xa[i+2]-xa[i+1];
	const double ydiff_i=ya[i+1]-ya[i];
	const double ydiff_ip1=ya[i+2]-ya[i+1];
	const double g_i=(h_i != 0.0) ? 1.0/h_i : 0.0;
	const double g_ip1=(h_ip1 != 0.0) ? 1.0/h_ip1 : 0.0;
	offdiag[i]=h_ip1;
	diag[i]=2.0*(h_ip1+h_i);
	g[i]=3.0*(ydiff_ip1*g_ip1- ydiff_i*g_i);
      }

      if (sys_size == 1) {

	c[1]= g[0]/ diag[0];
	return gsl_success;

      } else {

	double *cp1=c+1;
	o2scl_linalg::solve_tridiag_sym<double *,double *,double *,
	  double *,o2scl_linalg::uvector_4_mem,uvector>
	  (diag,offdiag,g,cp1,sys_size,p4m);
	return 0;

      }
    }
          
    /// Free allocated memory
    virtual int free() {
      p4m.free();
      std::free(c);
      std::free(g);
      std::free(diag);
      std::free(offdiag);
      this->mem_alloc=false;
      return 0;
    }

    /// Give the value of the function \f$ y(x=x_0) \f$ .
    virtual int interp(const vec_t &x_array, const vec_t &y_array, 
		       size_t size, double x, double &y) {
      
      double x_lo, x_hi;
      double dx;
      size_t index;
      
      index=this->sv->find(x);
  
      /* evaluate */
      x_hi=x_array[index+1];
      x_lo=x_array[index];
      dx=x_hi-x_lo;

      const double y_lo=y_array[index];
      const double y_hi=y_array[index+1];
      const double dy=y_hi-y_lo;
      double delx=x-x_lo;
      double b_i, c_i, d_i; 
      coeff_calc(c, dy, dx, index, &b_i, &c_i, &d_i);
      y=y_lo+delx*(b_i+delx*(c_i+delx*d_i));
      return gsl_success;
    }

    /// Give the value of the derivative \f$ y^{\prime}(x=x_0) \f$ .
    virtual int deriv(const vec_t &x_array, const vec_t &y_array, 
		      size_t size, double x, double &dydx) {

      double x_lo, x_hi;
      double dx;
      size_t index;
  
      index=this->sv->find(x);
  
      /* evaluate */
      x_hi=x_array[index+1];
      x_lo=x_array[index];
      dx=x_hi-x_lo;

      const double y_lo=y_array[index];
      const double y_hi=y_array[index+1];
      const double dy=y_hi-y_lo;
      double delx=x-x_lo;
      double b_i, c_i, d_i; 
      coeff_calc(c, dy, dx, index, &b_i, &c_i, &d_i);
      dydx=b_i+delx*(2.0*c_i+3.0*d_i*delx);
      return gsl_success;
    }

    /** \brief Give the value of the second derivative  
	\f$ y^{\prime \prime}(x=x_0) \f$ .
    */
    virtual int deriv2(const vec_t &x_array, const vec_t &y_array, 
		       size_t size, double x, double &d2ydx2) {

      double x_lo, x_hi;
      double dx;
      size_t index;
  
      index=this->sv->find(x);
  
      /* evaluate */
      x_hi=x_array[index+1];
      x_lo=x_array[index];
      dx=x_hi-x_lo;

      const double y_lo=y_array[index];
      const double y_hi=y_array[index+1];
      const double dy=y_hi-y_lo;
      double delx=x-x_lo;
      double b_i, c_i, d_i;
      coeff_calc(c, dy, dx, index, &b_i, &c_i, &d_i);
      d2ydx2=2.0*c_i+6.0*d_i*delx;
      return gsl_success;
    }

  public:
    
    /// Give the value of the integral \f$ \int_a^{b}y(x)~dx \f$ .
    virtual int integ(const vec_t &x_array, const vec_t &y_array, 
		      size_t size, double a, double b, double &result) {

      size_t i, index_a, index_b;
  
      double tmp;
      bool flip=false;
      if ((x_array[0]<x_array[size-1] && a>b) ||
	  (x_array[0]>x_array[size-1] && a<b)) {
	tmp=a;
	a=b;
	b=tmp;
	flip=true;
      }

      index_a=this->sv->find(a);
      index_b=this->sv->find(b);

      result=0.0;
  
      /* interior intervals */
      for(i=index_a; i<=index_b; i++) {

	const double x_hi=x_array[i+1];
	const double x_lo=x_array[i];
	const double y_lo=y_array[i];
	const double y_hi=y_array[i+1];
	const double dx=x_hi-x_lo;
	const double dy=y_hi-y_lo;

	if(dx != 0.0) {
	  double b_i, c_i, d_i; 
	  coeff_calc(c, dy, dx, i, &b_i, &c_i, &d_i);
	  if (i == index_a || i == index_b) {
	    double x1=(i == index_a) ? a : x_lo;
	    double x2=(i == index_b) ? b : x_hi;
	    result += this->integ_eval(y_lo, b_i, c_i, d_i, x_lo, x1, x2);
	  } else {
	    result += dx*(y_lo+dx*(0.5*b_i+
				   dx*(c_i/3.0+0.25*d_i*dx)));
	  }
	} else {
	  result=0.0;
	  // We don't throw an error here so smart_interp classes can
	  // attempt to recover.
	  return gsl_failure;
	}

      }
  
      if (flip) result*=-1.0;
      return gsl_success;
    }

    /// Return the type, \c "cspline_interp".
    virtual const char *type() { return "cspline_interp"; }

#ifndef DOXYGEN_INTERNAL

  private:

    cspline_interp<vec_t>(const cspline_interp<vec_t> &);
    cspline_interp<vec_t>& operator=(const cspline_interp<vec_t>&);

#endif

  };
  
  /** \brief Cubic spline interpolation with periodic 
      boundary conditions (GSL)
  */
  template<class vec_t> class cspline_peri_interp : 
  public cspline_interp<vec_t> {

  protected:

    /// Desc
    o2scl_linalg::uvector_5_mem p5m;
    
  public:
    
  cspline_peri_interp() : cspline_interp<vec_t>() {
      this->min_size=3;
    }
      
    virtual ~cspline_peri_interp() {}

    /// Allocate memory, assuming x and y have size \c size
    virtual int allocate(size_t size) {
      
      this->c=(double *)malloc(size*sizeof(double));
      if (this->c == 0) {
	O2SCL_ERR2_RET("Failed to allocate space for c",
		       " in cspline_interp::allocate().",gsl_enomem);
      }
      
      this->g=(double *)malloc(size*sizeof(double));
      if (this->g == 0) {
	std::free(this->c);
	O2SCL_ERR2_RET("Failed to allocate space for g",
		       " in cspline_interp::allocate().",gsl_enomem);
      }
      
      this->diag=(double *)malloc(size*sizeof(double));
      if (this->diag == 0) {
	std::free(this->g);
	std::free(this->c);
	O2SCL_ERR2_RET("Failed to allocate space for diag",
		       " in cspline_interp::allocate().",gsl_enomem);
      }
      
      this->offdiag=(double *)malloc(size*sizeof(double));
      if (this->offdiag == 0) {
	std::free(this->diag);
	std::free(this->g);
	std::free(this->c);
	O2SCL_ERR2_RET("Failed to allocate space for offdiag",
		       " in cspline_interp::allocate().",gsl_enomem);
      }

      p5m.allocate(size);

      this->mem_alloc=true;
      
      return 0;
    }

    /// Return the type, \c "cspline_peri_interp".
    virtual const char *type() { return "cspline_peri_interp"; }

    virtual int init(const vec_t &xa, const vec_t &ya, size_t size) {

      if (this->mem_alloc==false) {
	O2SCL_ERR2_RET("Memory allocation not performed in ",
		       "cspline_peri_interp::init().",gsl_efailed);
      }

      if (this->sv!=0) delete this->sv;
      this->sv=new search_vec<vec_t>(size,xa);

      double *lc=this->c;
      double *lg=this->g;
      double *ldiag=this->diag;
      double *loffdiag=this->offdiag;

      /// Periodic boundary conditions
	 
      size_t i;
      size_t num_points=size;
      /* Engeln-Mullges+Uhlig "n" */
      size_t max_index=num_points-1;  
      /* linear system is sys_size x sys_size */
      size_t sys_size=max_index;    
	
      if (sys_size == 2) {
	/* solve 2x2 system */
	  
	const double h0=xa[1]-xa[0];
	const double h1=xa[2]-xa[1];
	const double h2=xa[3]-xa[2];
	const double A=2.0*(h0+h1);
	const double B=h0+h1;
	double gx[2];
	double det;
	  
	gx[0]=3.0*((ya[2]-ya[1])/h1-(ya[1]-ya[0])/h0);
	gx[1]=3.0*((ya[1]-ya[2])/h2-(ya[2]-ya[1])/h1);
	  
	det=3.0*(h0+h1)*(h0+h1);
	lc[1]=( A*gx[0]-B*gx[1])/det;
	lc[2]=(-B*gx[0]+A*gx[1])/det;
	lc[0]=lc[2];
	  
	return gsl_success;

      } else {
	  
	for (i=0; i < sys_size-1; i++) {
	  const double h_i=xa[i+1]-xa[i];
	  const double h_ip1=xa[i+2]-xa[i+1];
	  const double ydiff_i=ya[i+1]-ya[i];
	  const double ydiff_ip1=ya[i+2]-ya[i+1];
	  const double g_i=(h_i != 0.0) ? 1.0/h_i : 0.0;
	  const double g_ip1=(h_ip1 != 0.0) ? 1.0/h_ip1 : 0.0;
	  loffdiag[i]=h_ip1;
	  ldiag[i]=2.0*(h_ip1+h_i);
	  lg[i]=3.0*(ydiff_ip1*g_ip1-ydiff_i*g_i);
	}
	  
	i=sys_size-1;
	{
	  const double h_i=xa[i+1]-xa[i];
	  const double h_ip1=xa[1]-xa[0];
	  const double ydiff_i=ya[i+1]-ya[i];
	  const double ydiff_ip1=ya[1]-ya[0];
	  const double g_i=(h_i != 0.0) ? 1.0/h_i : 0.0;
	  const double g_ip1=(h_ip1 != 0.0) ? 1.0/h_ip1 : 0.0;
	  loffdiag[i]=h_ip1;
	  ldiag[i]=2.0*(h_ip1+h_i);
	  lg[i]=3.0*(ydiff_ip1*g_ip1-ydiff_i*g_i);
	}

	double *cp1=lc+1;
	o2scl_linalg::solve_cyc_tridiag_sym<double *,double *,double *,
	  double *,o2scl_linalg::uvector_5_mem,uvector>
	  (ldiag,loffdiag,lg,cp1,sys_size,p5m);
	lc[0]=lc[max_index];
	return gsl_success;

      }

    }

    /// Free allocated memory
    virtual int free() {
      p5m.free();
      std::free(this->c);
      std::free(this->g);
      std::free(this->diag);
      std::free(this->offdiag);
      this->mem_alloc=false;
      return 0;
    }

#ifndef DOXYGEN_INTERNAL

  private:

    cspline_peri_interp<vec_t>(const cspline_peri_interp<vec_t> &);
    cspline_peri_interp<vec_t>& operator=(const cspline_peri_interp<vec_t>&);

#endif

  };

  /** \brief Akima spline interpolation (GSL)
      
      This class uses natural boundary conditions, where the second
      derivative vanishes at each end point. Extrapolation effectively
      assumes that the second derivative is linear outside of the
      endpoints. Use \ref akima_peri_interp for perodic boundary
      conditions.
  */
  template<class vec_t> class akima_interp : public base_interp<vec_t> {
    
#ifndef DOXYGEN_INTERNAL
    
  protected:

    /// \name Storage for Akima spline interpolation
    //@{
    double *b;
    double *c;
    double *d;
    double *um;
    //@}

    /// For initializing the interpolation
    void akima_calc(const vec_t &x_array, size_t size, double m[]) {
      
      for(size_t i=0; i < (size-1); i++) {

	const double NE=fabs(m[i+1]-m[i])+fabs(m[i-1]-m[i-2]);
	
	if (NE == 0.0) {
	  b[i]=m[i];
	  c[i]=0.0;
	  d[i]=0.0;
	} else {
	  const double h_i=x_array[i+1]-x_array[i];
	  const double NE_next=fabs(m[i+2]-m[i+1])+
	    fabs(m[i]-m[i-1]);
	  const double alpha_i=fabs(m[i-1]-m[i-2])/NE;
	  double alpha_ip1;
	  double tL_ip1;
	  if (NE_next == 0.0) {
	    tL_ip1=m[i];
	  } else {
	    alpha_ip1=fabs(m[i]-m[i-1])/NE_next;
	    tL_ip1=(1.0-alpha_ip1)*m[i]+alpha_ip1*m[i+1];
	  }
	  b[i]=(1.0-alpha_i)*m[i-1]+alpha_i*m[i];
	  c[i]=(3.0*m[i]-2.0*b[i]-tL_ip1)/h_i;
	  d[i]=(b[i]+tL_ip1-2.0*m[i])/(h_i*h_i);
	}
      }
    }
    
#endif
    
  public:

    /** \brief Create a base interpolation object with or without
	periodic boundary conditions
    */
    akima_interp() {
      this->min_size=5;
    }

    virtual ~akima_interp() {}
    
    /// Allocate memory, assuming x and y have size \c size
    virtual int allocate(size_t size) {
      
      b=(double *)malloc(size*sizeof (double));
      if (b==0) {
	O2SCL_ERR2_RET("Failed to allocate space for c",
		       " in akima_interp::allocate().",gsl_enomem);
      }
      
      c=(double *)malloc(size*sizeof (double));
      if (c==0) {
	std::free(b);
	O2SCL_ERR2_RET("Failed to allocate space for g",
		       " in akima_interp::allocate().",gsl_enomem);
      }
      
      d=(double *)malloc(size*sizeof (double));
      if (d==0) {
	std::free(b);
	std::free(c);
	O2SCL_ERR2_RET("Failed to allocate space for diag",
		       " in akima_interp::allocate().",gsl_enomem);
      }
      
      um=(double *)malloc((size+4)*sizeof (double));
      if (um==0) {
	std::free(b);
	std::free(c);
	std::free(d);
	O2SCL_ERR2_RET("Failed to allocate space for offdiag",
		       " in akima_interp::allocate().",gsl_enomem);
      }

      this->mem_alloc=true;

      return 0;
    }
    
    /// Initialize interpolation routine
    virtual int init(const vec_t &xa, const vec_t &ya, size_t size) {
      
      if (this->mem_alloc==false) {
	O2SCL_ERR2_RET("Memory allocation not performed in ",
		       "akima_interp::init().",gsl_efailed);
      }

      if (this->sv!=0) delete this->sv;
      this->sv=new search_vec<vec_t>(size,xa);

      /// Non-periodic boundary conditions

      double *m=um+2;
      size_t i;
      for (i=0; i <= size-2; i++) {
	m[i]=(ya[i+1]-ya[i])/(xa[i+1]-xa[i]);
      }
	
      /* non-periodic boundary conditions */
      m[-2]=3.0*m[0]-2.0*m[1];
      m[-1]=2.0*m[0]-m[1];
      m[size-1]=2.0*m[size-2]-m[size-3];
      m[size]=3.0*m[size-2]-2.0*m[size-3];
	
      akima_calc (xa, size, m);

      return 0;
    }
          
    /// Free allocated memory
    virtual int free() {
      std::free(b);
      std::free(c);
      std::free(d);
      std::free(um);
      this->mem_alloc=false;
      return 0;
    }

    /// Give the value of the function \f$ y(x=x_0) \f$ .
    virtual int interp(const vec_t &x_array, const vec_t &y_array, 
		       size_t size, double x, double &y) {
      
      size_t index;
      
      index=this->sv->find(x);
  
      const double x_lo=x_array[index];
      const double delx=x-x_lo;
      const double bb=b[index];
      const double cc=c[index];
      const double dd=d[index];
      y=y_array[index]+delx*(bb+delx*(cc+dd*delx));
      return gsl_success;
    }

    /// Give the value of the derivative \f$ y^{\prime}(x=x_0) \f$ .
    virtual int deriv(const vec_t &x_array, const vec_t &y_array, 
		      size_t size, double x, double &dydx) {

      size_t index;
  
      index=this->sv->find(x);

      double x_lo=x_array[index];
      double delx=x-x_lo;
      double bb=b[index];
      double cc=c[index];
      double dd=d[index];
      dydx=bb+delx*(2.0*cc+3.0*dd*delx);
      return gsl_success;

    }

    /** \brief Give the value of the second derivative  
	\f$ y^{\prime \prime}(x=x_0) \f$ .
    */
    virtual int deriv2(const vec_t &x_array, const vec_t &y_array, 
		       size_t size, double x, double &d2ydx2) {

      size_t index;
  
      index=this->sv->find(x);
  
      const double x_lo=x_array[index];
      const double delx=x-x_lo;
      const double cc=c[index];
      const double dd=d[index];
      d2ydx2=2.0*cc+6.0*dd*delx;
      return gsl_success;

    }

    /// Give the value of the integral \f$ \int_a^{b}y(x)~dx \f$ .
    virtual int integ(const vec_t &x_array, const vec_t &y_array, 
		      size_t size, double aa, double bb, 
		      double &result) {

      size_t i, index_a, index_b;
  
      double tmp;
      bool flip=false;
      if ((x_array[0]<x_array[size-1] && aa>bb) ||
	  (x_array[0]>x_array[size-1] && aa<bb)) {
	tmp=aa;
	aa=bb;
	bb=tmp;
	flip=true;
      }

      index_a=this->sv->find(aa);
      index_b=this->sv->find(bb);

      result=0.0;
  
      /* interior intervals */

      for(i=index_a; i<=index_b; i++) {

	const double x_hi=x_array[i+1];
	const double x_lo=x_array[i];
	const double y_lo=y_array[i];
	const double dx=x_hi-x_lo;

	if (dx != 0.0) {
	  
	  if (i==index_a || i==index_b) {
	    double x1=(i==index_a) ? aa : x_lo;
	    double x2=(i==index_b) ? bb : x_hi;
	    result += this->integ_eval(y_lo,b[i],c[i],d[i],x_lo,x1,x2);
	  } else {
	    result+=dx*(y_lo+dx*(0.5*b[i]+dx*(c[i]/3.0+0.25*d[i]*dx)));
	  }

	} else {
	  result=0.0;
	  // We don't throw an error here so smart_interp classes can
	  // attempt to recover.
	  return gsl_failure;
	}
      }
  
      if (flip) result*=-1.0;
      return gsl_success;
    }

    /// Return the type, \c "akima_interp".
    virtual const char *type() { return "akima_interp"; }

#ifndef DOXYGEN_INTERNAL

  private:

    akima_interp<vec_t>(const akima_interp<vec_t> &);
    akima_interp<vec_t>& operator=(const akima_interp<vec_t>&);

#endif

  };
  
  /** \brief Akima spline interpolation with periodic  
      boundary conditions (GSL)
  */
  template<class vec_t> class akima_peri_interp : 
  public akima_interp<vec_t> {
    
  public:
    
  akima_peri_interp() : akima_interp<vec_t>() {
    }
    
    virtual ~akima_peri_interp() {}
    
    /// Return the type, \c "akima_peri_interp".
    virtual const char *type() { return "akima_peri_interp"; }

    /// Initialize interpolation routine
    virtual int init(const vec_t &xa, const vec_t &ya, size_t size) {
      
      if (this->mem_alloc==false) {
	O2SCL_ERR2_RET("Memory allocation not performed in ",
		       "akima_peri_interp::init().",gsl_efailed);
      }

      if (this->sv!=0) delete this->sv;
      this->sv=new search_vec<vec_t>(size,xa);

      /// Periodic boundary conditions
      
      double *m=this->um+2;
      size_t i;
      // Form the required set of divided differences
      for (i=0;i<=size-2;i++) {
	m[i]=(ya[i+1]-ya[i])/(xa[i+1]-xa[i]);
      }
      
      /* periodic boundary conditions */
      m[-2]=m[size-1-2];
      m[-1]=m[size-1-1];
      m[size-1]=m[0];
      m[size]=m[1];
      
      akima_calc (xa,size, m);
      
      return 0;
    }
      
#ifndef DOXYGEN_INTERNAL

  private:

    akima_peri_interp<vec_t>(const akima_peri_interp<vec_t> &);
    akima_peri_interp<vec_t>& operator=(const akima_peri_interp<vec_t>&);

#endif

  };

  /** \brief A base interpolation object manager [abstract base]
   */
  template<class vec_t> class base_interp_mgr {

  public:

    virtual ~base_interp_mgr() {}

    /// Create a new interpolation object
    virtual base_interp<vec_t> *new_interp()=0;

    /// Deallocate an interpolation object
    virtual int free_interp(base_interp<vec_t> *b) {
      delete b;
      return 0;
    }
  };
  
  /** \brief A base interpolation object manager template 
   */
  template<class vec_t, template<class> class interp_t> class def_interp_mgr :
  public base_interp_mgr<vec_t> {
    
  public:

    virtual ~def_interp_mgr() {}

    /// Create a new interpolation object
    virtual base_interp<vec_t> *new_interp() {
      interp_t<vec_t> *b=new interp_t<vec_t>;
      return b;
    }

  };

  /** \brief Interpolation class 
      
      Interpolation of \ref ovector like objects is performed with the
      default template parameters, and \ref array_interp is provided
      for simple interpolation on C-style arrays.

      The type of interpolation to be performed can be specified using
      the set_type() function or in the constructor.  The default is
      cubic splines with natural boundary conditions.
  */
  template<class vec_t=ovector_const_view> class o2scl_interp {
    
#ifndef DOXYGEN_INTERNAL
    
  protected:
  
  /// Pointer to base interpolation object
  base_interp<vec_t> *itp;
  
  /// Pointer to base interpolation manager
  base_interp_mgr<vec_t> *bim1;
  
#endif
      
  public:
      
  /** \brief Default base interpolation object (cubic spline with
      natural boundary conditions)
  */
  def_interp_mgr<vec_t,cspline_interp> dim1;
  
  /// Create with base interpolation object \c it
  o2scl_interp(base_interp_mgr<vec_t> &it) {
    bim1=&it;
    itp=bim1->new_interp();
  }

  /** \brief Create an interpolator using the default
      cubic spline interpolation
  */
  o2scl_interp() {
    bim1=&dim1;
    itp=bim1->new_interp();
  }

  virtual ~o2scl_interp() {
    bim1->free_interp(itp);
  }

  /// Give the value of the function \f$ y(x=x_0) \f$ .
  virtual double interp(const double x0, size_t n, const vec_t &x, 
			const vec_t &y) {
    
    double ret=0.0;
    
    if (n<itp->min_size) {
      O2SCL_ERR("Vector size too small in interp().",gsl_edom);
    }
      
    int r1=itp->allocate(n);
    if (r1==0) {
      int r2=itp->init(x,y,n);
      if (r2==0) {
	itp->interp(x,y,n,x0,ret);
      }
      itp->free();
    }
	  
    return ret;
  }		      
    
  /// Give the value of the derivative \f$ y^{\prime}(x=x_0) \f$ .
  virtual double deriv(const double x0, size_t n, const vec_t &x, 
		       const vec_t &y) {
      
    double ret=0.0;
      
    if (n<itp->min_size) {
      O2SCL_ERR("Vector size too small in deriv().",gsl_edom);
    }
      
    int r1=itp->allocate(n);
    if (r1==0) {
      int r2=itp->init(x,y,n);
      if (r2==0) {
	itp->deriv(x,y,n,x0,ret);
      }
      itp->free();
    }
      
    return ret;
  }		      
    
  /** \brief Give the value of the second derivative  
      \f$ y^{\prime \prime}(x=x_0) \f$ .
  */
  virtual double deriv2(const double x0, size_t n, const vec_t &x, 
			const vec_t &y) {
      
    double ret=0.0;
      
    if (n<itp->min_size) {
      O2SCL_ERR("Vector size too small in deriv2().",gsl_edom);
    }
      
    // Increasing case

    int r1=itp->allocate(n);
    if (r1==0) {
      int r2=itp->init(x,y,n);
      if (r2==0) {
	itp->deriv2(x,y,n,x0,ret);
      }
      itp->free();
    }

    return ret;
  }		      
    
  /// Give the value of the integral \f$ \int_a^{b}y(x)~dx \f$ .
  virtual double integ(const double x1, const double x2, size_t n, 
		       const vec_t &x, const vec_t &y) {
      
    double ret=0.0;
      
    if (n<itp->min_size) {
      O2SCL_ERR("Vector size too small in integ().",gsl_edom);
    }
      
    // Increasing case
	    
    int r1=itp->allocate(n);
    if (r1==0) {
      int r2=itp->init(x,y,n);
      if (r2==0) {
	itp->integ(x,y,n,x1,x2,ret);
      }
      itp->free();
    }

    return ret;
  }		      
    
  /// Set base interpolation object
  int set_type(base_interp_mgr<vec_t> &it) {
    bim1->free_interp(itp);
    bim1=&it;
    itp=bim1->new_interp();
    return 0;
  }
      
#ifndef DOXYGEN_INTERNAL

  private:

  o2scl_interp<vec_t>(const o2scl_interp<vec_t> &);
  o2scl_interp<vec_t>& operator=(const o2scl_interp<vec_t>&);

#endif

  };
  
  /** \brief Interpolation class for pre-specified vector
      
      This interpolation class is intended to be sufficiently general
      to handle any vector type. Interpolation of \ref ovector like
      objects is performed with the default template parameters, and
      \ref array_interp_vec is provided for simple interpolation on
      C-style arrays.

      This class does not double check the vector to ensure that 
      all of the intervals for the abcissa are all increasing or
      all decreasing. 

      The type of interpolation to be performed can be specified 
      using the set_type() function. The default is cubic splines
      with natural boundary conditions. 
  */
  template<class vec_t=ovector_const_view> class o2scl_interp_vec {
      
#ifndef DOXYGEN_INTERNAL
      
  protected:
      
  /// Base interpolation object
  base_interp<vec_t> *itp;

  /// The interpolation manager
  base_interp_mgr<vec_t> *bim;
      
  /// True if the original array was increasing
  bool inc;

  /// Pointer to the user-specified x vector
  const vec_t *lx;

  /// Pointer to the user-specified x vector
  const vec_t *ly;

  /// Size of user-specified vectors
  size_t ln;

#endif
      
  public:
    
  /// Create with base interpolation object \c it
  o2scl_interp_vec(base_interp_mgr<vec_t> &it, 
		   size_t n, const vec_t &x, const vec_t &y) {

    bim=&it;
    itp=bim->new_interp();
    ln=0;
    lx=0;
    ly=0;
	
    if (n<itp->min_size) {
      O2SCL_ERR("Vector size too small in o2scl_interp_vec().",gsl_edom);
    }
      
    size_t iend=n-1;

    if (x[0]==x[iend]) {
      O2SCL_ERR("Vector endpoints equal in o2scl_interp_vec().",
		gsl_einval);
    }
      
    int r1=itp->allocate(n);
    if (r1==0) {
      int r2=itp->init(x,y,n);
      if (r2==0) {
	ln=n;
	lx=&x;
	ly=&y;
	inc=true;
      } else {
	itp->free();
      }
    }
      
  }
      
  virtual ~o2scl_interp_vec() {
    if (ln>0) {
      itp->free();
    }
    bim->free_interp(itp);
  }
    
  /// Give the value of the function \f$ y(x=x_0) \f$ .
  virtual double interp(const double x0) {
    double ret=0.0;
    if (ln>0) {
      itp->interp(*lx,*ly,ln,x0,ret);
      return ret;
    }
    O2SCL_ERR2_RET("No data to interpolate in ",
		   "o2scl_interp_vec::interp().",gsl_efailed);
    return ret;
  }		      
    
  /// Give the value of the derivative \f$ y^{\prime}(x=x_0) \f$ .
  virtual double deriv(const double x0) {
    double ret=0.0;
    if (ln>0) {
      itp->deriv(*lx,*ly,ln,x0,ret);
      return ret;
    }
    O2SCL_ERR2_RET("No data to interpolate in ",
		   "o2scl_interp_vec::deriv().",gsl_efailed);
    return ret;
  }		      
    
  /** \brief Give the value of the second derivative  
      \f$ y^{\prime \prime}(x=x_0) \f$ .
  */
  virtual double deriv2(const double x0) {
    double ret=0.0;
    if (ln>0) {
      itp->deriv2(*lx,*ly,ln,x0,ret);
      return ret;
    }
    O2SCL_ERR2_RET("No data to interpolate in ",
		   "o2scl_interp_vec::deriv2().",gsl_efailed);
    return ret;
  }		      
    
  /// Give the value of the integral \f$ \int_a^{b}y(x)~dx \f$ .
  virtual double integ(const double x1, const double x2) {
    double ret=0.0;
    if (ln>0) {
      itp->integ(*lx,*ly,ln,x1,x2,ret);
      return ret;
    }
    O2SCL_ERR2_RET("No data to interpolate in ",
		   "o2scl_interp_vec::integ().",gsl_efailed);
    return ret;
  }		      
    
  /// Set base interpolation object
  int set_type(base_interp_mgr<vec_t> &it) {
    bim=&it;
    return 0;
  }
  
#ifndef DOXYGEN_INTERNAL

  private:
  
  o2scl_interp_vec<vec_t>(const o2scl_interp_vec<vec_t> &);
  o2scl_interp_vec<vec_t>& operator=(const o2scl_interp_vec<vec_t>&);
    
  
#endif
  
  };
  
  /** \brief A specialization of o2scl_interp for C-style double arrays
   */
  template<size_t n> class array_interp : public o2scl_interp<double[n]> {
    
  public:
    
    /// Create with base interpolation objects \c it and \c rit
  array_interp(base_interp_mgr<double[n]> &it) 
    : o2scl_interp<double[n]>(it) {}
    
    /** \brief Create an interpolator using the default base
	interpolation objects
    */
  array_interp() : o2scl_interp<double[n]>() {}
    
  };
  
  /** \brief A specialization of o2scl_interp_vec for C-style arrays
   */
  template<class arr_t> class array_interp_vec : 
  public o2scl_interp_vec<arr_t> {
    
  public:
    
    /// Create with base interpolation object \c it
  array_interp_vec(base_interp_mgr<arr_t> &it, 
		   size_t nv, const arr_t &x, const arr_t &y) :
    o2scl_interp_vec<arr_t>(it,nv,x,y) {}
      
  };
  
#ifndef DOXYGENP
}
#endif

#endif
