/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/misc.h>

using namespace std;
using namespace o2scl;

std::string o2scl::binary_to_hex(std::string s) {
  std::string t="";
  char nums[16]={'0','1','2','3','4','5','6','7',
		 '8','9','A','B','C','D','E','F'};
  for(size_t i=0;i<s.length();i++) {
    bool found=false;
    if (s[i]=='1' || s[i]=='0') {
      if (i+1<s.length() && (s[i+1]=='1' || s[i+1]=='0')) {
	if (i+2<s.length() && (s[i+2]=='1' || s[i+2]=='0')) {
	  if (i+3<s.length() && (s[i+3]=='1' || s[i+3]=='0')) {
	    found=true;
	    int cnt=0;
	    if (s[i]=='1') cnt+=8;
	    if (s[i+1]=='1') cnt+=4;
	    if (s[i+2]=='1') cnt+=2;
	    if (s[i+3]=='1') cnt++;
	    t+=nums[cnt];
	    i+=3;
	  }
	}
      }
    }
    if (found==false) {
      t+=s[i];
    }
  }
  return t;
}

double o2scl::fermi_function(double E, double mu, double T, double limit) {
  double ret, x=(E-mu)/T;
  
  if (x>limit) {
    ret=0.0;
  } else if (x<-limit) {
    ret=1.0;
  } else {
    ret=1.0/(1.0+exp(x));
  }
  return ret;
}

int o2scl::count_words(string str) {
  string st;
  istringstream *is=new istringstream(str.c_str());
  int ctr=0;
  while((*is) >> st) ctr++;
  delete is;

  return ctr;
}

int o2scl::remove_whitespace(std::string &s) {
  // The index 'i' must be an integer so we can do 'i--' below
  for(int i=0;i<((int)s.length());i++) {
    if (s[i]==9 || s[i]==10 || s[i]==11 || s[i]==12 || s[i]==13 
	|| s[i]==32) {
      s=s.substr(0,i)+s.substr(i+1,s.length()-i-1);
      i--;
    }
  }
  return 0;
}

int o2scl::gsl_calloc_gvecs(size_t nv, size_t msize, const char *names[], 
			    gsl_vector ***ptrs, std::string func_name) {

  for(size_t i=0;i<nv;i++) {
    (*(ptrs[i]))=gsl_vector_calloc(msize);
    if ((*(ptrs[i]))==0) {
      for(size_t j=0;j<i;j++) {
	gsl_vector_free((*(ptrs[j])));
      }
      std::string s="Failed to allocate space for ";
      s+=names[i];
      s+=" in"+(func_name+'.');
      O2SCL_ERR_RET(s.c_str(),gsl_enomem);
    }
  }

  return gsl_success;
}

double o2scl::quadratic_extremum_x(double x1, double x2, double x3,
				 double y1, double y2, double y3) {
  
  if (x1==x2 || x2==x3 || x1==x3) {
    O2SCL_ERR2_RET("Two abscissae cannot be equal in function ",
		   "quadratic_extremum().",gsl_einval);
  }
    
  /*
    Start with:
    y1=a x1^2 + b x1 + c
    y2=a x2^2 + b x2 + c
    y3=a x3^2 + b x3 + c
      
    Eliminate 'c':
    (y1-y2)=a(x1^2-x2^2)+b(x1-x2)
    (y3-y2)=a(x3^2-x2^2)+b(x3-x2)

    Eliminate 'b':
    (x3-x2)*(y1-y2)=a*(x1^2-x2^2)*(x3-x2)+b*(x1-x2)*(x3-x2)
    (x1-x2)*(y3-y2)=a*(x3^2-x2^2)*(x1-x2)+b*(x3-x2)*(x1-x2)

    Alternatively, eliminate 'c' with:
    (y2-y1)=a(x2^2-x1^2)+b(x2-x1)
    (y3-y1)=a(x3^2-x1^2)+b(x3-x1)

    Eliminate 'b':
    (x3-x1)*(y2-y1)=a(x2^2-x1^2)*(x3-x1)+b(x2-x1)*(x3-x1)
    (x2-x1)*(y3-y1)=a(x3^2-x1^2)*(x2-x1)+b(x3-x1)*(x2-x1)
  */
    
  double a,b,c,den=(x1*x1-x2*x2)*(x3-x2)-(x3*x3-x2*x2)*(x1-x2);
  if (den==0.0) {
    den=(x2*x2-x1*x1)*(x3-x1)-(x3*x3-x1*x1)*(x2-x1);
    a=((x3-x1)*(y2-y1)-(x2-x2)*(y3-y1))/den;
  } else {
    a=((x3-x2)*(y1-y2)-(x1-x2)*(y3-y2))/den;
  }
  b=(y1-y2-a*(x1*x1-x2*x2))/(x1-x2);
  c=y2-a*x2*x2-b*x2;
    
  return -b/2.0/a;
}

double o2scl::quadratic_extremum_y(double x1, double x2, double x3,
				   double y1, double y2, double y3) {
  
  if (x1==x2 || x2==x3 || x1==x3) {
    O2SCL_ERR2_RET("Two abscissae cannot be equal in function ",
		   "quadratic_extremum().",gsl_einval);
  }
    
  double a,b,c,den=(x1*x1-x2*x2)*(x3-x2)-(x3*x3-x2*x2)*(x1-x2);
  if (den==0.0) {
    den=(x2*x2-x1*x1)*(x3-x1)-(x3*x3-x1*x1)*(x2-x1);
    a=((x3-x1)*(y2-y1)-(x2-x2)*(y3-y1))/den;
  } else {
    a=((x3-x2)*(y1-y2)-(x1-x2)*(y3-y2))/den;
  }
  b=(y1-y2-a*(x1*x1-x2*x2))/(x1-x2);
  c=y2-a*x2*x2-b*x2;
    
  return c-b*b/4.0/a;    
}
