/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_MULTI_FUNCT_H
#define O2SCL_MULTI_FUNCT_H

#include <string>
#include <o2scl/err_hnd.h>
#include <o2scl/ovector_tlate.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief Multi-dimensional function [abstract base]

      This class generalizes one function of several variables,
      i.e. \f$ y(x_0,x_1,...,x_{nv-1}) \f$ where \c nv is the number
      of variables in the function y. 


      This class is one of a large number of function object classes
      in \o2 designed to provide a mechanism for the user to 
      supply functions to solvers, minimizers, integrators, etc.
      See \ref funct_section for a general description.
      
  */
  template<class vec_t=ovector_base> class multi_funct {
    
    public:  

    multi_funct() {}

    virtual ~multi_funct() {}

    /** \brief Compute a function \c y of \c nv variables stored in \c x
	with parameter \c pa.
    */
    virtual double operator()(size_t nv, const vec_t &x)=0;

#ifndef DOXYGENP

    private:

    multi_funct(const multi_funct &);
    multi_funct& operator=(const multi_funct&);

#endif

  };

  /** \brief Function pointer to a multi-dimensional function
   */
  template<class vec_t=ovector_base>
    class multi_funct_fptr : public multi_funct<vec_t> {
    
    public:

    /** \brief Specify the function pointer
     */
    multi_funct_fptr(double (*fp)(size_t nv, const vec_t &x)) {
      fptr=fp;
    }

    virtual ~multi_funct_fptr() {};

    /** \brief Compute a function \c y of \c nv variables stored in \c x
	with parameter \c pa.
    */
    virtual double operator()(size_t nv, const vec_t &x) {
      return fptr(nv,x);
    }

#ifndef DOXYGEN_INTERNAL

    protected:
    
    /// Store the function pointer
    double (*fptr)(size_t nv, const vec_t &x);

    multi_funct_fptr() {}

#ifndef DOXYGENP
#endif

    private:

    multi_funct_fptr(const multi_funct_fptr &);
    multi_funct_fptr& operator=(const multi_funct_fptr&);

#endif

  };

  /** \brief Function pointer to a multi-dimensional function
   */
  template<class param_t, class vec_t=ovector_base>
    class multi_funct_fptr_param : public multi_funct<vec_t> {
    
    public:

    /** \brief Specify the function pointer
     */
    multi_funct_fptr_param(double (*fp)(size_t nv, const vec_t &x,
					param_t &), param_t &pa) {
      fptr=fp;
      pp=&pa;
    }

    virtual ~multi_funct_fptr_param() {};

    /** \brief Compute a function \c y of \c nv variables stored in \c x
	with parameter \c pa.
    */
    virtual double operator()(size_t nv, const vec_t &x) {
      return fptr(nv,x,*pp);
    }

#ifndef DOXYGEN_INTERNAL

    protected:
    
    /// Store the function pointer
    double (*fptr)(size_t nv, const vec_t &x, param_t &);

    /// Parameter
    param_t *pp;

    multi_funct_fptr_param() {}

#ifndef DOXYGENP
#endif

    private:

    multi_funct_fptr_param(const multi_funct_fptr_param &);
    multi_funct_fptr_param& operator=(const multi_funct_fptr_param&);

#endif

  };

  /** \brief Member function pointer to a multi-dimensional function
   */
  template<class tclass, class vec_t=ovector_base>
    class multi_funct_mfptr : public multi_funct<vec_t> {
    public:
  
    /** \brief Specify the member function pointer
     */
    multi_funct_mfptr(tclass *tp, double (tclass::*fp)
		      (size_t nv, const vec_t &x)) {
      tptr=tp;
      fptr=fp;
    }
  
    virtual ~multi_funct_mfptr() {}
  
    /** \brief Compute a function \c y of \c nv variables stored in \c x
	with parameter \c pa.
    */
    virtual double operator()(size_t nv, const vec_t &x) {
      return (*tptr.*fptr)(nv,x);
    }

#ifndef DOXYGEN_INTERNAL
  
    protected:
  
    /// Store the function pointer
    double (tclass::*fptr)(size_t nv, const vec_t &x);

    /// Store a pointer to the class instance
    tclass *tptr;
  
#ifndef DOXYGENP
#endif

    private:

    multi_funct_mfptr(const multi_funct_mfptr &);
    multi_funct_mfptr& operator=(const multi_funct_mfptr&);

#endif

  };
  
  /** \brief Member function pointer to a multi-dimensional function
   */
  template<class tclass, class param_t, class vec_t=ovector_base>
    class multi_funct_mfptr_param : public multi_funct<vec_t> {
    public:
  
    /** \brief Specify the member function pointer
     */
    multi_funct_mfptr_param(tclass *tp, double (tclass::*fp)
			    (size_t nv, const vec_t &x, param_t &),
			    param_t &pa) {
      tptr=tp;
      fptr=fp;
      pp=&pa;
    }
  
    virtual ~multi_funct_mfptr_param() {}
  
    /** \brief Compute a function \c y of \c nv variables stored in \c x
	with parameter \c pa.
    */
    virtual double operator()(size_t nv, const vec_t &x) {
      return (*tptr.*fptr)(nv,x,*pp);
    }

#ifndef DOXYGEN_INTERNAL
  
    protected:
  
    /// Store the function pointer
    double (tclass::*fptr)(size_t nv, const vec_t &x, param_t &);

    /// Parameter
    param_t *pp;

    /// Store a pointer to the class instance
    tclass *tptr;
  
#ifndef DOXYGENP
#endif

    private:

    multi_funct_mfptr_param(const multi_funct_mfptr_param &);
    multi_funct_mfptr_param& operator=(const multi_funct_mfptr_param&);

#endif

  };
  
  /** \brief Const member function pointer to a multi-dimensional function
   */
  template<class tclass, class vec_t=ovector_base>
    class multi_funct_cmfptr : public multi_funct<vec_t> {

    public:
  
    /** \brief Specify the member function pointer
     */
    multi_funct_cmfptr(tclass *tp, double (tclass::*fp)
		       (size_t nv, const vec_t &x) const) {
			
      tptr=tp;
      fptr=fp;
    }
  
    virtual ~multi_funct_cmfptr() {}
  
    /** \brief Compute a function \c y of \c nv variables stored in \c x
	with parameter \c pa.
    */
    virtual double operator()(size_t nv, const vec_t &x) {
      return (*tptr.*fptr)(nv,x);
    }

#ifndef DOXYGEN_INTERNAL
  
    protected:
  
    /// Store the function pointer
    double (tclass::*fptr)(size_t nv, const vec_t &x) const;

    /// Store a pointer to the class instance
    tclass *tptr;
  
#ifndef DOXYGENP
#endif

    private:

    multi_funct_cmfptr(const multi_funct_cmfptr &);
    multi_funct_cmfptr& operator=(const multi_funct_cmfptr&);

#endif

  };

#ifndef DOXYGENP
}
#endif

#endif
