/*
  -------------------------------------------------------------------
  
  Copyright (C) 2011-2012, Andrew W. Steiner
  
  This file is part of O2scl. It was adapted from the shared_ptr
  implementation of aegis, available at aegis.sourceforge.net.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
//
// aegis - project change supervisor
// Copyright (C) 2007, 2008 Peter Miller
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License 
// as published by the Free Software Foundation; either version 3 
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public 
// License along with this program. If not, see
// <http://www.gnu.org/licenses/>.

#ifndef O2SCL_SHARED_PTR_H
#define O2SCL_SHARED_PTR_H

#include <cassert>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief The internal shared pointer class used if no TR1 or Boost
      implementation is available

      See also \ref o2_shared_ptr . 

      This class was based on the Aegis <tt>shared_ptr</tt> class from
      http:://aegis.sourceforge.net, but is implemented in \o2 using a
      template struct rather than a define constant.
  */
  template<class T> class o2_int_shared_ptr {

  private:

    /** \brief Reference count

	Documentation from Aegis:
	\verbatim
	The reference_count instance variable is used to remember the
	location of the reference count. By having the reference count
	separate from the subject, we can cope with compatible pointers,
	not just exact pointers.
      
	This is not ideal because it allocates huge numbers of small
	objects. Some heap implementations go slowly when faced with
	many small allocations. Some heap implementations waste a lot of
	memory when faced with many small allocations.
	\endverbatim
    */
    long *reference_count;

    /** \brief Object being pointed to

	Documentation from Aegis:
	\verbatim
	The subject instance variable is used to remember the location
	of the object being reference counted.  By having it separate
	from the reference count, we can skip one indirection.
	\endverbatim
    */
    T *subject;

  public:

    /** \brief Create an empty shared_ptr
     */
    o2_int_shared_ptr() : reference_count(0),subject(0) {
      assert(valid());
    }
  
    /** \brief Create a shared pointer
      
	This constructor permits initialization from any compatable
	pointer where \c Y must be a complete type. 
    */
    template<class Y> explicit o2_int_shared_ptr(Y *rhs) :
    reference_count(0), subject(0) {
      assert(valid());
      if (rhs) {
	reference_count = new long(1);
	subject = rhs;
      }
      assert(valid());
    }
  
    ~o2_int_shared_ptr() {
      assert(valid());
      if (reference_count) {
	--*reference_count;
	if (*reference_count <= 0) {
	  delete subject;
	  subject = 0;
	  delete reference_count;
	  reference_count = 0;
	}
      }
    }
  
    /** \brief Copy constructor.
     */
  o2_int_shared_ptr(const o2_int_shared_ptr &rhs) :
    reference_count(0), subject(0) {
      assert(rhs.valid());
      if (rhs.subject) {
	assert(rhs.reference_count);
	reference_count = rhs.reference_count;
	subject = rhs.subject;
	++*reference_count;
      }
      assert(valid());
    }
  
    template <class Y> friend class o2_int_shared_ptr;
  
    /** \brief Copy constructor for compatible pointer types
     */
    template<class Y> o2_int_shared_ptr(const o2_int_shared_ptr<Y> &rhs) :
    reference_count(0), subject(0) {
      assert(rhs.valid());
      if (rhs.subject) {
	assert(rhs.reference_count);
	reference_count = rhs.reference_count;
	subject = rhs.subject;
	++*reference_count;
      }
      assert(valid());
    }
  
    /** \brief Test if the shared_ptr object is valid
      
	This method verifies that the internal state of 
	the shared pointer object is valid.
    */
    bool valid() const {
      return ((reference_count != 0) ? (*reference_count > 0 && 
					subject != 0) : (subject == 0));
    }
  
    /** \brief Swap

	Swap the contents of two o2_int_shared_ptr<T> objects.
	This method is more efficient than manually swapping 
	shared_ptr objects using a copy constructor or temporary.
    */
    void swap(o2_int_shared_ptr &rhs) {
      assert(valid());
      assert(rhs.valid());
      long *temp_reference_count = reference_count;
      T *temp_subject = subject;
      reference_count = rhs.reference_count;
      subject = rhs.subject;
      rhs.reference_count = temp_reference_count;
      rhs.subject = temp_subject;
    }

    /** \brief Drop the reference and set the pointer to zero
     */
    void reset() {
      // swap with a NULL pointer
      o2_int_shared_ptr tmp;
      this->swap(tmp);
    }

    /** \brief Assignment operator
     */
    o2_int_shared_ptr &operator=(const o2_int_shared_ptr &rhs) {
      assert(valid());
      assert(rhs.valid());
      if (this != &rhs) {
	// Documentation from Aegis:
	// 
	// In case you haven't seen the swap() technique to
	// implement copy assignment before, here's what it does:
	//
	// 1) Create a temporary o2_int_shared_ptr<> instance via the
	//    copy constructor, thereby increasing the reference
	//    count of the source object.
	//
	// 2) Swap the internal object pointers of *this and the
	//    temporary o2_int_shared_ptr<>.  After this step, *this
	//    already contains the new pointer, and the old pointer
	//    is now managed by temp.
	//
	// 3) The destructor of temp is executed, thereby
	//    unreferencing the old object pointer.
	//
	// This technique is described in Herb Sutter's "Exceptional
	// C++", and has a number of advantages over conventional
	// approaches:
	//
	// - Code reuse by calling the copy constructor.
	// - Strong exception safety for free.
	// - Self assignment is handled implicitly.
	// - Simplicity.
	// - It just works and is hard to get wrong; i.e. you can
	//   use it without even thinking about it to implement
	//   copy assignment where ever the object data is managed
	//   indirectly via a pointer, which is very common.
	//
	o2_int_shared_ptr temp(rhs);
	this->swap(temp);
	assert(valid());
	assert(rhs.valid());
      }
      return *this;
    }

    /** \brief Assignment operator for compatible pointer types
     */
    template<class Y> o2_int_shared_ptr &operator=
      (const o2_int_shared_ptr<Y> &rhs) {
      assert(valid());
      assert(rhs.valid());
      if (this != (o2_int_shared_ptr *)&rhs) {
	o2_int_shared_ptr tmp(rhs);
	this->swap(tmp);
	assert(valid());
	assert(rhs.valid());
      }
      return *this;
    }
  
    /** \brief Dereference operator 
     */
    T &operator*() const {
      assert(subject != 0);
      return *subject;
    }

    /** \brief Pointing at operator
     */
    T *operator->() const {
      assert(subject != 0);
      return subject;
    }
    
    /** \brief Return a pointer to the object
     */
    T *get() const {
      return subject;
    }

    /** \brief Return true if the pointer is not zero.
     */
    operator bool() const {
      return (subject != 0);
    }

    /** \brief Logical not operator
	
	Return true if and only if the pointer is zero.
    */
    bool operator!() const {
      return (subject == 0);
    }

    /** \brief Test equality
     */
    inline bool operator==(const o2_int_shared_ptr &rhs) const {
      return (subject == rhs.get());
    }

    /** \brief Test equality for compatible pointer types
     */
    template<class U> inline bool operator==
      (const o2_int_shared_ptr<U> &rhs) const {
      return (subject == rhs.get());
    }

    /** \brief Test inequality
     */
    inline bool operator!=(o2_int_shared_ptr &rhs) const {
      return (subject != rhs.get());
    }

    /** \brief Test inequality for compatible pointer types
     */
    template<class U> inline bool operator!=(o2_int_shared_ptr<U> &rhs) const {
      return (subject != rhs.get());
    }

    /** \brief Less than operator
     */
    inline bool operator<(o2_int_shared_ptr &rhs) const {
      return (subject < rhs.get());
    }

    /** \brief Less than operator for compatible pointer types
     */
    template<class U> inline bool operator<(o2_int_shared_ptr<U> &rhs) const {
      return (subject < rhs.get());
    }

  };

#ifndef DOXYGENP
}
#endif

#ifdef DOXYGENP

/** \brief A struct to provide the shared_ptr type

    This object exists in order to provide the shared_ptr template
    type used in \o2. The full specification of a shared pointer
    in \o2 for an object of type \c T is thus 
    \verbatim
    o2scl::o2_shared_ptr<T>::type
    \endverbatim
    In a default \o2 installation, \ref type (as given below)
    is a typedef of
    \verbatim
    std::tr1::shared_ptr<T>
    \endverbatim
    If <tt>O2SCL_NO_TR1_MEMORY</tt> is defined, then \ref type is
    a typedef of \ref o2_int_shared_ptr, unless <tt>O2SCL_HAVE_BOOST</tt>
    is defined, in which case \ref type is a typedef of
    \verbatim
    boost::shared_ptr<T>
    \endverbatim

    See also the discussion at http://www.gotw.ca/gotw/079.htm . This
    struct won't be necessary when C++ allows template typedef's as
    part of the C++11 standard http://en.wikipedia.org/wiki/C%2B%2B11
    , but very few compilers have implemented this standard yet.
*/
template<class T> struct o2_shared_ptr {
  /// The actual shared_ptr type
  typedef std::tr1::shared_ptr<T> type;
};

#else

// -------------------------------------------------------------------
// Define the o2_shared_ptr struct according to the installation settings

// AWS - 11/29/11: I can't remember if the #include statements have
// to be outside the o2scl namespace, but I make sure they're
// outside just in case it matters.

#ifndef O2SCL_NO_TR1_MEMORY

#include <tr1/memory>
namespace o2scl {
  template<class T> struct o2_shared_ptr {
    typedef std::tr1::shared_ptr<T> type;
  };
}

#else

#if O2SCL_HAVE_BOOST

#include <boost/shared_ptr.hpp>
namespace o2scl {
  template<class T> struct o2_shared_ptr {
    typedef boost::shared_ptr<T> type;
  };
}

#else

namespace o2scl {
  template<class T> struct o2_shared_ptr {
    typedef o2_int_shared_ptr<T> type;
  };
}

// end of if O2SCL_HAVE_BOOST
#endif

// end of if O2SCL_NO_TR1_MEMORY
#endif

// end of ifdef DOXYGENP
#endif

// -------------------------------------------------------------------

// end of ifdef O2SCL_SHARED_PTR_H
#endif 
