/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_SMART_INTERP_H
#define O2SCL_SMART_INTERP_H

/** \file smart_interp.h
    \brief File for "smart" interpolation routines

    In addition to the smart interpolation routines, this file
    contains the template functions \ref vector_find_level(), \ref
    vector_integ_linear() and \ref vector_invert_enclosed_sum().
*/

#include <o2scl/interp.h>
#include <o2scl/vector.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief Smart interpolation class
      
      This class can semi-intelligently handle arrays which are
      not-well formed outside the interpolation region. In particular,
      if an initial interpolation or derivative calculation fails, the
      arrays are searched for the largest neighborhood around the
      point \c x for which an interpolation or differentiation will
      likely produce a finite result.
  */
  template<class vec_t=ovector_const_view, 
    class svec_t=ovector_const_subvector> class smart_interp {
      
#ifndef DOXYGEN_INTERNAL
    
  protected:
  
  /// \name Storage internally created subvectors
  //@{
  bool sxalloc;
  const svec_t *sx;
  const svec_t *sy;
  //@}

  /// \name Pointers to interpolation objects
  //@{
  base_interp<vec_t> *rit1;
  base_interp<svec_t> *rit3;
  //@}

  /// \name Pointers to interpolation managers
  //@{
  base_interp_mgr<vec_t> *bim1;
  base_interp_mgr<svec_t> *bim3;
  //@}

#endif

  public:
    
  /** \brief Create a new interpolation object with the specified 
      interpolation managers
  */
  smart_interp(base_interp_mgr<vec_t> &im1,
	       base_interp_mgr<svec_t> &im3) {
    
    sxalloc=false;
    sx=0;
    sy=0;

    bim1=&im1;
    bim3=&im3;

    rit1=bim1->new_interp();
    rit3=bim3->new_interp();
  };

  /** \brief Create an interpolation object with the default 
      cubic spline interpolation managers
  */
  smart_interp() {
    sxalloc=false;
    sx=0;
    sy=0;

    bim1=&dim1;
    bim3=&dim3;

    rit1=bim1->new_interp();
    rit3=bim3->new_interp();
  };
    
  virtual ~smart_interp() {
    if (sxalloc) {
      delete sx;
      delete sy;
    } 

    /*
      We don't want to use "bim1->free_interp(rit1)" here instead
      of "delete rit1" because the pointer "bim1" may refer to 
      a user-defined interpolation manager, and it may have gone
      out of scope earlier.
    */
    delete rit1;
    delete rit3;
  }

  /** \brief Give the value of the function \f$ y(x=x_0) \f$ .

      \todo After calling find_subset, I think we might need
      to double check that \c nn is larger than the minimum
      interpolation size.
  */
  virtual double interp(const double x0, size_t n, const vec_t &x, 
			const vec_t &y) {
      
    bool to_subset=false;
    double ret=0.0;
    int err0, err1, err2;
      
    if (x[0]==x[n-1]) {
      to_subset=true;
    }

    if (n<rit1->min_size) {
      O2SCL_ERR("Vector size too small in interp().",gsl_edom);
      return 0.0;
    }
    err0=rit1->allocate(n);
    if (err0==0) {
      err1=rit1->init(x,y,n);
      if (err1==0) {
	err2=rit1->interp(x,y,n,x0,ret);
	if (err2!=0) to_subset=true;
      } else {
	to_subset=true;
      }
      rit1->free();
    }
      
    if (!gsl_finite(ret)) to_subset=true;
      
    if (to_subset) {
	
      size_t nn=0;
      int retfs=find_subset(x0,x0,n,x,y,nn);
      
      if (retfs!=0 || nn<=1) {
	O2SCL_ERR("Interpolation failed in interp().",gsl_efailed);
	return 0;
      }

      if (nn>=rit3->min_size) {
	err0=rit3->allocate(nn);
	if (err0==0) {
	  err1=rit3->init(*sx,*sy,nn);
	  if (err1==0) {
	    err2=rit3->interp(*sx,*sy,nn,x0,ret);
	  }		
	  rit3->free();
	}
      } else {
	O2SCL_ERR("Interpolation failed in interp().",gsl_efailed);
      }
      
    }
    
    return ret;
  }		      
  
  /// Give the value of the derivative \f$ y^{prime}(x=x_0) \f$ .
  virtual double deriv(const double x0, size_t n, const vec_t &x, 
		       const vec_t &y) {
    
    bool increasing=true, to_subset=false;
    double ret=0.0;
    int err0, err1, err2;
      
    if (x[0]==x[n-1]) {
      to_subset=true;
    }

    if (n<rit1->min_size) {
	O2SCL_ERR("Vector size too small in deriv().",gsl_edom);
	return 0.0;
      }
      err0=rit1->allocate(n);
      if (err0==0) {
	err1=rit1->init(x,y,n);
	if (err1==0) {
	  err2=rit1->deriv(x,y,n,x0,ret);
	  if (err2!=0) to_subset=true;
	} else {
	  to_subset=true;
	}
	rit1->free();
      }
	  
    if (!gsl_finite(ret)) to_subset=true;
      
    if (to_subset) {
	
      size_t nn=0;
      int retfs=find_subset(x0,x0,n,x,y,nn);
	
      if (retfs!=0 || nn<=1) {
	O2SCL_ERR("Interpolation failed in deriv().",gsl_efailed);
	return 0;
      }

      if (nn>=rit3->min_size) {
	err0=rit3->allocate(nn);
	if (err0==0) {
	  err1=rit3->init(*sx,*sy,nn);
	  if (err1==0) {
	    err2=rit3->deriv(*sx,*sy,nn,x0,ret);
	  }		
	  rit3->free();
	}
      } else {
	O2SCL_ERR("Interpolation failed in deriv().",gsl_efailed);
      }
      
    }
    
    return ret;
  }		      

  /** \brief Give the value of the second derivative  
      \f$ y^{prime \prime}(x=x_0) \f$ .
  */
  virtual double deriv2(const double x0, size_t n, const vec_t &x, 
			const vec_t &y) {
      
    bool increasing=true, to_subset=false;
    double ret=0.0;
    int err0, err1, err2;
      
    if (x[0]==x[n-1]) {
      to_subset=true;
    }

    if (n<rit1->min_size) {
	O2SCL_ERR("Vector size too small in deriv2().",gsl_edom);
	return 0.0;
      }
      err0=rit1->allocate(n);
      if (err0==0) {
	err1=rit1->init(x,y,n);
	if (err1==0) {
	  err2=rit1->deriv2(x,y,n,x0,ret);
	  if (err2!=0) to_subset=true;
	} else {
	  to_subset=true;
	}
	rit1->free();
      }
	  
    if (!gsl_finite(ret)) to_subset=true;
      
    if (to_subset) {
	
      size_t nn=0;
      int retfs=find_subset(x0,x0,n,x,y,nn);
	
      if (retfs!=0 || nn<=1) {
	O2SCL_ERR("Interpolation failed in deriv2().",gsl_efailed);
	return 0;
      }

      if (nn>=rit3->min_size) {
	err0=rit3->allocate(nn);
	if (err0==0) {
	  err1=rit3->init(*sx,*sy,nn);
	  if (err1==0) {
	    err2=rit3->deriv2(*sx,*sy,nn,x0,ret);
	  }		
	  rit3->free();
	}
      } else {
	O2SCL_ERR("Interpolation failed in deriv2().",gsl_efailed);
      }
      
    }
    
    return ret;
  }		      

  /// Give the value of the integral \f$ \int_a^{b}y(x)~dx \f$ .
  virtual double integ(const double a, const double b, 
		       size_t n, const vec_t &x, const vec_t &y) {
			   
    bool increasing=true, to_subset=false;
    double ret=0.0;
    int err0, err1, err2;
      
    if (x[0]==x[n-1]) {
      to_subset=true;
    }

    if (n<rit1->min_size) {
	O2SCL_ERR("Vector size too small in integ().",gsl_edom);
	return 0.0;
      }
      err0=rit1->allocate(n);
      if (err0==0) {
	err1=rit1->init(x,y,n);
	if (err1==0) {
	  err2=rit1->integ(x,y,n,a,b,ret);
	  if (err2!=0) to_subset=true;
	} else {
	  to_subset=true;
	}
	rit1->free();
      }
	  
    if (!gsl_finite(ret)) to_subset=true;
      
    if (to_subset) {
	
      size_t nn=0;
      int retfs=find_subset(a,b,n,x,y,nn);
	
      if (retfs!=0 || nn<=1) {
	O2SCL_ERR("Interpolation failed in integ().",gsl_efailed);
	return 0;
      }


      if (nn>=rit3->min_size) {
	err0=rit3->allocate(nn);
	if (err0==0) {
	  err1=rit3->init(*sx,*sy,nn);
	  if (err1==0) {
	    err2=rit3->integ(*sx,*sy,nn,a,b,ret);
	  }		
	  rit3->free();
	}
      } else {
	O2SCL_ERR("Interpolation failed in integ().",gsl_efailed);
      }
      
    }
    
    return ret;
  }		      

  /// \name Default interpolation managers
  //@{
  def_interp_mgr<vec_t,cspline_interp> dim1;
  def_interp_mgr<svec_t,cspline_interp> dim3;
  //@}

#ifndef DOXGYEN_INTERNAL

  protected:
  
  /** \brief A lookup function for generic vectors
   */
  size_t local_lookup(size_t n, const vec_t &x, double x0) {
    size_t row=0, i=0;
    while(!gsl_finite(x[i]) && i<n-1) i++;
    if (i==n-1) {
      return 0;
    }
    double best=x[i], bdiff=fabs(x[i]-x0);
    for(;i<n;i++) {
      if (gsl_finite(x[i]) && fabs(x[i]-x0)<bdiff) {
	row=i;
	best=x[i];
	bdiff=fabs(x[i]-x0);
      }
    }
    return row;
  }

  /** \brief Try to find the largest monotonic and finite region
      around the desired location

      This function tries to find the largest monotonic region
      enclosing both \c a and \c b in the vector \c x. If it succeeds,
      it returns \ref gsl_success, and if it fails, it returns \ref
      gsl_efailed. It does not call the error handler.

      \todo The error handling is a bit off here, as it can return
      a non-zero value even with there is no real "error". We should
      just make a new bool reference paramter. 
  */
  int find_subset(const double a, const double b, size_t sz, const vec_t &x, 
		  const vec_t &y, size_t &nsz) 
  {
      
    size_t left, right;

    bool increasing=true;
    if (x[0]>x[sz-1]) increasing=false;
      
    // Lookup the point
    size_t row=local_lookup(sz,x,a);
    if (!gsl_finite(x[row]) || !gsl_finite(y[row])) {
      return gsl_efailed;
    }
    size_t row2=local_lookup(sz,x,b);
    if (!gsl_finite(x[row2]) || !gsl_finite(y[row2])) {
      return gsl_efailed;
    }
  
    // Get a neighboring point if necessary
    if (row2==row) {
      row2=row++;
      if (row2>=sz) row2=row-1;
    }

    // Arrange
    if (row<row2) {
      left=row;
      right=row2;
    } else {
      left=row2;
      right=row;
    }
      
    // Establish whether we're increasing or decreasing
    increasing=true;
    if (row2<row && x[row]<x[row2]) increasing=false;
    if (row2>row && x[row2]<x[row]) increasing=false;

    // Check region between row and row2 

    if (increasing) {
      size_t lo, hi;
      if (row2<row) {
	lo=row2;
	hi=row;
      } else {
	lo=row;
	hi=row2;
      }
      for(size_t i=lo;i<hi-1;i++) {
	if (x[i]>x[i+1]) return gsl_efailed;
      }
    } else {
      size_t lo, hi;
      if (row2<row) {
	lo=row2;
	hi=row;
      } else {
	lo=row;
	hi=row2;
      }
      for(size_t i=lo;i<hi-1;i++) {
	if (x[i]<x[i+1]) return gsl_efailed;
      }
    }

    // Increasing case
    
    if (increasing) {
      
      while(left>0 && x[left-1]<x[left] && 
	    gsl_finite(x[left-1]) && gsl_finite(y[left-1])) left--;
      while(right<sz-1 && x[right]<x[right+1] &&
	    gsl_finite(x[right+1]) && gsl_finite(y[right+1])) right++;
	
      nsz=right-left+1;

      if (sxalloc) {
	delete sx;
	delete sy;
      }
      sx=new svec_t(x,left,nsz);
      sy=new svec_t(y,left,nsz);
      sxalloc=true;
	
    } else {

      while(left>0 && x[left-1]>x[left] && 
	    gsl_finite(x[left-1]) && gsl_finite(y[left-1])) left--;
      while(right<sz-1 && x[right]>x[right+1] &&
	    gsl_finite(x[right+1]) && gsl_finite(y[right+1])) right++;
	
      nsz=right-left+1;
	
      if (sxalloc==true) {
	delete sx;
	delete sy;
      }
      sx=new svec_t(x,left,nsz);
      sy=new svec_t(y,left,nsz);
      sxalloc=true;

    }
      
    return 0;
  }

#endif

  };

  /** \brief Smart interpolation class with pre-specified vectors
      
      This class can semi-intelligently handle arrays which are
      not-well formed outside the interpolation region. In particular,
      if an initial interpolation or derivative calculation fails, the
      arrays are searched for the largest neighborhood around the
      point \c x for which an interpolation or differentiation will
      likely produce a finite result.

      \future Properly implement handling of non-monotonic regions in
      the derivative functions as well as the interpolation function.
  */
  template<class vec_t, class svec_t, class alloc_vec_t, class alloc_t>
    class smart_interp_vec {

#ifndef DOXYGEN_INTERNAL
    
  protected:

    /// If true, then \ref sx and \ref sy have been allocated
    bool sxalloc;
    /// Storage for internally created subvector
    svec_t *sx;
    /// Storage for internally created subvector
    svec_t *sy;

    /// Pointer to base interpolation object
    base_interp<vec_t> *rit1;
    /// Pointer to base interpolation object
    base_interp<svec_t> *rit2;

    /// Pointer to base interpolation manager
    base_interp_mgr<vec_t> *bim1;

    /// Pointer to base interpolation manager
    base_interp_mgr<svec_t> *bim2;
    
    /// Memory allocator for objects of type \c alloc_vec_t
    alloc_t ao;

    /// True if the user-specified x vector is increasing
    bool inc;

    /// Pointer to user-specified vector
    const vec_t *lx;
    /// Pointer to user-specified vector
    const vec_t *ly;

    /// Size of user-specifed vector
    size_t ln;
    
#endif

  public:

    /// Create with base interpolation objects \c it and \c rit
    smart_interp_vec(size_t n, const vec_t &x, const vec_t &y) {
      sx=0;
      sy=0;
      sxalloc=false;

      bim1=&dim1;
      bim2=&dim2;

      rit1=bim1->new_interp();
      rit2=bim2->new_interp();
      ln=0;
      
      if (n<rit1->min_size || n<rit2->min_size) {
	O2SCL_ERR("Vector size too small in o2scl_interp_vec().",gsl_edom);
      } else {

	  int r1=rit1->allocate(n);
	  if (r1==0) {
	    rit1->init(x,y,n);
	    ln=n;
	    lx=&x;
	    ly=&y;
	    inc=true;
	}

      } 

    }

    /// Create with base interpolation objects \c it and \c rit
    smart_interp_vec(base_interp_mgr<vec_t> &it1, 
		     base_interp_mgr<svec_t> &it2,
		     size_t n, const vec_t &x, const vec_t &y) {
      sx=0;
      sy=0;
      sxalloc=false;

      bim1=&it1;
      bim2=&it2;

      rit1=bim1->new_interp();
      rit2=bim2->new_interp();
      ln=0;
	
      if (n<rit1->min_size || n<rit2->min_size) {
	O2SCL_ERR("Vector size too small in o2scl_interp_vec().",gsl_edom);
      } else {

	  int r1=rit1->allocate(n);
	  if (r1==0) {
	    rit1->init(x,y,n);
	    ln=n;
	    lx=&x;
	    ly=&y;
	    inc=true;
	  }

      } 
	
    }

    virtual ~smart_interp_vec() {
      if (sxalloc) {
	delete sx;
	delete sy;
      }
      
      if (ln>0) {
	rit1->free();
      }
      
      /*
	We don't want to use "bim1->free_interp(rit1)" here instead
	of "delete rit1" because the pointer "bim1" may refer to 
	a user-defined interpolation manager, and it may have gone
	out of scope earlier.
      */
      delete rit1;
      delete rit2;
    }

    /// Give the value of the function \f$ y(x=x_0) \f$ .
    virtual double operator()(const double x0) {
      return interp(x0);
    }

    /// Give the value of the function \f$ y(x=x_0) \f$ .
    virtual double interp(const double x0) {
      
      double ret=0.0;
      int err;
      if (ln>0) {
	
	err=rit1->interp(*lx,*ly,ln,x0,ret);
	if (err!=gsl_success) {
	  size_t nn=0;
	  int retfs=find_inc_subset(x0,ln,*lx,*ly,nn);
	  
	  if (retfs==0 && nn>1 && nn>=rit2->min_size) {
	    rit2->allocate(nn);
	    rit2->init(*sx,*sy,nn);
	    err=rit2->interp(*sx,*sy,nn,x0,ret);
	    rit2->free();
	  } else {
	    O2SCL_ERR2("Interpolation failed in ",
		       "smart_interp_vec::interp().",gsl_efailed);
	    return 0.0;
	  }
	  
	}
	
      }
      
      return ret;
    }
    
    /// Give the value of the derivative \f$ y^{prime}(x=x_0) \f$ .
    virtual double deriv(const double x0) {
      double ret=0.0;
      if (ln>0) {
	rit1->deriv(*lx,*ly,ln,x0,ret);
      }
      return ret;
    }		      
    
    /** \brief Give the value of the second derivative  
	\f$ y^{prime \prime}(x=x_0) \f$ .
    */
    virtual double deriv2(const double x0) {
      double ret=0.0;
      if (ln>0) {
	rit1->deriv2(*lx,*ly,ln,x0,ret);
      }
      return ret;
    }		      
    
    /// Give the value of the integral \f$ \int_a^{b}y(x)~dx \f$ .
    virtual double integ(const double x1, const double x2) {
      double ret=0.0;
      if (ln>0) {
	rit1->integ(*lx,*ly,ln,x1,x2,ret);
      }
      return ret;
    }		      

    /// Default interpolation manager
    def_interp_mgr<vec_t,cspline_interp> dim1;

    /// Default interpolation manager
    def_interp_mgr<svec_t,cspline_interp> dim2;
      
#ifndef DOXYGEN_INTERNAL

  protected:

    /// A lookup function for generic vectors
    size_t local_lookup(size_t n, const vec_t &x, double x0) {
      size_t row=0, i=0;
      while(!gsl_finite(x[i]) && i<n-1) i++;
      if (i==n-1) {
	return 0;
      }
      double best=x[i], bdiff=fabs(x[i]-x0);
      for(;i<n;i++) {
	if (gsl_finite(x[i]) && fabs(x[i]-x0)<bdiff) {
	  row=i;
	  best=x[i];
	  bdiff=fabs(x[i]-x0);
	}
      }
      return row;
    }

    /** \brief Try to find the largest monotonically increasing and
	finite region around the desired location

	This function looks through the vector \c x near the element
	closest to \c x0 to find the largest possible monotonic
	region. If it succeeds, it returns \ref gsl_success, and if it
	fails, it returns \ref gsl_efailed. It does not call the error
	handler.

	\todo Variables \c row and \c row appear to be unused? (2/17/11)
    */
    int find_inc_subset(const double x0, size_t sz, const vec_t &x, 
			const vec_t &y, size_t &nsz) {

      size_t row=local_lookup(sz,x,x0), row2=row++;
      size_t left=row, right=row;
      
      if (row2>sz) row2=row-1;

      if (!gsl_finite(x[row]) || !gsl_finite(y[row])) {
	return gsl_efailed;
      }
  
      // Increasing case
      
      while(left>0 && x[left-1]<x[left] && 
	    gsl_finite(x[left-1]) && gsl_finite(y[left-1])) left--;
      while(right<sz-1 && x[right]<x[right+1] &&
	    gsl_finite(x[right+1]) && gsl_finite(y[right+1])) right++;
      
      nsz=right-left+1;
      
      if (sxalloc) {
	delete sx;
	delete sy;
      } 
      sx=new svec_t(x,left,nsz);
      sy=new svec_t(y,left,nsz);
      sxalloc=true;
      
      return 0;
    }

#endif

  };

  // sm_interp typedef 
  typedef smart_interp<ovector_const_view,
    ovector_const_subvector> sm_interp;
  
  // sm_interp_vec typedef
  typedef smart_interp_vec<ovector_const_view,ovector_const_subvector,ovector,
    ovector_alloc> sm_interp_vec;
  
  /** \brief A specialization of smart_interp for C-style double arrays
   */
  template<size_t n> class sma_interp :
  public smart_interp<double[n],array_const_subvector> {
    
  public:
    
    /** \brief Create an interpolation object with user-specified
	interpolation types
    */
  sma_interp(base_interp_mgr<double[n]> &it1, 
	     base_interp_mgr<array_const_subvector> &it3) :
    smart_interp<double[n],array_const_subvector>(it1,it3) {
    }
    
    /// Create an interpolation object with the default interpolation types
  sma_interp() : smart_interp<double[n],array_const_subvector>() {
    }

  };
  
  /** \brief A specialization of smart_interp_vec for C-style double arrays
   */
  template<class arr_t> class sma_interp_vec : 
  public smart_interp_vec<arr_t,array_const_subvector,arr_t,
    array_alloc<arr_t> >
      {
	
      public:
    
    
    /** \brief Create an interpolation object with user-specified 
	interpolation types
    */
      sma_interp_vec(base_interp_mgr<arr_t> &it, 
		     base_interp_mgr<array_const_subvector> &it2,
		     size_t n, const arr_t &x, const arr_t &y) :
	smart_interp_vec<arr_t,array_const_subvector,arr_t,
	  array_alloc<arr_t> >(it,it2,n,x,y) {
	}
	
    /// Create an interpolation object with the default interpolation types
      sma_interp_vec(size_t n, const arr_t &x, const arr_t &y) :    
	smart_interp_vec<arr_t,array_const_subvector,arr_t,
	  array_alloc<arr_t> >(n,x,y) {
	}
	
      };
  
  /** \brief Perform inverse linear interpolation

      This function performs inverse linear interpolation of the data
      defined by \c x and \c y, finding all points in \c x which have
      the property \f$ \mathrm{level} = y(x) \f$. All points for which
      this relation holds are put into the vector \c locs. The
      previous information contained in vector \c locs before the
      function call is destroyed.

      This is the 1-dimensional analog of finding contour levels as
      the \ref contour class does for 2 dimensions.
  */
  template<class vec_t, class vec2_t> int vector_find_level
    (double level, size_t n, vec_t &x, vec2_t &y, ovector &locs) {

    if (n<=1) {
      O2SCL_ERR2_RET("Need at least two data points in ",
		     "vector_find_level().",gsl_einval);
    }
    
    // Ensure that the location vector is empty
    locs.free();

    // Find intersections
    for(size_t i=0;i<n-1;i++) {

      if ((y[i]<level && y[i+1]>=level) ||
	  (y[i]>level && y[i+1]<=level)) {
	// For each intersection, add the location using linear 
	// interpolation
	double x0=x[i]+(x[i+1]-x[i])*(level-y[i])/(y[i+1]-y[i]);
	locs.push_back(x0);
      }
    }

    return 0;
  }

  /// Compute the integral over <tt>y(x)</tt> using linear interpolation
  template<class vec_t> double vector_integ_linear(size_t n, vec_t &x, 
						   vec_t &y) {

    // Linear interpolation managers
    def_interp_mgr<vec_t,linear_interp> lint1;
    
    // Interpolation object
    o2scl_interp<vec_t> si(lint1);

    // Compute full integral
    double total=si.integ(x[0],x[n-1],n,x,y);

    return total;
  }

  /** \brief Compute the endpoints which enclose the regions whose
      integral is equal to \c sum

      Defining a new function, \f$ g(y_0) \f$ which takes as input any
      y-value, \f$ y_0 \f$ from the function \f$ y(x) \f$ (specified
      with the parameters \c x and \c y) and outputs the integral of
      the function \f$ y(x) \f$ over all regions where \f$ y(x) > y_0
      \f$. This function inverts \f$ g(y) \f$, taking the value
      of an integral as input, and returns the corresponding y-value
      in the variable \c lev. 

      In order to make sure that the intepretation of the integral is
      unambiguous, this function requires that the first and last values
      of \c y are equal, i.e. <tt>y[0]==y[n-1]</tt>. 

      This function is particularly useful, for example, in computing 
      the region which defines 68\% around a peak of data, thus 
      providing approximate \f$ 1~\sigma \f$ limits. 

      Linear interpolation is used to describe the function \f$ g \f$,
      and the precision of this function is limited by this assumption.
      This function may also sometimes fail if \c sum is very close to
      the minimum or maximum value of the function \f$ g \f$. 

      \comment
      Note that the two vector types for x and y must be the
      same in order to use o2scl_interp.
      \endcomment
  */
  template<class vec_t> int vector_invert_enclosed_sum
    (double sum, size_t n, vec_t &x, vec_t &y, double &lev) {
    
    if (n<=1) {
      O2SCL_ERR2_RET("Need at least two data points in ",
		     "vector_invert_enclosed_sum().",gsl_einval);
    }

    if (y[0]!=y[n-1]) {
      O2SCL_ERR2_RET("The first and last y-values must be equal in ",
		     "vector_invert_enclosed_sum().",gsl_einval);
    }

    // Construct a sorted list of function values 
    ovector ysort(n);
    vector_copy(n,y,ysort);
    vector_sort<ovector,double>(ysort.size(),ysort);

    // Create list of y-values to perform y-value and integral
    // interpolation. We choose values in between the grid points to
    // ensure that we don't accidentally land precisely on a peak or
    // valley.
    ovector ylist;
    for(size_t i=0;i<ysort.size()-1;i++) {
      if (ysort[i]!=ysort[i+1]) {
	ylist.push_back((ysort[i+1]+3.0*ysort[i])/4.0);
	ylist.push_back((ysort[i+1]*3.0+ysort[i])/4.0);
      }
    }
    
    // Linear interpolation managers
    def_interp_mgr<vec_t,linear_interp> lint1;
    def_interp_mgr<ovector,linear_interp> mint1;

    // Interpolation objects. We need two, one for the
    // user-specified vector type, and one for ovectors
    o2scl_interp<vec_t> si(lint1);
    o2scl_interp<ovector> si2(mint1);
    
    // Construct vectors for interpolation
    ovector xi, yi;
    
    size_t nfail=0;

    for(size_t k=0;k<ylist.size();k++) {
      double lev_tmp=ylist[k];
      ovector locs;
      vector_find_level(lev_tmp,n,x,y,locs);
      if ((locs.size()%2)!=0) {
	nfail++;
      } else {
	double sum_temp=0.0;
	for(size_t i=0;i<locs.size()/2;i++) {
	  double x0=locs[2*i];
	  double x1=locs[2*i+1];
	  sum_temp+=si.integ(x0,x1,n,x,y);
	}
	xi.push_back(sum_temp);
	yi.push_back(lev_tmp);
      }
    }
    if (nfail>10) {
      O2SCL_ERR2_RET("More than 10 failures in ",
		     "vector_invert_enclosed_sum().",gsl_einval);
    }

    lev=si2.interp(sum,xi.size(),xi,yi);
    
    return 0;
  }

#ifndef DOXYGENP
}
#endif

#endif
