/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/table_units.h>

using namespace std;
using namespace o2scl;

table_units::table_units(int cmaxlines) : table(cmaxlines) {
  cup=&def_cu;
}

table_units::~table_units() {
  utree.clear();
}

int table_units::delete_column(string scol) {

  // Find the tree iterator for the element we want to erase
  aiter it=atree.find(scol);
  if (it==atree.end()) {
    O2SCL_ERR((((string)"Column '")+scol+
	       " not found in delete_column().").c_str(),
	      gsl_enotfound);
    return gsl_enotfound;
  }

  // Remove the unit for the specified column
  if (get_unit(scol).length()>0) remove_unit(scol);

  // Find the corresponding list iterator
  aviter vit=alist.begin();
  vit+=it->second.index;
  
  // Find the last element in the list and it's corresponding table
  // entry. Change it's index to the index of the column to be
  // deleted.
  alist[alist.size()-1]->second.index=it->second.index;

  // Erase the elements from the list and the tree
  atree.erase(it);
  alist.erase(vit);

  // Reset the list to reflect the proper iterators
  reset_list();

  return gsl_success;
}

table_units::table_units(const table &t) : table(t.get_nlines()) {
  
  // Copy constants 
  size_t nc=t.get_nconsts();
  for(size_t i=0;i<nc;i++) {
    string name;
    double val;
    t.get_constant(i,name,val);
    constants.insert(make_pair(name,val));
  }
  
  // Copy the columns and data
  nlines=t.get_nlines();
  maxlines=nlines;
  for(size_t i=0;i<t.get_ncolumns();i++) {

    // Column name
    string cname=t.get_column_name(i);

    // Insert column into tree
    col s;
    s.dat=new ovector(nlines);
    s.owner=true;
    s.index=atree.size();
    atree.insert(make_pair(cname,s));

    // Insert in iterator index
    aiter it=atree.find(cname);
    alist.push_back(it);
    
    // Fill the data
    for(size_t j=0;j<t.get_nlines();j++) {
      (*it->second.dat)[j]=t.get(cname,j);
    }
    
  }

  intp_set=false;
  bim1=&dim1;
  bim2=&dim2;

  return;
}

table_units::table_units(const table_units &t) : table(t.get_nlines()) {
  
  // Copy constants 
  constants=t.constants;
  
  // Copy the columns and data
  nlines=t.get_nlines();
  maxlines=nlines;
  for(size_t i=0;i<t.get_ncolumns();i++) {

    // Column name
    string cname=t.get_column_name(i);

    // Insert column into tree
    col s;
    s.dat=new ovector(nlines);
    s.owner=true;
    s.index=atree.size();
    atree.insert(make_pair(cname,s));

    // Insert in iterator index
    aiter it=atree.find(cname);
    alist.push_back(it);
    
    // Insert in unit list
    utree.insert(make_pair(cname,t.get_unit(cname)));
    
    // Fill the data
    for(size_t j=0;j<t.get_nlines();j++) {
      (*it->second.dat)[j]=t.get(cname,j);
    }
    
  }

  intp_set=false;
  bim1=&dim1;
  bim2=&dim2;

  return;
}

table_units &table_units::operator=(const table &t) {
  
  clear_table();
  constants.clear();
  utree.clear();

  // Copy constants 
  size_t nc=t.get_nconsts();
  for(size_t i=0;i<nc;i++) {
    string name;
    double val;
    t.get_constant(i,name,val);
    constants.insert(make_pair(name,val));
  }
  
  // Copy the columns and data
  nlines=t.get_nlines();
  for(size_t i=0;i<t.get_ncolumns();i++) {

    // Column name
    string cname=t.get_column_name(i);

    // Insert column into tree
    col s;
    s.dat=new ovector(nlines);
    s.owner=true;
    s.index=atree.size();
    atree.insert(make_pair(cname,s));

    // Insert in iterator index
    aiter it=atree.find(cname);
    alist.push_back(it);
    
    // Fill the data
    for(size_t j=0;j<t.get_nlines();j++) {
      (*it->second.dat)[j]=t.get(cname,j);
    }
    
  }

  return *this;
}

table_units &table_units::operator=(const table_units &t) {
  
  clear_table();
  constants.clear();
  utree.clear();

  // Copy constants 
  constants=t.constants;
  
  // Copy the columns and data
  nlines=t.get_nlines();
  for(size_t i=0;i<t.get_ncolumns();i++) {

    // Column name
    string cname=t.get_column_name(i);

    // Insert column into tree
    col s;
    s.dat=new ovector(nlines);
    s.owner=true;
    s.index=atree.size();
    atree.insert(make_pair(cname,s));

    // Insert in iterator index
    aiter it=atree.find(cname);
    alist.push_back(it);
    
    // Insert in unit list
    utree.insert(make_pair(cname,t.get_unit(cname)));
    
    // Fill the data
    for(size_t j=0;j<t.get_nlines();j++) {
      (*it->second.dat)[j]=t.get(cname,j);
    }
    
  }

  return *this;
}

double table_units::get_conv(std::string old_unit, std::string new_unit) {
  return cup->convert(old_unit,new_unit,1.0);
}

int table_units::summary(ostream *out, int ncol) const {

  if (constants.size()==1) {
    (*out) << "1 constant:" << endl;
  } else {
    (*out) << constants.size() << " constants:" << endl;
  }
  std::map<std::string,double>::const_iterator mit;
  for(mit=constants.begin();mit!=constants.end();mit++) {
    (*out) << mit->first << " " << mit->second << endl;
  }

  // Output number of columns and preprend column numbers
  int nh=get_ncolumns(), nh2;
  if (nh==1) {
    (*out) << "1 column: " << endl;
  } else {
    (*out) << nh << " columns: " << endl;
  }
  string *h=new string[nh];
  for(int i=0;i<nh;i++) {
    h[i]=itos(i)+". "+get_column_name(i)+" ["+
      get_unit(get_column_name(i))+"]";
  }
  
  vector<string> h2;
  // Convert to string with width 'ncol'
  screenify(nh,h,h2,ncol);
  nh2=h2.size();
  
  // Output column names
  for(int i=0;i<nh2;i++) {
    (*out) << h2[i] << endl;
  }
  
  if (get_nlines()==0) (*out) << "No lines of data." << endl;
  else if (get_nlines()==1) (*out) << "One line of data." << endl;
  (*out) << get_nlines() << " lines of data." << endl;
  
  delete[] h;
  
  return 0;
}

int table_units::convert_to_unit(std::string scol, std::string unit,
				 bool err_on_fail) {

  // Find unit entry
  uiter it=utree.find(scol);
  if (it==utree.end()) {
    if (err_on_fail) {
      O2SCL_ERR_RET((((string)"Column '")+scol+
		     "' not found in table_units::convert_to_unit().").c_str(),
		    gsl_enotfound);
    } else {
      return gsl_enotfound;
    }
  };

  // If the units are equal, just do nothing
  if (it->second==unit) return gsl_success;

  // Find column of data
  aiter at=atree.find(scol);
  if (at==atree.end()) {
    if (err_on_fail) {
      O2SCL_ERR_RET((((string)"Column '")+scol+
		     "' not found in table_units::convert_to_unit().").c_str(),
		    gsl_enotfound);
    } else {
      return gsl_enotfound;
    }
  };

  // Perform conversion
  ovector_base *vec=at->second.dat;
  double conv=cup->convert(it->second,unit,1.0);
  for(size_t i=0;i<get_nlines();i++) {
    (*vec)[i]*=conv;
  }

  // Set new unit entry
  it->second=unit;

  return gsl_success;
}

std::string table_units::get_unit(string scol) const {
  
  uciter it=utree.find(scol);
  if (it==utree.end()) {
    // Not found in unit entry, look for column of data
    aciter at=atree.find(scol);
    if (at==atree.end()) {
      O2SCL_ERR((((string)"Column '")+scol+
		 "' not found in table_units::get_unit().").c_str(),
		gsl_enotfound);
    } else {
      return "";
    }
  }

  return it->second;
}

int table_units::remove_unit(string scol) {
  
  uiter it=utree.find(scol);
  if (it==utree.end()) {
    O2SCL_ERR((((string)"Column '")+scol+
	       "' not found in table_units::get_unit().").c_str(),
	      gsl_enotfound);
  }
  
  if (utree.size()==0) {
    O2SCL_ERR("No units specified in table_units::remove_unit().",
	      gsl_efailed);
  }

  utree.erase(it);
  return gsl_success;
}

int table_units::set_unit(string scol, string unit) {
  
  uiter it=utree.find(scol);
  if (it==utree.end()) {
    aiter at=atree.find(scol);
    if (at==atree.end()) {
      O2SCL_ERR_RET((((string)"Column '")+scol+
		     "' not found in table_units::set_unit().").c_str(),
		    gsl_enotfound);
    }
    utree.insert(make_pair(scol,unit));
  } else {
    it->second=unit;
  }

  return gsl_success;
}

