/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_TENSOR_H
#define O2SCL_TENSOR_H

/** \file tensor.h
    \brief File for definitions of tensors
*/

#include <iostream>
#include <cstdlib>
#include <string>
#include <fstream>
#include <sstream>

#include <gsl/gsl_matrix.h>
#include <gsl/gsl_ieee_utils.h>

#include <o2scl/err_hnd.h>
#include <o2scl/uvector_tlate.h>
#include <o2scl/umatrix_tlate.h>
#include <o2scl/smart_interp.h>

#ifndef DOXYGENP
namespace o2scl {
#endif
  
  /** \brief Tensor class with arbitrary dimensions

      The elements of a tensor are typically specified as a list of
      <tt>size_t</tt> numbers with length equal to the tensor rank.
      For a rank-4 tensor named \c t, the element 
      <tt>t[1][2][0][3]</tt> can be obtained with something similar to 
      \code
      size_t ix[4]={1,2,0,3};
      double x=t.get(ix);
      \endcode

      Empty tensors have zero rank.

      Slices of tensors are subsets obtained from fixing the index of
      several dimensions, while letting other dimensions vary. For a
      slice with one dimension not fixed, see \ref vector_slice(). For
      a slice with two dimensions not fixed, see \ref matrix_slice().

      For I/O with tensors, see \ref hdf_io.h .

      \future Could implement arithmetic operators + and - and some
      different products. 

      \future Implement copies to and from vector
      and matrices 

      \future Implement tensor contractions, i.e. tensor
      = tensor * tensor 

      \future Consider making a template type to replace double,
      e.g. <tt>value_t</tt>.

      \future Could be interesting to write an iterator for this class.

      \comment
      A tensor is defined to be empty if it's rank is zero. A tensor's
      rank should be zero if and only if no memory has been allocated
      for it.

      I could have chosen to make set() functions virtual functions
      instead of templates. However, many of these functions may be
      templates anyway when I generalize this class to hold floats or
      ints rather than just doubles. 
      \endcomment
  */
  class tensor {

#ifndef DOXYGEN_INTERNAL

  protected:

    /// The data
    uvector data;

    /// A rank-sized array of the sizes of each dimension
    uvector_size_t size;

    /// Rank
    size_t rk;

#endif

  public:
    
    /// Create an empty tensor with zero rank
    tensor() {
      rk=0;
    }

    /** \brief Create a tensor of rank \c rank with sizes given in \c dim
	
	The parameter \c dim must be a pointer to an array of sizes
	with length \c rank. If the user requests any of the sizes to
	be zero, this constructor will call the error handler, create
	an empty tensor, and will allocate no memory.
    */
    template<class size_vec_t> 
      tensor(size_t rank, const size_vec_t &dim) {
      if (rank==0) {
	data=0;
	size=0;
      } else {
	for(size_t i=0;i<rk;i++) {
	  if (dim[i]==0) {	  
	    rk=0;
	    O2SCL_ERR2("Requested zero size with non-zero rank in ",
		       "tensor::tensor<>(size_t, size_vec_t &)",gsl_einval);
	  }
	}
	rk=rank;
	size.allocate(rk);
	size_t tot=1;
	for(size_t i=0;i<rk;i++) {
	  size[i]=dim[i];
	  tot*=size[i];
	}
	data.allocate(tot);
      }
    }

    ~tensor() {
    }

    /// \name Set functions
    //@{
    /// Set the element indexed by \c index to value \c val
    template<class size_vec_t>
      void set(const size_vec_t &index, double val) {
#if O2SCL_NO_RANGE_CHECK
#else
      if (rk==0) {
	O2SCL_ERR("Empty tensor in set().",gsl_einval);
	return;
      }
      if (index[0]>=size[0]) {
	O2SCL_ERR("Index greater than size in set().",gsl_eindex);
	return;
      }
#endif
      size_t ix=index[0];
      for(size_t i=1;i<rk;i++) {
#if O2SCL_NO_RANGE_CHECK
#else
	if (index[i]>=size[i]) {
	  O2SCL_ERR("Index greater than size in get().",gsl_eindex);
	  return;
	}
#endif
	ix*=size[i];
	ix+=index[i];
      }
      data[ix]=val;
      return;
    }

    /// Set all elements in a tensor to some fixed value
    void set_all(double x) {
      for(size_t i=0;i<total_size();i++) data[i]=x;
      return;
    }
    //@}

    /// \name Get functions
    //@{
    /// Get the element indexed by \c index
    template<class size_vec_t> double &get(const size_vec_t &index) {
#if O2SCL_NO_RANGE_CHECK
#else
      if (rk==0) {
	O2SCL_ERR("Empty tensor in get().",gsl_einval);
	return data[0];
      }
      if (index[0]>=size[0]) {
	O2SCL_ERR("Index greater than size in get().",gsl_eindex);
	return data[0];
      }
#endif
      size_t ix=index[0];
      for(size_t i=1;i<rk;i++) {
#if O2SCL_NO_RANGE_CHECK
#else
	if (index[i]>=size[i]) {
	  O2SCL_ERR("Index greater than size in get().",gsl_eindex);
	  return data[0];
	}
#endif
	ix*=size[i];
	ix+=index[i];
      }
      return data[ix];
    }

    /// Get a const reference to the element indexed by \c index
    template<class size_vec_t> 
      double const &get(const size_vec_t &index) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (rk==0) {
	O2SCL_ERR("Empty tensor in get().",gsl_einval);
	return data[0];
      }
      if (index[0]>=size[0]) {
	O2SCL_ERR("Index greater than size in get().",gsl_eindex);
	return data[0];
      }
#endif
      size_t ix=index[0];
      for(size_t i=1;i<rk;i++) {
#if O2SCL_NO_RANGE_CHECK
#else
	if (index[i]>=size[i]) {
	  O2SCL_ERR("Index greater than size in get().",gsl_eindex);
	  return data[0];
	}
#endif
	ix*=size[i];
	ix+=index[i];
      }
      return data[ix];
    }
    //@}
    
    /// \name Slice functions
    //@{
    /** \brief Fix all but one index to create a vector

	This fixes all of the indices to the values given in \c index
	except for the index number \c ix, and returns the
	corresponding vector, whose length is equal to the size of the
	tensor in that index. The value <tt>index[ix]</tt> is ignored.

	For example, for a rank 3 tensor allocated with
	\code
	tensor t;
	size_t dim[3]={3,4,5};
	t.allocate(3,dim);
	\endcode
	the following code
	\code
	size_t index[3]={1,0,3};
	ovector_view v=t.vector_slice(1,index);
	\endcode
	Gives a vector \c v of length 4 which refers to the values
	<tt>t(1,0,3)</tt>, <tt>t(1,1,3)</tt>, <tt>t(1,2,3)</tt>, and
	<tt>t(1,3,3)</tt>.
    */
    template<class size_vec_t> 
      ovector_array_stride vector_slice(size_t ix, const size_vec_t &index) {
      if (ix+1>rk) {
	O2SCL_ERR("Specified index to large in tensor::vector_slice()",
		  gsl_einval);
      }
      size_t start;
      if (ix==0) start=0;
      else start=index[0];
      for(size_t i=1;i<rk;i++) {
	start*=size[i];
	if (i!=ix) start+=index[i];
      }
      size_t stride=1;
      for(size_t i=ix+1;i<rk;i++) stride*=size[i];
      return ovector_array_stride(size[ix],&(data[start]),stride);
    }

    /** \brief Fix all but two indices to create a matrix

	This fixes all of the indices to the values given in \c index
	except for the index number \c ix and the last index, and
	returns the corresponding matrix, whose size is equal to the
	size of the tensor in the two indices which are not fixed.

	Since the last index is always fixed, the parameter \c ix
	is limited to be \f$ \in [0,\mathrm{rank}-2] \f$.
    */
    template<class size_vec_t> 
      omatrix_array matrix_slice(size_t ix, const size_vec_t &index) {
      if (ix+1>=rk) {
	O2SCL_ERR("Specified index to large in tensor::matrix_slice()",
		  gsl_einval);
      }
      size_t start=0, tot=size[0];
      for(size_t i=1;i<rk;i++) {
	tot*=size[i];
      }
      // This O(rank^2) algorithm could possibly 
      // be replaced with a O(rank) one in the future
      for(size_t i=0;i<rk-1;i++) {
	if (i!=ix) {
	  size_t adj=index[i];
	  for(size_t j=i+1;j<rk;j++) adj*=size[j];
	  start+=adj;
	}
      }
      size_t tda=1;
      if (ix==rk-1) {
	tda=0;
      } else {
	for(size_t i=ix+1;i<rk;i++) tda*=size[i];
      }
      return omatrix_array(&(data[start]),tda,size[ix],size[rk-1]);
    }
    //@}

    /// \name Memory allocation
    //@{
    /** \brief Allocate space for a tensor of rank \c rank with sizes
	given in \c dim
	
	The parameter \c dim must be a pointer to an array of sizes
	with length \c rank. 

	If memory was previously allocated, it will be freed before
	the new allocation.

	If the user requests any of the sizes to be zero, this
	function will call the error handler and will allocate no
	memory. If memory was previously allocated, the tensor is left
	unmodified and no deallocation is performed.
    */
    template<class size_vec_t> 
      void allocate(size_t rank, const size_vec_t &dim) {
      for(size_t i=0;i<rk;i++) {
	if (dim[i]==0) {
	  O2SCL_ERR2("Requested zero size in tensor::",
		     "tensor(size_t, size_t *)",gsl_einval);
	}
      }
      if (rk>0) free();
      rk=rank;
      size.allocate(rk);
      size_t tot=1;
      for(size_t i=0;i<rk;i++) {
	size[i]=dim[i];
	tot*=size[i];
      }
      data.allocate(tot);
      return;
    }

    /// Free allocated space (also sets rank to zero)
    void free() {
      if (rk>0) {
	size.free();
	data.free();
	rk=0;
      }
      return;
    }
    //@}

    /// \name Size functions
    //@{
    /// Return the rank of the tensor
    size_t get_rank() const { return rk; }

    /// Returns the size of the ith index
    size_t get_size(size_t i) const { 
      if (i<rk) {
	return size[i]; 
      }
      O2SCL_ERR_RET("Requested index exceeding rank in get_size()",
		    gsl_einval);
    }
    
    /// Return the full array of sizes
    const uvector_size_t &get_size_arr() const {
      return size;
    }

    /// Return the full array of sizes
    const uvector &get_data() const {
      return data;
    }

#ifndef DOXYGENP
    // For HDF I/O. Undocumented. Probably should be replaced by
    // friend function or something else?
    uvector &get_data_non_const() {
      return data;
    }
#endif

    /** \brief Returns the size of the tensor (the product of 
	the sizes over every index)
    */
    size_t total_size() const { 
      if (rk==0) return 0;
      size_t tot=1;
      for(size_t i=0;i<rk;i++) tot*=size[i];
      return tot;
    }
    //@}

    /// \name Index manipulation
    //@{
    /// Pack the indices into a single array index
    template<class size_vec_t> 
      size_t pack_indices(const size_vec_t &index) {
      if (rk==0) {
	O2SCL_ERR("Empty tensor in pack_indices().",gsl_einval);
	return 0;
      }
      if (index[0]>=size[0]) {
	O2SCL_ERR("Index greater than size in pack_indices().",gsl_eindex);
	return 0;
      }
      size_t ix=index[0];
      for(size_t i=1;i<rk;i++) {
	if (index[i]>=size[i]) {
	  O2SCL_ERR("Index greater than size in pack_indices().",gsl_eindex);
	  return 0;
	}
	ix*=size[i];
	ix+=index[i];
      }
      return ix;
    }
    
    /// Unpack the single array index into indices
    template<class size_vec_t> 
      void unpack_indices(size_t ix, size_vec_t &index) {
      if (ix>total_size()) {
	O2SCL_ERR("Index greater than total size in unpack_indices().",
		  gsl_eindex);
	return;
      }
      size_t ix2, sub_size;
      for(size_t i=0;i<rk;i++) {
	if (i==rk-1) {
	  index[i]=ix;
	} else {
	  sub_size=1;
	  for(size_t j=i+1;j<rk;j++) sub_size*=size[j];
	  index[i]=ix/sub_size;
	  // (Remember we're doing integer arithmetic here.)
	  ix-=sub_size*(ix/sub_size);
	}
      }
      return;
    }
    //@}

  };

  /** \brief Tensor class with arbitrary dimensions with a grid

      This tensor class allows one to assign the indexes to 
      numerical scales, so that n-dimensional interpolation can
      be performed. To set the grid, use set_grid() and then
      interpolation can be done using interpolate().
      
      By convention, member functions ending in the _val 
      suffix return the closest grid-point to some user-specified
      values. 

      \comment
      I like how hist_new only allows you to set the
      grid if it's the same size as the old grid or the tensor
      is empty. This same practice could be applied here, to
      force the user to clear the tensor before resetting the grid.
      (10/24/11 - Actually, maybe this is a bad idea, because
      this class is more analogous to ovector, which doesn't
      behave this way.)
      \endcomment

      \future Only allocate space for grid if it is set
      \future Could implement arithmetic operators + and - and some
      different products.
      \future Consider creating a set_grid() function which
      takes grids from an object like hist_grid. Maybe make a 
      constructor for a tensor_grid object which just takes 
      as input a set of grids?
  */
  class tensor_grid : public tensor {

#ifndef DOXYGEN_INTERNAL

  protected:

    /// A rank-sized set of arrays for the grid points
    uvector grid;

    /// If true, the grid has been set by the user
    bool grid_set;
    
    /// The interpolation manager
    base_interp_mgr<uvector_base> *bim1;

    /// The subvector interpolation manager
    base_interp_mgr<uvector_const_subvector> *bim2;

#endif

  public:
    
    /// Create an empty tensor with zero rank
  tensor_grid() : tensor() {
      grid_set=false;
      bim1=&dim1;
      bim2=&dim2;
    }
    
    /** \brief Create a tensor of rank \c rank with sizes given in \c dim
	
	The parameter \c dim must be a pointer to an array of sizes
	with length \c rank. If the user requests any of the sizes to
	be zero, this constructor will call the error handler, create
	an empty tensor, and will allocate no memory.
    */
    template<class size_vec_t> 
      tensor_grid(size_t rank, const size_vec_t &dim) : 
    tensor(rank,dim) {
      grid_set=false;
      for(size_t i=0;i<rk;i++) {
	if (dim[i]==0) {
	  rk=0;
	  O2SCL_ERR("Requested zero size in tensor::tensor(size_t, size_t *)",
		    gsl_einval);
	}
      }
    }

    ~tensor_grid() {
    }

    /// \name Set functions
    //@{
    /// Set the element closest to grid point \c grdp to value \c val
    template<class vec_t> 
      int set_val(const vec_t &grdp, double val) {

      // Find indices
      uvector_size_t index(rk);
      for(size_t i=0;i<rk;i++) index[i]=lookup_grid(i,grdp[i]);
      
      // Pack
      size_t ix=index[0];
      for(size_t i=1;i<rk;i++) {
	ix*=size[i];
	ix+=index[i];
      }

      // Set value
      data[ix]=val;

      return 0;
    }

    /** \brief Set the element closest to grid point \c grdp to value 
	\c val
	
	The parameters \c closest and \c grdp may be identical,
	allowing one to update a vector \c grdp with the
	closest grid point.
    */
    template<class vec_t, class vec2_t> 
      int set_val(const vec_t &grdp, double val, vec2_t &closest) {

      // Find indices
      uvector_size_t index(rk);
      for(size_t i=0;i<rk;i++) {
	index[i]=lookup_grid_val(i,grdp[i],closest[i]);
      }
      
      // Pack
      size_t ix=index[0];
      for(size_t i=1;i<rk;i++) {
	ix*=size[i];
	ix+=index[i];
      }

      // Set value
      data[ix]=val;

      return 0;
    }
    //@}

    /// \name Get functions
    //@{
    /// Get the element closest to grid point \c grdp 
    template<class vec_t> 
      double get_val(const vec_t &grdp) {
      
      // Find indices
      uvector_size_t index(rk);
      for(size_t i=0;i<rk;i++) index[i]=lookup_grid(i,grdp[i]);

      // Pack
      size_t ix=index[0];
      for(size_t i=1;i<rk;i++) {
	ix*=size[i];
	ix+=index[i];
      }

      // Set value
      return data[ix];
    }

    /// Get the element closest to grid point \c grdp to value \c val
    template<class vec_t, class vec2_t> 
      double get_val(const vec_t &grdp, const vec2_t &closest) {
      
      // Find indices
      uvector_size_t index(rk);
      for(size_t i=0;i<rk;i++) {
	index[i]=lookup_grid_val(i,grdp[i],closest[i]);
      }
      
      // Pack
      size_t ix=index[0];
      for(size_t i=1;i<rk;i++) {
	ix*=size[i];
	ix+=index[i];
      }

      // Set value
      return data[ix];
    }
    //@}
    
    /// \name Memory allocation
    //@{
    /** \brief Allocate space for a tensor of rank \c rank with sizes
	given in \c dim
	
	The parameter \c dim must be a pointer to an array of sizes
	with length \c rank. 

	If memory was previously allocated, it will be freed before
	the new allocation and previously specified grid data will 
	be lost.

	If the user requests any of the sizes to be zero, this
	function will call the error handler and will allocate no
	memory. If memory was previously allocated, the tensor is left
	unmodified and no deallocation is performed.
    */
    template<class size_vec_t>
      int allocate(size_t rank, const size_vec_t &dim) {
      for(size_t i=0;i<rk;i++) {
	if (dim[i]==0) {
	  O2SCL_ERR2_RET("Requested zero size in ",
			 "tensor::tensor(size_t,size_vec_t &)",gsl_einval);
	}
      }
      if (rk>0) free();
      rk=rank;
      size.allocate(rk);
      size_t tot=1;
      for(size_t i=0;i<rk;i++) {
	size[i]=dim[i];
	tot*=size[i];
      }
      data.allocate(tot);
      return 0;
    }

    /// Free allocated space (also sets rank to zero)
    int free() {
      if (rk>0) {
	data.free();
	grid.free();
	size.free();
	rk=0;
      }
      return 0;
    }
    //@}

    /// \name Grid manipulation
    //@{
    /// Return true if the grid has been set
    bool is_grid_set() const { return grid_set; }

    /** \brief Set the grid

	The grid must be specified for all of the dimensions at
	once. Denote \f$ (\mathrm{size})_0 \f$ as the size
	of the first dimension, \f$ (\mathrm{size})_1 \f$ as the
	size of the second dimesion, and so on. Then the 
	first \f$ (\mathrm{size})_0 \f$ entries in \c grid must
	be the grid for the first dimension, the next 
	\f$ (\mathrm{size})_1 \f$ entries must be the grid
	for the second dimension, and so on. Thus \c grid
	must be a vector of size
	\f[
	\sum_{i=0}^{\mathrm{rank}} (\mathrm{size})_i
	\f]

	Note that the grid is copied so the function argument may
	be destroyed by the user after calling set_grid() without
	affecting the tensor grid. 

	\future Define a more generic interface for matrix types
    */
    template<class vec_t>
      int set_grid(const vec_t &grid_vec) {
      if (rk==0) {
	O2SCL_ERR2_RET("Tried to set grid for empty tensor in ",
		       "tensor_grid::set_grid().",gsl_einval);
      }
      size_t ngrid=0;
      for(size_t i=0;i<rk;i++) ngrid+=size[i];
      grid.allocate(ngrid);
      for(size_t i=0;i<ngrid;i++) {
	grid[i]=grid_vec[i];
      }
      grid_set=true;
      return 0;
    }

    /// Lookup index for grid closest to \c val
    size_t lookup_grid(size_t i, double val) {
      if (i<rk && grid_set) {
	size_t istart=0;
	for(size_t j=0;j<i;j++) istart+=size[j];
	size_t best=0;
	double min=fabs(grid[istart]-val);
	for(size_t j=istart;j<istart+size[i];j++) {
	  if (fabs(grid[j]-val)<min) {
	    best=j;
	    min=fabs(grid[j]-val);
	  }
	}
	return best;
      }
      return 0;
    }

    /// Lookup jth value on the ith grid
    double get_grid(size_t i, size_t j) const {
      if (i<rk && grid_set) {
	size_t istart=0;
	for(size_t k=0;k<i;k++) istart+=size[k];
	return grid[istart+j];
      }
      return 0;
    }

    /** \brief Lookup indices for grid closest point to \c vals

        The values in \c vals are not modified by this function.
	
	\comment
	This function must have a different name than 
	lookup_grid() because the template types cause
	confusion between the two functions.
	\endcomment
    */
    template<class vec_t, class size_vec_t>
      int lookup_grid_vec(const vec_t &vals, 
			  const size_vec_t &indices) const {
      for(size_t k=0;k<rk;k++) {
	indices[k]=lookup_grid(k,vals[k]);
      }
      return 0;
    }

    /// Lookup index for grid closest to \c val, returning the grid point
    size_t lookup_grid_val(size_t i, double &val, double &val2) {
      if (i<rk && grid_set) {
	size_t istart=0;
	for(size_t j=0;j<i;j++) istart+=size[j];
	size_t best=0;
	double min=fabs(grid[istart]-val);
	val2=grid[istart];
	for(size_t j=istart;j<istart+size[i];j++) {
	  if (fabs(grid[j]-val)<min) {
	    best=j;
	    min=fabs(grid[j]-val);
	    val2=grid[j];
	  }
	}
	return best;
      }
      return 0;
    }
  
    /// \name Interpolation
    //@{
    def_interp_mgr<uvector_base,linear_interp> dim1;
    def_interp_mgr<uvector_const_subvector,linear_interp> dim2;

    /// Set interpolation managers
    int set_interp(base_interp_mgr<uvector_base> &bi1,
		   base_interp_mgr<uvector_const_subvector> &bi2) {
      bim1=&bi1;
      bim2=&bi2;
      return 0;
    }

    /** \brief Interpolate values \c vals into the tensor, 
	returning the result
      
	This is a quick and dirty implementation of n-dimensional
	interpolation by recursive application of the 1-dimensional
	routine from \ref smart_interp_vec, using the base
	interpolation object specified in the template parameter \c
	base_interp_t. This will be slow for sufficiently large data
	sets.

	\future Maybe make this a template as well?

	\future It should be straightforward to improve the scaling of
	this algorithm significantly by creating a "window" of local
	points around the point of interest. This could be done easily
	by constructing an initial subtensor. However, this should
	probably be superceded by a more generic alternative which
	avoids explicit use of the 1-d interpolation types.
    */
    double interpolate(double *vals) {
      typedef smart_interp_vec<uvector_base,uvector_const_subvector,
	uvector,uvector_alloc> interp_t;
      
      if (rk==1) {
	
	interp_t si(size[0],grid,data);
	return si.interp(vals[0]);

      } else {
	
	// Get total number of interpolations at this level
	size_t ss=1;
	for(size_t i=1;i<rk;i++) ss*=size[i];

	// Create space for y vectors and interpolators
	std::vector<uvector> yvec(ss);
	std::vector<interp_t *> si(ss);
	for(size_t i=0;i<ss;i++) {
	  yvec[i].allocate(size[0]);
	}
	
	// Create space for interpolation results
	tensor_grid tdat;
	uvector_size_t_subvector size_new(size,1,rk-1);
	tdat.allocate(rk-1,size_new);

	// Set grid for temporary tensor
	uvector_subvector grid_new(grid,size[0],grid.size()-size[0]);
	tdat.set_grid(grid_new);
	
	// Create starting coordinate and counter
	uvector_size_t co(rk);
	for(size_t i=0;i<rk;i++) co[i]=0;
	size_t cnt=0;

	// Loop over every interpolation
	bool done=false;
	while(done==false) {

	  // Fill yvector with the appropriate data
	  for(size_t i=0;i<size[0];i++) {
	    co[0]=i;
	    yvec[cnt][i]=get(co);
	  }
	  
	  si[cnt]=new interp_t(*bim1,*bim2,size[0],grid,yvec[cnt]);
	  
	  uvector_size_t_subvector co2(co,1,rk-1);
	  tdat.set(co2,si[cnt]->interp(vals[0]));

	  // Go to next interpolation
	  cnt++;
	  co[rk-1]++;
	  // carry if necessary
	  for(int j=rk-1;j>0;j--) {
	    if (co[j]>=size[j]) {
	      co[j]=0;
	      co[j-1]++;
	    }
	  }

	  // Test if done
	  if (cnt==ss) done=true;

	  // End of while loop
	}
      
	// Now call the next level of interpolation
	double res=tdat.interpolate(vals+1);
	
	tdat.free();
	for(size_t i=0;i<ss;i++) {
	  delete si[i];
	}

	return res;
      }
    }
    //@}

  };

  /** \brief Rank 1 tensor
   */
  class tensor1 : public tensor {
  public:
    
    /// Create an empty tensor
  tensor1() : tensor() {}

    /// Create a rank 1 tensory of size \c sz
  tensor1(size_t sz) : tensor() {
      uvector_size_t sizex(1);
      sizex[0]=sz;
      allocate(1,sizex);
    }
	
    /// Get the element indexed by \c ix
    double &get(size_t ix) { return tensor::get(&ix); }

    /// Get the element indexed by \c ix
    const double &get(size_t ix) const { return tensor::get(&ix); }
	
    /// Set the element indexed by \c index to value \c val
    void set(size_t index, double val) 
    { tensor::set(&index,val); }
    
    /** \brief Set the element indexed by \c index to value \c val

	(We have to explicitly provide this version since the set()
	function is overloaded in this child of \ref tensor.)
    */
    void set(size_t *index, double val) {
      tensor::set(index,val);
    }

    /// Get an element using array-like indexing
    double &operator[](size_t ix) { return this->data[ix]; }
    
    /// Get an element using operator()
    double &operator()(size_t ix) { return this->data[ix]; }
  };

  /** \brief Rank 1 tensor with a grid
   */
  class tensor_grid1 : public tensor_grid {
     
  public:
     
    /// Create an empty tensor
  tensor_grid1() : tensor_grid() {}
      
    /// Create a rank 2 tensor of size \c (sz,sz2,sz3)
  tensor_grid1(size_t sz) : tensor_grid() {
      this->rk=1;
      this->size.allocate(1);
      this->size[0]=sz;
      this->data.allocate(sz);
      this->grid_set=false;
    }
    
    ~tensor_grid1() {
    }
   
    /// Get the element indexed by \c (ix1)
    double &get(size_t ix1) { 
      size_t sz[1]={ix1};
      return tensor_grid::get(sz); 
    }

    /// Get the element indexed by \c (ix1)
    const double &get(size_t ix1) const { 
      size_t sz[1]={ix1};
      return tensor_grid::get(sz); 
    }
 
    /// Set the element indexed by \c (ix1) to value \c val
    void set(size_t ix1, double val) {
      size_t sz[1]={ix1};
      tensor_grid::set(sz,val); 
    }

    /// Interpolate \c x and return the results
    double interp(double x) {
      return interpolate(&x);
    }
  };

  /** \brief Rank 2 tensor
   */
  class tensor2 : public tensor {
  public:

    /// Create an empty tensor
  tensor2() : tensor() {}

    /// Create a rank 2 tensor of size \c (sz,sz2)
  tensor2(size_t sz, size_t sz2) : tensor() {
      this->rk=2;
      this->size.allocate(2);
      this->size[0]=sz;
      this->size[1]=sz2;
      size_t tot=sz*sz2;
      this->data.allocate(tot);
    }
	
    /// Get the element indexed by \c (ix1,ix2)
    double &get(size_t ix1, size_t ix2) { 
      size_t sz[2]={ix1,ix2};
      return tensor::get(sz); 
    }

    /// Get the element indexed by \c (ix1,ix2)
    const double &get(size_t ix1, size_t ix2) const { 
      size_t sz[2]={ix1,ix2};
      return tensor::get(sz); 
    }

    /// Set the element indexed by \c (ix1,ix2) to value \c val
    void set(size_t ix1, size_t ix2, double val) {
      size_t sz[2]={ix1,ix2};
      tensor::set(sz,val); 
      return;
    }

    /** \brief Set the element indexed by \c index to value \c val

	(We have to explicitly provide this version since the set()
	function is overloaded in this child of \ref tensor.)
    */
    void set(size_t *index, double val) {
      tensor::set(index,val);
      return;
    }

    /// Get the element indexed by \c (ix1,ix2)
    double &operator()(size_t ix, size_t iy) 
    { return this->data[ix*this->size[1]+iy]; }
  };
  
  /** \brief Rank 2 tensor with a grid
   */
  class tensor_grid2 : public tensor_grid {
     
  public:
     
    /// Create an empty tensor
  tensor_grid2() : tensor_grid() {}

    /// Create a rank 2 tensor of size \c (sz,sz2)
  tensor_grid2(size_t sz, size_t sz2) : tensor_grid() {
      this->rk=2;
      this->size.allocate(2);
      this->size[0]=sz;
      this->size[1]=sz2;
      size_t tot=sz*sz2;
      this->data.allocate(tot);
      this->grid_set=false;
    }
   
    ~tensor_grid2() {
    }
   
    /// Get the element indexed by \c (ix1,ix2)
    double &get(size_t ix1, size_t ix2) { 
      size_t sz[2]={ix1,ix2};
      return tensor_grid::get(sz); 
    }

    /// Get the element indexed by \c (ix1,ix2)
    const double &get(size_t ix1, size_t ix2) const { 
      size_t sz[2]={ix1,ix2};
      return tensor_grid::get(sz); 
    }
 
    /// Set the element indexed by \c (ix1,ix2) to value \c val
    void set(size_t ix1, size_t ix2, double val) {
      size_t sz[2]={ix1,ix2};
      tensor_grid::set(sz,val); 
      return;
    }

    /// Interpolate \c (x,y) and return the results
    double interp(double x, double y) {
      double arr[2]={x,y};
      return interpolate(arr);
    }
  };
  
  /** \brief Rank 3 tensor
   */
  class tensor3 : public tensor {
  public:

    /// Create an empty tensor
  tensor3() : tensor() {}

    /// Create a rank 3 tensor of size \c (sz,sz2,sz3)
  tensor3(size_t sz, size_t sz2, size_t sz3) : tensor() {
      this->rk=3;
      this->size.allocate(3);
      this->size[0]=sz;
      this->size[1]=sz2;
      this->size[2]=sz3;
      size_t tot=sz*sz2*sz3;
      this->data.allocate(tot);
    }
	
    /// Get the element indexed by \c (ix1,ix2,ix3)
    double &get(size_t ix1, size_t ix2, size_t ix3) { 
      size_t sz[3]={ix1,ix2,ix3};
      return tensor::get(sz); 
    }

    /// Get the element indexed by \c (ix1,ix2,ix3)
    const double &get(size_t ix1, size_t ix2, size_t ix3) const { 
      size_t sz[3]={ix1,ix2,ix3};
      return tensor::get(sz); 
    }

    /// Set the element indexed by \c (ix1,ix2,ix3) to value \c val
    void set(size_t ix1, size_t ix2, size_t ix3, double val) {
      size_t sz[3]={ix1,ix2, ix3};
      tensor::set(sz,val); 
      return;
    }

    /** \brief Set the element indexed by \c index to value \c val

	(We have to explicitly provide this version since the set()
	function is overloaded in this child of \ref tensor.)
    */
    void set(size_t *index, double val) {
      tensor::set(index,val);
      return;
    }
  };
  
  /** \brief Rank 3 tensor with a grid
   */
  class tensor_grid3 : public tensor_grid {
     
  public:
     
    /// Create an empty tensor
  tensor_grid3() : tensor_grid () {}

    /// Create a rank 3 tensor of size \c (sz,sz2,sz3)
  tensor_grid3(size_t sz, size_t sz2, size_t sz3) : tensor_grid () {
      this->rk=3;
      this->size.allocate(3);
      this->size[0]=sz;
      this->size[1]=sz2;
      this->size[2]=sz3;
      size_t tot=sz*sz2*sz3;
      this->data.allocate(tot);
      this->grid_set=false;
    }
   
    ~tensor_grid3() {
    }
   
    /// Get the element indexed by \c (ix1,ix2,ix3)
    double &get(size_t ix1, size_t ix2, size_t ix3) { 
      size_t sz[3]={ix1,ix2,ix3};
      return tensor_grid::get(sz); 
    }
 
    /// Get the element indexed by \c (ix1,ix2,ix3)
    const double &get(size_t ix1, size_t ix2, size_t ix3) const { 
      size_t sz[3]={ix1,ix2,ix3};
      return tensor_grid::get(sz); 
    }
 
    /// Set the element indexed by \c (ix1,ix2,ix3) to value \c val
    void set(size_t ix1, size_t ix2, size_t ix3, double val) {
      size_t sz[3]={ix1,ix2, ix3};
      tensor_grid::set(sz,val); 
      return;
    }

    /// Interpolate \c (x,y,z) and return the results
    double interp(double x, double y, double z) {
      double arr[3]={x,y,z};
      return interpolate(arr);
    }
  };
  
  /** \brief Rank 4 tensor
   */
  class tensor4 : public tensor {

  public:

    /// Create an empty tensor
  tensor4() : tensor() {}

    /// Create a rank 4 tensor of size \c (sz,sz2,sz3,sz4)
  tensor4(size_t sz, size_t sz2, size_t sz3, size_t sz4) : 
    tensor() {
      this->rk=4;
      this->size.allocate(4);
      this->size[0]=sz;
      this->size[1]=sz2;
      this->size[2]=sz3;
      this->size[3]=sz4;
      size_t tot=sz*sz2*sz3*sz4;
      this->data.allocate(tot);
    }
	
    /// Get the element indexed by \c (ix1,ix2,ix3,ix4)
    double &get(size_t ix1, size_t ix2, size_t ix3, size_t ix4) { 
      size_t sz[4]={ix1,ix2,ix3,ix4};
      return tensor::get(sz); 
    }

    /// Get the element indexed by \c (ix1,ix2,ix3,ix4)
    const double &get(size_t ix1, size_t ix2, size_t ix3, 
		      size_t ix4) const { 
      size_t sz[4]={ix1,ix2,ix3,ix4};
      return tensor::get(sz); 
    }

    /// Set the element indexed by \c (ix1,ix2,ix3,ix4) to value \c val
    void set(size_t ix1, size_t ix2, size_t ix3, size_t ix4, 
	    double val) {
      size_t sz[4]={ix1,ix2,ix3,ix4};
      tensor::set(sz,val); 
      return;
    }

    /** \brief Set the element indexed by \c index to value \c val

	(We have to explicitly provide this version since the set()
	function is overloaded in this child of \ref tensor.)
    */
    void set(size_t *index, double val) {
      tensor::set(index,val);
      return;
    }
  };

  /** \brief Rank 4 tensor with a grid
   */
  class tensor_grid4 : public tensor_grid {
     
  public:
     
    /// Create an empty tensor
  tensor_grid4() : tensor_grid () {}

    /// Create a rank 4 tensor of size \c (sz,sz2,sz3,sz4)
  tensor_grid4(size_t sz, size_t sz2, size_t sz3,
	       size_t sz4) : tensor_grid () {
      this->rk=4;
      this->size.allocate(4);
      this->size[0]=sz;
      this->size[1]=sz2;
      this->size[2]=sz3;
      this->size[3]=sz4;
      size_t tot=sz*sz2*sz3*sz4;
      this->data.allocate(tot);
      this->grid_set=false;
    }
   
    ~tensor_grid4() {
    }
   
    /// Get the element indexed by \c (ix1,ix2,ix3,ix4)
    double &get(size_t ix1, size_t ix2, size_t ix3, size_t ix4) { 
      size_t sz[4]={ix1,ix2,ix3,ix4};
      return tensor_grid::get(sz); 
    }
    
    /// Get the element indexed by \c (ix1,ix2,ix3,ix4)
    const double &get(size_t ix1, size_t ix2, size_t ix3,
		      size_t ix4) const { 
      size_t sz[4]={ix1,ix2,ix3,ix4};
      return tensor_grid::get(sz); 
    }
    
    /// Set the element indexed by \c (ix1,ix2,ix3,ix4) to value \c val
    void set(size_t ix1, size_t ix2, size_t ix3, size_t ix4,
	    double val) {
      size_t sz[4]={ix1,ix2,ix3,ix4};
      tensor_grid::set(sz,val); 
      return;
    }

    /// Interpolate \c (x,y,z,a) and return the results
    double interp(double x, double y, double z, double a) {
      double arr[4]={x,y,z,a};
      return interpolate(arr);
    }
  };
  
#ifndef DOXYGENP
}
#endif

#endif



