/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_TENSOR_OLD_H
#define O2SCL_TENSOR_OLD_H

/** \file tensor_old.h
    \brief File for definitions of tensor_olds
*/

#include <iostream>
#include <cstdlib>
#include <string>
#include <fstream>
#include <sstream>

#include <gsl/gsl_matrix.h>
#include <gsl/gsl_ieee_utils.h>

#include <o2scl/err_hnd.h>
#include <o2scl/uvector_tlate.h>
#include <o2scl/umatrix_tlate.h>
#include <o2scl/smart_interp.h>

#ifndef DOXYGENP
namespace o2scl {
#endif
  
  /** \brief Tensor_Old class with arbitrary dimensions

      The elements of a tensor_old are typically specified as a list of
      <tt>size_t</tt> numbers with length equal to the tensor_old rank.
      For a rank-4 tensor_old named \c t, the element 
      <tt>t[1][2][0][3]</tt> can be obtained with something similar to 
      \code
      size_t ix[4]={1,2,0,3};
      double x=t.get(ix);
      \endcode

      Empty tensor_olds have zero rank.

      Slices of tensor_olds are subsets obtained from fixing the index of
      several dimensions, while letting other dimensions vary. For a
      slice with one dimension not fixed, see \ref vector_slice(). For
      a slice with two dimensions not fixed, see \ref matrix_slice().

      For I/O with tensor_olds, see \ref hdf_io.h .

      \future Could implement arithmetic operators + and - and some
      different products. 

      \future Implement copies to and from vector
      and matrices 

      \future Implement tensor_old contractions, i.e. tensor_old
      = tensor_old * tensor_old 

      \future Consider making a template type to replace double,
      e.g. <tt>value_t</tt>.

      \future Could be interesting to write an iterator for this class.

      \comment
      A tensor_old is defined to be empty if it's rank is zero. A tensor_old's
      rank should be zero if and only if no memory has been allocated
      for it.

      I could have chosen to make set() functions virtual functions
      instead of templates. However, many of these functions may be
      templates anyway when I generalize this class to hold floats or
      ints rather than just doubles. 
      \endcomment
  */
  class tensor_old {

#ifndef DOXYGEN_INTERNAL

  protected:

    /// The data
    double *data;

    /// A rank-sized array of the sizes of each dimension
    size_t *size;

    /// Rank
    size_t rk;

#endif

  public:
    
    /// Create an empty tensor_old with zero rank
    tensor_old() {
      rk=0;
      data=0;
      size=0;
    }

    /** \brief Create a tensor_old of rank \c rank with sizes given in \c dim
	
	The parameter \c dim must be a pointer to an array of sizes
	with length \c rank. If the user requests any of the sizes to
	be zero, this constructor will call the error handler, create
	an empty tensor_old, and will allocate no memory.
    */
    template<class uint_vec_t> 
      tensor_old(size_t rank, const uint_vec_t &dim) {
      if (rank==0) {
	data=0;
	size=0;
      } else {
	for(size_t i=0;i<rk;i++) {
	  if (dim[i]==0) {	  
	    rk=0;
	    O2SCL_ERR2("Requested zero size with non-zero rank in ",
		       "tensor_old::tensor_old<>(size_t, uint_vec_t &)",gsl_einval);
	  }
	}
	rk=rank;
	size=new size_t[rk];
	size_t tot=1;
	for(size_t i=0;i<rk;i++) {
	  size[i]=dim[i];
	  tot*=size[i];
	}
	data=new double[tot];
      }
    }

    ~tensor_old() {
      if (rk>0) {
	delete[] size;
	delete[] data;
	rk=0;
      }
    }

    /// \name Set functions
    //@{
    /// Set the element indexed by \c index to value \c val
    template<class uint_vec_t>
      void set(const uint_vec_t &index, double val) {
#if O2SCL_NO_RANGE_CHECK
#else
      if (rk==0) {
	O2SCL_ERR("Empty tensor_old in set().",gsl_einval);
	return;
      }
      if (index[0]>=size[0]) {
	O2SCL_ERR("Index greater than size in set().",gsl_eindex);
	return;
      }
#endif
      size_t ix=index[0];
      for(size_t i=1;i<rk;i++) {
#if O2SCL_NO_RANGE_CHECK
#else
	if (index[i]>=size[i]) {
	  O2SCL_ERR("Index greater than size in get().",gsl_eindex);
	  return;
	}
#endif
	ix*=size[i];
	ix+=index[i];
      }
      data[ix]=val;
      return;
    }

    /// Set all elements in a tensor_old to some fixed value
    void set_all(double x) {
      for(size_t i=0;i<total_size();i++) data[i]=x;
      return;
    }
    //@}

    /// \name Get functions
    //@{
    /// Get the element indexed by \c index
    template<class uint_vec_t> double &get(const uint_vec_t &index) {
#if O2SCL_NO_RANGE_CHECK
#else
      if (rk==0) {
	O2SCL_ERR("Empty tensor_old in get().",gsl_einval);
	return data[0];
      }
      if (index[0]>=size[0]) {
	O2SCL_ERR("Index greater than size in get().",gsl_eindex);
	return data[0];
      }
#endif
      size_t ix=index[0];
      for(size_t i=1;i<rk;i++) {
#if O2SCL_NO_RANGE_CHECK
#else
	if (index[i]>=size[i]) {
	  O2SCL_ERR("Index greater than size in get().",gsl_eindex);
	  return data[0];
	}
#endif
	ix*=size[i];
	ix+=index[i];
      }
      return data[ix];
    }

    /// Get a const reference to the element indexed by \c index
    template<class uint_vec_t> 
      double const &get(const uint_vec_t &index) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (rk==0) {
	O2SCL_ERR("Empty tensor_old in get().",gsl_einval);
	return data[0];
      }
      if (index[0]>=size[0]) {
	O2SCL_ERR("Index greater than size in get().",gsl_eindex);
	return data[0];
      }
#endif
      size_t ix=index[0];
      for(size_t i=1;i<rk;i++) {
#if O2SCL_NO_RANGE_CHECK
#else
	if (index[i]>=size[i]) {
	  O2SCL_ERR("Index greater than size in get().",gsl_eindex);
	  return data[0];
	}
#endif
	ix*=size[i];
	ix+=index[i];
      }
      return data[ix];
    }
    //@}
    
    /// \name Slice functions
    //@{
    /** \brief Fix all but one index to create a vector

	This fixes all of the indices to the values given in \c index
	except for the index number \c ix, and returns the
	corresponding vector, whose length is equal to the size of the
	tensor_old in that index. The value <tt>index[ix]</tt> is ignored.

	For example, for a rank 3 tensor_old allocated with
	\code
	tensor_old t;
	size_t dim[3]={3,4,5};
	t.tensor_old_allocate(3,dim);
	\endcode
	the following code
	\code
	size_t index[3]={1,0,3};
	ovector_view v=t.vector_slice(1,index);
	\endcode
	Gives a vector \c v of length 4 which refers to the values
	<tt>t(1,0,3)</tt>, <tt>t(1,1,3)</tt>, <tt>t(1,2,3)</tt>, and
	<tt>t(1,3,3)</tt>.
    */
    template<class uint_vec_t> 
      ovector_array_stride vector_slice(size_t ix, const uint_vec_t &index) {
      if (ix+1>rk) {
	O2SCL_ERR("Specified index to large in tensor_old::vector_slice()",
		  gsl_einval);
      }
      size_t start;
      if (ix==0) start=0;
      else start=index[0];
      for(size_t i=1;i<rk;i++) {
	start*=size[i];
	if (i!=ix) start+=index[i];
      }
      size_t stride=1;
      for(size_t i=ix+1;i<rk;i++) stride*=size[i];
      return ovector_array_stride(size[ix],data+start,stride);
    }

    /** \brief Fix all but two indices to create a matrix

	This fixes all of the indices to the values given in \c index
	except for the index number \c ix and the last index, and
	returns the corresponding matrix, whose size is equal to the
	size of the tensor_old in the two indices which are not fixed.

	Since the last index is always fixed, the parameter \c ix
	is limited to be \f$ \in [0,\mathrm{rank}-2] \f$.
    */
    template<class uint_vec_t> 
      omatrix_array matrix_slice(size_t ix, const uint_vec_t &index) {
      if (ix+1>=rk) {
	O2SCL_ERR("Specified index to large in tensor_old::matrix_slice()",
		  gsl_einval);
      }
      size_t start=0, tot=size[0];
      for(size_t i=1;i<rk;i++) {
	tot*=size[i];
      }
      // This O(rank^2) algorithm could possibly 
      // be replaced with a O(rank) one in the future
      for(size_t i=0;i<rk-1;i++) {
	if (i!=ix) {
	  size_t adj=index[i];
	  for(size_t j=i+1;j<rk;j++) adj*=size[j];
	  start+=adj;
	}
      }
      size_t tda=1;
      if (ix==rk-1) {
	tda=0;
      } else {
	for(size_t i=ix+1;i<rk;i++) tda*=size[i];
      }
      return omatrix_array(data+start,tda,size[ix],size[rk-1]);
    }
    //@}

    /// \name Memory allocation
    //@{
    /** \brief Allocate space for a tensor_old of rank \c rank with sizes
	given in \c dim
	
	The parameter \c dim must be a pointer to an array of sizes
	with length \c rank. 

	If memory was previously allocated, it will be freed before
	the new allocation.

	If the user requests any of the sizes to be zero, this
	function will call the error handler and will allocate no
	memory. If memory was previously allocated, the tensor_old is left
	unmodified and no deallocation is performed.
    */
    template<class uint_vec_t> 
      void tensor_old_allocate(size_t rank, const uint_vec_t &dim) {
      for(size_t i=0;i<rk;i++) {
	if (dim[i]==0) {
	  O2SCL_ERR2("Requested zero size in tensor_old::",
		     "tensor_old(size_t, size_t *)",gsl_einval);
	}
      }
      if (rk>0) tensor_old_free();
      rk=rank;
      size=new size_t[rk];
      size_t tot=1;
      for(size_t i=0;i<rk;i++) {
	size[i]=dim[i];
	tot*=size[i];
      }
      data=new double[tot];
      return;
    }

    /// Free allocated space (also sets rank to zero)
    void tensor_old_free() {
      if (rk>0) {
	delete[] size;
	delete[] data;
	rk=0;
      }
      return;
    }
    //@}

    /// \name Size functions
    //@{
    /// Return the rank of the tensor_old
    size_t get_rank() const { return rk; }

    /// Returns the size of the ith index
    size_t get_size(size_t i) const { 
      if (i<rk) {
	return size[i]; 
      }
      O2SCL_ERR_RET("Requested index exceeding rank in get_size()",
		    gsl_einval);
    }
    
    /// Return the full array of sizes
    const size_t *get_size_arr() const {
      return size;
    }

    /// Return the full array of sizes
    const double *get_data() const {
      return data;
    }

#ifndef DOXYGENP
    // For HDF I/O. Undocumented. Probably should be replaced by
    // friend function or something else?
    double *get_data_non_const() {
      return data;
    }
#endif

    /** \brief Returns the size of the tensor_old (the product of 
	the sizes over every index)
    */
    size_t total_size() const { 
      if (rk==0) return 0;
      size_t tot=1;
      for(size_t i=0;i<rk;i++) tot*=size[i];
      return tot;
    }
    //@}

    /// \name Index manipulation
    //@{
    /// Pack the indices into a single array index
    template<class uint_vec_t> 
      size_t pack_indices(const uint_vec_t &index) {
      if (rk==0) {
	O2SCL_ERR("Empty tensor_old in pack_indices().",gsl_einval);
	return 0;
      }
      if (index[0]>=size[0]) {
	O2SCL_ERR("Index greater than size in pack_indices().",gsl_eindex);
	return 0;
      }
      size_t ix=index[0];
      for(size_t i=1;i<rk;i++) {
	if (index[i]>=size[i]) {
	  O2SCL_ERR("Index greater than size in pack_indices().",gsl_eindex);
	  return 0;
	}
	ix*=size[i];
	ix+=index[i];
      }
      return ix;
    }
    
    /// Unpack the single array index into indices
    template<class uint_vec_t> 
      void unpack_indices(size_t ix, uint_vec_t &index) {
      if (ix>total_size()) {
	O2SCL_ERR("Index greater than total size in unpack_indices().",
		  gsl_eindex);
	return;
      }
      size_t ix2, sub_size;
      for(size_t i=0;i<rk;i++) {
	if (i==rk-1) {
	  index[i]=ix;
	} else {
	  sub_size=1;
	  for(size_t j=i+1;j<rk;j++) sub_size*=size[j];
	  index[i]=ix/sub_size;
	  // (Remember we're doing integer arithmetic here.)
	  ix-=sub_size*(ix/sub_size);
	}
      }
      return;
    }
    //@}

  };

  /** \brief Tensor_Old class with arbitrary dimensions with a grid

      This tensor_old class allows one to assign the indexes to 
      numerical scales, so that n-dimensional interpolation can
      be performed. To set the grid, use set_grid() and then
      interpolation can be done using interpolate().
      
      By convention, member functions ending in the _val 
      suffix return the closest grid-point to some user-specified
      values. 

      \comment
      I like how hist_new only allows you to set the
      grid if it's the same size as the old grid or the tensor_old
      is empty. This same practice could be applied here, to
      force the user to clear the tensor_old before resetting the grid.
      (10/24/11 - Actually, maybe this is a bad idea, because
      this class is more analogous to ovector, which doesn't
      behave this way.)
      \endcomment

      \future Only allocate space for grid if it is set
      \future Could implement arithmetic operators + and - and some
      different products.
      \future Consider creating a set_grid() function which
      takes grids from an object like hist_grid. Maybe make a 
      constructor for a tensor_old_grid object which just takes 
      as input a set of grids?
  */
  class tensor_old_grid : public tensor_old {

#ifndef DOXYGEN_INTERNAL

  protected:

    /// A rank-sized set of arrays for the grid points
    double **grd;

    /// If true, the grid has been set by the user
    bool grid_set;
    
    /// The interpolation manager
    base_interp_mgr<double *> *bim1;

    /// The subvector interpolation manager
    base_interp_mgr<array_const_subvector> *bim2;

#endif

  public:
    
    /// Create an empty tensor_old with zero rank
  tensor_old_grid() : tensor_old() {
      grd=0;
      grid_set=false;
      bim1=&dim1;
      bim2=&dim2;
    }
    
    /** \brief Create a tensor_old of rank \c rank with sizes given in \c dim
	
	The parameter \c dim must be a pointer to an array of sizes
	with length \c rank. If the user requests any of the sizes to
	be zero, this constructor will call the error handler, create
	an empty tensor_old, and will allocate no memory.
    */
    template<class uint_vec_t> 
      tensor_old_grid(size_t rank, const uint_vec_t &dim) : tensor_old(rank,dim) {
      grid_set=false;
      for(size_t i=0;i<rk;i++) {
	if (dim[i]==0) {
	  rk=0;
	  O2SCL_ERR("Requested zero size in tensor_old::tensor_old(size_t, size_t *)",
		    gsl_einval);
	}
      }
      grd=new double *[rk];
      for(size_t i=0;i<rk;i++) {
	grd[i]=new double[size[i]];
      }
    }

    ~tensor_old_grid() {
      if (rk>0) {
	for(size_t i=0;i<rk;i++) delete[] grd[i];
	delete[] size;
	delete[] data;
	rk=0;
      }
    }

    /// \name Set functions
    //@{
    /// Set the element closest to grid point \c grdp to value \c val
    template<class vec_t> 
      int set_val(const vec_t &grdp, double val) {

      // Find indices
      size_t *index=new size_t[rk];
      for(size_t i=0;i<rk;i++) index[i]=lookup_grid(i,grdp[i]);
      
      // Pack
      size_t ix=index[0];
      for(size_t i=1;i<rk;i++) {
	ix*=size[i];
	ix+=index[i];
      }

      // Delete memory for indices
      delete[] index;

      // Set value
      data[ix]=val;

      return 0;
    }

    /** \brief Set the element closest to grid point \c grdp to value 
	\c val
	
	The parameters \c closest and \c grdp may be identical,
	allowing one to update a vector \c grdp with the
	closest grid point.
    */
    template<class vec_t, class vec2_t> 
      int set_val(const vec_t &grdp, double val, vec2_t &closest) {

      // Find indices
      size_t *index=new size_t[rk];
      for(size_t i=0;i<rk;i++) {
	index[i]=lookup_grid_val(i,grdp[i],closest[i]);
      }
      
      // Pack
      size_t ix=index[0];
      for(size_t i=1;i<rk;i++) {
	ix*=size[i];
	ix+=index[i];
      }

      // Delete memory for indices
      delete[] index;

      // Set value
      data[ix]=val;

      return 0;
    }
    //@}

    /// \name Get functions
    //@{
    /// Get the element closest to grid point \c grdp 
    template<class vec_t> 
      double get_val(const vec_t &grdp) {
      
      // Find indices
      size_t *index=new size_t[rk];
      for(size_t i=0;i<rk;i++) index[i]=lookup_grid(i,grdp[i]);

      // Pack
      size_t ix=index[0];
      for(size_t i=1;i<rk;i++) {
	ix*=size[i];
	ix+=index[i];
      }

      // Delete memory for indices
      delete[] index;

      // Set value
      return data[ix];
    }

    /// Get the element closest to grid point \c grdp to value \c val
    template<class vec_t, class vec2_t> 
      double get_val(const vec_t &grdp, const vec2_t &closest) {
      
      // Find indices
      size_t *index=new size_t[rk];
      for(size_t i=0;i<rk;i++) {
	index[i]=lookup_grid_val(i,grdp[i],closest[i]);
      }
      
      // Pack
      size_t ix=index[0];
      for(size_t i=1;i<rk;i++) {
	ix*=size[i];
	ix+=index[i];
      }

      // Delete memory for indices
      delete[] index;

      // Set value
      return data[ix];
    }
    //@}
    
    /// \name Memory allocation
    //@{
    /** \brief Allocate space for a tensor_old of rank \c rank with sizes
	given in \c dim
	
	The parameter \c dim must be a pointer to an array of sizes
	with length \c rank. 

	If memory was previously allocated, it will be freed before
	the new allocation and previously specified grid data will 
	be lost.

	If the user requests any of the sizes to be zero, this
	function will call the error handler and will allocate no
	memory. If memory was previously allocated, the tensor_old is left
	unmodified and no deallocation is performed.
    */
    template<class uint_vec_t>
      int tensor_old_allocate(size_t rank, const uint_vec_t &dim) {
      for(size_t i=0;i<rk;i++) {
	if (dim[i]==0) {
	  O2SCL_ERR2_RET("Requested zero size in ",
			 "tensor_old::tensor_old(size_t,uint_vec_t &)",gsl_einval);
	}
      }
      if (rk>0) tensor_old_free();
      rk=rank;
      size=new size_t[rk];
      grd=new double *[rk];
      size_t tot=1;
      for(size_t i=0;i<rk;i++) {
	size[i]=dim[i];
	grd[i]=new double[size[i]];
	tot*=size[i];
      }
      data=new double[tot];
      return 0;
    }

    /// Free allocated space (also sets rank to zero)
    int tensor_old_free() {
      if (rk>0) {
	for(size_t i=0;i<rk;i++) delete[] grd[i];
	delete[] size;
	delete[] grd;
	delete[] data;
	rk=0;
      }
      return 0;
    }
    //@}

    /// \name Grid manipulation
    //@{
    /// Return true if the grid has been set
    bool is_grid_set() const { return grid_set; }

    /** \brief Set the grid

	The grid must be specified for all of the dimensions at
	once. Denote \f$ (\mathrm{size})_0 \f$ as the size
	of the first dimension, \f$ (\mathrm{size})_1 \f$ as the
	size of the second dimesion, and so on. Then the 
	first \f$ (\mathrm{size})_0 \f$ entries in \c grid must
	be the grid for the first dimension, the next 
	\f$ (\mathrm{size})_1 \f$ entries must be the grid
	for the second dimension, and so on. Thus \c grid
	must be a vector of size
	\f[
	\sum_{i=0}^{\mathrm{rank}} (\mathrm{size})_i
	\f]

	Note that the grid is copied so the function argument may
	be destroyed by the user after calling set_grid() without
	affecting the tensor_old grid. 

	\future Define a more generic interface for matrix types
    */
    template<class vec_t>
      int set_grid(const vec_t &grid) {
      if (rk==0) {
	O2SCL_ERR2_RET("Tried to set grid for empty tensor_old in ",
		       "tensor_old_grid::set_grid().",gsl_einval);
      }
      size_t ix=0;
      for(size_t i=0;i<rk;i++) {
	for(size_t j=0;j<size[i];j++) {
	  grd[i][j]=grid[ix];
	  ix++;
	}
      }
      grid_set=true;
      return 0;
    }

#ifndef DOXGYENP
    // Temporary(?). Undocumented for now. 
    int set_grid_old(double **vals) {
      for(size_t i=0;i<rk;i++) {
	for(size_t j=0;j<size[i];j++) {
	  grd[i][j]=vals[i][j];
	}
      }
      grid_set=true;
      return 0;
    }
#endif

    /// Lookup index for grid closest to \c val
    size_t lookup_grid(size_t i, double val) {
      if (i<rk && grid_set) {
	size_t best=0;
	double min=fabs(grd[i][0]-val);
	for(size_t j=0;j<size[i];j++) {
	  if (fabs(grd[i][j]-val)<min) {
	    best=j;
	    min=fabs(grd[i][j]-val);
	  }
	}
	return best;
      }
      return 0;
    }

    /// Lookup index for grid closest to \c val
    double get_grid(size_t i, size_t j) const {
      if (i<rk && grid_set) {
	return grd[i][j];
      }
      return 0;
    }

    /** \brief Lookup indices for grid closest point to \c vals

        The values in \c vals are not modified by this function.
	
	\comment
	This function must have a different name than 
	lookup_grid() because the template types cause
	confusion between the two functions.
	\endcomment
    */
    template<class vec_t, class uint_vec_t>
      int lookup_grid_vec(const vec_t &vals, 
			  const uint_vec_t &indices) const {
      for(size_t k=0;k<rk;k++) {
	indices[k]=0;
	double min=fabs(grd[k][0]-vals[k]);
	for(size_t j=0;j<size[k];j++) {
	  if (fabs(grd[k][j]-vals[k])<min) {
	    indices[k]=j;
	    min=fabs(grd[k][j]-vals[k]);
	  }
	}
      }
      return 0;
    }

    /// Lookup index for grid closest to \c val, returning the grid point
    size_t lookup_grid_val(size_t i, double &val, double &val2) {
      if (i<rk && grid_set) {
	size_t best=0;
	double min=fabs(grd[i][0]-val);
	val2=grd[i][0];
	for(size_t j=0;j<size[i];j++) {
	  if (fabs(grd[i][j]-val)<min) {
	    best=j;
	    min=fabs(grd[i][j]-val);
	    val2=grd[i][j];
	  }
	}
	return best;
      }
      return 0;
    }
  
    /// \name Interpolation
    //@{
    def_interp_mgr<double *,linear_interp> dim1;
    def_interp_mgr<array_const_subvector,linear_interp> dim2;

    /// Set interpolation managers
    int set_interp(base_interp_mgr<double *> &bi1,
		   base_interp_mgr<array_const_subvector> &bi2) {
      bim1=&bi1;
      bim2=&bi2;
      return 0;
    }

    /** \brief Interpolate values \c vals into the tensor_old, 
	returning the result
      
	This is a quick and dirty implementation of n-dimensional
	interpolation by recursive application of the 1-dimensional
	routine from \ref smart_interp_vec, using the base
	interpolation object specified in the template parameter \c
	base_interp_t. This will be slow for sufficiently large data
	sets.

	\future Maybe make this a template as well?

	\future It should be straightforward to improve the scaling of
	this algorithm significantly by creating a "window" of local
	points around the point of interest. This could be done easily
	by constructing an initial subtensor_old. However, this should
	probably be superceded by a more generic alternative which
	avoids explicit use of the 1-d interpolation types.
    */
    double interpolate(double *vals) {
      typedef smart_interp_vec<double *,array_const_subvector,
	double *, pointer_alloc<double> > interp_t;
      
      if (rk==1) {
	
	sma_interp_vec<double *> si(size[0],grd[0],data);
	return si.interp(vals[0]);

      } else {
	
	// Get total number of interpolations at this level
	size_t ss=1;
	for(size_t i=1;i<rk;i++) ss*=size[i];

	// Create space for y vectors and interpolators
	double **yvec=new double *[ss];
	interp_t **si=new interp_t *[ss];
	for(size_t i=0;i<ss;i++) yvec[i]=new double[size[0]];
	
	// Create space for interpolation results
	tensor_old_grid tdat;
	tdat.tensor_old_allocate(rk-1,size+1);

	// Set grid for temporary tensor_old
	tdat.set_grid_old(grd+1);

	// Create starting coordinate and counter
	size_t *co=new size_t[rk];
	for(size_t i=0;i<rk;i++) co[i]=0;
	size_t cnt=0;

	// Loop over every interpolation
	bool done=false;
	while(done==false) {

	  // Fill yvector with the appropriate data
	  for(size_t i=0;i<size[0];i++) {
	    co[0]=i;
	    yvec[cnt][i]=get(co);
	  }
	
	  si[cnt]=new interp_t(*bim1,*bim2,size[0],grd[0],yvec[cnt]);
	  
	  tdat.set(co+1,si[cnt]->interp(vals[0]));

	  // Go to next interpolation
	  cnt++;
	  co[rk-1]++;
	  // carry if necessary
	  for(int j=rk-1;j>0;j--) {
	    if (co[j]>=size[j]) {
	      co[j]=0;
	      co[j-1]++;
	    }
	  }

	  // Test if done
	  if (cnt==ss) done=true;

	  // End of while loop
	}
      
	// Now call the next level of interpolation
	double res=tdat.interpolate(vals+1);
	
	tdat.tensor_old_free();
	for(size_t i=0;i<ss;i++) {
	  delete[] yvec[i];
	  delete si[i];
	}
	delete[] co;
	delete[] si;
	delete[] yvec;

	return res;
      }
    }
    //@}

  };

  /** \brief Rank 1 tensor_old
   */
  class tensor_old1 : public tensor_old {
  public:
    
    /// Create an empty tensor_old
  tensor_old1() : tensor_old() {}

    /// Create a rank 1 tensor_oldy of size \c sz
  tensor_old1(size_t sz) : tensor_old(1,&sz) {}
	
    /// Get the element indexed by \c ix
    double &get(size_t ix) { return tensor_old::get(&ix); }

    /// Get the element indexed by \c ix
    const double &get(size_t ix) const { return tensor_old::get(&ix); }
	
    /// Set the element indexed by \c index to value \c val
    void set(size_t index, double val) 
    { tensor_old::set(&index,val); }
    
    /** \brief Set the element indexed by \c index to value \c val

	(We have to explicitly provide this version since the set()
	function is overloaded in this child of \ref tensor_old.)
    */
    void set(size_t *index, double val) {
      tensor_old::set(index,val);
    }

    /// Get an element using array-like indexing
    double &operator[](size_t ix) { return this->data[ix]; }
    
    /// Get an element using operator()
    double &operator()(size_t ix) { return this->data[ix]; }
  };

  /** \brief Rank 2 tensor_old with a grid
   */
  class tensor_old_grid1 : public tensor_old_grid {
     
  public:
     
    /// Create an empty tensor_old
  tensor_old_grid1() : tensor_old_grid() {}
      
    /// Create a rank 2 tensor_old of size \c (sz,sz2,sz3)
  tensor_old_grid1(size_t sz) : tensor_old_grid() {
      this->rk=1;
      this->size=new size_t[1];
      this->grd=new double *[1];
      this->size[0]=sz;
      this->grd[0]=new double[sz];
      size_t tot=sz;
      this->data=new double[tot];
      this->grid_set=false;
    }
   
    ~tensor_old_grid1() {
      if (this->rk>0) {
	for(size_t i=0;i<this->rk;i++) {
	  delete[] this->grd[i];
	}
	delete[] this->size;
	delete[] this->grd;
	delete[] this->data;
	this->rk=0;
      }
    }
   
    /// Get the element indexed by \c (ix1)
    double &get(size_t ix1) { 
      size_t sz[1]={ix1};
      return tensor_old_grid ::get(sz); 
    }

    /// Get the element indexed by \c (ix1)
    const double &get(size_t ix1) const { 
      size_t sz[1]={ix1};
      return tensor_old_grid ::get(sz); 
    }
 
    /// Set the element indexed by \c (ix1) to value \c val
    void set(size_t ix1, double val) {
      size_t sz[1]={ix1};
      tensor_old_grid ::set(sz,val); 
    }

    /// Interpolate \c x and return the results
    double interp(double x) {
      return interpolate(&x);
    }
  };

  /** \brief Rank 2 tensor_old
   */
  class tensor_old2 : public tensor_old {
  public:

    /// Create an empty tensor_old
  tensor_old2() : tensor_old() {}

    /// Create a rank 2 tensor_old of size \c (sz,sz2)
  tensor_old2(size_t sz, size_t sz2) : tensor_old() {
      this->rk=2;
      this->size=new size_t[2];
      this->size[0]=sz;
      this->size[1]=sz2;
      size_t tot=sz*sz2;
      this->data=new double[tot];
    }
	
    /// Get the element indexed by \c (ix1,ix2)
    double &get(size_t ix1, size_t ix2) { 
      size_t sz[2]={ix1,ix2};
      return tensor_old::get(sz); 
    }

    /// Get the element indexed by \c (ix1,ix2)
    const double &get(size_t ix1, size_t ix2) const { 
      size_t sz[2]={ix1,ix2};
      return tensor_old::get(sz); 
    }

    /// Set the element indexed by \c (ix1,ix2) to value \c val
    void set(size_t ix1, size_t ix2, double val) {
      size_t sz[2]={ix1,ix2};
      tensor_old::set(sz,val); 
      return;
    }

    /** \brief Set the element indexed by \c index to value \c val

	(We have to explicitly provide this version since the set()
	function is overloaded in this child of \ref tensor_old.)
    */
    void set(size_t *index, double val) {
      tensor_old::set(index,val);
      return;
    }

    /// Get the element indexed by \c (ix1,ix2)
    double &operator()(size_t ix, size_t iy) 
    { return this->data[ix*this->size[1]+iy]; }
  };
  
  /** \brief Rank 2 tensor_old with a grid
   */
  class tensor_old_grid2 : public tensor_old_grid {
     
  public:
     
    /// Create an empty tensor_old
  tensor_old_grid2() : tensor_old_grid() {}

    /// Create a rank 2 tensor_old of size \c (sz,sz2)
  tensor_old_grid2(size_t sz, size_t sz2) : tensor_old_grid() {
      this->rk=2;
      this->size=new size_t[2];
      this->grd=new double *[2];
      this->size[0]=sz;
      this->size[1]=sz2;
      this->grd[0]=new double[sz];
      this->grd[1]=new double[sz2];
      size_t tot=sz*sz2;
      this->data=new double[tot];
      this->grid_set=false;
    }
   
    ~tensor_old_grid2() {
      if (this->rk>0) {
	for(size_t i=0;i<this->rk;i++) {
	  delete[] this->grd[i];
	}
	delete[] this->size;
	delete[] this->grd;
	delete[] this->data;
	this->rk=0;
      }
    }
   
    /// Get the element indexed by \c (ix1,ix2)
    double &get(size_t ix1, size_t ix2) { 
      size_t sz[2]={ix1,ix2};
      return tensor_old_grid ::get(sz); 
    }

    /// Get the element indexed by \c (ix1,ix2)
    const double &get(size_t ix1, size_t ix2) const { 
      size_t sz[2]={ix1,ix2};
      return tensor_old_grid ::get(sz); 
    }
 
    /// Set the element indexed by \c (ix1,ix2) to value \c val
    void set(size_t ix1, size_t ix2, double val) {
      size_t sz[2]={ix1,ix2};
      tensor_old_grid ::set(sz,val); 
      return;
    }

    /// Interpolate \c (x,y) and return the results
    double interp(double x, double y) {
      double arr[2]={x,y};
      return interpolate(arr);
    }
  };
  
  /** \brief Rank 3 tensor_old
   */
  class tensor_old3 : public tensor_old {
  public:

    /// Create an empty tensor_old
  tensor_old3() : tensor_old() {}

    /// Create a rank 3 tensor_old of size \c (sz,sz2,sz3)
  tensor_old3(size_t sz, size_t sz2, size_t sz3) : tensor_old() {
      this->rk=3;
      this->size=new size_t[3];
      this->size[0]=sz;
      this->size[1]=sz2;
      this->size[2]=sz3;
      size_t tot=sz*sz2*sz3;
      this->data=new double[tot];
    }
	
    /// Get the element indexed by \c (ix1,ix2,ix3)
    double &get(size_t ix1, size_t ix2, size_t ix3) { 
      size_t sz[3]={ix1,ix2,ix3};
      return tensor_old::get(sz); 
    }

    /// Get the element indexed by \c (ix1,ix2,ix3)
    const double &get(size_t ix1, size_t ix2, size_t ix3) const { 
      size_t sz[3]={ix1,ix2,ix3};
      return tensor_old::get(sz); 
    }

    /// Set the element indexed by \c (ix1,ix2,ix3) to value \c val
    void set(size_t ix1, size_t ix2, size_t ix3, double val) {
      size_t sz[3]={ix1,ix2, ix3};
      tensor_old::set(sz,val); 
      return;
    }

    /** \brief Set the element indexed by \c index to value \c val

	(We have to explicitly provide this version since the set()
	function is overloaded in this child of \ref tensor_old.)
    */
    void set(size_t *index, double val) {
      tensor_old::set(index,val);
      return;
    }
  };
  
  /** \brief Rank 3 tensor_old with a grid
   */
  class tensor_old_grid3 : public tensor_old_grid {
     
  public:
     
    /// Create an empty tensor_old
  tensor_old_grid3() : tensor_old_grid () {}

    /// Create a rank 3 tensor_old of size \c (sz,sz2,sz3)
  tensor_old_grid3(size_t sz, size_t sz2, size_t sz3) : tensor_old_grid () {
      this->rk=3;
      this->size=new size_t[3];
      this->grd=new double *[3];
      this->size[0]=sz;
      this->size[1]=sz2;
      this->size[2]=sz3;
      this->grd[0]=new double[sz];
      this->grd[1]=new double[sz2];
      this->grd[2]=new double[sz3];
      size_t tot=sz*sz2*sz3;
      this->data=new double[tot];
      this->grid_set=false;
    }
   
    ~tensor_old_grid3() {
      if (this->rk>0) {
	for(size_t i=0;i<this->rk;i++) {
	  delete[] this->grd[i];
	}
	delete[] this->size;
	delete[] this->grd;
	delete[] this->data;
	this->rk=0;
      }
    }
   
    /// Get the element indexed by \c (ix1,ix2,ix3)
    double &get(size_t ix1, size_t ix2, size_t ix3) { 
      size_t sz[3]={ix1,ix2,ix3};
      return tensor_old_grid ::get(sz); 
    }
 
    /// Get the element indexed by \c (ix1,ix2,ix3)
    const double &get(size_t ix1, size_t ix2, size_t ix3) const { 
      size_t sz[3]={ix1,ix2,ix3};
      return tensor_old_grid ::get(sz); 
    }
 
    /// Set the element indexed by \c (ix1,ix2,ix3) to value \c val
    void set(size_t ix1, size_t ix2, size_t ix3, double val) {
      size_t sz[3]={ix1,ix2, ix3};
      tensor_old_grid ::set(sz,val); 
      return;
    }

    /// Interpolate \c (x,y,z) and return the results
    double interp(double x, double y, double z) {
      double arr[3]={x,y,z};
      return interpolate(arr);
    }
  };
  
  /** \brief Rank 4 tensor_old
   */
  class tensor_old4 : public tensor_old {

  public:

    /// Create an empty tensor_old
  tensor_old4() : tensor_old() {}

    /// Create a rank 4 tensor_old of size \c (sz,sz2,sz3,sz4)
  tensor_old4(size_t sz, size_t sz2, size_t sz3, size_t sz4) : 
    tensor_old() {
      this->rk=4;
      this->size=new size_t[4];
      this->size[0]=sz;
      this->size[1]=sz2;
      this->size[2]=sz3;
      this->size[3]=sz4;
      size_t tot=sz*sz2*sz3*sz4;
      this->data=new double[tot];
    }
	
    /// Get the element indexed by \c (ix1,ix2,ix3,ix4)
    double &get(size_t ix1, size_t ix2, size_t ix3, size_t ix4) { 
      size_t sz[4]={ix1,ix2,ix3,ix4};
      return tensor_old::get(sz); 
    }

    /// Get the element indexed by \c (ix1,ix2,ix3,ix4)
    const double &get(size_t ix1, size_t ix2, size_t ix3, 
		      size_t ix4) const { 
      size_t sz[4]={ix1,ix2,ix3,ix4};
      return tensor_old::get(sz); 
    }

    /// Set the element indexed by \c (ix1,ix2,ix3,ix4) to value \c val
    void set(size_t ix1, size_t ix2, size_t ix3, size_t ix4, 
	    double val) {
      size_t sz[4]={ix1,ix2,ix3,ix4};
      tensor_old::set(sz,val); 
      return;
    }

    /** \brief Set the element indexed by \c index to value \c val

	(We have to explicitly provide this version since the set()
	function is overloaded in this child of \ref tensor_old.)
    */
    void set(size_t *index, double val) {
      tensor_old::set(index,val);
      return;
    }
  };

  /** \brief Rank 4 tensor_old with a grid
   */
  class tensor_old_grid4 : public tensor_old_grid {
     
  public:
     
    /// Create an empty tensor_old
  tensor_old_grid4() : tensor_old_grid () {}

    /// Create a rank 4 tensor_old of size \c (sz,sz2,sz3,sz4)
  tensor_old_grid4(size_t sz, size_t sz2, size_t sz3,
	       size_t sz4) : tensor_old_grid () {
      this->rk=4;
      this->size=new size_t[4];
      this->grd=new double *[4];
      this->size[0]=sz;
      this->size[1]=sz2;
      this->size[2]=sz3;
      this->size[3]=sz4;
      this->grd[0]=new double[sz];
      this->grd[1]=new double[sz2];
      this->grd[2]=new double[sz3];
      this->grd[3]=new double[sz4];
      size_t tot=sz*sz2*sz3*sz4;
      this->data=new double[tot];
      this->grid_set=false;
    }
   
    ~tensor_old_grid4() {
      if (this->rk>0) {
	for(size_t i=0;i<this->rk;i++) {
	  delete[] this->grd[i];
	}
	delete[] this->size;
	delete[] this->grd;
	delete[] this->data;
	this->rk=0;
      }
    }
   
    /// Get the element indexed by \c (ix1,ix2,ix3,ix4)
    double &get(size_t ix1, size_t ix2, size_t ix3, size_t ix4) { 
      size_t sz[4]={ix1,ix2,ix3,ix4};
      return tensor_old_grid ::get(sz); 
    }
    
    /// Get the element indexed by \c (ix1,ix2,ix3,ix4)
    const double &get(size_t ix1, size_t ix2, size_t ix3,
		      size_t ix4) const { 
      size_t sz[4]={ix1,ix2,ix3,ix4};
      return tensor_old_grid ::get(sz); 
    }
    
    /// Set the element indexed by \c (ix1,ix2,ix3,ix4) to value \c val
    void set(size_t ix1, size_t ix2, size_t ix3, size_t ix4,
	    double val) {
      size_t sz[4]={ix1,ix2,ix3,ix4};
      tensor_old_grid ::set(sz,val); 
      return;
    }

    /// Interpolate \c (x,y,z,a) and return the results
    double interp(double x, double y, double z, double a) {
      double arr[4]={x,y,z,a};
      return interpolate(arr);
    }
  };
  
#ifndef DOXYGENP
}
#endif

#endif



