/*
  -------------------------------------------------------------------
  
  Copyright (C) 2011-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_UNIFORM_GRID_H
#define O2SCL_UNIFORM_GRID_H

#include <o2scl/err_hnd.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief A class representing a uniform linear or logarithmic grid

      \note This class has no public constructors and is to be
      instantiated through its children.
   */
  template<class data_t=double> class uniform_grid {
    
    private:

    // No public default constructor
    uniform_grid() {}
    
    protected:
  
    /// The low-side of the first bin
    data_t g_start;
    /// The high-side of the last bin
    data_t g_end;
    /// The width of each bin
    data_t g_width;
    /// The number of bins
    size_t g_n_bins;
    /// If true, use a logarithmic scale
    bool g_log;

    /** \brief Construct a grid with specified values

        \note This function is not public because it might create
	grids that are non-sensical. We require users to create grid
	objects using one of the children which don't allow
	non-sensical grids.
     */
    uniform_grid(data_t start, data_t end, data_t width, size_t n_bins, 
		 bool log=false) {
      if (n_bins==0) {
	O2SCL_ERR("Requested zero bins in grid::grid().",gsl_einval);
      }
      g_start=start;
      g_end=end;
      g_width=width;
      g_n_bins=n_bins;
      g_log=log;
    }

    public:

    /** \brief Get the number of bins (regions in between grid points)
     */
    size_t get_nbins() {
      return g_n_bins;
    }

    /** \brief Get the number of points in the grid (always get_nbins()+1)
     */
    size_t get_npoints() {
      return g_n_bins+1;
    }

    /** \brief Return true if the grid is logarithmic
     */
    bool is_log() {
      return g_log;
    }
    
    /** \brief Fill a vector with the specified grid
     */
    template<class vec_t> void vector(vec_t &v) {
      if (!g_log) {
	if (g_start<g_end) {
	  v[0]=g_start;
	  for(size_t i=1;i<g_n_bins;i++) {
	    v[i]=g_start+i*g_width;
	  }
	  v[g_n_bins]=g_end;
	} else {
	  v[0]=g_start;
	  for(size_t i=1;i<g_n_bins;i++) {
	    v[i]=g_start-i*g_width;
	  }
	  v[g_n_bins]=g_end;
	}
      } else {
	if (g_start<g_end) {
	  v[0]=g_start;
	  for(size_t i=1;i<g_n_bins;i++) {
	    v[i]=g_start*std::pow(g_width,((data_t)i));
	  }
	  v[g_n_bins]=g_end;
	} else {
	  v[0]=g_start;
	  for(size_t i=1;i<g_n_bins;i++) {
	    v[i]=g_start/std::pow(g_width,((data_t)i));
	  }
	  v[g_n_bins]=g_end;
	}
      }
    }

    /** \brief Get the grid point with index \c i 
	(\f$ i \in [0,\mathrm{n_bins}] \f$)
     */
    const data_t operator[](size_t i) const {
      if (!g_log) {
	if (g_start<g_end) {
	  if (i==0) return g_start;
	  else if (i==g_n_bins) return g_end;
	  return g_start+i*g_width;
	} else {
	  if (i==0) return g_start;
	  else if (i==g_n_bins) return g_end;
	  return g_start-i*g_width;
	}
      }
      if (g_start<g_end) {
	if (i==0) return g_start;
	else if (i==g_n_bins) return g_end;
	return g_start*std::pow(g_width,((data_t)i));
      }
      if (i==0) return g_start;
      else if (i==g_n_bins) return g_end;
      return g_start/std::pow(g_width,((data_t)i));
    }

  };

  /** \brief Linear grid with fixed number of bins and fixed endpoint
   */
  template<class data_t=double> class uniform_grid_end : 
  public uniform_grid<data_t> {
    public:
    uniform_grid_end(data_t start, data_t end, size_t n_bins) : 
    uniform_grid<data_t>(start,end,(end-start)/((data_t)n_bins),
			 n_bins,false) {
    }
  };
  
  /** \brief Linear grid with fixed number of bins and fixed bin size
   */
  template<class data_t=double> class uniform_grid_width : 
  public uniform_grid<data_t> {
    public:
    uniform_grid_width(data_t start, data_t width, size_t n_bins) :
    uniform_grid<data_t>(start,start+n_bins*width,width,n_bins,false) {
    }
  };

  /** \brief Linear grid with fixed endpoint and fixed bin size
   */
  template<class data_t=double> class uniform_grid_end_width : 
  public uniform_grid<data_t> {
    public:
    uniform_grid_end_width(data_t start, data_t end, data_t width) :
    uniform_grid<data_t>(start,end,width,
			 ((size_t)((end-start)/width*1.01)),false) {
    }
  };
  
  /** \brief Logarithmic grid with fixed number of bins and fixed endpoint
   */
  template<class data_t=double> class uniform_grid_log_end : 
  public uniform_grid<data_t> {
    public:
    uniform_grid_log_end(data_t start, data_t end, size_t n_bins) :
    uniform_grid<data_t>(start,end,std::pow(end/start,1.0/((data_t)n_bins)),
		 n_bins,true) {
    }
  };
  
  /** \brief Logarithmic grid with fixed number of bins and fixed bin size
   */
  template<class data_t=double> class uniform_grid_log_width : 
  public uniform_grid<data_t> {
    public:
    uniform_grid_log_width(data_t start, data_t width, size_t n_bins) :
    uniform_grid<data_t>(start,start*std::pow(width,n_bins),
			 width,n_bins,true) {
    }
  };
  
  /** \brief Logarithmic grid with fixed endpoint and fixed bin size
   */
  template<class data_t=double> class uniform_grid_log_end_width : 
    public uniform_grid<data_t> {
    public:
    uniform_grid_log_end_width(data_t start, data_t end, data_t width) :
    uniform_grid<data_t>(start,end,width,
		 ((size_t)(log(end/start)/log(width)*1.01)),true) {
    }
  };

#ifndef DOXYGENP
}
#endif

#endif
