/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef GEN_SN_EOS_H
#define GEN_SN_EOS_H

#include <cmath>
#include <iostream>
#include <fstream>
#include <o2scl/constants.h>
#include <o2scl/tensor.h>
#include <o2scl/table.h>
#include <o2scl/eff_boson.h>
#include <o2scl/rel_fermion.h>
#include <o2scl/eff_fermion.h>
#include <o2scl/test_mgr.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief A base class for the supernova EOSs [abstract]

      This class is experimental.

      The EOSs are stored in a set of \ref tensor_grid3 objects
      on grids with density in \f$ \mathrm{fm}^{-3}
      \f$, electron fraction (unitless) and temperature in MeV.

      Some tabulated EOSs do not store data for \ref gen_sn_eos::E,
      \ref gen_sn_eos::F, \ref gen_sn_eos::S, and \ref gen_sn_eos::P.
      In that case, the grid is set for these objects but the data is
      set to zero. To compute these from the data after loading the
      EOS table, use \ref gen_sn_eos::compute_eg().

      The function \ref load() loads the entire EOS into memory.
      Memory allocation is automatically performed by \ref load(),
      but not deallocated until free() or the destructor is called.

      After loading, you can interpolate the EOS by using 
      \ref tensor_grid3::interp directly. For example,
      the following returns the mass number at an arbitrary
      baryon density, electron fraction, and temperature
      assuming the table is stored in <tt>skm.dat</tt>:
      \verbatim
      ls_eos ls;
      ls.load("skm.dat");
      double nb=0.01, Ye=0.2, T=10.0;
      cout << A.interp(nb,Ye,T) << endl;
      \endverbatim
      Interpolation for all EOSs is linear by default.

      \todo Ensure all chemical potentials are based on the
      same rest masses?
      \todo Allow logarithmic grids for any of nb, Ye, or T. 
      \future Create a \ref table object, possibly using 
      tensor_grid::vector_slice. 
      \future Show how matrix_slice and vector_slice can be used
      with this object.
      \future Could this be a child of hadronic_eos_temp and
      then directly used in cold_nstar()?
      \future Add option to load and store a separate lepton/photon EOS
      \future Add muons and/or pions
  */
  class gen_sn_eos {

  public:

    gen_sn_eos();
    
    virtual ~gen_sn_eos();

    /// \name Load table
    //@{
    /// Load table from filename \c fname
    virtual void load(std::string fname)=0;
    /// If true, a EOS table was successfully loaded (default false)
    bool loaded;
    /// True if thermodynamics with leptons has been loaded
    bool with_leptons_loaded;
    /// True if baryon-only thermodynamics has been loaded
    bool baryons_only_loaded;
    //@}

    /// \name Grid and data sizes
    //@{
    /// Size of baryon density grid
    size_t n_nB;
    /// Size of electron fraction grid
    size_t n_Ye;
    /// Size of temperature grid
    size_t n_T;
    /// Number of additional data sets
    size_t n_oth;
    /// Number of base data sets
    static const size_t n_base=16;
    //@}

    /// \name Data
    //@{
    /// Total free energy per baryon in MeV
    tensor_grid3 F;
    /** \brief Free energy per baryon without lepton and photon 
	contributions in MeV
    */
    tensor_grid3 Fint;
    /// Total internal energy per baryon in MeV
    tensor_grid3 E;
    /** \brief Internal energy per baryon without lepton and photon 
	contributions in MeV
    */
    tensor_grid3 Eint;
    /// Total pressure in \f$ \mathrm{MeV}/\mathrm{fm}^3 \f$
    tensor_grid3 P;
    /** \brief Pressure without lepton and photon contributions
	 in \f$ \mathrm{MeV}/\mathrm{fm}^3 \f$
    */
    tensor_grid3 Pint;
    /// Total entropy per baryon
    tensor_grid3 S;
    /// Entry per baryon without lepton and photon contributions
    tensor_grid3 Sint;
    /// Neutron chemical potential in MeV
    tensor_grid3 mun;
    /// Proton chemical potential in MeV
    tensor_grid3 mup;
    /// Proton number
    tensor_grid3 Z;
    /// Mass number
    tensor_grid3 A;
    /// Neutron fraction
    tensor_grid3 Xn;
    /// Proton fraction
    tensor_grid3 Xp;
    /// Alpha particle fraction
    tensor_grid3 Xalpha;
    /// Fraction of heavy nuclei
    tensor_grid3 Xnuclei;
    /// Other data sets
    tensor_grid3 other[20];
    /// List of pointers to data
    tensor_grid3 *arr[n_base+20];
    //@}

    /// \name Memory allocation
    //@{
    /// Allocate memory
    int alloc();

    /// Free allocated memory
    int free();
    //@}

    /// \name Interpolation
    //@{
    /** \brief Set interpolation managers
     */
    int set_interp(base_interp_mgr<uvector_base> &bi1,
		   base_interp_mgr<uvector_const_subvector> &bi2);

    /// Default interpolation object
    def_interp_mgr<uvector_base,linear_interp> dim1;
    
    /// Default interpolation object
    def_interp_mgr<uvector_const_subvector,linear_interp> dim2;
    //@}

    /// \name Electron and photon contribution
    //@{
    /// Photon
    boson photon;
    /// Electron;
    fermion electron;
    /// Desc
    rel_fermion relf;
    /// Desc
    eff_boson effb;
    /** \brief Compute the electron and photon contribution for the full
	grid

	This function computes the data for <tt>E, P, S,</tt> and
	<tt>F</tt> by adding electrons and photons to the baryon
	contributions stored in <tt>Eint, Pint, Sint,</tt> and
	<tt>Fint</tt>.

	The electron contribution to the internal energy and free
	energy computed by this function includes the electron rest
	mass.
    */
    int compute_eg();
    //@}

    /** \brief Test the free energy and store results in \c tm

	This checks that the data in \c Fint 
	is consistent with that in \c Eint and \c Sint .
     */
    int check_free_energy(test_mgr &tm);
    
    /** \brief Verbosity parameter (default 1)
     */
    int verbose;

    /** \brief Compute energy per baryon of matter in beta equilibrium
	at zero temperature at a fixed grid point [abstract]

	Given an index \c i for the baryon grid, between 0 and \ref
	n_nB (inclusive), this computes the properties of matter in
	beta equilibrium at zero temperature by finding the electron
	fraction grid point which minimizes \ref E. The baryon density
	is returned in \c nb, the energy per baryon in \c E_beta, the
	pressure in \c P_beta, the electron fraction in \c Ye_beta,
	the proton number in \c Z_beta and the mass number of the
	nucleus in \c A_beta.

	Some tables explicitly contain zero-temperature data which is
	used when it is available. Otherwise, linear interpolation is
	used to extrapolate down to zero from the lowest temperature 
	grid points.
    */
    virtual void beta_eq_T0(size_t i, double &nb, double &E_beta, 
			    double &P_beta, double &Ye_beta,
			    double &Z_beta, double &A_beta)=0;

    /** \brief Compute properties of matter in beta equilibrium
	at s=4
    */
    virtual void beta_eq_sfixed(size_t i, double entr,
				double &nb, double &E_beta, 
				double &P_beta, double &Ye_beta,
				double &Z_beta, double &A_beta,
				double &T_beta) {

      if (loaded==false) {
	O2SCL_ERR2("No data loaded in ",
		   "gen_sn_eos::beta_eq_s4().",gsl_einval);
      }
      if (i>=n_nB) {
	O2SCL_ERR2("Too high for baryon grid in ",
		   "gen_sn_eos::beta_eq_s4().",gsl_einval);
      }
      if (with_leptons_loaded==false) {
	compute_eg();
      }
      // Get baryon density from grid
      nb=E.get_grid(0,i);
      double Emin=0.0;
      // The electron fraction grid point corresponding to the minimum
      size_t j_found=0;
      // The temperature grid point for s=4 at the Ye grid point j_found
      size_t k_min_j=0;
      for(size_t j=0;j<n_Ye;j++) {
	// For each electron fraction, we need to find the 
	// grid point surrounding s=4
	bool found=false;
	size_t k_found=0;
	for(size_t k=0;k<n_T-1;k++) {
	  if (S.get(i,j,k)<entr && S.get(i,j,k+1)>=entr) {
	    k_found=k; 
	    k=n_T;
	    found=true;
	  }
	}
	if (found==false) {
	  if (S.get(i,j,0)<entr) {
	    k_found=n_T-2;
	  }
	}
	if (j==0) {
	  Emin=E.get(i,j,k_found)+(E.get(i,j,k_found+1)-E.get(i,j,k_found))*
	    (entr-S.get(i,j,k_found))/(S.get(i,j,k_found+1)-
				       S.get(i,j,k_found));
	} else {
	  double Ethis=E.get(i,j,k_found)+
	    (E.get(i,j,k_found+1)-E.get(i,j,k_found))*
	    (entr-S.get(i,j,k_found))/(S.get(i,j,k_found+1)-
				       S.get(i,j,k_found));
	  if (Ethis<Emin) {
	    j_found=j;
	    Emin=Ethis;
	    k_min_j=k_found;
	  }
	}
      }
      // Interpolate final results
      E_beta=Emin;
      P_beta=P.get(i,j_found,k_min_j)+
	(P.get(i,j_found,k_min_j+1)-P.get(i,j_found,k_min_j))*
	(entr-S.get(i,j_found,k_min_j))/
	(S.get(i,j_found,k_min_j+1)-S.get(i,j_found,k_min_j));
      Z_beta=Z.get(i,j_found,k_min_j)+
	(Z.get(i,j_found,k_min_j+1)-Z.get(i,j_found,k_min_j))*
	(entr-S.get(i,j_found,k_min_j))/
	(S.get(i,j_found,k_min_j+1)-S.get(i,j_found,k_min_j));
      A_beta=A.get(i,j_found,k_min_j)+
	(A.get(i,j_found,k_min_j+1)-A.get(i,j_found,k_min_j))*
	(entr-S.get(i,j_found,k_min_j))/
	(S.get(i,j_found,k_min_j+1)-S.get(i,j_found,k_min_j));
      Ye_beta=E.get_grid(1,j_found);
      T_beta=E.get_grid(2,k_min_j)+
	(E.get_grid(2,k_min_j+1)-E.get_grid(2,k_min_j))*
	(entr-S.get(i,j_found,k_min_j))/
	(S.get(i,j_found,k_min_j+1)-S.get(i,j_found,k_min_j));
	
      return;
    }

  };

  /** \brief The Lattimer-Swesty supernova EOS 

      This class is experimental.

      \note \o2 Does not contain the Lattimer-Swesty EOS, only
      provides some code to manipulate it. This class is designed
      to be used with the files <tt>ls.dat, sk1.dat, ska.dat</tt>
      and <tt>skm.dat</tt> as provided on Jim Lattimer's website,
      http://www.astro.sunysb.edu/lattimer/EOS/main.html .

      Note that the tables on this website are different
      than what is generated from the LS Fortran code. See
      \ref oo_eos to read O'Connor and Ott's tables
      generated from the LS Fortran code.

      The four models are
      - LS (K=370, Sv=31)
      - SKI' (K=371, Sv=30.4)
      - SKa (K=263, Sv=34.5)
      - SKM* (K=217, Sv=31.4) 

      \note In the original table, the full internal energy per baryon
      (data section 4 of 26) is apparently based on a rest mass of
      \f$ Y_e m_p + (1-Y_e) m_n \f$, while the baryon part of the
      internal energy per baryon (data section 13 of 26) is based
      on a rest mass of \f$ m_n \f$. This means that 
      \f[
      E - E_{\mathrm{int}} = E_{\mathrm{eg}} - Y_e (m_n - m_p)
      \f]
      where \f$ E_{\mathrm{eg}} \f$ is the energy per baryon of
      electrons and photons. In order to keep things consistent with
      the other EOS tables, when the EOS table is loaded, \ref
      gen_sn_eos::Eint is rescaled to a rest mass of \f$ Y_e m_p +
      (1-Y_e) m_n \f$ .

      See also the documentation at \ref gen_sn_eos. 

      See \ref Lattimer91 and \ref Lattimer85.

      \todo There are still a few points for which the electron/photon
      EOS seems to be off. 
  */
  class ls_eos : public gen_sn_eos {

  public:
    
    /// Filling factor for nuclei
    tensor_grid3 &fill;
    /// Baryon number density inside nuclei in \f$ \mathrm{fm}^{-3} \f$
    tensor_grid3 &nb_in;
    /// Derivative of pressure with respect to baryon density
    tensor_grid3 &dPdn;
    /// Derivative of pressure with respect to temperature
    tensor_grid3 &dPdT;
    /// Derivative of pressure with respect to electron fraction
    tensor_grid3 &dPdY;
    /// Derivative of entropy with respect to temperature
    tensor_grid3 &dsdT;
    /// Derivative of entropy with respect to electron fraction
    tensor_grid3 &dsdY;
    /// Number of neutrons in skin
    tensor_grid3 &Nskin;
    /// Baryon density outside nuclei in \f$ \mathrm{fm}^{-3} \f$
    tensor_grid3 &nb_out;
    /// Proton fraction outside nuclei
    tensor_grid3 &x_out;
    /** \brief Out of whackness parameter, 
	\f$ \mu_n-\mu_p-\mu_e+1.293~\mathrm{MeV} \f$, in MeV
    */
    tensor_grid3 &mu;

  ls_eos() :
    fill(other[0]),
      nb_in(other[1]),
      dPdn(other[2]),
      dPdT(other[3]),
      dPdY(other[4]),
      dsdT(other[5]),
      dsdY(other[6]),
      Nskin(other[7]),
      nb_out(other[8]),
      x_out(other[9]),
      mu(other[10]) {
    }

    /// Load table from filename \c fname
    virtual void load(std::string fname);

    /** \brief Check electrons and photons

	This checks that the electron and photon thermodynamics
	generated by \o2 is consistent with the data in 
	\c E, \c Eint, \c F, \c Fint, \c P, \c Pint, \c S,
	and \c Sint.
     */
    int check_eg(test_mgr &tm);

    /** \brief Compute properties of matter in beta equilibrium
	at zero temperature at a baryon density grid point
	
	This EOS table doesn't have T=0 results, so we extrapolate
	from the two low-temperature grid points.
    */
    virtual void beta_eq_T0(size_t i, double &nb, double &E_beta, 
			    double &P_beta, double &Ye_beta,
			    double &Z_beta, double &A_beta) {

      if (loaded==false || with_leptons_loaded==false) {
	O2SCL_ERR2("No data loaded in ",
		   "ls_eos::beta_eq_T0().",gsl_einval);
      }
      if (i>=n_nB) {
	O2SCL_ERR2("Too high for baryon grid in ",
		   "ls_eos::beta_eq_T0().",gsl_einval);
      }
      // Get baryon density from grid
      nb=E.get_grid(0,i);
      // Find the minima for the two low-temperature grid points
      double ET1=E.get(i,0,0), PT1=P.get(i,0,0);
      double ET2=E.get(i,0,1), PT2=P.get(i,0,1);
      double ZT1=Z.get(i,0,0), ZT2=Z.get(i,0,1);
      double AT1=A.get(i,0,0), AT2=A.get(i,0,1);
      Ye_beta=E.get_grid(1,0);
      for(size_t j=1;j<n_Ye;j++) {
	double Enew=E.get(i,j,0);
	if (Enew<ET1) {
	  ET1=Enew;
	  ET2=E.get(i,j,1);
	  PT1=P.get(i,j,0);
	  PT2=P.get(i,j,1);
	  ZT1=Z.get(i,j,0);
	  ZT2=Z.get(i,j,1);
	  AT1=A.get(i,j,0);
	  AT2=A.get(i,j,1);
	  Ye_beta=E.get_grid(1,j);
	}
      }
      // Now extrapolate to T=0
      E_beta=ET1-(ET2-ET1)*E.get_grid(2,0)/
	(E.get_grid(2,1)-E.get_grid(2,0));
      P_beta=PT1-(PT2-PT1)*P.get_grid(2,0)/
	(P.get_grid(2,1)-P.get_grid(2,0));
      Z_beta=ZT1-(ZT2-ZT1)*Z.get_grid(2,0)/
	(Z.get_grid(2,1)-Z.get_grid(2,0));
      A_beta=AT1-(AT2-AT1)*A.get_grid(2,0)/
	(A.get_grid(2,1)-A.get_grid(2,0));
	
      return;
    }

  };

  /** \brief The EOS tables from O'Connor and Ott

      This class reads the HDF5 EOS tables generated by E. O'Connor
      and C. Ott in \ref OConnor10. The tables are available from 

      http://stellarcollapse.org/equationofstate

      and are available under a creative commons
      attribution-noncommercial-share alike license. This \o2 code to
      read those tables is licensed (along with all \o2 code) under
      the GPLv3 license (with permission from Evan O'Connor).

      The original README file from O'Connor and Ott's EOSdriver
      code is available in the \o2 
      distribution in <tt>doc/o2scl/eos/extras/scollapse_README</tt>
      and is reproduced below
						
      \verbinclude scollapse_README
  */
  class oo_eos : public gen_sn_eos {

  public:
    
  oo_eos() :
    cs2(other[0]),
      dedt(other[1]),
      dpderho(other[2]),
      dpdrhoe(other[3]),
      gamma(other[4]),
      mu_e(other[5]),
      muhat(other[6]),
      munu(other[7]) {
    }

    /// Speed of sound in cm^2/s^2
    tensor_grid &cs2;
    /// C_V in erg/g/K
    tensor_grid &dedt;
    /// dpderho in dyn*g/cm^2/erg
    tensor_grid &dpderho;
    /// dpdrhoe in dyn cm^3/cm^2/g
    tensor_grid &dpdrhoe;
    /// Gamma
    tensor_grid &gamma;
    /// Electron chemical potential per baryon including rest mass
    tensor_grid &mu_e;
    /// mun - mup
    tensor_grid &muhat;
    /// mue - mun + mup
    tensor_grid &munu;
    /// The original mass density grid from the table in g/cm^3
    ovector rho;
    /// Energy shift for table storage in erg/g
    double energy_shift;

    /// Use the J. Lattimer et al. method for handling the chemical potentials
    static const size_t ls_mode=0;
    /// Use the H. Shen et al. method for handling the chemical potentials
    static const size_t stos_mode=1;
    
    /// Load table from filename \c fname
    virtual void load(std::string fname, size_t mode);

    /// Load table from filename \c fname
    virtual void load(std::string fname) {
      load(fname,0);
      return;
    }
    
    /** \brief Compute properties of matter in beta equilibrium
	at zero temperature at a baryon density grid point
	
	This EOS table doesn't have T=0 results, so we extrapolate
	from the two low-temperature grid points.
    */
    virtual void beta_eq_T0(size_t i, double &nb, double &E_beta, 
			    double &P_beta, double &Ye_beta,
			    double &Z_beta, double &A_beta) {

      if (loaded==false || baryons_only_loaded==false) {
	O2SCL_ERR2("No data loaded in ",
		   "oo_eos::beta_eq_T0().",gsl_einval);
      }
      if (i>=n_nB) {
	O2SCL_ERR2("Too high for baryon grid in ",
		   "oo_eos::beta_eq_T0().",gsl_einval);
      }
      if (with_leptons_loaded==false) {
	compute_eg();
      }
      // Get baryon density from grid
      nb=E.get_grid(0,i);
      // Find the minima for the two low-temperature grid points
      double ET1=E.get(i,0,0), PT1=P.get(i,0,0);
      double ET2=E.get(i,0,1), PT2=P.get(i,0,1);
      double ZT1=Z.get(i,0,0), ZT2=Z.get(i,0,1);
      double AT1=A.get(i,0,0), AT2=A.get(i,0,1);
      Ye_beta=E.get_grid(1,0);
      for(size_t j=1;j<n_Ye;j++) {
	double Enew=E.get(i,j,0);
	if (Enew<ET1) {
	  ET1=Enew;
	  ET2=E.get(i,j,1);
	  PT1=P.get(i,j,0);
	  PT2=P.get(i,j,1);
	  ZT1=Z.get(i,j,0);
	  ZT2=Z.get(i,j,1);
	  AT1=A.get(i,j,0);
	  AT2=A.get(i,j,1);
	  Ye_beta=E.get_grid(1,j);
	}
      }
      // Now extrapolate to T=0
      E_beta=ET1-(ET2-ET1)*E.get_grid(2,0)/
	(E.get_grid(2,1)-E.get_grid(2,0));
      P_beta=PT1-(PT2-PT1)*P.get_grid(2,0)/
	(P.get_grid(2,1)-P.get_grid(2,0));
      Z_beta=ZT1-(ZT2-ZT1)*Z.get_grid(2,0)/
	(Z.get_grid(2,1)-Z.get_grid(2,0));
      A_beta=AT1-(AT2-AT1)*A.get_grid(2,0)/
	(A.get_grid(2,1)-A.get_grid(2,0));
	
      return;
    }

  };
  
  /** \brief The Shen et al. supernova EOS
      
      This class is experimental.

      \note \o2 Does not contain the EOS, only provides some code to
      manipulate it. This class is designed to be used with the file
      which was originally called <tt>eos.tab</tt> and now referred to
      as <tt>eos1.tab</tt> and stored e.g. at
      http://user.numazu-ct.ac.jp/~sumi/eos/.

      In order to force the EOS to a uniform grid, linear
      interpolation is used to recast the variation in baryon density,
      choosing the grid in baryon density to be the same as the
      section in the table with T=0.1 MeV and \f$ Y_p = 0.1 \f$ for
      all temperature and proton fraction points.

      Also, the original EOS is tabulated for constant proton
      fraction, and this \o2 interface assumes that the electron
      fraction is equal to the proton fraction. Currently, this is a
      problem only at higher densities where muons might appear.

      The data for \ref gen_sn_eos::E, \ref gen_sn_eos::F, \ref
      gen_sn_eos::S, and \ref gen_sn_eos::P is not stored in the table
      but can be computed with \ref gen_sn_eos::compute_eg().

      See also the documentation at \ref gen_sn_eos. 

      See \ref Shen98 and \ref Shen98b .

      \note Thanks to Matthias Hempel for providing the correct
      temperature grid.

      \future Add the T=0 and Ye=0 data to this class
   */
  class stos_eos : public gen_sn_eos {

  public:

    /** \brief Logarithm of baryon number density in 
	\f$ \mathrm{g}/\mathrm{cm}^3 \f$
    */
    tensor_grid3 &log_rho;
    /// Baryon number density in \f$ \mathrm{fm}^{-3} \f$
    tensor_grid3 &nB;
    /// Logarithm of proton fraction
    tensor_grid3 &log_Y;
    /// Proton fraction
    tensor_grid3 &Yp;
    /// Nucleon effective mass in MeV
    tensor_grid3 &M_star;
    /// Fraction of quark matter
    tensor_grid3 &quark_frac;

  stos_eos() :
    log_rho(other[0]),
      nB(other[1]),
      log_Y(other[2]),
      Yp(other[3]),
      M_star(other[4]),
      quark_frac(other[5]) {
    }

    static const size_t orig_mode=0;
    static const size_t quark_mode=1;

    /// Load table from filename \c fname
    virtual void load(std::string fname) {
      load(fname,0);
    }
    
    /// Load table from filename \c fname
    virtual void load(std::string fname, size_t mode);

    /** \brief Compute properties of matter in beta equilibrium
	at zero temperature at a baryon density grid point

	This EOS table doesn't have T=0 results, so we extrapolate
	from the two low-temperature grid points.
    */
    virtual void beta_eq_T0(size_t i, double &nb, double &E_beta, 
			    double &P_beta, double &Ye_beta,
			    double &Z_beta, double &A_beta) {
      if (loaded==false || baryons_only_loaded==false) {
	O2SCL_ERR2("No data loaded in ",
		   "stos_eos::beta_eq_T0().",gsl_einval);
      }
      if (i>=n_nB) {
	O2SCL_ERR2("Too high for baryon grid in ",
		   "stos_eos::beta_eq_T0().",gsl_einval);
      }
      if (with_leptons_loaded==false) {
	compute_eg();
      }
      // Get baryon density from grid
      nb=E.get_grid(0,i);
      // Find the minima for the two low-temperature grid points
      double ET1=E.get(i,0,0), PT1=P.get(i,0,0);
      double ET2=E.get(i,0,1), PT2=P.get(i,0,1);
      double ZT1=Z.get(i,0,0), ZT2=Z.get(i,0,1);
      double AT1=A.get(i,0,0), AT2=A.get(i,0,1);
      Ye_beta=E.get_grid(1,0);
      for(size_t j=1;j<n_Ye;j++) {
	double Enew=E.get(i,j,0);
	if (Enew<ET1) {
	  ET1=Enew;
	  ET2=E.get(i,j,1);
	  PT1=P.get(i,j,0);
	  PT2=P.get(i,j,1);
	  ZT1=Z.get(i,j,0);
	  ZT2=Z.get(i,j,1);
	  AT1=A.get(i,j,0);
	  AT2=A.get(i,j,1);
	  Ye_beta=E.get_grid(1,j);
	}
      }
      // Now extrapolate to T=0
      E_beta=ET1-(ET2-ET1)*E.get_grid(2,0)/
	(E.get_grid(2,1)-E.get_grid(2,0));
      P_beta=PT1-(PT2-PT1)*P.get_grid(2,0)/
	(P.get_grid(2,1)-P.get_grid(2,0));
      Z_beta=ZT1-(ZT2-ZT1)*Z.get_grid(2,0)/
	(Z.get_grid(2,1)-Z.get_grid(2,0));
      A_beta=AT1-(AT2-AT1)*A.get_grid(2,0)/
	(A.get_grid(2,1)-A.get_grid(2,0));
	
      return;
    }
    
  };
  
  /** \brief A class to manipulate the G. Shen et al. EOS

      This class is experimental.

      \note \o2 Does not contain the EOS, only provides some code to
      manipulate it. This class was designed to be used with the FSU
      models given at
      http://cecelia.physics.indiana.edu/gang_shen_eos/FSU/fsu.html
      and stored in the files named <tt>FSU1.7eos1.01.dat</tt>,
      <tt>FSU2.1eos1.01.dat</tt>, <tt>FSU1.7eosb1.01.dat</tt>, and
      <tt>FSU2.1eosb1.01.dat</tt> corresponding to 
      \ref mode_17, \ref mode_21, \ref mode_17b and \ref mode_21b
      respectively. 

      See also the documentation at \ref gen_sn_eos. 

      \todo More documentation
   */
  class sht_eos : public gen_sn_eos {

  public:

    /// 1.7 solar masses with leptons and photons
    static const size_t mode_17=0;
    /// 2.1 solar masses with leptons and photons
    static const size_t mode_21=1;
    /// 1.7 solar masses without leptons and photons
    static const size_t mode_17b=2;
    /// 2.1 solar masses without leptons and photons
    static const size_t mode_21b=3;
    /// NL3 model with leptons and photons
    static const size_t mode_NL3=4;
    /// NL3 model with leptons and photons
    static const size_t mode_NL3b=5;

    /// Temperature in MeV
    tensor_grid3 &T;
    /// Proton fraction
    tensor_grid3 &Yp;
    /// Baryon number density in \f$ 1/\mathrm{fm}^3 \f$
    tensor_grid3 &nB;
    /// Electron chemical potential in MeV
    tensor_grid3 &mue;
    /// Nucleon effective mass in MeV
    tensor_grid3 &M_star;
    
    sht_eos() :
      T(other[0]),
      Yp(other[1]),
      nB(other[2]),
      mue(other[3]),
      M_star(other[4]) {
    }

    /// Load table from filename \c fname
    virtual void load(std::string fname, size_t mode);

    /// Load table from filename \c fname
    virtual void load(std::string fname) {
      load(fname,0);
    }

    /** \brief Compute properties of matter in beta equilibrium
	at zero temperature at a baryon density grid point
    */
    virtual void beta_eq_T0(size_t i, double &nb, double &E_beta, 
			    double &P_beta, double &Ye_beta,
			    double &Z_beta, double &A_beta) {
      if (loaded==false || (with_leptons_loaded==false && 
			    baryons_only_loaded==false)) {
	O2SCL_ERR2("No data loaded in ",
		   "sht_eos::beta_eq_T0().",gsl_einval);
      }
      if (i>=n_nB) {
	O2SCL_ERR2("Too high for baryon grid in ",
		   "sht_eos::beta_eq_T0().",gsl_einval);
      }
      if (with_leptons_loaded==false) {
	compute_eg();
      }
      // Look up baryon density from grid
      nb=E.get_grid(0,i);
      // Minimize over all electron fractions
      E_beta=E.get(i,0,0);
      P_beta=P.get(i,0,0);
      Ye_beta=E.get_grid(1,0);
      Z_beta=Z.get(i,0,0);
      A_beta=A.get(i,0,0);
      for(size_t j=1;j<n_Ye;j++) {
	double Enew=E.get(i,j,0);
	if (Enew<E_beta) {
	  E_beta=Enew;
	  P_beta=P.get(i,j,0);
	  Ye_beta=E.get_grid(1,j);
	  Z_beta=Z.get(i,j,0);
	  A_beta=A.get(i,j,0);
	}
      }
      return;
    }
    
  };
  
  /** \brief The Hempel et al. supernova EOSs
      
      This class is experimental.

      \note \o2 Does not contain the EOS, only provides some code to
      manipulate it. This class was designed to be used with the files
      <tt>dd2_frdm_eos_shen98format_v1.02.tab</tt>,
      <tt>fsg_roca_eos_shen98format_v1.0.tab</tt>, and
      <tt>nl3_lala_eos_shen98format_v1.0.tab</tt> as obtained from
      http://phys-merger.physik.unibas.ch/~hempel/eos.html.

      See also the documentation at \ref gen_sn_eos. 

      See \ref Hempel10 and \ref Hempel11.
   */
  class hfsl_eos : public gen_sn_eos {

  public:

    /** \brief Logarithm of baryon number density in 
	\f$ \mathrm{g}/\mathrm{cm}^3 \f$
    */
    tensor_grid3 &log_rho;
    /// Baryon number density in \f$ 1/\mathrm{fm}^3 \f$
    tensor_grid3 &nB;
    /// Logarithm of proton fraction
    tensor_grid3 &log_Y;
    /// Proton fraction
    tensor_grid3 &Yp;
    /// Nucleon effective mass in MeV
    tensor_grid3 &M_star;
    /// Mass number of light fragments
    tensor_grid3 &A_light;
    /// Proton number of light fragments
    tensor_grid3 &Z_light;

    hfsl_eos() :
      log_rho(other[0]),
      nB(other[1]),
      log_Y(other[2]),
      Yp(other[3]),
      M_star(other[4]),
      A_light(other[5]),
      Z_light(other[6]) {
    }
    
    /// Load table from filename \c fname
    virtual void load(std::string fname);

    /** \brief Compute properties of matter in beta equilibrium
	at zero temperature at a baryon density grid point
    */
    virtual void beta_eq_T0(size_t i, double &nb, double &E_beta, 
			    double &P_beta, double &Ye_beta,
			    double &Z_beta, double &A_beta) {
      if (loaded==false || baryons_only_loaded==false) {
	O2SCL_ERR2("No data loaded in ",
		   "hfsl_eos::beta_eq_T0().",gsl_einval);
      }
      if (i>=n_nB) {
	O2SCL_ERR2("Too high for baryon grid in ",
		   "hfsl_eos::beta_eq_T0().",gsl_einval);
      }
      if (with_leptons_loaded==false) {
	compute_eg();
      }
      // Get baryon density from grid
      nb=E.get_grid(0,i);
      // Find the minima for the two low-temperature grid points
      double ET1=E.get(i,0,0), PT1=P.get(i,0,0);
      double ET2=E.get(i,0,1), PT2=P.get(i,0,1);
      double ZT1=Z.get(i,0,0), ZT2=Z.get(i,0,1);
      double AT1=A.get(i,0,0), AT2=A.get(i,0,1);
      Ye_beta=E.get_grid(1,0);
      for(size_t j=1;j<n_Ye;j++) {
	double Enew=E.get(i,j,0);
	if (Enew<ET1) {
	  ET1=Enew;
	  ET2=E.get(i,j,1);
	  PT1=P.get(i,j,0);
	  PT2=P.get(i,j,1);
	  ZT1=Z.get(i,j,0);
	  ZT2=Z.get(i,j,1);
	  AT1=A.get(i,j,0);
	  AT2=A.get(i,j,1);
	  Ye_beta=E.get_grid(1,j);
	}
      }
      // Now extrapolate to T=0
      E_beta=ET1-(ET2-ET1)*E.get_grid(2,0)/
	(E.get_grid(2,1)-E.get_grid(2,0));
      P_beta=PT1-(PT2-PT1)*P.get_grid(2,0)/
	(P.get_grid(2,1)-P.get_grid(2,0));
      Z_beta=ZT1-(ZT2-ZT1)*Z.get_grid(2,0)/
	(Z.get_grid(2,1)-Z.get_grid(2,0));
      A_beta=AT1-(AT2-AT1)*A.get_grid(2,0)/
	(A.get_grid(2,1)-A.get_grid(2,0));
	
      return;
    }
    
  };
  
#ifndef DOXYGENP
}
#endif

#endif
