/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/tov_eos_fast.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;

tov_eos_fast::tov_eos_fast() {
  eos_set=false;
  verbose=1;
}

tov_eos_fast::~tov_eos_fast() {
}

void tov_eos_fast::interp(const ovector_base &x, const ovector_base &y, 
			  double xx, double &yy, int n1, int n2) {
  int lo,hi,k;
  
  lo=n1;
  hi=n2;
  while (hi-lo > 1) {
    k=(hi+lo) >> 1;
    if (x[k] > xx) hi=k;
    else lo=k;
  }
  if (x[hi]==x[lo]) yy=y[hi];
  else yy=y[lo]+(xx-x[lo])/(x[hi]-x[lo])*(y[hi]-y[lo]);
  
  return;
}

int tov_eos_fast::set_crust_core(size_t n_crust, ovector_base &crust_ed, 
				 ovector_base &crust_pr, 
				 ovector_base &crust_nb,
				 size_t n_core, ovector_base &core_ed, 
				 ovector_base &core_pr, 
				 ovector_base &core_nb) {
  
  if (check_eos) {
    
    for(size_t i=0;i<n_crust-1;i++) {
      if (crust_ed[i]>crust_ed[i+1]) {
	O2SCL_ERR2_RET("Crust energy density is decreasing in ",
		       "tov_eos_fast::read_table().",gsl_einval);
      }
      if (crust_pr[i]>crust_pr[i+1]) {
	O2SCL_ERR2_RET("Crust pressure is decreasing in ",
		       "tov_eos_fast::read_table().",gsl_einval);
      }
    }

    for(size_t i=0;i<n_core-1;i++) {
      if (core_ed[i]>core_ed[i+1]) {
	O2SCL_ERR2_RET("Core energy density is decreasing in ",
		       "tov_eos_fast::read_table().",gsl_einval);
      }
      if (core_pr[i]>core_pr[i+1]) {
	O2SCL_ERR2_RET("Core pressure is decreasing in ",
		       "tov_eos_fast::read_table().",gsl_einval);
      }
    }

  }

  edv.allocate(n_crust+n_core);
  prv.allocate(n_crust+n_core);
  nbv.allocate(n_crust+n_core);

  size_t k_crust=n_crust-1, k_core=0;

  double ed_crust_high=crust_ed[k_crust];
  double pr_crust_high=crust_pr[k_crust];
  double ed_core_low=core_ed[k_core];
  double pr_core_low=core_pr[k_core];

  cout << "ed: " << ed_crust_high << " " << ed_core_low << endl;
  cout << "pr: " << pr_crust_high << " " << pr_core_low << endl;
  cout << "k: " << k_crust << " " << k_core << endl;

  while (ed_crust_high>ed_core_low || pr_crust_high>pr_core_low) {
    if (k_crust==0 || k_core>n_core-2) {
      O2SCL_ERR2_RET("Failed to perform transition in ",
		     "tov_eos_fast::add_core_to_crust().",gsl_einval);
    }
    if ((pr_crust_high/ed_crust_high)>(pr_core_low/ed_core_low)) {
      k_core++;
    } else {
      k_crust--;
    }
    ed_crust_high=crust_ed[k_crust];
    pr_crust_high=crust_pr[k_crust];
    ed_core_low=core_ed[k_core];
    pr_core_low=core_pr[k_core];
    cout << "ed: " << ed_crust_high << " " << ed_core_low << endl;
    cout << "pr: " << pr_crust_high << " " << pr_core_low << endl;
    cout << "k: " << k_crust << " " << k_core << endl;
  }

  for(size_t j=0;j<=k_crust;j++) {
    edv[j]=crust_ed[j];
    prv[j]=crust_pr[j];
    nbv[j]=crust_nb[j];
    cout << j << " " << edv[j] << endl;
  }
  cout << endl;

  size_t j=k_crust+1;
  for(size_t k=k_core;k<n_core;k++) {
    edv[j]=core_ed[k];
    prv[j]=core_pr[k];
    nbv[j]=core_nb[k];
    cout << j << " " << edv[j] << endl;
    j++;
  }
  
  nlines=j;
  cout << nlines << endl;

  return 0;
}

int tov_eos_fast::get_eden(double pres, double &ed, double &nb) {
  
  interp(prv,edv,pres,ed,0,nlines-1);
  interp(prv,nbv,pres,nb,0,nlines-1);
  
  if (!gsl_finite(ed) || !gsl_finite(nb)) {
    
    string s="Energy density or baryon density not finite at pressure ";
    s+=dtos(pres)+" in tov_eos_fast::get_eden().";
    O2SCL_ERR_RET(s.c_str(),gsl_efailed);
    
  }
  
  return 0;
}

