/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_TOV_SOLVE_H
#define O2SCL_TOV_SOLVE_H

#include <o2scl/tov_eos.h>
#include <o2scl/smart_interp.h>
#include <o2scl/table_units.h>
#include <o2scl/shared_ptr.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief Class to solve the Tolman-Oppenheimer-Volkov equations

      \hline
      <b>Mathematical background:</b>

      In units where \f$ c=1 \f$, the TOV equations (i.e. Einstein's
      equations for a static spherically symmetric object) are
      \f[
      \frac{d m}{d r} = 4 \pi r^2 \varepsilon
      \f]
      \f[
      \frac{d P}{d r} = - \frac{G \varepsilon m}{r^2}
      \left( 1+\frac{P}{\varepsilon}\right)
      \left( 1+\frac{4 \pi P r^3}{m} \right)
      \left( 1-\frac{2 G m}{r}\right)^{-1}
      \f]
      where \f$r\f$ is the radial coordinate, \f$m(r)\f$ is the
      gravitational mass enclosed within a radius \f$r\f$, and
      \f$\varepsilon(r)\f$ and \f$P(r)\f$ are the energy density and
      pressure at \f$r\f$, and \f$G\f$ is the gravitational constant.
      The boundary conditions are \f$m(r=0)=0\f$ the condition
      \f$P(r=R)=0\f$ for some fixed radius \f$R\f$. These boundary
      conditions give a series of solutions to the TOV equations as a
      function of the radius, although they do not necessarily have a
      solution for all radii.

      The gravitational mass is given by
      \f[
      M_G = \int_0^R 4 \pi r^2 \varepsilon d r
      \f]

      The gravitational potential, \f$\Phi(r)\f$ can be determined from
      \f[
      \frac{d \Phi}{d r} = - \frac{1}{\varepsilon}
      \frac{ d P}{d r} \left(1+\frac{P}{\varepsilon}\right)^{-1}
      \f]

      The proper boundary condition for the gravitational potential
      is 
      \f[
      \Phi(r=R) = \frac{1}{2} \ln \left(1-\frac{2 G M}{R} \right)
      \f]
      which ensures that \f$ \phi(r) \rightarrow 0 \f$ as 
      \f$ r \rightarrow \infty \f$ .

      The surface gravity is
      \f[
      g = \frac{G m}{r^2}\left(1-\frac{2 G m}{r}\right)^{-1/2}
      \f]
      which is given in inverse kilometers and the redshift is
      \f[
      z = \left(1-\frac{2 G m}{r}\right)^{-1/2} - 1
      \f]
      which is unitless.

      The baryonic mass is typically defined by
      \f[
      M_B = \int_0^R 4 \pi r^2 n_B m_B 
      \left(1-\frac{2 G m}{r}\right)^{-1/2} d r
      \f]
      where \f$ n_B(r) \f$ is the baryon number density at radius 
      \f$ r \f$ and \f$ m_B \f$ is the mass of one baryon. Then the
      "binding energy" of the neutron star is defined to be \f$ M_B -
      M_G \f$ . If you can specify the product \f$ n_B m_B \f$ in the
      EOS (analogous to the rest mass energy density), it will compute
      the associated baryonic mass for you.

      In the case of slow rigid rotation with angular velocity
      \f$ \Omega \f$, the moment of inertia is
      \f[
      I = \frac{8 \pi}{3} \int_0^R dr~r^4\left(\varepsilon+P\right)
      e^{\Phi} \left(\frac{\bar{\omega}}{\Omega}\right)
      \left(1-\frac{2 G m}{r}\right)^{-1/2}
      \f]
      where the function \f$ \bar{\omega}(r) \f$ is the solution of 
      \f[
      \frac{d}{dr} \left( r^4 j \frac{d \bar{\omega}}{dr}\right)
      + 4 r^3 \frac{d j}{dr} \bar{\omega} = 0
      \f]
      and the function \f$ j(r) \f$ is defined by 
      \f[
      j = \left( 1-\frac{2 G m}{r} \right) e^{-\Phi}
      \f]
      The boundary conditions for \f$ \bar{\omega} \f$ are
      \f$ d \bar{\omega}/dr = 0 \f$ at \f$ r=0 \f$ and 
      \f[
      \bar{\omega}(R) = \Omega - \left(\frac{R}{3}\right)
      \left(\frac{d \bar{\omega}}{dr}\right)_{r=R} \, .
      \f]

      \hline
      <b>General usage notes:</b>

      The equation of state may be changed at any time, by specifying
      the appropriate tov_eos object 
      
      Screen output:
      - \c verbose=0 - Nothing
      - \c verbose=1 - Basic information
      - \c verbose=2 - For each profile computation, report solution
      information at every kilometer.
      - \c verbose=3 - Report profile information at every 20 grid points. A 
      keypress is required after each profile.

      \hline
      <b>Output tables:</b>

      The functions fixed() and max() produce output tables which
      represent the profile of the neutron star of the requested mass.
      The columns are
      - \c gm, the enclosed gravitational mass in solar masses
      - \c r, the radial coordinate in \f$ \mathrm{km} \f$
      - \c gp, the gravitational potential (unitless) when
      \ref calcgpot is true
      - \c bm, the baryonic mass in solar masses (when 
      \ref tov_eos::baryon_column is true). 
      - \c pr, the pressure
      - \c ed, the energy density
      - \c nb, the baryon density (if \ref tov_eos::baryon_column
      is true)
      - \c sg, the local surface gravity 
      (in \f$ \mathrm{g}/\mathrm{cm}^{2} \f$ )
      - \c rs, the local redshift (unitless),
      - \c dmdr, the derivative of the enclosed gravitational mass
      in \f$ \mathrm{M}_{\odot}/\mathrm{km} \f$
      - \c dlogpdr, the derivative of the natural logarithm of the 
      pressure
      - \c dgpdr, the derivative of the gravitational potential
      in \f$ 1/\mathrm{km} \f$ (if \ref calcgpot is true)
      - \c dbmdr, the derivative of the enclosed baryonic mass
      (if \ref tov_eos::baryon_column is true). \n
      The remaining columns are given by the user-defined columns from
      the equation of state as determined by \ref tov_eos::get_names_units()
      and \ref tov_eos::get_aux().

      The function mvsr() produces a different kind of output table
      corresponding to the mass versus radius curve. Some points on
      the curve may correspond to unstable branches.
      - \c gm, the total gravitational mass in solar masses
      - \c r, the radius in \f$ \mathrm{km} \f$
      - \c gp, the gravitational potential at the surface (unitless) when
      \ref calcgpot is true
      - \c bm, total the baryonic mass in solar masses (when 
      \ref tov_eos::baryon_column is true). 
      - \c pr, the central pressure
      - \c ed, the central energy density
      - \c nb, the central baryon density (if \ref tov_eos::baryon_column
      is true)
      - \c sg, the surface gravity 
      (in \f$ \mathrm{g}/\mathrm{cm}^{2} \f$ )
      - \c rs, the redshift at the surface,
      - \c dmdr, the derivative of the gravitational mass
      - \c dlogpdr, the derivative of the natural logarithm of the 
      pressure
      - \c dgpdr, the derivative of the gravitational potential
      in \f$ 1/\mathrm{km} \f$ (if \ref calcgpot is true)
      - \c dbmdr, the derivative of the enclosed baryonic mass
      (if \ref tov_eos::baryon_column is true). \n
      The remaining columns are given by the user-defined columns from
      the equation of state as determined by \ref tov_eos::get_names_units()
      and \ref tov_eos::get_aux().
      
      \hline
      <b>Accuracy:</b>

      The present code, as demonstrated in the tests, gives the
      correct central pressure and energy density of the analytical
      solution by Buchdahl to within less than 1 \part in \f$ 10^8 \f$.

      \hline
      <b>Other details and todos:</b>

      \note The function star_fun() returns <tt>gsl_efailed</tt>
      without calling the error handler in the case that the solver
      can recover gracefully from, for example, a negative pressure.
      
      \todo
      - baryon mass doesn't work for fixed() (This may be fixed. We should
      make sure it's tested.)
      - Combine maxoutsize and kmax?
      - Document column naming issues
      - Document surface gravity and redshift
      - Standardize \c xmev_kg, etc. 
      - Use convert_units? 
      - Double check that fixed() doesn't give a solution on the
      unstable branch
      - Ensure that this class copies over the units of the 
      user-specified columns to the table

      \comment
      To solve the calling of the error handler, I need to rewrite the
      adaptive stepper to simply stop if the base stepper makes a step
      which is too large. That demands creating a new version of
      gsl_astep::evolve_apply(). The question is how to communicate
      the information from the tov derivs() function to the new
      version of gsl_astep::evolve_apply(). An easy way would be to
      just create a new return value which indicates the step should
      be shortened, and just stop if the step becomes too short.
      Alternatively, instead of modifying evolve_apply(), we could
      just ensure the adaptive stepper bails out correctly if the
      stepper returns a nonzero value and then have the caller
      (starfun) take care of it. This would probably be better. We
      might have to make sure that gsl_astep does not set yout if the
      step fails. (3/27/09 - It looks like this might be fixed now?)
      \endcomment

      \hline
  */
  class tov_solve {

#ifndef DOXYGEN_INTERNAL

  protected:

    /// To convert MeV to kilograms
    double mev_kg;

    /// To convert \f$ MeV/fm{^3} \f$ to solar masses per cubic kilometer
    double mev_per_fm3_msun_km3;
    
    /** \brief Target mass
	
	Negative values to indicate a mass measured relative to the
	maximum mass. For example, if the EOS has a maximum mass of 2.0, then
	-0.15 will give the profile of a 1.85 solar mass star.
    */
    double tmass;

    /** \brief Ensure \c col does not match strings in \c cnames
	
	Underscores are added to \c col until it matches none of
	the strings in \c cnames.
    */
    int make_unique_name(std::string &col, 
			 std::vector<std::string> &cnames);
    
    /// The EOS
    tov_eos *te;

    /// True if the EOS has been set
    bool eos_set;

    /// \name User EOS units
    //@{
    /// Units for energy density
    std::string eunits;
    /// Units for pressure
    std::string punits;
    /// Units for baryon density
    std::string nunits;
    /// unit conversion factor for energy density (default 1.0)
    double efactor;
    /// unit conversion factor for pressure (default 1.0)
    double pfactor;
    /// unit conversion factor for baryon density (default 1.0)
    double nfactor;
    //@}
    
    /// maximum storage size (default 40000)
    int kmax;
    /// maximum size of output file (default 400)
    int maxoutsize;

    /** \brief Smallest allowed pressure for integration (default: -100)
   
	This quantity can't be much smaller than -100 since we need to
	compute numbers near \f$ e^{-presmin} \f$
    */
    double presmin;

    /// \name Integration storage
    //@{
    /** \brief ODE functions

	The vector \c rky[0] is the gravitational mass in solar
	masses, and the vector \c rky[1] is the natural logarithm of
	the pressure in solar masses per cubic kilometer. When \ref
	calcgpot is true, the next vector in this list is the
	gravitational potential (which is unitless), and when \ref
	tov_eos::baryon_column is true, the next vector in this 
	list is the baryonic mass in solar masses.
    */
    ovector rky[6];
    /// Radial coordinate (in kilometers)
    ovector rkx;
    /// The derivatives of the ODE functions
    ovector rkdydx[6];
    //@}
    
    /// Compute the angular velocity
    int ang_vel();

    /// Interpolation object for derivs_ang_vel()
    sm_interp smi;

    /// The output table
    o2_shared_ptr<table_units>::type out_table;

    /// The solver
    mroot<mm_funct<> > *mroot_ptr;

    /// The minimizer
    minimize<funct> *min_ptr;
    
    /// The adaptive stepper
    adapt_step<ode_funct<> > *as_ptr;

    /// The ODE step function
    virtual int derivs(double x, size_t nv, const ovector_base &y,
		       ovector_base &dydx);
    
    /// The ODE step function for the angular velocity
    virtual int derivs_ang_vel(double x, size_t nv, const ovector_base &y,
			       ovector_base &dydx);
    
    /** \brief Construct a stellar profile

	This function constructs a stellar profile in \ref out_table
	from the information in \ref rkx and \ref rky.
     */
    virtual int profile_out(double xx);
    
    /// The minimizer function to compute the maximum mass
    virtual double maxfun(double maxx);

    /** \brief The solver function to compute the stellar profile
     */
    virtual int starfun(size_t ndvar, const ovector_base &ndx, 
			ovector_base &ndy);
    
    /// The schwarzchild radius in km
    double schwarz_km;
    
#endif

  public:

    tov_solve();

    virtual ~tov_solve();
    
    /// \name Basic properties
    //@{
    /// mass
    double mass;
    /// radius
    double rad;
    /// baryonic mass
    double bmass;
    /// gravitational potential
    double gpot;
    //@}

    /** \name Solution parameters
	
	These parameters can be changed at any time.
    */
    //@{
    /// Use general relativistic version (default true)
    bool generel;
    /** \brief calculate the gravitational potential and the enclosed 
	baryon mass (default false)
    */
    bool calcgpot;
    /// smallest allowed radial stepsize (default 1.0e-4)
    double hmin;
    /// largest allowed radial stepsize (default 0.05)
    double hmax;
    /// initial radial stepsize (default 4.0e-3)
    double hstart;
    /// control for output (default 1)
    int verbose;
    /// maximum radius for integration in km (default 60)
    double maxradius;
    //@}

    /// \name Mass versus radius parameters

    //@{
    /// Beginning pressure (default 7.0e-7)
    double prbegin;
    /// Ending pressure (default 8.0e-3)
    double prend;
    /// Increment for pressure (default 1.1)
    double princ;
    /// Use 'princ' as a multiplier, not an additive increment (default true)
    bool logmode;
    /** \brief Guess for central pressure in solar masses per km3 
	(default \f$ 5.2 \times 10^{-5} \f$)

	This guess is used in the function fixed().
    */
    double prguess;
    /// Beginning pressure for maximum mass guess (default 7.0e-5)
    double max_begin;
    /// Ending pressure for maximum mass guess (default 5.0e-3)
    double max_end;
    /// Increment for pressure for maximum mass guess (default 1.3)
    double max_inc;
    //@}

    /// \name Basic operation
    //@{
    /// Set the EOS to use
    int set_eos(tov_eos &ter) {
      te=&ter;
      eos_set=true;
      return 0;
    }

    /// Set units
    int set_units(double s_efactor=1.0, double s_pfactor=1.0, 
		  double s_nbfactor=1.0);

    /** \brief Set units
	
	Valid entries for the units of energy density and pressure are:
	- "g/cm^3"
	- "erg/cm^3"
	- "MeV/fm^3"
	- "1/fm^4"
	- "Msun/km^3" (i.e. solar masses per cubic kilometer)
	
	Valid entries for the units of baryon density are:
	- "1/m^3"
	- "1/cm^3"
	- "1/fm^3"
    */
    int set_units(std::string eunits="", std::string punits="", 
		  std::string nunits="");

    /// Calculate the mass vs. radius curve
    virtual int mvsr();

    /// Calculate the profile of a star with fixed mass
    virtual int fixed(double d_tmass);

    /// Calculate the profile of the maximum mass star
    virtual int max();

    /** \brief Return the results data table
     */
    o2_shared_ptr<table_units>::type get_results() {
      return out_table;
    }

    /** \brief Check the solution (unfinished)
     */
    int solution_check();
    //@}
    
    /// \name Control numerical methods
    //@{
    /** \brief Set maximum storage for integration
	
	The variable \c s_kmax is the maximum number of radial
	integration stepsk while \c s_maxoutsize is the maximum number
	of points that will be output for any profile.
	
	If \c s_kmax is less than zero, there is no limit on
	the number of radial steps.
    */
    int set_kmax(int s_maxoutsize=400, int s_kmax=80000);

    /** \brief Set solver
    */
    int set_mroot(mroot<mm_funct<> > &s_mrp) 
    {mroot_ptr=&s_mrp; return 0;};
    
    /** \brief Set minimizer
    */
    int set_minimize(minimize<funct > &s_mp) 
    {min_ptr=&s_mp; return 0;};

    /** \brief Set the adaptive stepper
    */
    int set_stepper(adapt_step<ode_funct<> > &sap)
    {as_ptr=&sap; return 0;};
    //@}

    /// The default minimizer
    gsl_min_brent<funct > def_min;
    
    /// The default solver
    gsl_mroot_hybrids<mm_funct<> > def_solver;

    /// The default adaptive stepper
    gsl_astep<ode_funct<> > def_stepper;

    /// If true, compute the angular velocity (default false)
    bool compute_ang_vel;
    
    /// The angular velocity
    double cap_omega;

  };

#ifndef DOXYGENP
}
#endif

#endif


