/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
/* multifit/fsolver.c
 * 
 * Copyright (C) 1996, 1997, 1998, 1999, 2000, 2007 Brian Gough
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 */

#ifndef O2SCL_GSL_FIT_H
#define O2SCL_GSL_FIT_H

#include <o2scl/fit_base.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_multifit_nlin.h>
#include <o2scl/omatrix_tlate.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief Non-linear least-squares fitting class (GSL)
    
      The GSL-based fitting class using a Levenberg-Marquardt type
      algorithm. The algorithm stops when
      \f[
      |dx_i| < \mathrm{epsabs}+\mathrm{epsrel}\times|x_i|
      \f]
      where \f$dx\f$ is the last step and \f$x\f$ is the current
      position. If test_gradient is true, then additionally fit()
      requires that
      \f[
      \sum_i |g_i| < \mathrm{epsabs}
      \f]
      where \f$g_i\f$ is the \f$i\f$-th component of the gradient of
      the function \f$\Phi(x)\f$ where
      \f[
      \Phi(x) = || F(x) ||^2
      \f]

      Default template arguments
      - \c func_t - \ref fit_funct\<\>
      - \c vec_t - \ref ovector_base
      - \c alloc_vec_t - \ref ovector
      - \c bool_vec_t - <tt>bool *</tt>
      
      \todo Properly generalize other vector types than 
      ovector_base
      \todo Allow the user to specify the derivatives
      \todo Fix so that the user can specify automatic
      scaling of the fitting parameters, where the initial
      guess are used for scaling so that the fitting parameters
      are near unity.
  */
  template<class func_t=fit_funct<>, class vec_t=ovector_base, 
    class mat_t=omatrix_base, class bool_vec_t=bool *> class gsl_fit : 
  public fit_base<func_t,vec_t,mat_t> {

  protected:

#ifdef O2SCL_NEVER_DEFINED

  int covar_new(size_t n, const mat_t &J, double epsrel, 
		mat_t &covar) {

    double tolr;
      
    size_t i, j, k;
    size_t kmax = 0;
    
    alloc_mat_t r;
    alloc_vec_t tau, norm;
    permutation perm;
      
    am.allocate(r,n,n);
    av.allocate(tau,n);
    av.allocate(norm,n);
    perm.allocate(n);
      
    int signum = 0;
    gsl_matrix_memcpy (r, J);
    gsl_linalg_QRPT_decomp (r, tau, perm, &signum, norm);
  
    /* Form the inverse of R in the full upper triangle of R */

    tolr = epsrel * fabs(gsl_matrix_get(r, 0, 0));

    for (k = 0 ; k < n ; k++) {

      double rkk = r[k][k];

      if (fabs(rkk) <= tolr) {
	break;
      }

      r[k][k]=1.0/rkk;

      for (j = 0; j < k ; j++) {

	double t = r[j][k] / rkk;
	r[j][k]=0.0;

	for (i = 0; i <= j; i++) {

	  double rik = r[i][k];
	  double rij = r[i][k];
              
	  r[i][k]=rik - t * rij;
	}
      }
      kmax = k;
    }

    /* Form the full upper triangle of the inverse of R^T R in the full
       upper triangle of R */

    for (k = 0; k <= kmax ; k++) {

      for (j = 0; j < k; j++) {

	double rjk = r[j][k];

	for (i = 0; i <= j ; i++) {

	  double rij = r[i][j];
	  double rik = r[i][k];

	  r[i][j]= rij + rjk * rik;
	}
      }
      
      {
	double t = r[k][k];

	for (i = 0; i <= k; i++) {

	  double rik = r[i][k];

	  r[i][k]*=t;
	};
      }
    }

    /* Form the full lower triangle of the covariance matrix in the
       strict lower triangle of R and in w */

    for (j = 0 ; j < n ; j++) {

      size_t pj = gsl_permutation_get (perm, j);
      
      for (i = 0; i <= j; i++) {

	size_t pi = gsl_permutation_get (perm, i);

	double rij;

	if (j > kmax) {

	  gsl_matrix_set (r, i, j, 0.0);
	  rij = 0.0 ;

	} else {

	  rij = gsl_matrix_get (r, i, j);
	}

	if (pi > pj) {
	  gsl_matrix_set (r, pi, pj, rij); 

	} else if (pi < pj) {

	  gsl_matrix_set (r, pj, pi, rij);
	}

      }
      
      { 
	double rjj = r[j][j];
	covar[pj][pj]=rjj;
      }
    }

     
    /* symmetrize the covariance matrix */

    for (j = 0 ; j < n ; j++) {

      for (i = 0; i < j ; i++) {

	double rji = gsl_matrix_get (r, j, i);

	gsl_matrix_set (covar, j, i, rji);
	gsl_matrix_set (covar, i, j, rji);
      }
    }

    gsl_matrix_free (r);
    gsl_permutation_free (perm);
    gsl_vector_free (tau);
    gsl_vector_free (norm);

    return GSL_SUCCESS;
  }

#endif
    
  public:
    
  gsl_fit() {
    gsl_set_error_handler(err_hnd->gsl_hnd);
    max_iter=500;
    epsabs=1.0e-4;
    epsrel=1.0e-4;
    use_scaled=true;
    test_gradient=false;
  }

  virtual ~gsl_fit() {}
  
  /** \brief Fit the data specified in (xdat,ydat) to
      the function \c fitfun with the parameters in \c par.
	
      The covariance matrix for the parameters is returned in \c covar
      and the value of \f$ \chi^2 \f$ is returned in \c chi2.
  */
  virtual int fit(size_t ndat, vec_t &xdat, vec_t &ydat, vec_t &yerr,
		  size_t npar, vec_t &par, mat_t &covar, double &chi2,
		  func_t &fitfun) {

    int status, iter=0;

    gsl_vector *x=gsl_vector_alloc(npar);
    for(size_t i=0;i<npar;i++) gsl_vector_set(x,i,par[i]);

    // Create fitting function and parameters
    gsl_multifit_function_fdf f;
    gsl_vector *grad=gsl_vector_alloc(npar);
    func_par fpar={fitfun,ndat,&xdat,&ydat,&yerr,npar};
    f.f=func;
    f.df=dfunc;
    f.fdf=fdfunc;
    f.n=ndat;
    f.p=npar;
    f.params=&fpar;
     
    // Create fitting object
    gsl_multifit_fdfsolver *s;
    if (use_scaled) {
      s=gsl_multifit_fdfsolver_alloc
	(gsl_multifit_fdfsolver_lmsder,ndat,npar);
    } else {
      s=gsl_multifit_fdfsolver_alloc
      (gsl_multifit_fdfsolver_lmder,ndat,npar);
    }
    gsl_multifit_fdfsolver_set(s,&f,x);
  
    // Perform the fit
    do {
      iter++;
      status=gsl_multifit_fdfsolver_iterate(s);
    
      if (status) {
	break;
      }
      
      // The equivalent of 
      // 
      // status=gsl_multifit_test_delta(s->dx,s->x,epsabs,epsrel);
      //
      // with an additional calculation of the worst deviation for
      // printing iteration information in the variable 'max_ratio'
      
      double max_ratio=0.0;
      double val=1.0, tol=1.0;
      {
	if (epsrel<0.0 || epsabs<0.0) {
	  O2SCL_ERR2_RET("A tolerance is negative in ",
			 "gsl_fit::iterate().",gsl_ebadtol);
	}
	
	status=gsl_success;

	for (size_t i=0;i<x->size;i++) {
	  double xi=gsl_vector_get(s->x,i);
	  double dxi=gsl_vector_get(s->dx,i);
	  double tolerance=epsabs+epsrel*fabs(xi);
	  
	  if (fabs(dxi)>tolerance) {
	    status=gsl_continue;
	  }
	  if (fabs(dxi)/tolerance>max_ratio) {
	    max_ratio=fabs(dxi)/tolerance;
	    val=fabs(dxi);
	    tol=tolerance;
	  }
	}
      }

      if (this->verbose>0) {
	print_iter(npar,s->x,iter,val,tol);
      }
      
      // If requested, also test the gradient

      if (status!=gsl_continue && test_gradient) {

	// The equivalent of 
	// status=gsl_multifit_test_gradient(grad,epsabs);
	
	gsl_multifit_gradient(s->J,s->x,grad);
	double residual=0.0;
	for(size_t i=0;i<grad->size;i++) {
	  residual+=fabs(gsl_vector_get(grad,i));
	}
	if (residual<epsabs) status=gsl_success;
	else status=gsl_continue;
	
	if (this->verbose>0) {
	  print_iter(npar,s->x,iter,residual,epsabs);
	}
      }
      
    } while (status == gsl_continue && iter < max_iter);
    
    // Create covariance matrix and copy results to the 
    // user-specified vector
    gsl_matrix *gcovar=gsl_matrix_alloc(npar,npar);
    gsl_multifit_covar(s->J,0.0,gcovar);
    for(size_t i=0;i<npar;i++) {
      par[i]=gsl_vector_get(s->x,i);
      for(size_t j=0;j<npar;j++) {
	covar.set(i,j,gsl_matrix_get(gcovar,i,j));
      }
    }
    gsl_matrix_free(gcovar);
    
    // Compute chi squared
    chi2=gsl_blas_dnrm2(s->f);
    chi2*=chi2;
    
    // Free memory
    gsl_multifit_fdfsolver_free(s);
    gsl_vector_free(grad);
    
    return 0;
  }
  
  /// (default 500)
  int max_iter;
  
  /// (default 1.0e-4)
  double epsabs;

  /// (default 1.0e-4)
  double epsrel;

  /// If true, test the gradient also (default false)
  bool test_gradient;

  /** \brief Use the scaled routine if true (default true)
   */
  bool use_scaled;

  /// Return string denoting type ("gsl_fit")
  virtual const char *type() { return "gsl_fit"; }

#ifndef DOXYGEN_INTERNAL

  protected:

  /// Print the progress in the current iteration
  virtual int print_iter(size_t nv, gsl_vector *x, 
			 int iter, double val, double lim) {
    
    char ch;
    
    if (this->verbose<=0) return 0;
    
    std::cout << "gsl_fit iteration: " << iter << std::endl;
    std::cout << " par: ";
    for(size_t j=0;j<nv;j++) {
      std::cout << gsl_vector_get(x,j) << " ";
    }
    std::cout << std::endl;
    std::cout << " Val: " << val
    << " Lim: " << lim << std::endl;
    if (this->verbose>1) {
      std::cout << "Press a key and type enter to continue. ";
      std::cin >> ch;
    }
    
    return 0;
  }

  /** \brief A structure for passing to the functions func(), dfunc(),
      and fdfunc()
  */
  typedef struct {
    /// The function object 
    func_t &f;
    /// The number 
    int ndat;
    /// The x values 
    vec_t *xdat;
    /// The y values 
    vec_t *ydat;
    /// The y uncertainties
    vec_t *yerr;
    /// The number of parameters
    int npar;
  } func_par;

  /// Evaluate the function
  static int func(const gsl_vector *x, void *pa, gsl_vector *f) {
    func_par *fp=(func_par *)pa;
    int i, j;
    double yi;
    ovector xp(fp->npar);
    for(j=0;j<fp->npar;j++) xp[j]=gsl_vector_get(x,j);
    for(i=0;i<fp->ndat;i++) {
      yi=fp->f(fp->npar,xp,(*fp->xdat)[i]);
      gsl_vector_set(f,i,(yi-(*fp->ydat)[i])/(*fp->yerr)[i]);
    }

    return 0;
  }

  /// Evaluate the jacobian
  static int dfunc(const gsl_vector *x, void *pa, gsl_matrix *jac) {
    func_par *fp=(func_par *)pa;
    int i, j;
    double yi, ylo, yhi, xtemp, eps=1.0e-4;
    ovector xp(fp->npar);
  
    for(j=0;j<fp->npar;j++) xp[j]=gsl_vector_get(x,j);
    for(j=0;j<fp->npar;j++) {
      for(i=0;i<fp->ndat;i++) {
	xtemp=xp[j];
	xp[j]+=eps;
	yi=fp->f(fp->npar,xp,(*fp->xdat)[i]);
	yhi=(yi-(*fp->ydat)[i])/(*fp->yerr)[i];
	xp[j]=xtemp;
	yi=fp->f(fp->npar,xp,(*fp->xdat)[i]);
	ylo=(yi-(*fp->ydat)[i])/(*fp->yerr)[i];
	gsl_matrix_set(jac,i,j,(yhi-ylo)/eps);
      }
    }

    return 0;
  }

  /// Evaluate the function and the jacobian
  static int fdfunc(const gsl_vector *x, void *pa, gsl_vector *f,
		    gsl_matrix *jac) {
    func(x,pa,f);
    dfunc(x,pa,jac);
    return 0;
  }

#endif
  
  };

#ifndef DOXYGENP
}
#endif

#endif
