/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <o2scl/gsl_fit.h>
#include <o2scl/test_mgr.h>
#include <o2scl/columnify.h>

using namespace std;
using namespace o2scl;

int expb_f(const gsl_vector *x, void *params, gsl_vector *f);
int expb_df(const gsl_vector *x, void *params, gsl_matrix *J);
int expb_fdf(const gsl_vector *x, void *params,
	     gsl_vector *f, gsl_matrix *J);
int print_state(size_t iter, gsl_multifit_fdfsolver *s);
double func(size_t np, const ovector_base &p, double x);

struct data {
  size_t n;
  double *y;
  double *sigma;
};

int main(void) {
  cout.setf(ios::scientific);
  test_mgr t;
  t.set_output_level(1);

  //----------------------------------------------------------------
  // First do the GSL example the usual way:
  //----------------------------------------------------------------

  double fact;
  int status;
  size_t i, iter=0;
  const size_t n=40;
  const size_t p=3;
  gsl_matrix *covar=gsl_matrix_alloc(p,p);
  double xdat[40], y[40], sigma[40];
  struct data d={n,y,sigma};
  double x_init[3]={1.0,0.0,0.0};
  gsl_vector_view x=gsl_vector_view_array(x_init,p);

  gsl_multifit_function_fdf f;
  f.f=&expb_f;
  f.df=&expb_df;
  f.fdf=&expb_fdf;
  f.n=n;
  f.p=p;
  f.params=&d;

  // Create data to be fitted

  const gsl_rng_type *type;
  gsl_rng *r;
  gsl_rng_env_setup();
  type=gsl_rng_default;
  r=gsl_rng_alloc(type);

  for (i=0;i<n;i++) {
    double tt=i;
    xdat[i]=tt;
    y[i]=1.0+5*exp(-0.1*tt)+gsl_ran_gaussian(r,0.1);
    sigma[i]=0.1;
  };

  // Create fitting object

  const gsl_multifit_fdfsolver_type *T;
  gsl_multifit_fdfsolver *s;
  T=gsl_multifit_fdfsolver_lmsder;
  s=gsl_multifit_fdfsolver_alloc(T,n,p);
  gsl_multifit_fdfsolver_set(s,&f,&x.vector);

  do {
    iter++;
    status=gsl_multifit_fdfsolver_iterate (s);
    if (status)
      break;
    
    status=gsl_multifit_test_delta(s->dx,s->x,1e-4,1e-4);
  }
  while (status == gsl_continue && iter < 500);

  gsl_multifit_covar (s->J, 0.0, covar);
  
  double chi=gsl_blas_dnrm2(s->f);
  fact=1.0;
  if (chi*chi/(n-p)>fact) fact=chi*chi/(n-p);

  cout << endl;
  cout << "GSL results: " << endl;
  cout << "Parameters: " << gsl_vector_get(s->x,0) << " "
       << gsl_vector_get(s->x,1) << " "
       << gsl_vector_get(s->x,2) << endl;
  cout << "Errors    : " << sqrt(gsl_matrix_get(covar,0,0))*fact << " "
       << sqrt(gsl_matrix_get(covar,1,1))*fact << " "
       << sqrt(gsl_matrix_get(covar,2,2))*fact << endl;
  cout << "chi^2/dof : " << chi*chi/(n-p) << endl;
  cout << "Covariance matrix: " << endl;
  for(size_t ik=0;ik<3;ik++) {
    for(size_t jk=0;jk<3;jk++) {
      cout << gsl_matrix_get(covar,ik,jk) << " ";
    }
    cout << endl;
  }
  cout << endl;

  cout << "-----------------------------------------------" << endl;

  //----------------------------------------------------------------
  // Next, the same fit using the O2scl wrapper.
  //----------------------------------------------------------------
  
  fit_funct_fptr<ovector_base> ff(func);
  gsl_fit<fit_funct<ovector_base>,ovector_base,omatrix_base> gf;
  
  double chi2;
  void *vp=0;
  omatrix mycovar(3,3);
  ovector_array axdat(40,xdat);
  ovector_array ay(40,y);
  ovector_array asigma(40,sigma);
  ovector_array ax_init(3,x_init);
  gf.fit(40,axdat,ay,asigma,3,ax_init,mycovar,chi2,ff);
  
  fact=1.0;
  if (chi2/(n-p)>fact) fact=chi2/(n-p);

  cout << endl;
  cout << "O2scl results: " << endl;
  cout << "Parameters: " 
       << x_init[0] << " " << x_init[1] << " " << x_init[2] << endl;
  cout << "Errors    : " << sqrt(mycovar[0][0])*fact << " " 
       << sqrt(mycovar[1][1])*fact << " " 
       << sqrt(mycovar[2][2])*fact << endl;
  cout << "chi^2/dof : " << chi2/(n-p) << endl;

  cout << "Covariance matrix: " << endl;
  matrix_out(cout,mycovar,3,3);
  cout << endl;
  
  t.test_rel(x_init[0],gsl_vector_get(s->x,0),1.0e-5,"fit0");
  t.test_rel(x_init[1],gsl_vector_get(s->x,1),1.0e-5,"fit1");
  t.test_rel(x_init[2],gsl_vector_get(s->x,2),1.0e-5,"fit2");
  t.test_rel(sqrt(mycovar[0][0]),sqrt(gsl_matrix_get(covar,0,0)),
	     1.0e-3,"err0");
  t.test_rel(sqrt(mycovar[1][1]),sqrt(gsl_matrix_get(covar,1,1)),
	     1.0e-3,"err1");
  t.test_rel(sqrt(mycovar[2][2]),sqrt(gsl_matrix_get(covar,2,2)),
	     1.0e-3,"err2");
  t.test_rel(chi2,chi*chi,1.0e-6,"chi2");
  for(size_t ik=0;ik<3;ik++) {
    for(size_t jk=0;jk<3;jk++) {
      t.test_rel(mycovar[ik][jk],gsl_matrix_get(covar,ik,jk),1.0e-2,"cv");
    }
  }

  cout << "-----------------------------------------------" << endl;

  //----------------------------------------------------------------
  // Try the unscaled algorithm
  //----------------------------------------------------------------
  
  gf.use_scaled=false;
  x_init[0]=1.0;
  x_init[1]=0.0;
  x_init[2]=0.0;
  gf.fit(40,axdat,ay,asigma,3,ax_init,mycovar,chi2,ff);
  gf.use_scaled=true;

  fact=1.0;
  if (chi2/(n-p)>fact) fact=chi2/(n-p);

  cout << endl;
  cout << "Unscaled O2scl results: " << endl;
  cout << "Parameters: " 
       << x_init[0] << " " << x_init[1] << " " << x_init[2] << endl;
  cout << "Errors    : " << sqrt(mycovar[0][0])*fact << " " 
       << sqrt(mycovar[1][1])*fact << " " 
       << sqrt(mycovar[2][2])*fact << endl;
  cout << "chi^2/dof : " << chi2/(n-p) << endl;
  cout << endl;

  t.test_rel(x_init[0],gsl_vector_get(s->x,0),1.0e-5,"fit0");
  t.test_rel(x_init[1],gsl_vector_get(s->x,1),1.0e-5,"fit1");
  t.test_rel(x_init[2],gsl_vector_get(s->x,2),1.0e-5,"fit2");
  t.test_rel(sqrt(mycovar[0][0]),sqrt(gsl_matrix_get(covar,0,0)),
	     1.0e-3,"err0");
  t.test_rel(sqrt(mycovar[1][1]),sqrt(gsl_matrix_get(covar,1,1)),
	     1.0e-3,"err1");
  t.test_rel(sqrt(mycovar[2][2]),sqrt(gsl_matrix_get(covar,2,2)),
	     1.0e-3,"err2");
  t.test_rel(chi2,chi*chi,1.0e-6,"chi2");
  
  cout << "-----------------------------------------------" << endl;

  //----------------------------------------------------------------
  // Clean up
  //----------------------------------------------------------------

  gsl_multifit_fdfsolver_free(s);
  
  t.report();

  return 0;
}

int expb_f(const gsl_vector *x, void *params, gsl_vector *f) {

  size_t n=((struct data *)params)->n;
  double *y=((struct data *)params)->y;
  double *sigma=((struct data *)params)->sigma;
  
  double A=gsl_vector_get(x,0);
  double lambda=gsl_vector_get(x,1);
  double b=gsl_vector_get(x,2);

  size_t i;

  for (i=0;i<n;i++) {
    double t=i;
    double Yi=A*exp(-lambda*t)+b;
    gsl_vector_set(f,i,(Yi-y[i])/sigma[i]);
  }

  return gsl_success;
}

int expb_df(const gsl_vector *x, void *params, gsl_matrix *J) {
  size_t n=((struct data *)params)->n;
  double *sigma=((struct data *)params)->sigma;

  double A=gsl_vector_get(x,0);
  double lambda=gsl_vector_get(x,1);

  size_t i;
  for (i=0;i<n;i++) {
    double t=i;
    double s=sigma[i];
    double e=exp(-lambda*t);
    gsl_matrix_set(J,i,0,e/s);
    gsl_matrix_set(J,i,1,-t*A*e/s);
    gsl_matrix_set(J,i,2,1/s);
  }
  return gsl_success;
}

int expb_fdf(const gsl_vector *x, void *params, gsl_vector *f, 
	     gsl_matrix *J) {
  expb_f(x,params,f);
  expb_df(x,params,J);
  return gsl_success;
}

double func(size_t np, const ovector_base &p, double x) {
  return p[0]*exp(-p[1]*x)+p[2];
}


