/*
  -------------------------------------------------------------------

  Copyright (C) 2006-2012, Andrew W. Steiner

  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/err_hnd.h>
#include <o2scl/hdf_file.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_hdf;

hdf_file::hdf_file() {
  file=0;
  current=0;
  file_open=false;
}

hdf_file::~hdf_file() {
  if (file_open) {
    H5Fclose(file);
  }
}

int hdf_file::open(std::string fname) {
      
  H5E_BEGIN_TRY
    {
      file=H5Fopen(fname.c_str(),H5F_ACC_RDWR,H5P_DEFAULT);
    } 
  H5E_END_TRY 
    if(file < 0) {
      file=H5Fcreate(fname.c_str(),H5F_ACC_TRUNC,H5P_DEFAULT,H5P_DEFAULT);
    }
  if (file<0) {
    O2SCL_ERR_RET("Open and create file failed in hdf_file::open().",
		  gsl_efailed);
  }
  file_open=true;
  current=file;
  return 0;
}

int hdf_file::close() {
  if (file_open) {
    herr_t status=H5Fclose(file);
    file_open=false;
    current=0;
  } else {
    O2SCL_ERR_RET("No file to close in hdf_file::close().",gsl_einval);
  }
  return 0;
}
    
hid_t hdf_file::get_file_id() {
  if (!file_open) {
    O2SCL_ERR_RET("No file open in hdf_file::get_file_id().",-1);
  }
  return file;
}

hid_t hdf_file::get_current_id() {
  return current;
}

int hdf_file::set_current_id(hid_t cur) {
  current=cur;
  return 0;
}

int hdf_file::setc(std::string name, char c) { 

  hid_t dset, space=0;
  bool space_alloc=false;

  H5E_BEGIN_TRY
    {
      // See if the dataspace already exists first
      dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
    } 
  H5E_END_TRY 
#ifdef O2SCL_NEVER_DEFINED
    {
    }
#endif
      
  // If it doesn't exist, create it
  if (dset<0) {
	
    // Create the dataspace
    hsize_t dims=1;
    space=H5Screate_simple(1,&dims,0);
    dset=H5Dcreate(current,name.c_str(),H5T_STD_I8LE,space,H5P_DEFAULT,
		   H5P_DEFAULT,H5P_DEFAULT);
    if (dset<0) {
      H5Sclose(space);
      O2SCL_ERR2_RET("Failed to create dataspace in ",
		     "hdf_file::setc().",gsl_einval);
    }
    space_alloc=true;

  }

  // Write the data 
  int status=H5Dwrite(dset,H5T_NATIVE_CHAR,H5S_ALL,
		      H5S_ALL,H5P_DEFAULT,&c);
      
  int status2=H5Dclose(dset);
  if (space_alloc) status2=H5Sclose(space);

  // Throw exception if write failed
  if (status<0) {
    O2SCL_ERR2_RET("Failed to write data in ",
		   "hdf_file::setc().",gsl_einval);
  }
      
  return 0;
}
    
int hdf_file::setd(std::string name, double d) { 

  hid_t dset, space=0;
  bool space_alloc=false;

  H5E_BEGIN_TRY
    {
      // See if the dataspace already exists first
      dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
    } 
  H5E_END_TRY 
#ifdef O2SCL_NEVER_DEFINED
    {
    }
#endif
      
  // If it doesn't exist, create it
  if (dset<0) {
	
    // Create the dataspace
    hsize_t dims=1;
    space=H5Screate_simple(1,&dims,0);
    dset=H5Dcreate(current,name.c_str(),H5T_IEEE_F64LE,space,H5P_DEFAULT,
		   H5P_DEFAULT,H5P_DEFAULT);
    if (dset<0) {
      O2SCL_ERR2_RET("Failed to create dataspace in ",
		     "hdf_file::setd().",gsl_einval);
    }
    space_alloc=true;

  }

  // Write the data 
  int status=H5Dwrite(dset,H5T_NATIVE_DOUBLE,H5S_ALL,
		      H5S_ALL,H5P_DEFAULT,&d);
      
  status=H5Dclose(dset);
  if (space_alloc) status=H5Sclose(space);
      
  return 0;
}
    
int hdf_file::setf(std::string name, float f) { 

  hid_t dset, space=0;
  bool space_alloc=false;

  H5E_BEGIN_TRY
    {
      // See if the dataspace already exists first
      dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
    } 
  H5E_END_TRY 
#ifdef O2SCL_NEVER_DEFINED
    {
    }
#endif
      
  // If it doesn't exist, create it
  if (dset<0) {
	
    // Create the dataspace
    hsize_t dims=1;
    space=H5Screate_simple(1,&dims,0);
    dset=H5Dcreate(current,name.c_str(),H5T_IEEE_F32LE,space,H5P_DEFAULT,
		   H5P_DEFAULT,H5P_DEFAULT);
    if (dset<0) {
      O2SCL_ERR2_RET("Failed to create dataspace in ",
		     "hdf_file::setf().",gsl_einval);
    }
    space_alloc=true;

  }

  // Write the data 
  int status=H5Dwrite(dset,H5T_NATIVE_FLOAT,H5S_ALL,
		      H5S_ALL,H5P_DEFAULT,&f);
      
  status=H5Dclose(dset);
  if (space_alloc) status=H5Sclose(space);
      
  return 0;
}
    
int hdf_file::seti(std::string name, int i) { 

  hid_t dset, space=0;
  bool space_alloc=false;

  H5E_BEGIN_TRY
    {
      // See if the dataspace already exists first
      dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
    } 
  H5E_END_TRY 
#ifdef O2SCL_NEVER_DEFINED
    {
    }
#endif
      
  // If it doesn't exist, create it
  if (dset<0) {
	
    // Create the dataspace
    hsize_t dims=1;
    space=H5Screate_simple(1,&dims,0);
    dset=H5Dcreate(current,name.c_str(),H5T_STD_I32LE,space,H5P_DEFAULT,
		   H5P_DEFAULT,H5P_DEFAULT);
    if (dset<0) {
      O2SCL_ERR2_RET("Failed to create dataspace in ",
		     "hdf_file::seti().",gsl_einval);
    }
    space_alloc=true;

  }

  // Write the data 
  int status=H5Dwrite(dset,H5T_NATIVE_INT,H5S_ALL,
		      H5S_ALL,H5P_DEFAULT,&i);
      
  status=H5Dclose(dset);
  if (space_alloc) status=H5Sclose(space);
      
  return 0;
}
    
int hdf_file::sets(std::string name, std::string s) { 
  return setc_arr(name,s.length(),s.c_str());
}
    
int hdf_file::sets_fixed(std::string name, std::string s) { 

  hid_t dset, space=0, filetype=0;
  hsize_t str_size=0;
  
  H5E_BEGIN_TRY
    {
      // See if the dataspace already exists first
      dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
    } 
  H5E_END_TRY 
#ifdef O2SCL_NEVER_DEFINED
    {
    }
#endif
      
  // If it doesn't exist, create it
  if (dset<0) {
	
    // Create the dataspace
    hsize_t dims=1;
    space=H5Screate_simple(1,&dims,0);
    
    // Create the file type
    filetype=H5Tcopy(H5T_C_S1);
    herr_t statusx=H5Tset_size(filetype,s.length()+1);
    
    dset=H5Dcreate(current,name.c_str(),filetype,space,H5P_DEFAULT,
		   H5P_DEFAULT,H5P_DEFAULT);
    if (dset<0) {
      O2SCL_ERR2_RET("Failed to create dataspace in ",
		     "hdf_file::setd().",gsl_einval);
    }
    str_size=s.length()+1;
    
  } else {

    filetype=H5Dget_type(dset);
    str_size=H5Tget_size(filetype);
    
    hsize_t dims[1];
    space=H5Dget_space(dset);
    int ndims=H5Sget_simple_extent_dims(space,dims,0);
    if (ndims!=1 || dims[0]!=1) {
      O2SCL_ERR2_RET("Incorrect dimensions in sets_fixed","",gsl_einval);
    }
    if (str_size<s.length()+1) {
      O2SCL_ERR2_RET("Not enough space in sets_fixed().","",gsl_einval);
    }

  }

  char *c=new char[str_size];
  for(size_t i=0;i<str_size-1;i++) {
    c[i]=s[i];
  }
  c[str_size-1]='\0';

  // Create the memory type
  hid_t memtype=H5Tcopy(H5T_C_S1);
  herr_t status=H5Tset_size(memtype,str_size);

  // Write the data 
  status=H5Dwrite(dset,memtype,H5S_ALL,
		  H5S_ALL,H5P_DEFAULT,c);
  
  status=H5Dclose(dset);
  status=H5Sclose(space);
  status=H5Tclose(memtype);
  status=H5Tclose(filetype);
      
  delete[] c;

  return 0;
}
    
int hdf_file::gets_fixed(std::string name, std::string &s) { 

  hid_t dset, space=0, filetype=0;
  
  H5E_BEGIN_TRY
    {
      // See if the dataspace already exists first
      dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
    } 
  H5E_END_TRY 
#ifdef O2SCL_NEVER_DEFINED
    {
    }
#endif
      
  // If it doesn't exist, create it
  if (dset<0) {
    O2SCL_ERR_RET((((string)"Dataspace named '")+name+
                   "' not found in hdf_file::gets_fixed().").c_str(),
		  gsl_einval);
	
  }

  filetype=H5Dget_type(dset);
  size_t str_size=H5Tget_size(filetype);

  hsize_t dims[1];
  space=H5Dget_space(dset);
  int ndims=H5Sget_simple_extent_dims(space,dims,0);
  if (ndims!=1 || dims[0]!=1) {
    O2SCL_ERR2_RET("Incorrect dimensions in sets_fixed","",gsl_einval);
  }
  
  char *c=new char[str_size];
  
  // Create the memory type
  hid_t memtype=H5Tcopy(H5T_C_S1);
  herr_t status=H5Tset_size(memtype,str_size);
  
  // Write the data 
  herr_t status2=H5Dread(dset,memtype,H5S_ALL,
			 H5S_ALL,H5P_DEFAULT,c);
  
  // Make sure it's terminated properly
  c[str_size-1]='\0';
  s=c;
  
  herr_t status3=H5Dclose(dset);
  status=H5Sclose(space);
  status=H5Tclose(memtype);
  status=H5Tclose(filetype);

  delete[] c;
      
  return 0;
}
    
int hdf_file::gets_def_fixed(std::string name, std::string def,
			     std::string &s) { 

  hid_t dset, space=0, filetype=0;
  
  H5E_BEGIN_TRY
    {
      // See if the dataspace already exists first
      dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
    } 
  H5E_END_TRY 
#ifdef O2SCL_NEVER_DEFINED
    {
    }
#endif
      
  // If it doesn't exist, return the default string
  if (dset<0) {
    s=def;
    return 0;
  }

  filetype=H5Dget_type(dset);
  size_t str_size=H5Tget_size(filetype);

  hsize_t dims[1];
  space=H5Dget_space(dset);
  int ndims=H5Sget_simple_extent_dims(space,dims,0);
  if (ndims!=1 || dims[0]!=1) {
    O2SCL_ERR2_RET("Incorrect dimensions in sets_fixed","",gsl_einval);
  }
  
  char *c=new char[str_size];
  
  // Create the memory type
  hid_t memtype=H5Tcopy(H5T_C_S1);
  herr_t status=H5Tset_size(memtype,str_size);
  
  // Write the data 
  herr_t status2=H5Dread(dset,memtype,H5S_ALL,
			 H5S_ALL,H5P_DEFAULT,c);
  
  // Make sure it's terminated properly
  c[str_size-1]='\0';
  s=c;
  
  herr_t status3=H5Dclose(dset);
  status=H5Sclose(space);
  status=H5Tclose(memtype);
  status=H5Tclose(filetype);

  delete[] c;
      
  return 0;
}
    
int hdf_file::getc(std::string name, char &c) {
      
  // Open the data space
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
  if (dset<0) {
    O2SCL_ERR_RET((((string)"Dataspace named '")+name+
		   "' not found in hdf_file::getc().").c_str(),gsl_einval);
  }

  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_CHAR,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,&c);
  if (status<0) {
    O2SCL_ERR_RET("Could not read dataspace in hdf_file::getc().",
		  gsl_einval);
  }

  status=H5Dclose(dset);

  return 0;
}

int hdf_file::getd(std::string name, double &d) {
      
  // Open the data space
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
  if (dset<0) {
    O2SCL_ERR_RET((((string)"Dataspace named '")+name+
		   "' not found in hdf_file::getd().").c_str(),gsl_einval);
  }

  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_DOUBLE,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,&d);
  if (status<0) {
    O2SCL_ERR_RET("Could not read dataspace in hdf_file::getd().",
		  gsl_einval);
  }

  status=H5Dclose(dset);

  return 0;
}

int hdf_file::getf(std::string name, float &f) {
      
  // Open the data space
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
  if (dset<0) {
    O2SCL_ERR_RET((((string)"Dataspace named '")+name+
		   "' not found in hdf_file::getf().").c_str(),gsl_einval);
  }

  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_FLOAT,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,&f);
  if (status<0) {
    O2SCL_ERR_RET("Could not read dataspace in hdf_file::getf().",
		  gsl_einval);
  }

  status=H5Dclose(dset);

  return 0;
}

int hdf_file::geti(std::string name, int &i) {
      
  // Open the data space
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
  if (dset<0) {
    O2SCL_ERR_RET((((string)"Dataspace named '")+name+
		   "' not found in hdf_file::geti().").c_str(),gsl_einval);
  }

  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_INT,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,&i);
  if (status<0) {
    O2SCL_ERR_RET("Could not read dataspace in hdf_file::geti().",
		  gsl_einval);
  }

  status=H5Dclose(dset);

  return 0;
}

int hdf_file::gets(std::string name, std::string &s) {
      
  // Open the data space
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
  if (dset<0) {
    O2SCL_ERR_RET((((string)"Dataspace named '")+name+
		   "' not found in hdf_file::gets().").c_str(),gsl_einval);
  }

  // Get space requirements, to make sure they coincide
  // with the size specified by the user
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[1];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);
  if (ndims!=1) {
    O2SCL_ERR2_RET("Dataspace has incorrect number of dimensions ",
		   "in hdf_file::gets().",gsl_einval);
  }

  // Allocate memory
  char *c=new char[dims[0]];

  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_CHAR,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,c);
  if (status<0) {
    O2SCL_ERR_RET("Could not read dataspace in hdf_file::gets().",
		  gsl_einval);
  }

  // Close the dataset
  status=H5Dclose(dset);

  // Copy to the string object
  s="";
  for(size_t i=0;i<dims[0];i++) s+=c[i];

  // Delete char memory
  delete[] c;

  return 0;
}

int hdf_file::getc_def(std::string name, char def, char &c) {
   
  hid_t dset=0;
   
  H5E_BEGIN_TRY {
    // See if the dataspace already exists first
    dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
  }
  H5E_END_TRY
#ifdef O2SCL_NEVER_DEFINED
    {
    }
#endif
  
  // Not found, return default
  if (dset<0) {
    c=def;
    return gsl_success;
  }

  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_CHAR,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,&c);
  if (status<0) {
    O2SCL_ERR_RET("Could not read dataspace in hdf_file::getc_def().",
		  gsl_einval);
  }

  status=H5Dclose(dset);

  return 0;
}

int hdf_file::getd_def(std::string name, double def, double &d) {
      
  hid_t dset=0;
   
  H5E_BEGIN_TRY {
    // See if the dataspace already exists first
    dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
  }
  H5E_END_TRY
#ifdef O2SCL_NEVER_DEFINED
    {
    }
#endif
  
  // Not found, return default
  if (dset<0) {
    d=def;
    return gsl_success;
  }

  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_DOUBLE,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,&d);
  if (status<0) {
    O2SCL_ERR_RET("Could not read dataspace in hdf_file::getd_def().",
		  gsl_einval);
  }

  status=H5Dclose(dset);

  return 0;
}

int hdf_file::getf_def(std::string name, float def, float &f) {
      
  hid_t dset=0;
   
  H5E_BEGIN_TRY {
    // See if the dataspace already exists first
    dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
  }
  H5E_END_TRY
#ifdef O2SCL_NEVER_DEFINED
    {
    }
#endif
  
  // Not found, return default
  if (dset<0) {
    f=def;
    return gsl_success;
  }

  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_FLOAT,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,&f);
  if (status<0) {
    O2SCL_ERR_RET("Could not read dataspace in hdf_file::getf_def().",
		  gsl_einval);
  }

  status=H5Dclose(dset);

  return 0;
}

int hdf_file::geti_def(std::string name, int def, int &i) {
      
  hid_t dset=0;
   
  H5E_BEGIN_TRY {
    // See if the dataspace already exists first
    dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
  }
  H5E_END_TRY
#ifdef O2SCL_NEVER_DEFINED
    {
    }
#endif
  
  // Not found, return default
  if (dset<0) {
    i=def;
    return gsl_success;
  }

  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_INT,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,&i);
  if (status<0) {
    O2SCL_ERR_RET("Could not read dataspace in hdf_file::geti_def().",
		  gsl_einval);
  }

  status=H5Dclose(dset);

  return 0;
}

int hdf_file::gets_def(std::string name, std::string def, std::string &s) {
      
  // Open the data space
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
  if (dset<0) {
    s=def;
    return gsl_success;
  }

  // Get space requirements, to make sure they coincide
  // with the size specified by the user
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[1];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);
  if (ndims!=1) {
    O2SCL_ERR2_RET("Dataspace has incorrect number of dimensions ",
		   "in hdf_file::gets().",gsl_einval);
  }

  // Allocate memory
  char *c=new char[dims[0]];

  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_CHAR,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,c);
  if (status<0) {
    O2SCL_ERR_RET("Could not read dataspace in hdf_file::gets().",
		  gsl_einval);
  }

  // Close the dataset
  status=H5Dclose(dset);

  // Copy to the string object
  s="";
  for(size_t i=0;i<dims[0];i++) s+=c[i];

  // Delete char memory
  delete[] c;

  return 0;
}

int hdf_file::setc_arr_fixed(std::string name, size_t n, const char *c) { 

  hid_t dset, space;
  bool space_alloc=false;

  H5E_BEGIN_TRY
    {
      // See if the dataspace already exists first
      dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
    } 
  H5E_END_TRY 
#ifdef O2SCL_NEVER_DEFINED
    {
    }
#endif
      
  // If it doesn't exist, create it
  if (dset<0) {
	
    // Create the dataspace
    hsize_t dims=n;
    space=H5Screate_simple(1,&dims,0);
    dset=H5Dcreate(current,name.c_str(),H5T_STD_I8LE,space,H5P_DEFAULT,
		   H5P_DEFAULT,H5P_DEFAULT);
    space_alloc=true;

  } else {
    
    // Get space requirements, to make sure they coincide
    // with the size specified by the user
    space=H5Dget_space(dset);  
    hsize_t dims[1];
    int ndims=H5Sget_simple_extent_dims(space,dims,0);
    if (ndims!=1 || dims[0]!=n) {
      O2SCL_ERR2_RET("Incompatible dataspace size or dimensions ",
		     "in hdf_file::setc_arr_fixed().",gsl_einval);
    }
    
  }

  // Write the data 
  int status=H5Dwrite(dset,H5T_NATIVE_CHAR,H5S_ALL,
		      H5S_ALL,H5P_DEFAULT,c);
      
  status=H5Dclose(dset);
  if (space_alloc) status=H5Sclose(space);
      
  return 0;
}

int hdf_file::setd_arr_fixed(std::string name, size_t n, const double *d) { 
  
  hid_t dset, space;
  bool space_alloc=false;

  H5E_BEGIN_TRY
    {
      // See if the dataspace already exists first
      dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
    } 
  H5E_END_TRY 
#ifdef O2SCL_NEVER_DEFINED
    {
    }
#endif
      
  // If it doesn't exist, create it
  if (dset<0) {
	
    // Create the dataspace
    hsize_t dims=n;
    space=H5Screate_simple(1,&dims,0);
    dset=H5Dcreate(current,name.c_str(),H5T_IEEE_F64LE,space,H5P_DEFAULT,
		   H5P_DEFAULT,H5P_DEFAULT);
    space_alloc=true;

  } else {
    
    // Get space requirements, to make sure they coincide
    // with the size specified by the user
    space=H5Dget_space(dset);  
    hsize_t dims[1];
    int ndims=H5Sget_simple_extent_dims(space,dims,0);
    if (ndims!=1 || dims[0]!=n) {
      O2SCL_ERR2_RET("Incompatible dataspace size or dimensions ",
		     "in hdf_file::setd_arr_fixed().",gsl_einval);
    }

  }

  // Write the data 
  int status=H5Dwrite(dset,H5T_NATIVE_DOUBLE,H5S_ALL,
		      H5S_ALL,H5P_DEFAULT,d);
      
  status=H5Dclose(dset);
  if (space_alloc) status=H5Sclose(space);
      
  return 0;
}

int hdf_file::setf_arr_fixed(std::string name, size_t n, const float *f) { 

  hid_t dset, space;
  bool space_alloc=false;

  H5E_BEGIN_TRY
    {
      // See if the dataspace already exists first
      dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
    } 
  H5E_END_TRY 
#ifdef O2SCL_NEVER_DEFINED
    {
    }
#endif
      
  // If it doesn't exist, create it
  if (dset<0) {
	
    // Create the dataspace
    hsize_t dims=n;
    space=H5Screate_simple(1,&dims,0);
    dset=H5Dcreate(current,name.c_str(),H5T_IEEE_F32LE,space,H5P_DEFAULT,
		   H5P_DEFAULT,H5P_DEFAULT);
    space_alloc=true;

  } else {
    
    // Get space requirements, to make sure they coincide
    // with the size specified by the user
    space=H5Dget_space(dset);  
    hsize_t dims[1];
    int ndims=H5Sget_simple_extent_dims(space,dims,0);
    if (ndims!=1 || dims[0]!=n) {
      O2SCL_ERR2_RET("Incompatible dataspace size or dimensions ",
		     "in hdf_file::setf_arr_fixed().",gsl_einval);
    }
    
  }

  // Write the data 
  int status=H5Dwrite(dset,H5T_NATIVE_FLOAT,H5S_ALL,
		      H5S_ALL,H5P_DEFAULT,f);
      
  status=H5Dclose(dset);
  if (space_alloc) status=H5Sclose(space);
      
  return 0;
}

int hdf_file::seti_arr_fixed(std::string name, size_t n, const int *i) { 

  hid_t dset, space;
  bool space_alloc=false;

  H5E_BEGIN_TRY
    {
      // See if the dataspace already exists first
      dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
    } 
  H5E_END_TRY 
#ifdef O2SCL_NEVER_DEFINED
    {
    }
#endif
      
  // If it doesn't exist, create it
  if (dset<0) {
	
    // Create the dataspace
    hsize_t dims=n;
    space=H5Screate_simple(1,&dims,0);
    dset=H5Dcreate(current,name.c_str(),H5T_STD_I32LE,space,H5P_DEFAULT,
		   H5P_DEFAULT,H5P_DEFAULT);
    space_alloc=true;

  } else {
    
    // Get space requirements, to make sure they coincide
    // with the size specified by the user
    space=H5Dget_space(dset);  
    hsize_t dims[1];
    int ndims=H5Sget_simple_extent_dims(space,dims,0);
    if (ndims!=1 || dims[0]!=n) {
      O2SCL_ERR2_RET("Incompatible dataspace size or dimensions ",
		     "in hdf_file::seti_arr_fixed().",gsl_einval);
    }
    
  }

  // Write the data 
  int status=H5Dwrite(dset,H5T_NATIVE_INT,H5S_ALL,
		      H5S_ALL,H5P_DEFAULT,i);
      
  status=H5Dclose(dset);
  if (space_alloc) status=H5Sclose(space);
      
  return 0;
}
    
int hdf_file::setc_arr(std::string name, size_t n, const char *c) { 
  
  hid_t dset, space, dcpl=0;
  bool chunk_alloc=false;

  H5E_BEGIN_TRY
    {
      // See if the dataspace already exists first
      dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
    } 
  H5E_END_TRY 
#ifdef O2SCL_NEVER_DEFINED
    {
    }
#endif
      
  // If it doesn't exist, create it
  if (dset<0) {
    
    // Create the dataspace
    hsize_t dims=n;
    hsize_t max=H5S_UNLIMITED;
    space=H5Screate_simple(1,&dims,&max);

    // Set chunk with size determined by def_chunk()
    dcpl=H5Pcreate(H5P_DATASET_CREATE);
    hsize_t chunk=def_chunk(n);
    int status2=H5Pset_chunk(dcpl,1,&chunk);

    // Create the dataset
    dset=H5Dcreate(current,name.c_str(),H5T_STD_I8LE,space,H5P_DEFAULT,
		   dcpl,H5P_DEFAULT);
    chunk_alloc=true;

  } else {
    
    // Get current dimensions
    space=H5Dget_space(dset);  
    hsize_t dims;
    int ndims=H5Sget_simple_extent_dims(space,&dims,0);

    // Set error if this dataset is more than 1-dimensional
    if (ndims!=1) {
      O2SCL_ERR2_RET("Tried to set a multidimensional dataset with an ",
		     "array in hdf_file::setc_arr().",gsl_einval);
    }

    // If necessary, extend the dataset
    if (n!=dims) {
      hsize_t new_dims=n;
      int status3=H5Dset_extent(dset,&new_dims);
    }
    
  }

  // Write the data 
  int status=H5Dwrite(dset,H5T_NATIVE_CHAR,H5S_ALL,
		      H5S_ALL,H5P_DEFAULT,c);
  
  status=H5Dclose(dset);
  status=H5Sclose(space);
  if (chunk_alloc) {
    status=H5Pclose(dcpl);
  }
      
  return 0;
}

int hdf_file::setd_arr(std::string name, size_t n, const double *d) { 
  
  hid_t dset, space, dcpl=0;
  bool chunk_alloc=false;

  H5E_BEGIN_TRY
    {
      // See if the dataspace already exists first
      dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
    } 
  H5E_END_TRY 
#ifdef O2SCL_NEVER_DEFINED
    {
    }
#endif
      
  // If it doesn't exist, create it
  if (dset<0) {
    
    // Create the dataspace
    hsize_t dims=n;
    hsize_t max=H5S_UNLIMITED;
    space=H5Screate_simple(1,&dims,&max);

    // Set chunk with size determined by def_chunk()
    dcpl=H5Pcreate(H5P_DATASET_CREATE);
    hsize_t chunk=def_chunk(n);
    int status2=H5Pset_chunk(dcpl,1,&chunk);

    // Create the dataset
    dset=H5Dcreate(current,name.c_str(),H5T_IEEE_F64LE,space,H5P_DEFAULT,
		   dcpl,H5P_DEFAULT);
    chunk_alloc=true;

  } else {
    
    // Get current dimensions
    space=H5Dget_space(dset);  
    hsize_t dims;
    int ndims=H5Sget_simple_extent_dims(space,&dims,0);

    // Set error if this dataset is more than 1-dimensional
    if (ndims!=1) {
      O2SCL_ERR2_RET("Tried to set a multidimensional dataset with an ",
		     "array in hdf_file::setd_arr().",gsl_einval);
    }

    // If necessary, extend the dataset
    if (n!=dims) {
      hsize_t new_dims=n;
      int status3=H5Dset_extent(dset,&new_dims);
    }
    
  }

  // Write the data 
  int status=H5Dwrite(dset,H5T_NATIVE_DOUBLE,H5S_ALL,
		      H5S_ALL,H5P_DEFAULT,d);
  
  status=H5Dclose(dset);
  status=H5Sclose(space);
  if (chunk_alloc) {
    status=H5Pclose(dcpl);
  }
      
  return 0;
}

int hdf_file::setf_arr(std::string name, size_t n, const float *f) { 
  
  hid_t dset, space, dcpl=0;
  bool chunk_alloc=false;

  H5E_BEGIN_TRY
    {
      // See if the dataspace already exists first
      dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
    } 
  H5E_END_TRY 
#ifdef O2SCL_NEVER_DEFINED
    {
    }
#endif
      
  // If it doesn't exist, create it
  if (dset<0) {
    
    // Create the dataspace
    hsize_t dims=n;
    hsize_t max=H5S_UNLIMITED;
    space=H5Screate_simple(1,&dims,&max);

    // Set chunk with size determined by def_chunk()
    dcpl=H5Pcreate(H5P_DATASET_CREATE);
    hsize_t chunk=def_chunk(n);
    int status2=H5Pset_chunk(dcpl,1,&chunk);

    // Create the dataset
    dset=H5Dcreate(current,name.c_str(),H5T_IEEE_F32LE,space,H5P_DEFAULT,
		   dcpl,H5P_DEFAULT);
    chunk_alloc=true;

  } else {
    
    // Get current dimensions
    space=H5Dget_space(dset);  
    hsize_t dims;
    int ndims=H5Sget_simple_extent_dims(space,&dims,0);

    // Set error if this dataset is more than 1-dimensional
    if (ndims!=1) {
      O2SCL_ERR2_RET("Tried to set a multidimensional dataset with an ",
		     "array in hdf_file::setf_arr().",gsl_einval);
    }

    // If necessary, extend the dataset
    if (n!=dims) {
      hsize_t new_dims=n;
      int status3=H5Dset_extent(dset,&new_dims);
    }
    
  }

  // Write the data 
  int status=H5Dwrite(dset,H5T_NATIVE_FLOAT,H5S_ALL,
		      H5S_ALL,H5P_DEFAULT,f);
  
  status=H5Dclose(dset);
  status=H5Sclose(space);
  if (chunk_alloc) {
    status=H5Pclose(dcpl);
  }
      
  return 0;
}

int hdf_file::seti_arr(std::string name, size_t n, const int *i) { 
  
  hid_t dset, space, dcpl=0;
  bool chunk_alloc=false;

  H5E_BEGIN_TRY
    {
      // See if the dataspace already exists first
      dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
    } 
  H5E_END_TRY 
#ifdef O2SCL_NEVER_DEFINED
    {
    }
#endif
      
  // If it doesn't exist, create it
  if (dset<0) {
    
    // Create the dataspace
    hsize_t dims=n;
    hsize_t max=H5S_UNLIMITED;
    space=H5Screate_simple(1,&dims,&max);

    // Set chunk with size determined by def_chunk()
    dcpl=H5Pcreate(H5P_DATASET_CREATE);
    hsize_t chunk=def_chunk(n);
    int status2=H5Pset_chunk(dcpl,1,&chunk);

    // Create the dataset
    dset=H5Dcreate(current,name.c_str(),H5T_STD_I32LE,space,H5P_DEFAULT,
		   dcpl,H5P_DEFAULT);
    chunk_alloc=true;

  } else {
    
    // Get current dimensions
    space=H5Dget_space(dset);  
    hsize_t dims;
    int ndims=H5Sget_simple_extent_dims(space,&dims,0);

    // Set error if this dataset is more than 1-dimensional
    if (ndims!=1) {
      O2SCL_ERR2_RET("Tried to set a multidimensional dataset with an ",
		     "array in hdf_file::seti_arr().",gsl_einval);
    }

    // If necessary, extend the dataset
    if (n!=dims) {
      hsize_t new_dims=n;
      int status3=H5Dset_extent(dset,&new_dims);
    }
    
  }

  // Write the data 
  int status=H5Dwrite(dset,H5T_NATIVE_INT,H5S_ALL,
		      H5S_ALL,H5P_DEFAULT,i);
  
  status=H5Dclose(dset);
  status=H5Sclose(space);
  if (chunk_alloc) {
    status=H5Pclose(dcpl);
  }
      
  return 0;
}

int hdf_file::getc_arr(std::string name, size_t n, char *c) {
      
  // See if the dataspace already exists first
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);

  // Get space requirements, to make sure they coincide
  // with the size specified by the user
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[1];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);
  
  if (dims[0]!=n) {
    O2SCL_ERR_RET("Incompatible size in hdf_file::getc_arr().",
		  gsl_einval);
  }

  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_CHAR,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,c);

  status=H5Dclose(dset);

  return 0;
}

int hdf_file::getd_arr(std::string name, size_t n, double *d) {
      
  // See if the dataspace already exists first
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);

  // Get space requirements, to make sure they coincide
  // with the size specified by the user
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[1];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);
  
  if (dims[0]!=n) {
    string str="Asked for size "+itos(n)+" but file has size "+
      itos(dims[0])+" in hdf_file::getd_arr().";
    O2SCL_ERR_RET(str.c_str(),gsl_einval);
  }

  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_DOUBLE,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,d);

  status=H5Dclose(dset);

  return 0;
}

int hdf_file::getf_arr(std::string name, size_t n, float *f) {
      
  // See if the dataspace already exists first
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);

  // Get space requirements, to make sure they coincide
  // with the size specified by the user
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[1];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);
  
  if (dims[0]!=n) {
    O2SCL_ERR_RET("Incompatible size in hdf_file::getc_arr().",
		  gsl_einval);
  }

  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_FLOAT,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,f);

  status=H5Dclose(dset);

  return 0;
}

int hdf_file::geti_arr(std::string name, size_t n, int *i) {
      
  // See if the dataspace already exists first
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);

  // Get space requirements, to make sure they coincide
  // with the size specified by the user
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[1];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);
  
  if (dims[0]!=n) {
    O2SCL_ERR_RET("Incompatible size in hdf_file::getc_arr().",
		  gsl_einval);
  }

  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_INT,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,i);

  status=H5Dclose(dset);

  return 0;
}

int hdf_file::getc_arr_alloc(std::string name, size_t &n, char *c) {
      
  // See if the dataspace already exists first
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);

  // Get space requirements, to allocate memory
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[1];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);
  
  n=dims[0];
  c=new char[n];

  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_CHAR,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,c);

  status=H5Dclose(dset);

  return 0;
}

int hdf_file::getd_arr_alloc(std::string name, size_t &n, double *d) {
      
  // See if the dataspace already exists first
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);

  // Get space requirements, to allocate memory
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[1];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);
  
  n=dims[0];
  d=new double[n];

  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_DOUBLE,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,d);

  status=H5Dclose(dset);

  return 0;
}

int hdf_file::getf_arr_alloc(std::string name, size_t &n, float *f) {
      
  // See if the dataspace already exists first
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);

  // Get space requirements, to allocate memory
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[1];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);
  
  n=dims[0];
  f=new float[n];

  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_FLOAT,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,f);

  status=H5Dclose(dset);

  return 0;
}

int hdf_file::geti_arr_alloc(std::string name, size_t &n, int *i) {
      
  // See if the dataspace already exists first
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);

  // Get space requirements, to allocate memory
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[1];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);
  
  n=dims[0];
  i=new int[n];

  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_INT,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,i);

  status=H5Dclose(dset);

  return 0;
}

hid_t hdf_file::open_group(std::string path) {
  if (!file_open) {
    O2SCL_ERR_RET("File not opened in hdf_file::open_group().",
		  gsl_einval);
  }
  if (current<=0) {
    O2SCL_ERR_RET("Invalid current HDF5 id in hdf_file::open_group().",
		  gsl_einval);
  }
  hid_t group;
  H5E_BEGIN_TRY
    {
      group=H5Gopen(current,path.c_str(),H5P_DEFAULT);
    }
  H5E_END_TRY 
    if (group<0) {
      group=H5Gcreate(current,path.c_str(),H5P_DEFAULT,
		      H5P_DEFAULT,H5P_DEFAULT);
    }
  if (group<0) {
    O2SCL_ERR2_RET("Failed to open or create group in ",
		   "hdf_file::open_group().",gsl_einval);
  }
  return group;
}

hid_t hdf_file::open_group(hid_t init_id, std::string path) {
  hid_t group;
  H5E_BEGIN_TRY
    {
      group=H5Gopen(init_id,path.c_str(),H5P_DEFAULT);
    }
  H5E_END_TRY 
    if (group<0) {
      group=H5Gcreate(init_id,path.c_str(),H5P_DEFAULT,
		      H5P_DEFAULT,H5P_DEFAULT);
    }
  if (group<0) {
    O2SCL_ERR2_RET("Failed to open or create group in ",
		   "hdf_file::open_group().",gsl_einval);
  }
  return group;
}

int hdf_file::setd_vec(std::string name, const std::vector<double> &v) {
  return setd_arr(name,v.size(),&v[0]);
}

int hdf_file::setd_vec(std::string name, const ovector_const_view &v) {
  if (v.stride()==1) {
    return setd_arr(name,v.size(),&v[0]);
  } 
  
  double *d=new double[v.size()];
  for(size_t j=0;j<v.size();j++) d[j]=v[j];
  int ret=setd_arr(name,v.size(),d);
  delete[] d;

  return ret;
}

int hdf_file::setd_vec(std::string name, const uvector_const_view &v) {
  return setd_arr(name,v.size(),&v[0]);
}

int hdf_file::seti_vec(std::string name, const std::vector<int> &v) {
  return seti_arr(name,v.size(),&v[0]);
}

int hdf_file::seti_vec(std::string name, const ovector_int_const_view &v) {
  if (v.stride()==1) {
    return seti_arr(name,v.size(),&v[0]);
  } 
  
  int *d=new int[v.size()];
  for(size_t j=0;j<v.size();j++) d[j]=v[j];
  int ret=seti_arr(name,v.size(),d);
  delete[] d;

  return ret;
}

int hdf_file::seti_vec(std::string name, const uvector_int_const_view &v) {
  return seti_arr(name,v.size(),&v[0]);
}

int hdf_file::getd_vec(std::string name, std::vector<double> &v) {
      
  // See if the dataspace already exists first
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
  
  // Get space requirements, to make sure they coincide
  // with the size specified by the user
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[1];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);
  
  herr_t status;

  if (dims[0]>0) {

  v.reserve(dims[0]);

  // Read the data
  status=H5Dread(dset,H5T_NATIVE_DOUBLE,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,&v[0]);

  }
  
  status=H5Dclose(dset);

  return 0;
}

int hdf_file::getd_vec(std::string name, ovector &v) {
      
  // See if the dataspace already exists first
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
  
  // Get space requirements, to make sure they coincide
  // with the size specified by the user
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[1];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);

  herr_t status;

  if (dims[0]>0) {
    v.allocate(dims[0]);
    
    // Read the data
    status=H5Dread(dset,H5T_NATIVE_DOUBLE,H5S_ALL,H5S_ALL,
		       H5P_DEFAULT,&v[0]);
  }

  status=H5Dclose(dset);

  return 0;
}

int hdf_file::getd_vec(std::string name, uvector &v) {
      
  // See if the dataspace already exists first
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
  
  // Get space requirements, to make sure they coincide
  // with the size specified by the user
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[1];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);

  herr_t status;

  if (dims[0]>0) {
    
    v.allocate(dims[0]);
    
    // Read the data
    status=H5Dread(dset,H5T_NATIVE_DOUBLE,H5S_ALL,H5S_ALL,
		       H5P_DEFAULT,&v[0]);

  }
  
  status=H5Dclose(dset);

  return 0;
}

int hdf_file::geti_vec(std::string name, std::vector<int> &v) {
      
  // See if the dataspace already exists first
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
  
  // Get space requirements, to make sure they coincide
  // with the size specified by the user
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[1];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);
  
  herr_t status;

  if (dims[0]>0) {
    
    v.reserve(dims[0]);
    
    // Read the data
    status=H5Dread(dset,H5T_NATIVE_INT,H5S_ALL,H5S_ALL,
		       H5P_DEFAULT,&v[0]);

  }
  
  status=H5Dclose(dset);

  return 0;
}

int hdf_file::geti_vec(std::string name, ovector_int &v) {
      
  // See if the dataspace already exists first
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
  
  // Get space requirements, to make sure they coincide
  // with the size specified by the user
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[1];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);
  
  herr_t status;

  if (dims[0]>0) {
    
    v.allocate(dims[0]);
    
    // Read the data
    status=H5Dread(dset,H5T_NATIVE_INT,H5S_ALL,H5S_ALL,
		       H5P_DEFAULT,&v[0]);
    
  }

  status=H5Dclose(dset);
  
  return 0;
}

int hdf_file::geti_vec(std::string name, uvector_int &v) {
      
  // See if the dataspace already exists first
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
  
  // Get space requirements, to make sure they coincide
  // with the size specified by the user
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[1];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);
  
  herr_t status;

  if (dims[0]>0) {
    
    v.allocate(dims[0]);
    
    // Read the data
    status=H5Dread(dset,H5T_NATIVE_INT,H5S_ALL,H5S_ALL,
		   H5P_DEFAULT,&v[0]);
    
  }
  
  status=H5Dclose(dset);

  return 0;
}

int hdf_file::gets_arr(std::string name, std::vector<std::string> &s) {
		  
  int *ip;
  char *cp;
  int nc, nw;

  // Open the group
  hid_t top=get_current_id();
  hid_t group=open_group(name);
  set_current_id(group);
  
  string o2t;
  gets_fixed("o2scl_type",o2t);
  if (o2t!="string[]") {
    set_current_id(top);
    O2SCL_ERR2_RET("The specified name does not refer to data which ",
		   "can be read by O2scl in hdf_file::gets_arr().",
		   gsl_efailed);
  }

  // Get number of words
  geti("nw",nw);

  if (nw>0) {
    
    // Get number of characters
    geti("nc",nc);
    
    if (nc>0) {
      
      // Allocate space for ip and cp
      ip=new int[nw];
      cp=new char[nc];
      
      // Get counter and data
      geti_arr("counter",nw,ip);
      getc_arr("data",nc,cp);
      
      // Copy the data over
      size_t ix=0;
      for(int i=0;i<nw;i++) {
	string tmp;
	for(int j=0;j<ip[i];j++) {
	  tmp+=cp[ix];
	  ix++;
	}
	s.push_back(tmp);
      }

      // Free allocations
      delete[] cp;
      delete[] ip;
      
    }
    
  }

  close_group(group);

  // Return file location
  set_current_id(top);

  return 0;
}

int hdf_file::sets_arr(std::string name, std::vector<std::string> &s) {

  // Create the group
  hid_t top=current;
  hid_t group=open_group(name);
  set_current_id(group);
  
  sets_fixed("o2scl_type","string[]");
  
  seti("nw",s.size());
  
  if (s.size()>0) {
    
    // Compute total length
    int nc=0;
    for(size_t i=0;i<s.size();i++) nc+=s[i].length();

    seti("nc",nc);

    if (nc>0) {
      
      // Copy strings over to a contiguous char *, and count lengths
      int *ip=new int[s.size()];
      char *cp=new char[nc];
      size_t ix=0;
      for(size_t i=0;i<s.size();i++) {
	ip[i]=s[i].length();
	for(size_t j=0;j<s[i].size();j++) {
	  cp[ix]=s[i][j];
	  ix++;
	}
      }
      
      // Set data
      seti_arr("counter",s.size(),ip);
      setc_arr("data",nc,cp);

      // Free allocations
      delete[] cp;
      delete[] ip;

    }

  }

  // Close the group
  close_group(group);
  
  // Return file location
  set_current_id(top);

  return 0;
}

int hdf_file::setd_mat(std::string name, const omatrix_const_view &m) {

  omatrix m2;

  bool mem_connected=true;
  if (m.tda()!=m.cols()) {
    mem_connected=false;
    m2=m;
  }
  
  hid_t dset, space, dcpl=0;
  bool chunk_alloc=false;

  H5E_BEGIN_TRY
    {
      // See if the dataspace already exists first
      dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
    } 
  H5E_END_TRY 
#ifdef O2SCL_NEVER_DEFINED
    {
    }
#endif
      
  // If it doesn't exist, create it
  if (dset<0) {
    
    // Create the dataspace
    hsize_t dims[2]={m.rows(),m.cols()};
    hsize_t max[2]={H5S_UNLIMITED,H5S_UNLIMITED};
    space=H5Screate_simple(2,dims,max);

    // Set chunk with size determined by def_chunk()
    dcpl=H5Pcreate(H5P_DATASET_CREATE);
    hsize_t chunk[2]={def_chunk(m.rows()),def_chunk(m.cols())};
    int status2=H5Pset_chunk(dcpl,2,chunk);

    // Create the dataset
    dset=H5Dcreate(current,name.c_str(),H5T_IEEE_F64LE,space,H5P_DEFAULT,
		   dcpl,H5P_DEFAULT);
    chunk_alloc=true;

  } else {
    
    // Get current dimensions
    space=H5Dget_space(dset);  
    hsize_t dims[2];
    int ndims=H5Sget_simple_extent_dims(space,dims,0);

    // Set error if this dataset is more than 1-dimensional
    if (ndims!=2) {
      O2SCL_ERR2_RET("Tried to set a non-matrix dataset with a ",
		     "matrix in hdf_file::setd_mat().",gsl_einval);
    }

    // If necessary, extend the dataset
    if (m.rows()!=dims[0] || m.cols()!=dims[1]) {
      hsize_t new_dims[2]={m.rows(),m.cols()};
      int status3=H5Dset_extent(dset,new_dims);
    }
    
  }

  // Write the data 
  int status;
  if (mem_connected) {
    status=H5Dwrite(dset,H5T_NATIVE_DOUBLE,H5S_ALL,
		    H5S_ALL,H5P_DEFAULT,&(m[0][0]));
  } else {
    status=H5Dwrite(dset,H5T_NATIVE_DOUBLE,H5S_ALL,
		    H5S_ALL,H5P_DEFAULT,&(m2[0][0]));
  }
  
  status=H5Dclose(dset);
  status=H5Sclose(space);
  if (chunk_alloc) {
    status=H5Pclose(dcpl);
  }
      
  return 0;
}

int hdf_file::setd_mat(std::string name, const umatrix_const_view &m) {
  
  hid_t dset, space, dcpl=0;
  bool chunk_alloc=false;

  H5E_BEGIN_TRY
    {
      // See if the dataspace already exists first
      dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
    } 
  H5E_END_TRY 
#ifdef O2SCL_NEVER_DEFINED
    {
    }
#endif
      
  // If it doesn't exist, create it
  if (dset<0) {
    
    // Create the dataspace
    hsize_t dims[2]={m.rows(),m.cols()};
    hsize_t max[2]={H5S_UNLIMITED,H5S_UNLIMITED};
    space=H5Screate_simple(2,dims,max);

    // Set chunk with size determined by def_chunk()
    dcpl=H5Pcreate(H5P_DATASET_CREATE);
    hsize_t chunk[2]={def_chunk(m.rows()),def_chunk(m.cols())};
    int status2=H5Pset_chunk(dcpl,2,chunk);

    // Create the dataset
    dset=H5Dcreate(current,name.c_str(),H5T_IEEE_F64LE,space,H5P_DEFAULT,
		   dcpl,H5P_DEFAULT);
    chunk_alloc=true;

  } else {
    
    // Get current dimensions
    space=H5Dget_space(dset);  
    hsize_t dims[2];
    int ndims=H5Sget_simple_extent_dims(space,dims,0);

    // Set error if this dataset is more than 1-dimensional
    if (ndims!=2) {
      O2SCL_ERR2_RET("Tried to set a non-matrix dataset with a ",
		     "matrix in hdf_file::setd_mat().",gsl_einval);
    }

    // If necessary, extend the dataset
    if (m.rows()!=dims[0] || m.cols()!=dims[1]) {
      hsize_t new_dims[2]={m.rows(),m.cols()};
      int status3=H5Dset_extent(dset,new_dims);
    }
    
  }

  // Write the data 
  int status=H5Dwrite(dset,H5T_NATIVE_DOUBLE,H5S_ALL,
		      H5S_ALL,H5P_DEFAULT,&(m[0][0]));
  
  status=H5Dclose(dset);
  status=H5Sclose(space);
  if (chunk_alloc) {
    status=H5Pclose(dcpl);
  }
      
  return 0;
}

int hdf_file::seti_mat(std::string name, const omatrix_int_const_view &m) {

  omatrix_int m2;

  bool mem_connected=true;
  if (m.tda()!=m.cols()) {
    mem_connected=false;
    m2=m;
  }
  
  hid_t dset, space, dcpl=0;
  bool chunk_alloc=false;

  H5E_BEGIN_TRY
    {
      // See if the dataspace already exists first
      dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
    } 
  H5E_END_TRY 
#ifdef O2SCL_NEVER_DEFINED
    {
    }
#endif
      
  // If it doesn't exist, create it
  if (dset<0) {
    
    // Create the dataspace
    hsize_t dims[2]={m.rows(),m.cols()};
    hsize_t max[2]={H5S_UNLIMITED,H5S_UNLIMITED};
    space=H5Screate_simple(2,dims,max);

    // Set chunk with size determined by def_chunk()
    dcpl=H5Pcreate(H5P_DATASET_CREATE);
    hsize_t chunk[2]={def_chunk(m.rows()),def_chunk(m.cols())};
    int status2=H5Pset_chunk(dcpl,2,chunk);

    // Create the dataset
    dset=H5Dcreate(current,name.c_str(),H5T_STD_I32LE,space,H5P_DEFAULT,
		   dcpl,H5P_DEFAULT);
    chunk_alloc=true;

  } else {
    
    // Get current dimensions
    space=H5Dget_space(dset);  
    hsize_t dims[2];
    int ndims=H5Sget_simple_extent_dims(space,dims,0);

    // Set error if this dataset is more than 1-dimensional
    if (ndims!=2) {
      O2SCL_ERR2_RET("Tried to set a non-matrix dataset with a ",
		     "matrix in hdf_file::seti_mat().",gsl_einval);
    }

    // If necessary, extend the dataset
    if (m.rows()!=dims[0] || m.cols()!=dims[1]) {
      hsize_t new_dims[2]={m.rows(),m.cols()};
      int status3=H5Dset_extent(dset,new_dims);
    }
    
  }

  // Write the data 
  int status;
  if (mem_connected) {
    status=H5Dwrite(dset,H5T_NATIVE_INT,H5S_ALL,
		    H5S_ALL,H5P_DEFAULT,&(m[0][0]));
  } else {
    status=H5Dwrite(dset,H5T_NATIVE_INT,H5S_ALL,
		    H5S_ALL,H5P_DEFAULT,&(m2[0][0]));
  }
  
  status=H5Dclose(dset);
  status=H5Sclose(space);
  if (chunk_alloc) {
    status=H5Pclose(dcpl);
  }
      
  return 0;
}

int hdf_file::seti_mat(std::string name, const umatrix_int_const_view &m) {
  
  hid_t dset, space, dcpl=0;
  bool chunk_alloc=false;

  H5E_BEGIN_TRY
    {
      // See if the dataspace already exists first
      dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
    } 
  H5E_END_TRY 
#ifdef O2SCL_NEVER_DEFINED
    {
    }
#endif
      
  // If it doesn't exist, create it
  if (dset<0) {
    
    // Create the dataspace
    hsize_t dims[2]={m.rows(),m.cols()};
    hsize_t max[2]={H5S_UNLIMITED,H5S_UNLIMITED};
    space=H5Screate_simple(2,dims,max);

    // Set chunk with size determined by def_chunk()
    dcpl=H5Pcreate(H5P_DATASET_CREATE);
    hsize_t chunk[2]={def_chunk(m.rows()),def_chunk(m.cols())};
    int status2=H5Pset_chunk(dcpl,2,chunk);

    // Create the dataset
    dset=H5Dcreate(current,name.c_str(),H5T_STD_I32LE,space,H5P_DEFAULT,
		   dcpl,H5P_DEFAULT);
    chunk_alloc=true;

  } else {
    
    // Get current dimensions
    space=H5Dget_space(dset);  
    hsize_t dims[2];
    int ndims=H5Sget_simple_extent_dims(space,dims,0);

    // Set error if this dataset is more than 1-dimensional
    if (ndims!=2) {
      O2SCL_ERR2_RET("Tried to set a non-matrix dataset with a ",
		     "matrix in hdf_file::seti_mat().",gsl_einval);
    }

    // If necessary, extend the dataset
    if (m.rows()!=dims[0] || m.cols()!=dims[1]) {
      hsize_t new_dims[2]={m.rows(),m.cols()};
      int status3=H5Dset_extent(dset,new_dims);
    }
    
  }

  // Write the data 
  int status=H5Dwrite(dset,H5T_NATIVE_INT,H5S_ALL,
		      H5S_ALL,H5P_DEFAULT,&(m[0][0]));
  
  status=H5Dclose(dset);
  status=H5Sclose(space);
  if (chunk_alloc) {
    status=H5Pclose(dcpl);
  }
      
  return 0;
}

int hdf_file::getd_mat(std::string name, omatrix &m) {
      
  // See if the dataspace already exists first
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
  
  // Get space requirements, to make sure they coincide
  // with the size specified by the user
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[2];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);

  if (ndims!=2) {
    O2SCL_ERR2_RET("Dimensions of dataspace do not match ",
		   "a matrix in hdf_file::getd_mat().",gsl_efailed);
  }
  
  m.allocate(dims[0],dims[1]);

  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_DOUBLE,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,&(m[0][0]));
  
  status=H5Dclose(dset);

  return 0;
}

int hdf_file::getd_ten3(std::string name, o2scl::tensor3 &t3) {
      
  // See if the dataspace already exists first
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
  
  // Get space requirements, to make sure they coincide
  // with the size specified by the user
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[3];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);

  if (ndims!=3) {
    O2SCL_ERR2_RET("Dimensions of dataspace do not match ",
		   "a rank 3 tensor in hdf_file::getd_ten3().",
		   gsl_efailed);
  }
  
  t3.free();
  t3.allocate(3,dims);

  double *start=&t3.get(0,0,0);

  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_DOUBLE,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,start);
  
  status=H5Dclose(dset);

  return 0;
}

int hdf_file::getd_mat(std::string name, umatrix &m) {
      
  // See if the dataspace already exists first
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
  
  // Get space requirements, to make sure they coincide
  // with the size specified by the user
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[2];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);
  
  if (ndims!=2) {
    O2SCL_ERR2_RET("Dimensions of dataspace do not match ",
		   "a matrix in hdf_file::getd_mat().",gsl_efailed);
  }
  
  m.allocate(dims[0],dims[1]);

  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_DOUBLE,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,&(m[0][0]));
  
  status=H5Dclose(dset);

  return 0;
}

int hdf_file::geti_mat(std::string name, omatrix_int &m) {
      
  // See if the dataspace already exists first
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
  
  // Get space requirements, to make sure they coincide
  // with the size specified by the user
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[2];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);
  
  if (ndims!=2) {
    O2SCL_ERR2_RET("Dimensions of dataspace do not match ",
		   "a matrix in hdf_file::geti_mat().",gsl_efailed);
  }
  
  m.allocate(dims[0],dims[1]);

  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_INT,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,&(m[0][0]));
  
  status=H5Dclose(dset);

  return 0;
}

int hdf_file::geti_mat(std::string name, umatrix_int &m) {
      
  // See if the dataspace already exists first
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
  
  // Get space requirements, to make sure they coincide
  // with the size specified by the user
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[2];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);
  
  if (ndims!=2) {
    O2SCL_ERR2_RET("Dimensions of dataspace do not match ",
		   "a matrix in hdf_file::geti_mat().",gsl_efailed);
  }
  
  m.allocate(dims[0],dims[1]);

  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_INT,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,&(m[0][0]));
  
  status=H5Dclose(dset);

  return 0;
}

int hdf_file::getd_vec_prealloc(std::string name, size_t n, double *d) {
      
  // See if the dataspace already exists first
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
      
  // Get space requirements, to make sure they coincide
  // with the size specified by the user
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[1];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);

  // Make sure sizes match
  if (dims[0]!=n) {
    O2SCL_ERR2_RET("Vector size in file doesn't match allocated ",
		   "size in hdf_file::getd_vec_prealloc().",
		   gsl_einval);
  }
      
  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_DOUBLE,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,d);
      
  status=H5Dclose(dset);
      
  return 0;
}

int hdf_file::geti_vec_prealloc(std::string name, size_t n, int *i) {
      
  // See if the dataspace already exists first
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
      
  // Get space requirements, to make sure they coincide
  // with the size specified by the user
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[1];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);

  // Make sure sizes match
  if (dims[0]!=n) {
    O2SCL_ERR2_RET("Vector size in file doesn't match allocated ",
		   "size in hdf_file::getd_vec_prealloc().",
		   gsl_einval);
  }
      
  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_INT,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,i);
      
  status=H5Dclose(dset);
      
  return 0;
}

int hdf_file::getd_mat_prealloc(std::string name, size_t n, 
				size_t m, double *d) {
      
  // See if the dataspace already exists first
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
      
  // Get space requirements, to make sure they coincide
  // with the size specified by the user
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[2];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);

  // Make sure sizes match
  if (dims[0]!=n || dims[1]!=m) {
    O2SCL_ERR2_RET("Matrix size in file doesn't match allocated ",
		   "size in hdf_file::getd_mat_prealloc().",
		   gsl_einval);
  }
      
  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_DOUBLE,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,d);
      
  status=H5Dclose(dset);
      
  return 0;
}

int hdf_file::geti_mat_prealloc(std::string name, size_t n, 
				size_t m, int *i) {
      
  // See if the dataspace already exists first
  hid_t dset=H5Dopen(current,name.c_str(),H5P_DEFAULT);
      
  // Get space requirements, to make sure they coincide
  // with the size specified by the user
  hid_t space=H5Dget_space(dset);  
  hsize_t dims[2];
  int ndims=H5Sget_simple_extent_dims(space,dims,0);

  // Make sure sizes match
  if (dims[0]!=n || dims[1]!=m) {
    O2SCL_ERR2_RET("Matrix size in file doesn't match allocated ",
		   "size in hdf_file::getd_mat_prealloc().",
		   gsl_einval);
  }
      
  // Read the data
  int status=H5Dread(dset,H5T_NATIVE_INT,H5S_ALL,H5S_ALL,
		     H5P_DEFAULT,i);
      
  status=H5Dclose(dset);
      
  return 0;
}
