/*
  -------------------------------------------------------------------

  Copyright (C) 2006-2012, Andrew W. Steiner

  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_HDF_FILE_H
#define O2SCL_HDF_FILE_H

#include <hdf5.h>

#include <o2scl/ovector_tlate.h>
#include <o2scl/uvector_tlate.h>
#include <o2scl/omatrix_tlate.h>
#include <o2scl/umatrix_tlate.h>
#include <o2scl/tensor.h>

#ifndef DOXYGENP
namespace o2scl_hdf {
#endif

  /** \brief Store data in an HDF5 file

      The member functions which write or get data from an HDF file
      begin with either <tt>get</tt> or <tt>set</tt>.  Where
      appropriate, the next character is either \c c for character, \c
      d for double, \c f for float, or \c i for int. 
      
      The get functions are:
      - getc() - get a single character
      - getc_def() - get a character or set to a default value
      - getc_arr() - get an array
      - getc_arr_alloc() - allocate memory with \c new and get an array

      The set functions are:
      - setc() - set a single character
      - setc_arr() - set an extensible array
      - setc_arr_fixed() - set a fixed array

      The string I/O functions:
      - gets() - get a string
      - gets_def() - get a string or set to a default value
      - gets_arr() - Get an extensible string
      - gets_fixed() - get a fixed-length string
      - sets() - set a string
      - sets_fixed() - set a fixed-length string
      - sets_arr() - set an extensible array of strings

      The vector I/O functions:
      - setd_vec() and seti_vec() - set a vector
      - getd_vec() and geti_vec() - get a vector
      - getd_vec_prealloc() and geti_vec_prealloc - 
      get a pre-allocated vector

      The matrix I/O functions:
      - setd_mat() and seti_mat() - set a matrix
      - getd_mat() and geti_mat() - get a matrix
      
      By default, vectors and matrices are written to HDF files in a
      chunked format, so their length can be changed later as
      necessary. The chunk size is chosen in \ref def_chunk() to be
      the closest power of 10 to the current vector size. 

      All files not closed by the user are closed in the destructor,
      but the destructor does not automatically close groups.

      \warning This class is still in development. Because of this,
      hdf5 files generated by this class may not be easily read by
      future versions. Later versions of \o2 will have stronger
      guarantees on backwards compatibility.

      \todo Error handling with HDF functions is a particular concern.
      Modify to throw \o2 exceptions when appropriate.

      \future Automatically close groups, e.g. by storing hid_t's in a
      stack?
  */
  class hdf_file {
    
#ifndef DOXYGEN_INTERNAL

  protected:

    /// File ID
    hid_t file;

    /// True if a file has been opened
    bool file_open;
    
    /// Current file or group location
    hid_t current;
    
    /** \brief Default chunk size

	Choose the closest power of 10 which is greater than or equal
	to 10 and less than or equal to \f$ 10^6 \f$.
    */
    virtual hsize_t def_chunk(size_t n) {
      size_t ch=(size_t)((1.0+1.0e-12)*
			 pow(10.0,floor(log10(((double)n))+0.5)));
      if (ch<10) ch=10;
      if (ch>1000000) ch=1000000;
      return ch;
    }
    
#endif
    
  public:

    hdf_file();

    virtual ~hdf_file();

    /// \name Open and close files
    //@{
    /// Open a file named \c fname
    int open(std::string fname);

    /// Close the file
    int close();
    //@}
    
    /// \name Manipulate ids
    //@{
    /// Get the current file id
    hid_t get_file_id();

    /// Set the current working id
    int set_current_id(hid_t cur);

    /// Retrieve the current working it
    hid_t get_current_id();
    //@}

    /// \name Set functions
    //@{
    /// Set a character named \c name to value \c d
    int setc(std::string name, char c);

    /// Set a double named \c name to value \c d
    int setd(std::string name, double d);
    
    /// Set a float named \c name to value \c d
    int setf(std::string name, float f);
    
    /// Set an integer named \c name to value \c d
    int seti(std::string name, int i);

    /** \brief Set a string named \c name to value \c s
	
	The string is stored in the HDF file as an extensible
	character array rather than a string.
    */
    int sets(std::string name, std::string s);

    /** \brief Set a string named \c name to value \c s
	
	This function stores <tt>s</tt> as a fixed-length string
	in the HDF file. If a dataset named \c name is already
	present, then \c s must not be longer than the string
	length already specified in the HDF file.
    */
    int sets_fixed(std::string name, std::string s);
    //@}

    /// \name Group manipulation
    //@{
    /** \brief Open a group relative to the location specified in 
	\c init_id

	\note In order to ensure that future objects are written to the
	newly-created group, the user must use set_current_id()
	using the newly-created group ID for the argument. 
    */
    hid_t open_group(hid_t init_id, std::string path);
    
    /** \brief Open a group relative to the current location

	\note In order to ensure that future objects are written to the
	newly-created group, the user must use set_current_id()
	using the newly-created group ID for the argument. 
     */
    hid_t open_group(std::string path); 

    /// Close a previously created group
    int close_group(hid_t group) {
      return H5Gclose(group);
    }
    //@}

    /** \name Vector set functions

	These functions automatically write all of the vector elements
	to the HDF file, if necessary extending the data that is
	already present.
    */
    //@{
    /// Set vector dataset named \c name with \c v
    int setd_vec(std::string name, const std::vector<double> &v);
    /** \brief Set vector dataset named \c name with \c v

        \note In the case that the stride of the vector \c v is not
	unity, this function requires copying the data twice: once from
	the vector to a pointer, and once from the pointer to the file.
	Keep this in mind when performing I/O with larger data sets.
    */
    int setd_vec(std::string name, const o2scl::ovector_const_view &v);
    /// Set vector dataset named \c name with \c v
    int setd_vec(std::string name, const o2scl::uvector_const_view &v);
    /// Set vector dataset named \c name with \c v
    int seti_vec(std::string name, const std::vector<int> &v);
    /** \brief Set vector dataset named \c name with \c v

        \note In the case that the stride of the vector \c v is not
	unity, this function requires copying the data twice: once from
	the vector to a pointer, and once from the pointer to the file.
	Keep this in mind when performing I/O with larger data sets.
    */
    int seti_vec(std::string name, const o2scl::ovector_int_const_view &v);
    /// Set vector dataset named \c name with \c v
    int seti_vec(std::string name, const o2scl::uvector_int_const_view &v);
    //@}

    /** \name Vector get functions

	These functions automatically free any previously allocated
	memory in <tt>v</tt> and then allocate the proper space
	required to read the information from the HDF file.
    */
    //@{
    /// Get vector dataset and place data in \c v
    int getd_vec(std::string name, std::vector<double> &v);
    /// Get vector dataset and place data in \c v
    int getd_vec(std::string name, o2scl::ovector &v);
    /// Get vector dataset and place data in \c v
    int getd_vec(std::string name, o2scl::uvector &v);
    /// Get vector dataset and place data in \c v
    int geti_vec(std::string name, std::vector<int> &v);
    /// Get vector dataset and place data in \c v
    int geti_vec(std::string name, o2scl::ovector_int &v);
    /// Get vector dataset and place data in \c v
    int geti_vec(std::string name, o2scl::uvector_int &v);
    //@}

    /** \name Matrix set functions

	These functions automatically write all of the vector elements
	to the HDF file, if necessary extending the data that is
	already present.
    */
    //@{
    /** \brief Set matrix dataset named \c name with \c m

        \note In the case that the trailing dimension (as reported by
	\ref omatrix_const_view_tlate::tda() ) of the matrix \c m is
	greater than the number of columns, this function requires
	copying the data twice: once from the matrix to a pointer, and
	once from the pointer to the file. Keep this in mind when
	performing I/O with larger data sets.
    */
    int setd_mat(std::string name, const o2scl::omatrix_const_view &m);
    /** \brief Set matrix dataset named \c name with \c m
    */
    int setd_mat(std::string name, const o2scl::umatrix_const_view &m);
    /** \brief Set matrix dataset named \c name with \c m

        \note In the case that the trailing dimension (as reported by
	\ref omatrix_const_view_tlate::tda() ) of the matrix \c m is
	greater than the number of columns, this function requires
	copying the data twice: once from the matrix to a pointer, and
	once from the pointer to the file. Keep this in mind when
	performing I/O with larger data sets.
    */
    int seti_mat(std::string name, const o2scl::omatrix_int_const_view &m);
    /** \brief Set matrix dataset named \c name with \c m
    */
    int seti_mat(std::string name, const o2scl::umatrix_int_const_view &m);
    //@}

    /** \name Matrix get functions

	These functions automatically free any previously allocated
	memory in <tt>m</tt> and then allocate the proper space
	required to read the information from the HDF file.
    */
    //@{
    /// Get matrix dataset and place data in \c m
    int getd_mat(std::string name, o2scl::omatrix &m);
    /// Get matrix dataset and place data in \c m
    int getd_mat(std::string name, o2scl::umatrix &m);
    /// Get matrix dataset and place data in \c m
    int geti_mat(std::string name, o2scl::omatrix_int &m);
    /// Get matrix dataset and place data in \c m
    int geti_mat(std::string name, o2scl::umatrix_int &m);
    //@}

    /// Desc
    int getd_ten3(std::string name, o2scl::tensor3 &t3);

    /// \name Array set functions
    //@{
    /// Set a character array named \c name of size \c n to value \c c
    int setc_arr(std::string name, size_t n, const char *c);
    
    /// Set a double array named \c name of size \c n to value \c d
    int setd_arr(std::string name, size_t n, const double *d);
    
    /// Set a float array named \c name of size \c n to value \c f
    int setf_arr(std::string name, size_t n, const float *f);

    /// Set a integer array named \c name of size \c n to value \c i
    int seti_arr(std::string name, size_t n, const int *i);

    /** \brief Set a vector of strings named \c name

        \devnote 
	String vectors are reformatted as a single character array, in
	order to allow each string to have different length and to
	make each string extensible. The size of the vector \c s is
	stored as an integer named <tt>nw</tt>.
    */
    int sets_arr(std::string name, std::vector<std::string> &s);
    //@}

    /** \name Fixed-length array set functions
	
	If a dataset named \c name is already present, then the
	user-specified array must not be longer than the array already
	present in the HDF file.
    */
    //@{
    /// Set a character array named \c name of size \c n to value \c c
    int setc_arr_fixed(std::string name, size_t n, const char *c);

    /// Set a double array named \c name of size \c n to value \c d
    int setd_arr_fixed(std::string name, size_t n, const double *c);

    /// Set a float array named \c name of size \c n to value \c f
    int setf_arr_fixed(std::string name, size_t n, const float *f);
    
    /// Set an integer array named \c name of size \c n to value \c i
    int seti_arr_fixed(std::string name, size_t n, const int *i);
    //@}
        
    /** \name Get functions

	If the specified object is not found, the \o2 error handler
	will be called.
     */
    //@{
    /// Get a character named \c name
    int getc(std::string name, char &c);
    
    /// Get a double named \c name
    int getd(std::string name, double &d);
    
    /// Get a float named \c name
    int getf(std::string name, float &f);

    /// Get a integer named \c name
    int geti(std::string name, int &i);

    /** \brief Get a string named \c name

	\note Strings are stored as character arrays and thus
	retrieving a string from a file requires loading the
	information from the file into a character array, and then
	copying it to the string. This will be slow for very long
	strings.
    */
    int gets(std::string name, std::string &s);
    
    /** \brief Get a string named \c name
     */
    int gets_fixed(std::string name, std::string &s);

    /** \brief Get a string named \c name
     */
    int gets_def_fixed(std::string name, std::string def, std::string &s);
    //@}

    /** \name Array get functions

	All of these functions (except gets_arr()) assume that the
	pointer allocated beforehand, and matches the size of the
	array in the HDF file. If the specified object is not found,
	the \o2 error handler will be called.
     */
    //@{
    /** \brief Get a character array named \c name of size \c n

	\note The pointer \c c must be allocated beforehand to 
	hold \c n entries, and \c n must match the size of the
	array in the HDF file.
     */
    int getc_arr(std::string name, size_t n, char *c);

    /** \brief Get a double array named \c name of size \c n 

	\note The pointer \c d must be allocated beforehand to 
	hold \c n entries, and \c n must match the size of the
	array in the HDF file.
     */
    int getd_arr(std::string name, size_t n, double *d);
    
    /** \brief Get a float array named \c name of size \c n 

	\note The pointer \c f must be allocated beforehand to 
	hold \c n entries, and \c n must match the size of the
	array in the HDF file.
     */
    int getf_arr(std::string name, size_t n, float *f);
    
    /** \brief Get an integer array named \c name of size \c n

	\note The pointer \c i must be allocated beforehand to 
	hold \c n entries, and \c n must match the size of the
	array in the HDF file.
     */
    int geti_arr(std::string name, size_t n, int *i);
    
    /** \brief Get a string array named \c name and store it in \c s
     */
    int gets_arr(std::string name, std::vector<std::string> &s);
    //@}
    
    /** \name Array get functions with memory allocation

        These functions allocate memory with \c new, which
	should be freed by the user with \c delete .
    */
    //@{
    /// Get a character array named \c name of size \c n
    int getc_arr_alloc(std::string name, size_t &n, char *c);

    /// Get a double array named \c name of size \c n 
    int getd_arr_alloc(std::string name, size_t &n, double *d);
    
    /// Get a float array named \c name of size \c n 
    int getf_arr_alloc(std::string name, size_t &n, float *f);
    
    /// Get an integer array named \c name of size \c n
    int geti_arr_alloc(std::string name, size_t &n, int *i);
    //@}
    
    /** \name Get functions with default values

	If the requested dataset is not found in the HDF file,
	the object is set to the specified default value
	and the error handler is not called.
     */
    //@{
    /// Get a character named \c name
    int getc_def(std::string name, char def, char &c);
    
    /// Get a double named \c name
    int getd_def(std::string name, double def, double &d);
    
    /// Get a float named \c name
    int getf_def(std::string name, float def, float &f);

    /// Get a integer named \c name
    int geti_def(std::string name, int def, int &i);

    /// Get a string named \c name
    int gets_def(std::string name, std::string def, std::string &s);
    //@}

    /** \name Get functions with pre-allocated pointer
     */
    //@{
    /// Get a double array \c d pre-allocated to have size \c n
    int getd_vec_prealloc(std::string name, size_t n, double *d);

    /// Get an integer array \c i pre-allocated to have size \c n
    int geti_vec_prealloc(std::string name, size_t n, int *i);

    /// Get a double matrix \c d pre-allocated to have size <tt>(n,m)</tt>
    int getd_mat_prealloc(std::string name, size_t n, size_t m, double *d);

    /// Get an integer matrix \c i pre-allocated to have size <tt>(n,m)</tt>
    int geti_mat_prealloc(std::string name, size_t n, size_t m, int *i);
    //@}
    
  };

#ifndef DOXYGENP
}
#endif

#endif
