/*
  -------------------------------------------------------------------

  Copyright (C) 2006-2012, Andrew W. Steiner

  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/hdf_io.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;
using namespace o2scl_hdf;

int o2scl_hdf::iterate_func(hid_t loc, const char *name, 
			    const H5L_info_t *inf, void *op_data) {
  
  iterate_parms *ip=(iterate_parms *)op_data;
  string type=ip->type;
  int verbose=ip->verbose;
  hdf_file &hf=*(ip->hf);
  hid_t top=hf.get_current_id();

  ip->found=false;

  H5O_info_t infobuf;
  herr_t status=H5Oget_info_by_name(loc,name,&infobuf,H5P_DEFAULT);
  
  // If it's a group
  if (infobuf.type==H5O_TYPE_GROUP) {
    
    if (verbose>0) {
      cout << "Found group with name " << name << "." << endl;
    }

    // Open the group and see if it's an O2scl object
    hid_t group=hf.open_group(name);
    hf.set_current_id(group);
    string otype;
    hf.gets_def_fixed("o2scl_type","",otype);
    hf.close_group(group);
    hf.set_current_id(top);
    
    if (verbose>0 && otype.length()!=0) {
      cout << "Group has o2scl_type " << otype << "." << endl;
    }

    if (otype.length()!=0 && otype==type) {
      if (verbose>0) {
	cout << "Found type " << type << "." << endl;
      }
      ip->found=true;
      ip->group_name=name;
      return 1;
    }
  }
  return 0;
}

int o2scl_hdf::find_group_by_type(hdf_file &hf, std::string type,
				  std::string &group_name,
				  int verbose) {
  iterate_parms ip={&hf,type,"",false,verbose};
  H5Literate(hf.get_current_id(),H5_INDEX_NAME,H5_ITER_NATIVE,
             0,iterate_func,&ip);
  if (ip.found) group_name=ip.group_name;
  return 0;
}

int o2scl_hdf::hdf_output(hdf_file &hf, const table &t, std::string name) {
  
  // Start group
  hid_t top=hf.get_current_id();
  hid_t group=hf.open_group(name);
  hf.set_current_id(group);
  
  // Add typename
  hf.sets_fixed("o2scl_type","table");

  // Restructure constants
  std::vector<std::string> cnames, cols;
  ovector cvalues;

  for(size_t i=0;i<t.get_nconsts();i++) {
    string n;
    double val;
    t.get_constant(i,n,val);
    cnames.push_back(n);
    cvalues.push_back(val);
  }

  // Restructure column names
  for(size_t i=0;i<t.get_ncolumns();i++) {
    cols.push_back(t.get_column_name(i));
  }

  // Output constants
  hf.sets_arr("con_names",cnames);
  hf.setd_vec("con_values",cvalues);
      
  // Output column names
  hf.sets_arr("col_names",cols);

  // Output unit info
  hf.seti("unit_flag",0);

  // Output number of lines
  hf.seti("nlines",t.get_nlines());

  // Create data group
  hid_t group2=hf.open_group("data");
  hf.set_current_id(group2);

  if (t.get_nlines()>0) {
    
    // Output data
    for(size_t i=0;i<t.get_ncolumns();i++) {
      // The actual vector is of size "maxlines", but we
      // only want to output the first "nlines" elements
      ovector_const_subvector sub(t[i],0,t.get_nlines());
      hf.setd_vec(t.get_column_name(i),sub);
    }

  }

  hf.close_group(group2);
  hf.close_group(group);

  // Return location to previous value
  hf.set_current_id(top);

  return 0;
}

int o2scl_hdf::hdf_input(hdf_file &hf, table &t, std::string name) {

  // Clear previous data
  t.clear_table();
  t.clear_constants();

  // If no name specified, find name of first group of specified type
  if (name.length()==0) {
    find_group_by_type(hf,"table",name);
    if (name.length()==0) {
      O2SCL_ERR2_RET("No object of type table found in ",
		     "o2scl_hdf::hdf_input().",gsl_efailed);
    }
  }

  // Open main group
  hid_t top=hf.get_current_id();
  hid_t group=hf.open_group(name);
  hf.set_current_id(group);

  // Check typename
  string type;
  hf.gets_fixed("o2scl_type",type);
  if (type!="table") {
    O2SCL_ERR2_RET("Typename in HDF group does not match ",
		   "class in table::hdf_input().",gsl_einval);
  }

  // Storage
  std::vector<std::string> cnames, cols;
  ovector cvalues;
      
  // Get constants
  hf.gets_arr("con_names",cnames);
  hf.getd_vec("con_values",cvalues);
  if (cnames.size()!=cvalues.size()) {
    O2SCL_ERR2_RET("Size mismatch between constant names and values ",
		   "in table::hdf_input().",gsl_einval);
  }
  for(size_t i=0;i<cnames.size();i++) {
    t.add_constant(cnames[i],cvalues[i]);
  }

  // Get column names
  hf.gets_arr("col_names",cols);
  for(size_t i=0;i<cols.size();i++) {
    t.new_column(cols[i]);
  }

  // Get number of lines
  int nlines;
  hf.geti("nlines",nlines);
  t.set_nlines(nlines);

  // Open data group
  hid_t group2=hf.open_group("data");
  hf.set_current_id(group2);

  if (nlines>0) {
    
    // Get data
    for(size_t i=0;i<t.get_ncolumns();i++) {
      ovector_base &v=t[i];
      double *d=&(v[0]);
      hf.getd_vec_prealloc(t.get_column_name(i),nlines,d);
    }

  }

  // Close groups
  hf.close_group(group2);
  hf.close_group(group);

  // Return location to previous value
  hf.set_current_id(top);

  t.check_synchro();

  return 0;
}

int o2scl_hdf::hdf_output(hdf_file &hf, const table_units &t, 
			  std::string name) {
  
  // Start group
  hid_t top=hf.get_current_id();
  hid_t group=hf.open_group(name);
  hf.set_current_id(group);
  
  // Add typename
  hf.sets_fixed("o2scl_type","table");

  // Restructure constants
  std::vector<std::string> cnames, cols, units;
  ovector cvalues;

  for(size_t i=0;i<t.get_nconsts();i++) {
    string n;
    double val;
    t.get_constant(i,n,val);
    cnames.push_back(n);
    cvalues.push_back(val);
  }

  // Restructure column names
  for(size_t i=0;i<t.get_ncolumns();i++) {
    cols.push_back(t.get_column_name(i));
  }

  // Output constants
  hf.sets_arr("con_names",cnames);
  hf.setd_vec("con_values",cvalues);
      
  // Output column names
  hf.sets_arr("col_names",cols);

  // Output unit info
  hf.seti("unit_flag",1);

  // Restructure units
  for(size_t i=0;i<t.get_ncolumns();i++) {
    units.push_back(t.get_unit(t.get_column_name(i)));
  }

  // Output units
  hf.sets_arr("units",units);

  // Output number of lines
  hf.seti("nlines",t.get_nlines());

  // Create data group
  hid_t group2=hf.open_group("data");
  hf.set_current_id(group2);

  if (t.get_nlines()>0) {
    
    // Output data
    for(size_t i=0;i<t.get_ncolumns();i++) {
      // The actual vector is of size "maxlines", but we
      // only want to output the first "nlines" elements
      ovector_const_subvector sub(t[i],0,t.get_nlines());
      hf.setd_vec(t.get_column_name(i),sub);
    }

  }
    
  hf.close_group(group2);
  hf.close_group(group);

  // Return location to previous value
  hf.set_current_id(top);

  return 0;
}

int o2scl_hdf::hdf_input(hdf_file &hf, table_units &t, std::string name) {

  // Clear previous data
  t.clear_table();
  t.clear_constants();

  // If no name specified, find name of first group of specified type
  if (name.length()==0) {
    find_group_by_type(hf,"table",name);
    if (name.length()==0) {
      O2SCL_ERR2_RET("No object of type table found in ",
		     "o2scl_hdf::hdf_input().",gsl_efailed);
    }
  }

  // Open main group
  hid_t top=hf.get_current_id();
  hid_t group=hf.open_group(name);
  hf.set_current_id(group);

  // Check typename
  string type;
  hf.gets_fixed("o2scl_type",type);
  if (type!="table") {
    O2SCL_ERR2_RET("Typename in HDF group does not match ",
		   "class in table::hdf_input().",gsl_einval);
  }

  // Storage
  std::vector<std::string> cnames, cols, units;
  ovector cvalues;
      
  // Get constants
  hf.gets_arr("con_names",cnames);
  hf.getd_vec("con_values",cvalues);
  if (cnames.size()!=cvalues.size()) {
    O2SCL_ERR2_RET("Size mismatch between constant names and values ",
		   "in table::hdf_input().",gsl_einval);
  }
  for(size_t i=0;i<cnames.size();i++) {
    t.add_constant(cnames[i],cvalues[i]);
  }
      
  // Get column names
  hf.gets_arr("col_names",cols);
  for(size_t i=0;i<cols.size();i++) {
    t.new_column(cols[i]);
  }

  // Get unit flag
  int uf;
  hf.geti("unit_flag",uf);

  // If present, get units
  if (uf>0) {
    hf.gets_arr("units",units);
    for(size_t i=0;i<units.size();i++) {
      t.set_unit(t.get_column_name(i),units[i]);
    }
  }

  // Get number of lines
  int nlines;
  hf.geti("nlines",nlines);
  t.set_nlines(nlines);

  // Open data group
  hid_t group2=hf.open_group("data");
  hf.set_current_id(group2);

  if (nlines>0) {
    
    // Get data
    for(size_t i=0;i<t.get_ncolumns();i++) {
      ovector_base &v=t[i];
      double *d=&(v[0]);
      hf.getd_vec_prealloc(t.get_column_name(i),nlines,d);
    }

  }

  hf.close_group(group2);
  hf.close_group(group);

  // Return location to previous value
  hf.set_current_id(top);

  t.check_synchro();

  return 0;
}

int o2scl_hdf::hdf_output(hdf_file &hf, const table3d &t, 
			  std::string name) {

  // Start group
  hid_t top=hf.get_current_id();
  hid_t group=hf.open_group(name);
  hf.set_current_id(group);
  
  // Add typename
  hf.sets_fixed("o2scl_type","table3d");

  string sx;
  hf.gets_fixed("o2scl_type",sx);

  // Restructure constants
  std::vector<std::string> cnames;
  ovector cvalues;

  for(size_t i=0;i<t.get_nconsts();i++) {
    string n;
    double val;
    t.get_constant(i,n,val);
    cnames.push_back(n);
    cvalues.push_back(val);
  }

  // Output constants
  hf.sets_arr("con_names",cnames);
  hf.setd_vec("con_values",cvalues);
      
  // Restructure slice names
  std::vector<std::string> slnames;
  for(int i=0;i<t.get_nslices();i++) {
    slnames.push_back(t.get_slice_name(i));
  }

  // Output slice names
  hf.sets_arr("slice_names",slnames);

  // Output unit info
  hf.seti("unit_flag",0);
  
  hf.seti("size_set",t.is_size_set());

  if (t.is_size_set()) {

    // Output grid
    hf.seti("numx",t.get_nx());
    hf.seti("numy",t.get_ny());
    
    hf.seti("xy_set",t.is_xy_set());
    
    if (t.is_xy_set()) {
      hf.sets("xname",t.get_x_name());
      hf.sets("yname",t.get_y_name());
      hf.setd_vec("xval",t.get_x_data());
      hf.setd_vec("yval",t.get_y_data());
    }
    
    // Create data group
    hid_t group2=hf.open_group("data");
    hf.set_current_id(group2);
    
    // Output data
    for(int i=0;i<t.get_nslices();i++) {
      hf.setd_mat(t.get_slice_name(i),t.get_slice(i));
    }
    
    hf.close_group(group2);

  }

  hf.close_group(group);

  // Return location to previous value
  hf.set_current_id(top);

  return 0;
}

int o2scl_hdf::hdf_input(hdf_file &hf, table3d &t, std::string name) {

  // Clear previous data
  t.clear_table();

  // If no name specified, find name of first group of specified type
  if (name.length()==0) {
    find_group_by_type(hf,"table3d",name);
    if (name.length()==0) {
      O2SCL_ERR2_RET("No object of type table found in ",
		     "o2scl_hdf::hdf_input().",gsl_efailed);
    }
  }

  // Open main group
  hid_t top=hf.get_current_id();
  hid_t group=hf.open_group(name);
  hf.set_current_id(group);

  // Check typename
  string type;
  hf.gets_fixed("o2scl_type",type);
  if (type!="table3d") {
    O2SCL_ERR2_RET("Typename in HDF group does not match ",
		   "class in table::hdf_input().",gsl_einval);
  }

  // Storage
  std::vector<std::string> cnames;
  ovector cvalues;

  // Get constants
  hf.gets_arr("con_names",cnames);
  hf.getd_vec("con_values",cvalues);
  for(size_t i=0;i<cnames.size();i++) {
    t.add_constant(cnames[i],cvalues[i]);
  }

  int size_set, xy_set;

  hf.geti("size_set",size_set);

  if (size_set>0) {
    
    int nx, ny;

    // Get grid
    hf.geti("numx",nx);
    hf.geti("numy",ny);
    
    hf.geti("xy_set",xy_set);
    
    if (xy_set>0) {
      
      string xname, yname;
      ovector xv, yv;
      hf.gets("xname",xname);
      hf.gets("yname",yname);
      hf.getd_vec("xval",xv);
      hf.getd_vec("yval",yv);
      t.set_xy(xname,nx,xv,yname,ny,yv);

    } else {

      t.set_size(nx,ny);

    }

    // Get slice names
    std::vector<std::string> slnames;
    hf.gets_arr("slice_names",slnames);
    for(size_t i=0;i<slnames.size();i++) {
      t.new_slice(slnames[i]);
    }

    // Open data group
    hid_t group2=hf.open_group("data");
    hf.set_current_id(group2);
    
    // Get data
    for(int i=0;i<t.get_nslices();i++) {
      omatrix &m=t.get_slice(i);
      double *mp=&(m[0][0]);
      hf.getd_mat_prealloc(t.get_slice_name(i),nx,ny,mp);
    }

    hf.close_group(group2);

  }

  hf.close_group(group);

  // Return location to previous value
  hf.set_current_id(top);

  return 0;
}

int o2scl_hdf::hdf_output(hdf_file &hf, const hist &h, std::string name) {
  
  // Start group
  hid_t top=hf.get_current_id();
  hid_t group=hf.open_group(name);
  hf.set_current_id(group);

  // Add typename
  hf.sets_fixed("o2scl_type","hist");

  // Add histogram
  hf.seti("size",h.size());
  hf.seti("rmode",h.get_rep_mode());
  hf.setd_vec("bins",h.get_bins());
  hf.setd_vec("weights",h.get_wgts());
  if (h.get_user_reps().size()>0) {
    hf.seti("ur_size",h.size());
    hf.setd_vec("ureps",h.get_user_reps());
  } else {
    hf.seti("ur_size",0);
  }

  // Close group
  hf.close_group(group);

  // Return location to previous value
  hf.set_current_id(top);

  return 0;
}

int o2scl_hdf::hdf_input(hdf_file &hf, hist &h, std::string name) {
  
  // Clear previous data
  h.clear();

  // If no name specified, find name of first group of specified type
  if (name.length()==0) {
    find_group_by_type(hf,"hist",name);
    if (name.length()==0) {
      O2SCL_ERR2_RET("No object of type hist found in ",
		     "o2scl_hdf::hdf_input().",gsl_efailed);
    }
  }

  // Open main group
  hid_t top=hf.get_current_id();
  hid_t group=hf.open_group(name);
  hf.set_current_id(group);

  // Check typename
  string type;
  hf.gets_fixed("o2scl_type",type);
  if (type!="hist") {
    O2SCL_ERR2_RET("Typename in HDF group does not match ",
		   "class in hdf_input().",gsl_einval);
  }

  int size, mode;
  ovector bins, weights, ur;

  // Get histogram
  hf.geti("size",size);
  hf.getd_vec("bins",bins);

  // Set bins
  h.set_bins(size+1,bins);

  // Get weights
  hf.getd_vec("weights",weights);
  for(int i=0;i<size;i++) {
    h.set_wgt_i(i,weights[i]);
  }

  // Get center mode
  hf.geti("rmode",mode);
  h.set_rep_mode(mode);

  // Get centers
  int ur_size;
  hf.geti("ur_size",ur_size);
  if (ur_size>0) {
    hf.getd_vec("ureps",ur);
    if (((size_t)mode)==hist::rmode_user) {
      h.set_reps(size,ur);
    }
  }

  // Close group
  hf.close_group(group);

  // Return location to previous value
  hf.set_current_id(top);

  return 0;
}

int o2scl_hdf::hdf_output(hdf_file &hf, const hist_2d &h, 
			  std::string name) {

  // Start group
  hid_t top=hf.get_current_id();
  hid_t group=hf.open_group(name);
  hf.set_current_id(group);

  // Add typename
  hf.sets_fixed("o2scl_type","hist_2d");

  // Add histogram
  hf.seti("size_x",h.size_x());
  hf.seti("size_y",h.size_y());
  hf.seti("xrmode",h.get_x_rep_mode());
  hf.seti("yrmode",h.get_y_rep_mode());
  hf.setd_vec("x_bins",h.get_x_bins());
  hf.setd_vec("y_bins",h.get_y_bins());
  hf.setd_mat("weights",h.get_wgts());

  if (h.get_user_reps_x().size()>0) {
    hf.seti("urx_size",h.get_user_reps_x().size());
    hf.setd_vec("uxreps",h.get_user_reps_x());
  } else {
    hf.seti("urx_size",0);
  }

  if (h.get_user_reps_y().size()>0) {
    hf.seti("ury_size",h.get_user_reps_y().size());
    hf.setd_vec("uyreps",h.get_user_reps_y());
  } else {
    hf.seti("ury_size",0);
  }

  // Close group
  hf.close_group(group);

  // Return location to previous value
  hf.set_current_id(top);

  return 0;
}

int o2scl_hdf::hdf_input(hdf_file &hf, hist_2d &h, std::string name) {

  h.clear();

  // If no name specified, find name of first group of specified type
  if (name.length()==0) {
    find_group_by_type(hf,"hist_2d",name);
    if (name.length()==0) {
      O2SCL_ERR2_RET("No object of type hist_2d found in ",
		     "o2scl_hdf::hdf_input().",gsl_efailed);
    }
  }
  
  // Open main group
  hid_t top=hf.get_current_id();
  hid_t group=hf.open_group(name);
  hf.set_current_id(group);

  // Check typename
  string type;
  hf.gets_fixed("o2scl_type",type);
  if (type!="hist_2d") {
    O2SCL_ERR2_RET("Typename in HDF group does not match ",
		   "class in hdf_input().",gsl_einval);
  }

  int size_x, size_y, mode_x, mode_y;
  ovector bins_x, bins_y, weights;

  // Get sizes
  hf.geti("size_x",size_x);
  hf.geti("size_y",size_y);

  // Get bins
  hf.getd_vec("x_bins",bins_x);
  hf.getd_vec("y_bins",bins_y);

  // Set bins
  h.set_bins(size_x,bins_x,size_y,bins_y);

  // Get modes
  hf.geti("xrmode",mode_x);
  hf.geti("yrmode",mode_y);

  // Set modes
  h.set_rep_mode(mode_x,mode_y);
  
  // Get weights
  hf.getd_mat("weights",h.get_wgts());

  // Get x reps
  int urx_size;
  ovector urx;
  hf.geti("urx_size",urx_size);
  if (urx_size>0) {
    hf.getd_vec("uxreps",urx);
    if (((size_t)mode_x)==hist_2d::rmode_user) {
      h.set_x_reps(urx_size,urx);
    }
  }

  // Get y reps
  int ury_size;
  ovector ury;
  hf.geti("ury_size",ury_size);
  if (ury_size>0) {
    hf.getd_vec("uyreps",ury);
    if (((size_t)mode_y)==hist_2d::rmode_user) {
      h.set_y_reps(ury_size,ury);
    }
  }

  // Close group
  hf.close_group(group);

  // Return location to previous value
  hf.set_current_id(top);

  return 0;
}

int o2scl_hdf::hdf_output(hdf_file &hf, const tensor &t, std::string name) {

  // Start group
  hid_t top=hf.get_current_id();
  hid_t group=hf.open_group(name);
  hf.set_current_id(group);

  // Add typename
  hf.sets_fixed("o2scl_type","tensor");

  // Add rank
  hf.seti("rank",t.get_rank());

  // Add dimensions
  vector<int> size_arr;
  for(size_t i=0;i<t.get_rank();i++) {
    size_arr.push_back(t.get_size(i));
  }
  hf.seti_vec("size",size_arr);

  // Add data
  const uvector &d=t.get_data();
  hf.setd_vec("data",d);

  // Close group
  hf.close_group(group);

  // Return location to previous value
  hf.set_current_id(top);

  return 0;
}

int o2scl_hdf::hdf_input(hdf_file &hf, tensor &t, std::string name) {

  // Free previously allocated memory
  t.free();

  // If no name specified, find name of first group of specified type
  if (name.length()==0) {
    find_group_by_type(hf,"tensor",name);
    if (name.length()==0) {
      O2SCL_ERR2_RET("No object of type tensor found in ",
		     "o2scl_hdf::hdf_input().",gsl_efailed);
    }
  }
  
  // Open main group
  hid_t top=hf.get_current_id();
  hid_t group=hf.open_group(name);
  hf.set_current_id(group);

  // Check typename
  string type;
  hf.gets_fixed("o2scl_type",type);
  if (type!="tensor") {
    O2SCL_ERR2_RET("Typename in HDF group does not match ",
		   "class in hdf_input().",gsl_einval);
  }

  int rank;
  hf.geti("rank",rank);
  int *size_i=new int[rank];
  hf.geti_vec_prealloc("size",rank,size_i);

  size_t *size_s=new size_t[rank];
  for(int k=0;k<rank;k++) size_s[k]=size_i[k];
  t.allocate(rank,size_s);

  delete[] size_i;
  delete[] size_s;

  uvector &d=t.get_data_non_const();
  hf.getd_vec("data",d);

  // Close group
  hf.close_group(group);

  // Return location to previous value
  hf.set_current_id(top);

  return 0;
}

int o2scl_hdf::hdf_output(hdf_file &hf, const tensor_grid &t, 
			  std::string name) {

  // Start group
  hid_t top=hf.get_current_id();
  hid_t group=hf.open_group(name);
  hf.set_current_id(group);

  // Add typename
  hf.sets_fixed("o2scl_type","tensor_grid");

  // Add rank
  hf.seti("rank",t.get_rank());

  // Add dimensions
  vector<int> size_arr;
  for(size_t i=0;i<t.get_rank();i++) {
    size_arr.push_back(t.get_size(i));
  }
  hf.seti_vec("size",size_arr);

  // Add data
  const uvector &d=t.get_data();
  hf.setd_vec("data",d);

  // Add grid
  if (t.is_grid_set()) hf.seti("grid_set",1);
  else hf.seti("grid_set",0);
    
  if (t.is_grid_set()) {
    ovector grid;
    for(size_t j=0;j<t.get_rank();j++) {
      for(size_t k=0;k<t.get_size(j);k++) {
	grid.push_back(t.get_grid(j,k));
      }
    }
    hf.setd_vec("grid",grid);
  }

  // Close group
  hf.close_group(group);

  // Return location to previous value
  hf.set_current_id(top);

  return 0;
}

int o2scl_hdf::hdf_input(hdf_file &hf, tensor_grid &t, std::string name) {

  // Free previously allocated memory
  t.free();

  // If no name specified, find name of first group of specified type
  if (name.length()==0) {
    find_group_by_type(hf,"tensor_grid",name);
    if (name.length()==0) {
      O2SCL_ERR2_RET("No object of type tensor found in ",
		     "o2scl_hdf::hdf_input().",gsl_efailed);
    }
  }
  
  // Open main group
  hid_t top=hf.get_current_id();
  hid_t group=hf.open_group(name);
  hf.set_current_id(group);

  // Check typename
  string type;
  hf.gets_fixed("o2scl_type",type);
  if (type!="tensor_grid") {
    O2SCL_ERR2_RET("Typename in HDF group does not match ",
		   "class in hdf_input().",gsl_einval);
  }

  // Get rank
  int rank;
  hf.geti("rank",rank);
  int *size_i=new int[rank];
  hf.geti_vec_prealloc("size",rank,size_i);

  size_t *size_s=new size_t[rank];
  for(int k=0;k<rank;k++) size_s[k]=size_i[k];
  t.allocate(rank,size_s);

  delete[] size_i;
  delete[] size_s;

  uvector &d=t.get_data_non_const();
  hf.getd_vec("data",d);

  // Get grid
  bool grid_set=false;
  int igrid_set;
  hf.geti("grid_set",igrid_set);
  if (igrid_set>0) grid_set=true;
  if (grid_set) {
    ovector ogrid;
    hf.getd_vec("grid",ogrid);
    size_t ix=0;
    double **grid=new double *[rank];
    for(size_t j=0;j<((size_t)rank);j++) {
      grid[j]=new double[t.get_size(j)];
      for(size_t k=0;k<t.get_size(j);k++) {
	grid[j][k]=ogrid[ix];
	ix++;
      }
    }
    t.set_grid(ogrid);
    for(size_t j=0;j<((size_t)rank);j++) {
      delete[] grid[j];
    }
    delete[] grid;
  }

  // Close group
  hf.close_group(group);

  // Return location to previous value
  hf.set_current_id(top);

  return 0;
}

