/*
  -------------------------------------------------------------------

  Copyright (C) 2006-2012, Andrew W. Steiner

  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_HDF_IO_H
#define O2SCL_HDF_IO_H

/** \file hdf_io.h
    \brief File for HDF I/O for \ref table and \ref table3d
*/

#include <hdf5.h>

#include <o2scl/table.h>
#include <o2scl/table_units.h>
#include <o2scl/table3d.h>
#include <o2scl/constants.h>
#include <o2scl/hist.h>
#include <o2scl/hist_2d.h>
#include <o2scl/tensor.h>

#include <o2scl/hdf_file.h>

#ifndef DOXYGENP
namespace o2scl_hdf {
#endif

  /// Output a \ref table to a HDF file
  int hdf_output(hdf_file &hf, const o2scl::table &t, std::string name);
  
  /// Input a \ref table from a HDF file
  int hdf_input(hdf_file &hf, o2scl::table &t, std::string name="");

  /** \brief Output a \ref table_units to a HDF file

      The unit conversion object, as specified in \ref
      table_units::set_convert() is not written to the HDF file. Thus
      when a \ref table_units object is read from a file, it always
      uses the default unit conversion object.
  */
  int hdf_output(hdf_file &hf, const o2scl::table_units &t, std::string name);
  
  /** \brief Input a \ref table_units from a HDF file

      The unit conversion object, as specified in \ref
      table_units::set_convert() is not written to the HDF file. Thus
      when a \ref table_units object is read from a file, it always
      uses the default unit conversion object.
  */
  int hdf_input(hdf_file &hf, o2scl::table_units &t, std::string name="");

  /// Output a \ref table3d to a HDF file
  int hdf_output(hdf_file &hf, const o2scl::table3d &t, std::string name);
  
  /// Input a \ref table3d from a HDF file
  int hdf_input(hdf_file &hf, o2scl::table3d &t, std::string name="");

  /// Output a \ref hist to a HDF file
  int hdf_output(hdf_file &hf, const o2scl::hist &t, std::string name);
  
  /// Input a \ref hist from a HDF file
  int hdf_input(hdf_file &hf, o2scl::hist &t, std::string name="");

  /// Output a \ref hist_2d to a HDF file
  int hdf_output(hdf_file &hf, const o2scl::hist_2d &t, std::string name);
  
  /// Input a \ref hist_2d from a HDF file
  int hdf_input(hdf_file &hf, o2scl::hist_2d &t, std::string name="");

  /// Output a \ref tensor to a HDF file
  int hdf_output(hdf_file &hf, const o2scl::tensor &t, std::string name);
  
  /// Input a \ref tensor from a HDF file
  int hdf_input(hdf_file &hf, o2scl::tensor &t, std::string name="");

  /// Output a \ref tensor to a HDF file
  int hdf_output(hdf_file &hf, const o2scl::tensor_grid &t, std::string name);
  
  /// Input a \ref tensor from a HDF file
  int hdf_input(hdf_file &hf, o2scl::tensor_grid &t, std::string name="");

  /// \name Helper struct and functions
  //@{
  /// A structure to pass information to and from \ref iterate_func()
  typedef struct {
    hdf_file *hf;
    std::string type;
    std::string group_name;
    bool found;
    int verbose;
  } iterate_parms;
  
  /** \brief Look at location \c loc in an HDF file for an \o2 object

      This is used by \ref find_group_by_type() where \c op_data
      is a pointer to an object of type \ref iterate_parms to 
      look for \o2 objects of a specified type without knowing
      the group name.
  */
  int iterate_func(hid_t loc, const char *name, 
			      const H5L_info_t *inf, void *op_data);
  
  /** \brief Look in hdf_file \c hf for an \o2 object of type \c type and 
      if found, set \c group_name to the associated object name

      This function returns 1 if an object of type \c type is found
      and 0 if it fails.
   */
  int find_group_by_type(hdf_file &hf, std::string type,
			 std::string &group_name, int verbose=0);
  //@}
  

#ifndef DOXYGENP
}
#endif

#endif


