/*
  -------------------------------------------------------------------

  Copyright (C) 2006-2012, Andrew W. Steiner

  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  -------------------------------------------------------------------
*/
#include <o2scl/test_mgr.h>
#include <o2scl/hdf_io.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_hdf;

int main(void) {

  cout.setf(ios::scientific);
  cout.precision(10);

  test_mgr t;
  t.set_output_level(1);

  // Test of table I/O
  {
    table tab, tab2;
    tab.add_constant("pi",acos(-1.0));
    tab.line_of_names("a b c");
    for(size_t i=0;i<10;i++) {
      double d=((double)i);
      double line[3]={d,sin(d),cos(d)};
      tab.line_of_data(3,line);
    }

    hdf_file hf;
    hf.open("table.o2");
    hdf_output(hf,tab,"table_test");
    hf.close();

    hf.open("table.o2");
    hdf_input(hf,tab2,"table_test");
    hf.close();

    t.test_gen(tab.get_nlines()==tab2.get_nlines(),"lines");
    t.test_gen(tab.get_ncolumns()==tab2.get_ncolumns(),"cols");
    t.test_gen(tab.get_nconsts()==tab2.get_nconsts(),"cols");
  }

  // Test of table_units I/O
  {
    table_units tab, tab2;
    tab.add_constant("pi",acos(-1.0));
    tab.line_of_names("a b c");
    tab.set_unit("a","m");
    tab.set_unit("b","cm");
    tab.set_unit("c","km");
    for(size_t i=0;i<10;i++) {
      double d=((double)i);
      double line[3]={d,sin(d),cos(d)};
      tab.line_of_data(3,line);
    }

    hdf_file hf;
    hf.open("table_units.o2");
    hdf_output(hf,tab,"table_test");
    hf.close();

    hf.open("table_units.o2");
    hdf_input(hf,tab2,"table_test");
    hf.close();

    t.test_gen(tab.get_nlines()==tab2.get_nlines(),"lines");
    t.test_gen(tab.get_ncolumns()==tab2.get_ncolumns(),"cols");
    t.test_gen(tab.get_nconsts()==tab2.get_nconsts(),"cols");
    t.test_gen(tab.get_unit("a")==tab2.get_unit("a"),"unit");
  }

  // Test of table3d I/O
  {
    table3d at, at2;

    // Set the grid size
    at.set_size(3,5);
    size_t nx1, ny1;
    at.get_size(nx1,ny1);
    
    // Add some slices
    int r1=at.new_slice("z1");
    t.test_gen(r1==0,"slice 1");
    int r2=at.new_slice("z2");
    t.test_gen(r2==0,"slice 2");
    
    // Add the data
    for(size_t i=0;i<nx1;i++) {
      for(size_t j=0;j<ny1;j++) {
	int er1=at.set(i,j,"z1",sqrt(((double)(i+j))));
	t.test_gen(er1==0,"set 1");
	int er2=at.set(i,j,"z2",sqrt(2.0*((double)(i+j))));
	t.test_gen(er2==0,"set 2");
      }
    }
    
    hdf_file hf;
    hf.open("table3d.o2");
    hdf_output(hf,at,"table3d_test");
    hf.close();
    hf.open("table3d.o2");
    hdf_input(hf,at2,"table3d_test");
    hf.close();

    t.test_gen(at.get_nx()==at2.get_nx(),"nx");
    t.test_gen(at.get_ny()==at2.get_ny(),"ny");
    t.test_gen(at.get_nslices()==at2.get_nslices(),"cols");
    t.test_gen(at.get_nconsts()==at2.get_nconsts(),"cols");
  }    

  // Test of tensor_grid3 I/O
  {
    tensor_grid3 tg(3,2,1), tg2;
    double grid[6]={4,5,6,7,8,9};
    tg.set_grid(grid);
    for(size_t j=0;j<3;j++) {
      for(size_t k=0;k<2;k++) {
	for(size_t ell=0;ell<1;ell++) {
	  tg.set(j,k,ell,((double)(j+k+ell)));
	}
      }
    }

    hdf_file hf;
    hf.open("tens_grid.o2");
    hdf_output(hf,tg,"tens_grid_test");
    hf.close();
    hf.open("tens_grid.o2");
    hdf_input(hf,tg2,"tens_grid_test");
    hf.close();

    t.test_gen(tg2.get_rank()==3,"rank");
    t.test_rel(tg2.get_grid(0,1),5.0,1.0e-12,"grid 1");
    t.test_rel(tg2.get_grid(2,0),9.0,1.0e-12,"grid 2");
    for(size_t j=0;j<3;j++) {
      for(size_t k=0;k<2;k++) {
	for(size_t ell=0;ell<1;ell++) {
	  double x=tg.get(j,k,ell);
	  t.test_rel(((double)(j+k+ell)),x,1.0e-12,"element");
	}
      }
    }
    
  }

  t.report();

  return 0;
}

