/*
  -------------------------------------------------------------------

  Copyright (C) 2006-2012, Andrew W. Steiner

  This file is part of O2scl.

  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#include <iostream>

#include <gsl/gsl_linalg.h>
#include <gsl/gsl_blas.h>

#include <o2scl/cholesky.h>
#include <o2scl/test_mgr.h>
#include <o2scl/ovector_tlate.h>
#include <o2scl/omatrix_tlate.h>
#include <o2scl/columnify.h>

using namespace std;
using namespace o2scl;

int main(void) {
  test_mgr t;
  t.set_output_level(2);

  cout.setf(ios::scientific);
  
  if (lib_settings.range_check()) {
    cout << "O2scl range checking on." << endl;
  } else {
    cout << "O2scl range checking off." << endl;
  }

  {
    using namespace o2scl_cblas;
    using namespace o2scl_linalg;

    omatrix gm1(5,5), om1(5,5), gm2(5,5), om2(5,5);
    ovector gv1(5), ov1(5), gv2(5), ov2(5), gv3(5), ov3(5);
    double mm1[5][5], d, mv1[5], mv2[5], mm2[5][5], mv3[5];
    for(size_t i=0;i<5;i++) {
      gv1[i]=cos((double)(i));
      ov1[i]=cos((double)(i));
      mv1[i]=cos((double)(i));
      for(size_t j=0;j<5;j++) {
	gm1[i][j]=1.0/(1.0+((double)i)+((double)j));
	om1[i][j]=1.0/(1.0+((double)i)+((double)j));
	mm1[i][j]=1.0/(1.0+((double)i)+((double)j));
	gm2[i][j]=1.0/(1.0+((double)i)+((double)j));
	om2[i][j]=1.0/(1.0+((double)i)+((double)j));
	mm2[i][j]=1.0/(1.0+((double)i)+((double)j));
      }
    }

    // Test cholesky decomposition

    gsl_linalg_cholesky_decomp(&gm1);
    cholesky_decomp(5,om1);
    cholesky_decomp(5,mm1);

    t.test_rel_mat(5,5,gm1,om1,1.0e-12,"cholesky decomp");
    t.test_rel_mat(5,5,om1,mm1,1.0e-12,"cholesky decomp 2");

    // Test solve 

    for(size_t i=0;i<5;i++) {
      gv1[i]=cos((double)(i));
      ov1[i]=cos((double)(i));
      mv1[i]=cos((double)(i));
      gv2[i]=cos((double)(i));
      ov2[i]=cos((double)(i));
      mv2[i]=cos((double)(i));
      for(size_t j=0;j<5;j++) {
	gm1[i][j]=1.0/(1.0+((double)i)+((double)j));
	om1[i][j]=1.0/(1.0+((double)i)+((double)j));
	mm1[i][j]=1.0/(1.0+((double)i)+((double)j));
	gm2[i][j]=1.0/(1.0+((double)i)+((double)j));
	om2[i][j]=1.0/(1.0+((double)i)+((double)j));
	mm2[i][j]=1.0/(1.0+((double)i)+((double)j));
      }
    }

    // -------------------------------------------------

    gsl_linalg_cholesky_decomp(&gm1);
    gsl_linalg_cholesky_solve(&gm1,&gv2,&gv1);

    dgemv(o2cblas_RowMajor,o2cblas_NoTrans,5,5,1.0,gm2,gv1,0.0,gv3);
    t.test_rel_arr(5,gv2,gv3,1.0e-10,"solve 1");
    
    // -------------------------------------------------

    cholesky_decomp(5,om1);
    cholesky_solve(5,om1,ov2,ov1);

    dgemv(o2cblas_RowMajor,o2cblas_NoTrans,5,5,1.0,om2,ov1,0.0,ov3);
    t.test_rel_arr(5,ov2,ov3,1.0e-10,"solve 2");

    // -------------------------------------------------

    cholesky_decomp(5,mm1);
    cholesky_solve(5,mm1,mv2,mv1);

    dgemv(o2cblas_RowMajor,o2cblas_NoTrans,5,5,1.0,mm2,mv1,0.0,mv3);
    t.test_rel_arr(5,mv2,mv3,1.0e-10,"solve 3");

    // -------------------------------------------------

    // Test invert

    for(size_t i=0;i<5;i++) {
      for(size_t j=0;j<5;j++) {
	gm1[i][j]=1.0/(1.0+i+j);
	om1[i][j]=1.0/(1.0+i+j);
	mm1[i][j]=1.0/(1.0+i+j);
	gm2[i][j]=1.0/(1.0+i+j);
	om2[i][j]=1.0/(1.0+i+j);
	mm2[i][j]=1.0/(1.0+i+j);
      }
    }
    gsl_linalg_cholesky_decomp(&gm1);
    gsl_linalg_cholesky_invert(&gm1);

    cholesky_decomp(5,om1);
    cholesky_invert<omatrix>(5,om1);
    
    cholesky_decomp(5,mm1);
    cholesky_invert<double[5][5]>(5,mm1);

    t.test_rel_mat(5,5,gm1,om1,1.0e-12,"cholesky invert 1");
    t.test_rel_mat(5,5,gm1,mm1,1.0e-12,"cholesky invert 2");
  }

  {
    using namespace o2scl_cblas_paren;
    using namespace o2scl_linalg_paren;

    omatrix gm1(5,5), om1(5,5), gm2(5,5), om2(5,5);
    ovector gv1(5), ov1(5), gv2(5), ov2(5), gv3(5), ov3(5);
      
    for(size_t i=0;i<5;i++) {
      gv1[i]=cos((double)(i));
      ov1[i]=cos((double)(i));
      gv2[i]=cos((double)(i));
      ov2[i]=cos((double)(i));
      for(size_t j=0;j<5;j++) {
	gm1[i][j]=1.0/(1.0+((double)i)+((double)j));
	om1[i][j]=1.0/(1.0+((double)i)+((double)j));
	gm2[i][j]=1.0/(1.0+((double)i)+((double)j));
	om2[i][j]=1.0/(1.0+((double)i)+((double)j));
      }
    }
      
    // Test cholesky decomposition
      
    gsl_linalg_cholesky_decomp(&gm1);
    cholesky_decomp(5,om1);
      
    t.test_rel_mat(5,5,gm1,om1,1.0e-12,"cholesky decomp");
      
    // Test solve 

    for(size_t i=0;i<5;i++) {
      gv1[i]=cos((double)(i));
      ov1[i]=cos((double)(i));
      gv2[i]=cos((double)(i));
      ov2[i]=cos((double)(i));
      for(size_t j=0;j<5;j++) {
	gm1[i][j]=1.0/(1.0+((double)i)+((double)j));
	om1[i][j]=1.0/(1.0+((double)i)+((double)j));
	gm2[i][j]=1.0/(1.0+((double)i)+((double)j));
	om2[i][j]=1.0/(1.0+((double)i)+((double)j));
      }
    }

    // -------------------------------------------------

    gsl_linalg_cholesky_decomp(&gm1);
    gsl_linalg_cholesky_solve(&gm1,&gv2,&gv1);

    dgemv(o2cblas_RowMajor,o2cblas_NoTrans,5,5,1.0,gm2,gv1,0.0,gv3);
    t.test_rel_arr(5,gv2,gv3,1.0e-10,"solve 1");
    
    // -------------------------------------------------

    cholesky_decomp(5,om1);
    cholesky_solve(5,om1,ov2,ov1);

    dgemv(o2cblas_RowMajor,o2cblas_NoTrans,5,5,1.0,om2,ov1,0.0,ov3);
    t.test_rel_arr(5,ov2,ov3,1.0e-10,"solve 2");

    // -------------------------------------------------

    // Test invert

    for(size_t i=0;i<5;i++) {
      for(size_t j=0;j<5;j++) {
	gm1[i][j]=1.0/(1.0+i+j);
	om1[i][j]=1.0/(1.0+i+j);
	gm2[i][j]=1.0/(1.0+i+j);
	om2[i][j]=1.0/(1.0+i+j);
      }
    }
    gsl_linalg_cholesky_decomp(&gm1);
    gsl_linalg_cholesky_invert(&gm1);

    cholesky_decomp(5,om1);
    cholesky_invert<omatrix>(5,om1);
    
    t.test_rel_mat(5,5,gm1,om1,1.0e-12,"cholesky invert 1");

  }

  t.report();
  return 0;
}

