/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.
  
  -------------------------------------------------------------------
*/
/* linalg/householder.c
 * 
 * Copyright (C) 1996, 1997, 1998, 1999, 2000, 2004, 
 * 2007 Gerard Jungman, Brian Gough
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 */
/** \file householder_base.h
    \brief File for Householder transformations

    \todo Better documentation for the Householder functions.
*/

#include <gsl/gsl_machine.h>

#include <o2scl/err_hnd.h>
#include <o2scl/cblas.h>
#include <o2scl/permutation.h>
#include <o2scl/vec_arith.h>

#ifdef DOXYGENP
namespace o2scl_linalg {
#endif
  
  /** \brief Replace the vector \c v with a Householder vector and a
      coefficient tau that annihilates <tt>v[1]</tt> through
      <tt>v[n-1]</tt> (inclusive)
      
      On exit, this function returns the value of \f$ \tau = 2/ (v^{T}
      v) \f$. If \c n is less than or equal to 1 then this function
      returns zero without calling the error handler.
  */
  template<class vec_t> 
    double householder_transform(const size_t n, vec_t &v) {

    if (n <= 1) {
      
      // tau=0 
      return 0.0;
      
    } else { 
      
      double alpha, beta, tau;
      double xnorm=O2SCL_CBLAS_NAMESPACE::dnrm2_subvec(n,v,1);
    
      if (xnorm == 0) {
	// tau=0 
	return 0.0;
      }
    
      alpha=O2SCL_IX(v,0);
      beta=-(alpha >= 0.0 ? +1.0 : -1.0)*hypot(alpha,xnorm);
      tau=(beta-alpha)/beta;
    
      double s=(alpha-beta);
      if (fabs(s)>GSL_DBL_MIN) {
	O2SCL_CBLAS_NAMESPACE::dscal_subvec(1.0/s,n,v,1);
	O2SCL_IX(v,0)=beta;
      } else {
	O2SCL_CBLAS_NAMESPACE::dscal_subvec(GSL_DBL_EPSILON/s,n,v,1);
	O2SCL_CBLAS_NAMESPACE::dscal_subvec(1.0/GSL_DBL_EPSILON,n,v,1);
	O2SCL_IX(v,0)=beta;
      }
      
      return tau;
    }
  }

  /** \brief Compute the Householder transform of a vector
      formed with \c n rows of a column of a matrix
      
      This performs a Householder transform of a vector defined by a
      column of a matrix \c A which starts at element
      <tt>A[ir][ic]</tt> and ends at element <tt>A[N-1][ic]</tt>.

      Used in QR_decomp().
  */
  template<class mat_t> 
    double householder_transform_subcol(mat_t &A, const size_t ir,
					const size_t ic, const size_t M) {
    
    if (M == ir+1) {
      /* tau=0 */
      return 0.0;
    } else { 
      double alpha, beta, tau;
      
      double xnorm=o2scl_cblas::dnrm2_subcol(A,ir+1,ic,M);
      
      if (xnorm == 0) {
	/* tau=0 */
	return 0.0;
      }
    
      alpha=O2SCL_IX2(A,ir,ic);
      beta=-(alpha >= 0.0 ? +1.0 : -1.0)*hypot(alpha,xnorm);
      tau=(beta-alpha)/beta;
      
      double s=(alpha-beta);
      if (fabs(s)>GSL_DBL_MIN) {
	o2scl_cblas::dscal_subcol(A,ir+1,ic,M,1.0/s);
	O2SCL_IX2(A,ir,ic)=beta;
      } else {
	o2scl_cblas::dscal_subcol(A,ir+1,ic,M,GSL_DBL_EPSILON/s);
	o2scl_cblas::dscal_subcol(A,ir+1,ic,M,1.0/GSL_DBL_EPSILON);
	O2SCL_IX2(A,ir,ic)=beta;
      }
    
      return tau;
    }
  }

  /** \brief Compute the Householder transform of a vector
      formed with the last \c n columns of a row of a matrix

      This performs a Householder transform of a vector 
      defined by a row of a matrix \c A which starts at element
      <tt>A[ir][ic]</tt> and ends at element <tt>A[ir][N-1]</tt>
  */
  template<class mat_t> 
    double householder_transform_subrow(mat_t &A, const size_t ir,
					const size_t ic, const size_t N) {
    
    if (N == ic+1) {
      /* tau=0 */
      return 0.0;
    } else { 
      double alpha, beta, tau;
      
      double xnorm=o2scl_cblas::dnrm2_subrow(A,ir,ic+1,N);
      
      if (xnorm == 0) {
	/* tau=0 */
	return 0.0;
      }
    
      alpha=O2SCL_IX2(A,ir,ic);
      beta=-(alpha >= 0.0 ? +1.0 : -1.0)*hypot(alpha,xnorm);
      tau=(beta-alpha)/beta;
      
      double s=(alpha-beta);
      if (fabs(s)>GSL_DBL_MIN) {
	o2scl_cblas::dscal_subrow(A,ir,ic+1,N,1.0/s);
	O2SCL_IX2(A,ir,ic)=beta;
      } else {
	o2scl_cblas::dscal_subrow(A,ir,ic+1,N,GSL_DBL_EPSILON/s);
	o2scl_cblas::dscal_subrow(A,ir,ic+1,N,1.0/GSL_DBL_EPSILON);
	O2SCL_IX2(A,ir,ic)=beta;
      }
    
      return tau;
    }
  }

  /** \brief Apply a Householder transformation \f$ (v,\tau) \f$ to 
      matrix \f$ A \f$ of size \f$ M \f$ by \f$ N \f$
  */
  template<class vec_t, class mat_t>
    void householder_hm(const size_t M, const size_t N, 
		       double tau, const vec_t &v, mat_t &A) {

    if (tau == 0.0) {
      return;
    }
    
    size_t i, j;
    
    for (j=0;j<N;j++) {
      
      /* Compute wj=Akj vk */
      double wj=O2SCL_IX2(A,0,j);
      
      for (i=1;i<M;i++) {
	wj+=O2SCL_IX2(A,i,j)*O2SCL_IX(v,i);
      }
      
      /* Aij=Aij-tau vi wj */
      
      /* i=0 */
      O2SCL_IX2(A,0,j)-=tau*wj;
      
      /* i=1 .. M-1 */
      for (i=1;i<M;i++) {
	O2SCL_IX2(A,i,j)-=tau*wj*O2SCL_IX(v,i);
      }
    }
    return;
  }

  /** \brief Apply a Householder transformation to submatrix of a
      larger matrix
      
      This applies a householder transformation <tt>(v,tau)</tt> to
      submatrix of \c M. The submatrix has \c nr rows and \c nc
      columns and starts at row \c ir of column \c ic of the original
      matrix \c M. The vector \c v is taken from a column of \c M2
      starting at row \c ir2 and column \c ic2.

      This function is used in \ref QR_decomp().
  */
  template<class mat_t>
    void householder_hm_sub(mat_t &M, const size_t ir,
			   const size_t ic, const size_t nr,
			   const size_t nc, const mat_t &M2,
			   const size_t ir2, const size_t ic2,
			   double tau) {

    if (tau == 0.0) {
      return;
    }
    
    size_t i, j, last=nr;
    
    for (j=ic;j<nc;j++) {
      
      /* Compute wj=Akj vk */
      double wj=O2SCL_IX2(M,ir,j);
      
      for (i=ir+1;i<last;i++) {
	wj+=O2SCL_IX2(M,i,j)*O2SCL_IX2(M2,i-ir+ir2,ic2);
      }
      
      /* Aij=Aij-tau vi wj */
      
      /* i=0 */
      O2SCL_IX2(M,ir,j)-=tau*wj;
      
      /* i=1 .. M-1 */
      for (i=ir+1;i<last;i++) {
	O2SCL_IX2(M,i,j)-=tau*wj*O2SCL_IX2(M2,i-ir+ir2,ic2);
      }
    }
    return;
  }

  /** \brief Special version of Householder transformation for
      QR_unpack()
  */
  template<class mat1_t, class mat2_t>
    void householder_hm_sub2(const size_t M, const size_t ic, 
			    double tau, const mat1_t &mv, mat2_t &A) {
    
    if (tau == 0.0) {
      return;
    }
    
    size_t i, j;
    
    for (j=ic;j<M;j++) {
      
      /* Compute wj=Akj vk */
      double wj=O2SCL_IX2(A,ic,j);
      
      for (i=ic+1;i<M;i++) {
	wj+=O2SCL_IX2(A,i,j)*O2SCL_IX2(mv,i,ic);
      }
      
      /* Aij=Aij-tau vi wj */
      
      /* i=0 */
      O2SCL_IX2(A,ic,j)-=tau*wj;
      
      /* i=1 .. M-1 */
      for (i=ic+1;i<M;i++) {
	O2SCL_IX2(A,i,j)-=tau*wj*O2SCL_IX2(mv,i,ic);
      }
    }
    return;
  }

  /** \brief Apply a Householder transformation \c v to vector \c w 
   */
  template<class vec_t, class vec2_t>
    void householder_hv(const size_t N, double tau, const vec_t &v, 
		       vec2_t &w) {
    
    if (tau==0) return;
    
    // compute d=v'w
    
    double d0=O2SCL_IX(w,0);
    double d1, d;
      
    d1=o2scl_cblas::ddot_subvec(N,v,w,1);
      
    d=d0+d1;
      
    // compute w=w-tau(v)(v'w)
    O2SCL_IX(w,0)-=tau*d;
    
    o2scl_cblas::daxpy_subvec(-tau*d,N,v,w,1);

    return;
  }

  /** \brief Apply a Householder transformation \c v to vector \c w 
      where \c v is stored as a column in a matrix \c A

      Used in QR_QTvec().
  */
  template<class mat_t, class vec_t>
    void householder_hv_subcol(const mat_t &A, vec_t &w, double tau,
			      const size_t ie, const size_t N) {
    
    if (tau==0) return;
    
    // compute d=v'w
    
    double d0=O2SCL_IX(w,ie);
    double d1, d;
      
    d1=o2scl_cblas::ddot_subcol(N,A,ie+1,ie,w);
      
    d=d0+d1;
      
    // compute w=w-tau(v)(v'w)
    O2SCL_IX(w,ie)-=tau*d;
    
    o2scl_cblas::daxpy_subcol(-tau*d,N,A,ie+1,ie,w);

    return;
  }

  /** \brief Apply a Householder transformation \f$ (v,\tau) \f$ to a
      matrix being build up from the identity matrix, using the first
      column of A as a Householder vector 
  */
  template<class mat_t> 
    void householder_hm1(const size_t M, const size_t N, 
			double tau, mat_t &A) {
    
    if (tau == 0) {
      size_t i,j;
      O2SCL_IX2(A,0,0)=1.0;
      for (j=1;j<N;j++) {
	O2SCL_IX2(A,0,j)=0.0;
      }
      for (i=1;i<M;i++) {
	O2SCL_IX2(A,i,0)=0.0;
      }
      return;
    }

    /* w=A' v */

    size_t i, j;
    for (j=1;j<N;j++) {
      double wj=0.0;
      for (i=1;i<M;i++) {
	wj+=O2SCL_IX2(A,i,j)*O2SCL_IX2(A,i,0);;
      }
      
      /* A=A-tau v w' */
      O2SCL_IX2(A,0,j)=-tau*wj;
      
      for (i=1;i<M;i++) {
	O2SCL_IX2(A,i,j)-=tau*wj*O2SCL_IX2(A,i,0);
      }
    }
    
    for (i=1;i<M;i++) {
      O2SCL_IX2(A,i,0)=-tau*O2SCL_IX2(A,i,0);
    }
    O2SCL_IX2(A,0,0)=1.0-tau;
    
    return;
  }
  
  /** \brief Apply the Householder transformation <tt>(v,tau)</tt> to
      the right-hand side of the matrix \c A.
  */
  template<class vec_t, class mat_t>
    void householder_mh(const size_t M, const size_t N,
		       double tau, const vec_t &v, mat_t &A) {
    
    if (tau==0.0) {
      return;
    }

    /* A=A-tau w  v' */
    size_t i, j;
    
    for (i=0;i<M;i++) {
      
      double wi=O2SCL_IX2(A,i,0);

      /* note, computed for v(0)=1 above */
      for (j=1;j<N;j++)  {
	wi+=O2SCL_IX2(A,i,j)*O2SCL_IX(v,j);
      }
      
      /* j=0 */
      O2SCL_IX2(A,i,0)-=tau*wi;
      
      /* j=1 .. N-1 */
      for (j=1;j<N;j++) {
	O2SCL_IX2(A,i,j)-=tau*wi*O2SCL_IX(v,j);
      }
    }
    
    return;
  }
  
  /** \brief Apply the Householder transformation <tt>(v,tau)</tt> to
      the right-hand side of the matrix \c A.
      
      This is unfinished but needed for the bidiagonalization
      functions.
  */
  template<class mat_t, class mat2_t>
    void householder_mh_sub(mat_t &M, const size_t ir, const size_t ic,
			    const size_t nr, const size_t nc, const mat2_t &M2,
			    const size_t ir2, const size_t ic2, double tau) {
    
    if (tau==0.0) {
      return;
    }

    /* A=A-tau w  v' */
    size_t i, j, last=ic+nc;
    
    for (i=ir;i<ir+nr;i++) {
      
      double wi=O2SCL_IX2(M,i,ic);

      /* note, computed for v(0)=1 above */
      for (j=ic+1;j<last;j++)  {
	wi+=O2SCL_IX2(M,i,j)*O2SCL_IX2(M2,ir2,j-ic+ic2);
      }
      
      /* j=0 */
      O2SCL_IX2(M,i,ic)-=tau*wi;
      
      /* j=1 .. N-1 */
      for (j=ic+1;j<last;j++) {
	O2SCL_IX2(M,i,j)-=tau*wi*O2SCL_IX2(M2,ir2,j-ic+ic2);
      }
    }

    return;
  }
  
#ifdef DOXYGENP
}
#endif
