/*
  -------------------------------------------------------------------

  Copyright (C) 2006-2012, Andrew W. Steiner

  This file is part of O2scl.

  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
/* linalg/svdstep.c
 * 
 * Copyright (C) 2007 Brian Gough
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */
/** \file svdstep_base.h
    \brief File for SVD decomposition
*/

#ifdef DOXYGENP
namespace o2scl_linalg {
#endif

  /** \brief Desc
      
      The parameter \c N is the size of \c d.
   */
  template<class vec_t, class vec2_t>
    int chop_small_elements(size_t N, vec_t &d, vec2_t &f) {
    
    double d_i=O2SCL_IX(d,0);
    
    size_t i;
    
    for (i = 0; i < N - 1; i++) {

      double f_i = O2SCL_IX(f,i);
      double d_ip1 = O2SCL_IX(d,i+1);
      
      if (fabs (f_i) < GSL_DBL_EPSILON * (fabs (d_i) + fabs (d_ip1))) {
	O2SCL_IX(f,i)=0.0;
      }
      d_i = d_ip1;
    }
    
    return o2scl::gsl_success;
  }
  
  /** \brief Desc
      
      The parameter \c n is the size of the vector \c d.

      Should be finished.
   */
  template<class vec_t, class vec2_t>
    double trailing_eigenvalue(size_t n, const vec_t &d, const vec_t &f) {
    
    double da = O2SCL_IX(d, n - 2);
    double db = O2SCL_IX(d, n - 1);
    double fa = (n > 2) ? O2SCL_IX(f, n - 3) : 0.0;
    double fb = O2SCL_IX(f, n - 2);
    
    double ta = da * da + fa * fa;
    double tb = db * db + fb * fb;
    double tab = da * fb;
    
    double dt = (ta - tb) / 2.0;
    
    double S = ta + tb;
    double da2 = da * da, db2 = db * db;
    double fa2 = fa * fa, fb2 = fb * fb;
    double P = (da2 * db2) + (fa2 * db2) + (fa2 * fb2);
    double D = hypot(dt, tab);
    double r1 = S/2 + D;
    
    double mu;
    if (dt >= 0) {
      /* tb < ta, choose smaller root */
      mu = (r1 > 0) ?  P / r1 : 0.0;
    } else {
      /* tb > ta, choose larger root */
      mu = r1;
    }
    return mu;
  }
  
  /** \brief Desc

      Should be finished.
   */
  int create_schur(double d0, double f0, double d1, double &c, 
		   double &s) {
    double apq = 2.0 * d0 * f0;
    
    if (d0 == 0 || f0 == 0) {
      c = 1.0;
      s = 0.0;
      return o2scl::gsl_success;
    }
    
    /* Check if we need to rescale to avoid underflow/overflow */
    if (fabs(d0) < GSL_SQRT_DBL_MIN || fabs(d0) > GSL_SQRT_DBL_MAX
	|| fabs(f0) < GSL_SQRT_DBL_MIN || fabs(f0) > GSL_SQRT_DBL_MAX
	|| fabs(d1) < GSL_SQRT_DBL_MIN || fabs(d1) > GSL_SQRT_DBL_MAX) {
      
      double scale;
      int d0_exp, f0_exp;
      frexp(d0, &d0_exp);
      frexp(f0, &f0_exp);
      /* Bring |d0*f0| into the range GSL_DBL_MIN to GSL_DBL_MAX */
      scale = ldexp(1.0, -(d0_exp + f0_exp)/4);
      d0 *= scale;
      f0 *= scale;
      d1 *= scale;
      apq = 2.0 * d0 * f0;
    }
    
    if (apq != 0.0) {
      double t;
      double tau = (f0*f0 + (d1 + d0)*(d1 - d0)) / apq;
      
      if (tau >= 0.0) {
	t = 1.0/(tau + hypot(1.0, tau));
      } else {
	t = -1.0/(-tau + hypot(1.0, tau));
      }
      
      c = 1.0 / hypot(1.0, t);
      s = t * (c);
    } else {
      c = 1.0;
      s = 0.0;
    }
    return o2scl::gsl_success;
  }
  
  /** \brief Desc
      
      The parameter \c M is the number of rows in \c U and \c N 
      is the number of rows in \c V.
  */
  template<class vec_t, class vec2_t, class mat_t, class mat2_t>
    int svd2(size_t M, size_t N, vec_t &d, vec2_t &f, mat_t &U, mat2_t &V) {

    size_t i;
    double c, s, a11, a12, a21, a22;
    
    double d0 = O2SCL_IX(d,0);
    double f0 = O2SCL_IX(f, 0);
    
    double d1 = O2SCL_IX(d, 1);
    
    if (d0 == 0.0) {
      
      /* Eliminate off-diagonal element in [0,f0;0,d1] to make [d,0;0,0] */
      o2scl_linalg::create_givens(f0,d1,c,s);
      
      /* compute B <= G^T B X,  where X = [0,1;1,0] */

      O2SCL_IX(d,0)=c * f0 - s * d1;
      O2SCL_IX(f,0)=s * f0 + c * d1;
      O2SCL_IX(d,1)=0.0;
      
      /* Compute U <= U G */
      
      for (i = 0; i < M; i++) {
	
	double Uip = O2SCL_IX2 (U, i, 0);
	double Uiq = O2SCL_IX2 (U, i, 1);
	O2SCL_IX2(U,i,0)=c * Uip - s * Uiq;
	O2SCL_IX2(U,i,1)=s * Uip + c * Uiq;
      }
      
      /* Compute V <= V X */

      double temp;
      for(size_t ik=0;ik<N;ik++) {
	temp=O2SCL_IX2(V,ik,0);
	O2SCL_IX2(V,ik,0)=O2SCL_IX2(V,ik,1);
	O2SCL_IX2(V,ik,1)=temp;
      }

      return o2scl::gsl_success;

    } else if (d1 == 0.0) {

      /* Eliminate off-diagonal element in [d0,f0;0,0] */

      o2scl_linalg::create_givens(d0,f0,c,s);

      /* compute B <= B G */

      O2SCL_IX (d, 0)= d0 * c - f0 * s;
      O2SCL_IX (f, 0)= 0.0;

      /* Compute V <= V G */

      for (i = 0; i < N; i++) {
	double Vip = O2SCL_IX2 (V, i, 0);
	double Viq = O2SCL_IX2 (V, i, 1);
	O2SCL_IX2 (V, i, 0)= c * Vip - s * Viq;
	O2SCL_IX2 (V, i, 1)= s * Vip + c * Viq;
      }
      
      return o2scl::gsl_success;

    } else {

      /* Make columns orthogonal, A = [d0, f0; 0, d1] * G */

      create_schur (d0, f0, d1, c, s);

      /* compute B <= B G */
      
      a11 = c * d0 - s * f0;
      a21 = - s * d1;
      
      a12 = s * d0 + c * f0;
      a22 = c * d1;
      
      /* Compute V <= V G */
      
      for (i = 0; i < N; i++) {
	
	double Vip = O2SCL_IX2 (V, i, 0);
	double Viq = O2SCL_IX2 (V, i, 1);
	O2SCL_IX2 (V, i, 0)= c * Vip - s * Viq;
	O2SCL_IX2 (V, i, 1)= s * Vip + c * Viq;
      }
      
      /* Eliminate off-diagonal elements, bring column with largest
         norm to first column */
      
      if (hypot(a11, a21) < hypot(a12,a22)) {
	
	double t1, t2;
	
	/* B <= B X */
	
	t1 = a11; a11 = a12; a12 = t1;
	t2 = a21; a21 = a22; a22 = t2;
	
	/* V <= V X */
	
	double temp;
	for(size_t ik=0;ik<N;ik++) {
	  temp=O2SCL_IX2(V,ik,0);
	  O2SCL_IX2(V,ik,0)=O2SCL_IX2(V,ik,1);
	  O2SCL_IX2(V,ik,1)=temp;
	}
      } 
      
      o2scl_linalg::create_givens(a11,a21,c,s);
      
      /* compute B <= G^T B */
      
      O2SCL_IX (d, 0)= c * a11 - s * a21;
      O2SCL_IX (f, 0)= c * a12 - s * a22;
      O2SCL_IX (d, 1)= s * a12 + c * a22;
      
      /* Compute U <= U G */
      
      for (i = 0; i < M; i++) {
	double Uip = O2SCL_IX2 (U, i, 0);
	double Uiq = O2SCL_IX2 (U, i, 1);
	O2SCL_IX2 (U, i, 0)= c * Uip - s * Uiq;
	O2SCL_IX2 (U, i, 1)= s * Uip + c * Uiq;
      }
      
      return o2scl::gsl_success;
    }
  }
  
  /** \brief Desc
      
      Should be finished.
   */
  template<class vec_t, class vec2_t, class mat_t>
    int chase_out_intermediate_zero(size_t M, size_t n, vec_t &d,
				    vec2_t &f, mat_t &U, size_t k0) {
    
    double c, s;
    double x, y;
    size_t k;
    
    x = O2SCL_IX(f, k0);
    y = O2SCL_IX(d, k0+1);
    
    for (k = k0; k < n - 1; k++) {
      
      o2scl_linalg::create_givens(y,-x,c,s);
      
      /* Compute U <= U G */
      for (size_t i = 0; i < M; i++) {
	double Uip = O2SCL_IX2 (U, i, k0);
	double Uiq = O2SCL_IX2 (U, i, k + 1);
	O2SCL_IX2 (U, i, k0)= c * Uip - s * Uiq;
	O2SCL_IX2 (U, i, k + 1)= s * Uip + c * Uiq;
      }
      
      /* compute B <= G^T B */
      
      O2SCL_IX (d, k + 1)= s * x + c * y;
      
      if (k == k0) {
        O2SCL_IX (f, k)= c * x - s * y ;
      }
      
      if (k < n - 2) {
	double z = O2SCL_IX(f, k + 1);
	O2SCL_IX (f, k + 1)= c * z; 
	
	x = -s * z ;
	y = O2SCL_IX(d, k + 2); 
      }
    }

    return o2scl::gsl_success;
  }

  /** \brief Desc
      
      Should be finished.
   */
  template<class vec_t, class vec2_t, class mat_t>
    int chase_out_trailing_zero(size_t N, size_t n, vec_t &d, 
				vec2_t &f, mat_t &V) {

    double c, s;
    double x, y;
    size_t k;
    
    x = O2SCL_IX(d, n - 2);
    y = O2SCL_IX(f, n - 2);
    
    for (k = n - 1; k-- > 0;) {
      
      o2scl_linalg::create_givens(x,y,c,s);
      
      /* Compute V <= V G where G = [c, s ; -s, c] */
      
      for (size_t i = 0; i < N; i++) {
	double Vip = O2SCL_IX2 (V, i, k);
	double Viq = O2SCL_IX2 (V, i, n - 1);
	O2SCL_IX2 (V, i, k)= c * Vip - s * Viq;
	O2SCL_IX2 (V, i, n - 1)= s * Vip + c * Viq;
      }
      
      /* compute B <= B G */
      
      O2SCL_IX (d, k)= c * x - s * y;
      
      if (k == n - 2)
        O2SCL_IX (f, k)= s * x + c * y ;
      
      if (k > 0) {
	double z = O2SCL_IX(f, k - 1);
	O2SCL_IX (f, k - 1)= c * z; 
	
	x = O2SCL_IX(d, k - 1); 
	y = s * z ;
      }
    }
    return o2scl::gsl_success;
  }
  
  /** \brief Desc
      
      Should be finished.

      The parameter \c M is the number of rows in \c U, \c N is the
      number of rows in \c V, and \c n is the length of the vector \c
      d.
   */
  template<class vec_t, class vec2_t, class mat_t, class mat2_t>
    int qrstep(size_t M, size_t N, size_t n, 
	       vec_t &d, vec2_t &f, mat_t &U, mat2_t &V) {
    
    double y, z;
    double ak, bk, zk, ap, bp, aq, bq;
    size_t i, k;
    
    if (n == 1) {
      /* shouldn't happen */
      return o2scl::gsl_success;
    }
    
    /* Compute 2x2 svd directly */
    
    if (n == 2) {
      svd2 (d, f, U, V);
      return o2scl::gsl_success;
    }
    
    /* Chase out any zeroes on the diagonal */
    
    for (i = 0; i < n - 1; i++) {
      double d_i = O2SCL_IX(d, i);
      if (d_i == 0.0) {
	chase_out_intermediate_zero (d, f, U, i);
	return o2scl::gsl_success;
      }
    }
    
    /* Chase out any zero at the end of the diagonal */
    {
      double d_nm1 = O2SCL_IX(d, n - 1);
      
      if (d_nm1 == 0.0) {
	chase_out_trailing_zero (d, f, V);
	return o2scl::gsl_success;
      }
    }
    

    /* Apply QR reduction steps to the diagonal and offdiagonal */
    
    {
      double d0 = O2SCL_IX(d, 0);
      double f0 = O2SCL_IX(f, 0);
      
      double d1 = O2SCL_IX(d, 1);
      double f1 = O2SCL_IX(f, 1);
      
      {
	double mu = trailing_eigenvalue (d, f);
	
	y = d0 * d0 - mu;
	z = d0 * f0;
      }
      
      /* Set up the recurrence for Givens rotations on a bidiagonal matrix */
      
      ak = 0;
      bk = 0;
      
      ap = d0;
      bp = f0;
      
      aq = d1;
      bq = f1;
    }
    
    for (k = 0; k < n - 1; k++) {

      double c, s;
      o2scl_linalg::create_givens(y,z,c,s);

      /* Compute V <= V G */
      
      for (i = 0; i < N; i++) {
	double Vip = O2SCL_IX2 (V, i, k);
	double Viq = O2SCL_IX2 (V, i, k + 1);
	O2SCL_IX2 (V, i, k)= c * Vip - s * Viq;
	O2SCL_IX2 (V, i, k + 1)= s * Vip + c * Viq;
      }
      
      /* compute B <= B G */
      
      {
        double bk1 = c * bk - s * z;

        double ap1 = c * ap - s * bp;
        double bp1 = s * ap + c * bp;
        double zp1 = -s * aq;

        double aq1 = c * aq;

        if (k > 0) {
	  O2SCL_IX (f, k - 1)= bk1;
	}
	
        ak = ap1;
        bk = bp1;
        zk = zp1;
	
        ap = aq1;
	
        if (k < n - 2) {
	  bp = O2SCL_IX(f, k + 1);
	} else {
	  bp = 0.0;
	}
	
        y = ak;
        z = zk;
      }
      
      o2scl_linalg::create_givens(y,z,c,s);
      
      /* Compute U <= U G */
      
      for (i = 0; i < M; i++) {
	double Uip = O2SCL_IX2 (U, i, k);
	double Uiq = O2SCL_IX2 (U, i, k + 1);
	O2SCL_IX2 (U, i, k)= c * Uip - s * Uiq;
	O2SCL_IX2 (U, i, k + 1)= s * Uip + c * Uiq;
      }
      
      /* compute B <= G^T B */
      
      {
        double ak1 = c * ak - s * zk;
        double bk1 = c * bk - s * ap;
        double zk1 = -s * bp;

        double ap1 = s * bk + c * ap;
        double bp1 = c * bp;

        O2SCL_IX (d, k)= ak1;

        ak = ak1;
        bk = bk1;
        zk = zk1;

        ap = ap1;
        bp = bp1;

        if (k < n - 2) {
	  aq = O2SCL_IX(d, k + 2);
	} else {
	  aq = 0.0;
	}

        y = bk;
        z = zk;
      }
    }

    O2SCL_IX (f, n - 2)= bk;
    O2SCL_IX (d, n - 1)= ap;

    return o2scl::gsl_success;
  }
  
#ifdef DOXYGENP
}
#endif
