/*
  -------------------------------------------------------------------

  Copyright (C) 2006-2012, Andrew W. Steiner

  This file is part of O2scl.

  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#include <iostream>

#include <gsl/gsl_linalg.h>

#include <o2scl/tridiag.h>
#include <o2scl/test_mgr.h>
#include <o2scl/ovector_tlate.h>

using namespace std;
using namespace o2scl;

// Test also a C-style pointer version to demonstrate that it
// can be used this way
template<class vec_t, class vec2_t, class vec3_t, class vec4_t,
	 class vec5_t> 
void solve_tridiag_nonsym_ptr(const vec_t &diag, const vec2_t &abovediag, 
			      const vec3_t &belowdiag, const vec4_t &rhs, 
			      vec5_t &x, size_t N) {
  o2scl_linalg::pointer_2_mem v2m;
  v2m.allocate(N);
  o2scl_linalg::solve_tridiag_nonsym<vec_t,vec2_t,vec3_t,vec4_t,vec5_t,
				     o2scl_linalg::pointer_2_mem,double *>
    (diag,abovediag,belowdiag,rhs,x,N,v2m);
  
  v2m.free();
  return;
}

int main(void) {
  test_mgr t;
  t.set_output_level(1);

  void *vpx=0;
  size_t tmp;

  cout.setf(ios::scientific);
  static const size_t arr_size=50;

  {
    using namespace o2scl_cblas;
    using namespace o2scl_linalg;

    // GSL
    ovector d(arr_size), od(arr_size-1), r(arr_size), s(arr_size);
  
    for(size_t i=0;i<arr_size;i++) {
      d[i]=i+1.0;
      if (i!=arr_size-1) od[i]=1.0;
      r[i]=1.0;
      s[i]=0.0;
    }
    int ret=gsl_linalg_solve_symm_tridiag(&d,&od,&r,&s);

    // O2scl with uvectors
    uvector du(arr_size), odu(arr_size-1), ru(arr_size), su(arr_size);
  
    for(size_t i=0;i<arr_size;i++) {
      du[i]=i+1.0;
      if (i!=arr_size-1) odu[i]=1.0;
      ru[i]=1.0;
      su[i]=0.0;
    }
    solve_tridiag_sym(du,odu,ru,su,arr_size);
  
    // O2scl with arrays
    double da[arr_size], oda[arr_size-1], ra[arr_size], sa[arr_size];
  
    for(size_t i=0;i<arr_size;i++) {
      da[i]=i+1.0;
      if (i!=arr_size-1) oda[i]=1.0;
      ra[i]=1.0;
      sa[i]=0.0;
    }
    solve_tridiag_sym<double[arr_size],
      double[arr_size-1]>(da,oda,ra,sa,arr_size);

    for(size_t i=0;i<arr_size;i++) {
      t.test_rel(s[i],su[i],1.0e-12,"o vs. u 1");
      t.test_rel(s[i],sa[i],1.0e-12,"o vs. a 1");
    }
  }

  {
    using namespace o2scl_cblas;
    using namespace o2scl_linalg;

    // GSL
    ovector d(arr_size), od(arr_size-1), r(arr_size), s(arr_size);
    ovector od2(arr_size-1);
  
    for(size_t i=0;i<arr_size;i++) {
      d[i]=i+1.0;
      if (i!=arr_size-1) od[i]=1.0;
      if (i!=arr_size-1) od2[i]=1.0;
      r[i]=1.0;
      s[i]=0.0;
    }
    int ret=gsl_linalg_solve_tridiag(&d,&od,&od2,&r,&s);

    // O2scl with uvectors
    uvector du(arr_size), odu(arr_size-1), ru(arr_size), su(arr_size);
    uvector odu2(arr_size-1);
  
    for(size_t i=0;i<arr_size;i++) {
      du[i]=i+1.0;
      if (i!=arr_size-1) odu[i]=1.0;
      if (i!=arr_size-1) odu2[i]=1.0;
      ru[i]=1.0;
      su[i]=0.0;
    }
    solve_tridiag_nonsym(du,odu,odu2,ru,su,arr_size);

    // O2scl with uvectors and internal pointers
    uvector dup(arr_size), odup(arr_size-1), rup(arr_size), sup(arr_size);
    uvector odup2(arr_size-1);
  
    for(size_t i=0;i<arr_size;i++) {
      dup[i]=i+1.0;
      if (i!=arr_size-1) odup[i]=1.0;
      if (i!=arr_size-1) odup2[i]=1.0;
      rup[i]=1.0;
      sup[i]=0.0;
    }
    solve_tridiag_nonsym_ptr(dup,odup,odup2,rup,sup,arr_size);
  
    // O2scl with arrays
    double da[arr_size], oda[arr_size-1], ra[arr_size], sa[arr_size];
    double oda2[arr_size-1];
  
    for(size_t i=0;i<arr_size;i++) {
      da[i]=i+1.0;
      if (i!=arr_size-1) oda[i]=1.0;
      if (i!=arr_size-1) oda2[i]=1.0;
      ra[i]=1.0;
      sa[i]=0.0;
    }
    solve_tridiag_nonsym<double[arr_size],
      double[arr_size-1]>(da,oda,oda2,ra,sa,arr_size);

    for(size_t i=0;i<arr_size;i++) {
      t.test_rel(s[i],su[i],1.0e-12,"o vs. u 2");
      t.test_rel(s[i],sup[i],1.0e-12,"o vs. up 2");
      t.test_rel(s[i],sa[i],1.0e-12,"o vs. a 2");
    }
  }

  {
    using namespace o2scl_cblas;
    using namespace o2scl_linalg;

    // GSL
    ovector d(arr_size), od(arr_size), r(arr_size), s(arr_size);
  
    for(size_t i=0;i<arr_size;i++) {
      d[i]=i+1.0;
      od[i]=1.0;
      r[i]=1.0;
      s[i]=0.0;
    }
    int ret=gsl_linalg_solve_symm_cyc_tridiag(&d,&od,&r,&s);

    // O2scl with uvectors
    uvector du(arr_size), odu(arr_size), ru(arr_size), su(arr_size);
  
    for(size_t i=0;i<arr_size;i++) {
      du[i]=i+1.0;
      odu[i]=1.0;
      ru[i]=1.0;
      su[i]=0.0;
    }
    solve_cyc_tridiag_sym(du,odu,ru,su,arr_size);
  
    // O2scl with arrays
    double da[arr_size], oda[arr_size], ra[arr_size], sa[arr_size];
  
    for(size_t i=0;i<arr_size;i++) {
      da[i]=i+1.0;
      oda[i]=1.0;
      ra[i]=1.0;
      sa[i]=0.0;
    }
    solve_cyc_tridiag_sym<double[arr_size]>(da,oda,ra,sa,arr_size);

    for(size_t i=0;i<arr_size;i++) {
      t.test_rel(s[i],su[i],1.0e-12,"o vs. u 3");
      t.test_rel(s[i],sa[i],1.0e-12,"o vs. a 3");
    }
  }

  {
    using namespace o2scl_cblas;
    using namespace o2scl_linalg;

    // GSL
    ovector d(arr_size), od(arr_size), r(arr_size), s(arr_size);
    ovector od2(arr_size);
  
    for(size_t i=0;i<arr_size;i++) {
      d[i]=i+1.0;
      od[i]=1.0;
      od2[i]=1.0;
      r[i]=1.0;
      s[i]=0.0;
    }
    int ret=gsl_linalg_solve_cyc_tridiag(&d,&od,&od2,&r,&s);

    // O2scl with uvectors
    uvector du(arr_size), odu(arr_size), ru(arr_size), su(arr_size);
    uvector odu2(arr_size);
  
    for(size_t i=0;i<arr_size;i++) {
      du[i]=i+1.0;
      odu[i]=1.0;
      odu2[i]=1.0;
      ru[i]=1.0;
      su[i]=0.0;
    }
    solve_cyc_tridiag_nonsym(du,odu,odu2,ru,su,arr_size);
  
    // O2scl with arrays
    double da[arr_size], oda[arr_size], ra[arr_size], sa[arr_size];
    double oda2[arr_size];
  
    for(size_t i=0;i<arr_size;i++) {
      da[i]=i+1.0;
      oda[i]=1.0;
      oda2[i]=1.0;
      ra[i]=1.0;
      sa[i]=0.0;
    }
    solve_cyc_tridiag_nonsym<double[arr_size]>(da,oda,oda2,ra,sa,arr_size);

    for(size_t i=0;i<arr_size;i++) {
      t.test_rel(s[i],su[i],1.0e-12,"o vs. u 4");
      t.test_rel(s[i],sa[i],1.0e-12,"o vs. a 4");
    }
  }

  {
    using namespace o2scl_cblas_paren;
    using namespace o2scl_linalg_paren;

    // GSL
    ovector d(arr_size), od(arr_size-1), r(arr_size), s(arr_size);
  
    for(size_t i=0;i<arr_size;i++) {
      d[i]=i+1.0;
      if (i!=arr_size-1) od[i]=1.0;
      r[i]=1.0;
      s[i]=0.0;
    }
    int ret=gsl_linalg_solve_symm_tridiag(&d,&od,&r,&s);

    // O2scl with uvectors
    uvector du(arr_size), odu(arr_size-1), ru(arr_size), su(arr_size);
  
    for(size_t i=0;i<arr_size;i++) {
      du[i]=i+1.0;
      if (i!=arr_size-1) odu[i]=1.0;
      ru[i]=1.0;
      su[i]=0.0;
    }
    solve_tridiag_sym(du,odu,ru,su,arr_size);
  
    for(size_t i=0;i<arr_size;i++) {
      t.test_rel(s[i],su[i],1.0e-12,"o vs. u 5");
    }
  }

  {
    using namespace o2scl_cblas_paren;
    using namespace o2scl_linalg_paren;

    // GSL
    ovector d(arr_size), od(arr_size-1), r(arr_size), s(arr_size);
    ovector od2(arr_size-1);
  
    for(size_t i=0;i<arr_size;i++) {
      d[i]=i+1.0;
      if (i!=arr_size-1) od[i]=1.0;
      if (i!=arr_size-1) od2[i]=1.0;
      r[i]=1.0;
      s[i]=0.0;
    }
    int ret=gsl_linalg_solve_tridiag(&d,&od,&od2,&r,&s);

    // O2scl with uvectors
    uvector du(arr_size), odu(arr_size-1), ru(arr_size), su(arr_size);
    uvector odu2(arr_size-1);
  
    for(size_t i=0;i<arr_size;i++) {
      du[i]=i+1.0;
      if (i!=arr_size-1) odu[i]=1.0;
      if (i!=arr_size-1) odu2[i]=1.0;
      ru[i]=1.0;
      su[i]=0.0;
    }
    solve_tridiag_nonsym(du,odu,odu2,ru,su,arr_size);
  
    for(size_t i=0;i<arr_size;i++) {
      t.test_rel(s[i],su[i],1.0e-12,"o vs. u 6");
    }
  }

  {
    using namespace o2scl_cblas_paren;
    using namespace o2scl_linalg_paren;

    // GSL
    ovector d(arr_size), od(arr_size), r(arr_size), s(arr_size);
  
    for(size_t i=0;i<arr_size;i++) {
      d[i]=i+1.0;
      od[i]=1.0;
      r[i]=1.0;
      s[i]=0.0;
    }
    int ret=gsl_linalg_solve_symm_cyc_tridiag(&d,&od,&r,&s);

    // O2scl with uvectors
    uvector du(arr_size), odu(arr_size), ru(arr_size), su(arr_size);
  
    for(size_t i=0;i<arr_size;i++) {
      du[i]=i+1.0;
      odu[i]=1.0;
      ru[i]=1.0;
      su[i]=0.0;
    }
    solve_cyc_tridiag_sym(du,odu,ru,su,arr_size);
  
    for(size_t i=0;i<arr_size;i++) {
      t.test_rel(s[i],su[i],1.0e-12,"o vs. u 7");
    }
  }

  {
    using namespace o2scl_cblas_paren;
    using namespace o2scl_linalg_paren;

    // GSL
    ovector d(arr_size), od(arr_size), r(arr_size), s(arr_size);
    ovector od2(arr_size);
  
    for(size_t i=0;i<arr_size;i++) {
      d[i]=i+1.0;
      od[i]=1.0;
      od2[i]=1.0;
      r[i]=1.0;
      s[i]=0.0;
    }
    int ret=gsl_linalg_solve_cyc_tridiag(&d,&od,&od2,&r,&s);

    // O2scl with uvectors
    uvector du(arr_size), odu(arr_size), ru(arr_size), su(arr_size);
    uvector odu2(arr_size);
  
    for(size_t i=0;i<arr_size;i++) {
      du[i]=i+1.0;
      odu[i]=1.0;
      odu2[i]=1.0;
      ru[i]=1.0;
      su[i]=0.0;
    }
    solve_cyc_tridiag_nonsym(du,odu,odu2,ru,su,arr_size);
  
    for(size_t i=0;i<arr_size;i++) {
      t.test_rel(s[i],su[i],1.0e-12,"o vs. u 8");
    }
  }

  t.report();
  return 0;
}

