/*
  -------------------------------------------------------------------
  
  Copyright (C) 2011-2012 Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/expect_val.h>

using namespace std;
using namespace o2scl;

expect_val::expect_val() {
  nblocks=1;
  nperblock=0;
  iblock=0;
  i=0;
}

expect_val::expect_val(size_t n_blocks, size_t n_per_block) {
  if (n_blocks==0) {
    O2SCL_ERR("Value of zero for n_blocks in expect_val::expect_val().",
	      gsl_einval);
  }
  nblocks=n_blocks;
  nperblock=n_per_block;
  iblock=0;
  i=0;
}

expect_val::~expect_val() {
}

expect_val::expect_val(const expect_val &ev) {
  nblocks=ev.nblocks;
  nperblock=ev.nperblock;
  iblock=ev.iblock;
  i=ev.i;
  name=ev.name;
  short_name=ev.short_name;
}

expect_val &expect_val::operator=(const expect_val &ev) {
  if (&ev!=this) {
    nblocks=ev.nblocks;
    nperblock=ev.nperblock;
    iblock=ev.iblock;
    i=ev.i;
    name=ev.name;
    short_name=ev.short_name;
  }
  return *this;
}

void expect_val::set_blocks(size_t n_blocks, size_t n_per_block) {
  if (n_blocks==0) {
    O2SCL_ERR("Value of zero for n_blocks in expect_val::set_blocks().",
	      gsl_einval);
  }
  if (i!=0 || iblock!=0) {
    O2SCL_ERR2("Cannot reblock before reset() in ",
	       "expect_val::set_blocks().",gsl_einval);
  }
  free();
  nblocks=n_blocks;
  nperblock=n_per_block;
  return;
}

void expect_val::get_blocks(size_t &n_blocks, size_t &n_per_block) const {
  n_blocks=nblocks;
  n_per_block=nperblock;
  return;
}

void expect_val::free() {
  return;
}

void expect_val::reset() {
  i=0;
  iblock=0;
  return;
}

void expect_val::get_block_indices(size_t &i_block, 
				   size_t &i_curr_block) const {
  i_block=iblock;
  i_curr_block=i;
  return;
}

bool expect_val::finished() const {
  if (nperblock>0) {
    if (iblock==nblocks) return true;
  } else {
    if (iblock==0 && i>0) return true;
  }
  return false;
}

double expect_val::progress() const {
  if (nperblock==0) {
    if (iblock==0) {
      if (i==0) return 0.0;
      else return 1.0;
    }
    return ((double)iblock)/((double)nblocks);
  }
  return ((double)(iblock*nperblock+i))/((double)nblocks)/
    ((double)(nperblock));
}

scalar_ev::scalar_ev() : expect_val() {
  vals.allocate(this->nblocks);
  vals.set_all(0.0);
  current=0.0;
  last.allocate(this->nblocks);
  last.set_all(0.0);
}

scalar_ev::scalar_ev(size_t n_blocks, size_t n_per_block) : 
  expect_val(n_blocks,n_per_block) {
  vals.allocate(n_blocks);
  vals.set_all(0.0);
  current=0.0;
  if (n_per_block==0) {
    last.allocate(n_blocks);
    last.set_all(0.0);
  }
}

scalar_ev::~scalar_ev() {
  free();
}

scalar_ev::scalar_ev(const scalar_ev &ev) : expect_val() {
  nblocks=ev.nblocks;
  nperblock=ev.nperblock;
  iblock=ev.iblock;
  i=ev.i;
  name=ev.name;
  short_name=ev.short_name;
  current=ev.current;
  vals=ev.vals;
  last=ev.last;
}

scalar_ev &scalar_ev::operator=(const scalar_ev &ev) {
  if (&ev!=this) {
    nblocks=ev.nblocks;
    nperblock=ev.nperblock;
    iblock=ev.iblock;
    i=ev.i;
    name=ev.name;
    short_name=ev.short_name;
  current=ev.current;
  vals=ev.vals;
  last=ev.last;
  }
  return *this;
}

void scalar_ev::set_blocks(size_t n_blocks, size_t n_per_block) {
  // The parent calls the error handler if reset() has not been
  // called and then calls free()
  expect_val::set_blocks(n_blocks,n_per_block);

  // Allocate vectors
  vals.allocate(nblocks);
  vals.set_all(0.0);
  if (n_per_block==0) {
    last.allocate(n_blocks);
    last.set_all(0.0);
  }
  return;
}

void scalar_ev::free() {
  expect_val::free();
  vals.free();
  last.free();
  return;
}

void scalar_ev::reset() {
  expect_val::reset();
  current=0.0;
  vals.set_all(0.0);
  if (last.size()>0) last.set_all(0.0);
  return;
}

void scalar_ev::add(double val) {

  // Handle varable block size case separately
  if (nperblock==0) {

    // Store in last vector and increment block index
    last[iblock]=val;
    iblock++;

    // If last vector is full, add to vals, increment index,
    // and reset block counter
    if (iblock==nblocks) {
      for(size_t j=0;j<nblocks;j++) {
	// Note that when i is zero, one might naively think the
	// previous value of vals[j] cancels out in this recurrence
	// relation and is ignored. However, if the previous value of
	// vals[j] is NaN, then the cancellation fails to occur.
	vals[j]+=(last[j]-vals[j])/(i+1);
      }
      i++;
      iblock=0;
    }

    return;
  }

  if (iblock==nblocks) {
    O2SCL_ERR("Too much data provided in scalar_ev::add().",
	      gsl_einval);
  }

  // Keep track of the rolling average and increment the index
  current+=(val-current)/(i+1);
  i++;

  // If the block is full, store the average and move to the next block
  if (i==nperblock) {
    vals[iblock]=current;
    iblock++;
    current=0.0;
    i=0;
  }

  return;
}

void scalar_ev::current_avg_stats(double &avg, double &std_dev, 
				  double &avg_err, size_t &m_block,
				  size_t &m_per_block) const {
  
  if (nperblock==0) {
	
    // Unspecified block size

    if (i<1) {

      if (iblock==0) {

	// No data
	O2SCL_ERR("No data in scalar_ev::current_avg_stats().",
		  gsl_einval);

      } else if (iblock==1) {

	// Only one data point
	m_block=1;
	m_per_block=1;
	avg=last[0];
	std_dev=0.0;
	avg_err=0.0;

	return;
      }
	  
      // Only one data point for some of the blocks
      m_block=iblock;
      m_per_block=1;
      avg=vector_mean(iblock,last);
      std_dev=vector_stddev(iblock,last);
      avg_err=std_dev/sqrt(((double)iblock));

      return;
    }

    // General case
    m_block=nblocks;
    m_per_block=i;
    avg=vector_mean(nblocks,vals);
    std_dev=vector_stddev(nblocks,vals);
    avg_err=std_dev/sqrt(((double)nblocks));

    return;

  } else if (iblock==0) {

    // We're blocking, but don't have any full blocks
    m_block=1;
    m_per_block=i;
    avg=current;
    std_dev=0.0;
    avg_err=0.0;

    return;

  } else if (iblock==1) {

    // We're blocking, but have only one full block so far
    m_block=1;
    m_per_block=nperblock;
    avg=vals[0];
    std_dev=0.0;
    avg_err=0.0;

    return;
  }

  // In the blocking case, report the average and std. dev.
  // for all the blocks which have been finished
  m_block=iblock;
  m_per_block=nperblock;
  avg=vector_mean(iblock,vals);
  std_dev=vector_stddev(iblock,vals);
  avg_err=std_dev/sqrt(((double)iblock));
      
  return;
}

void scalar_ev::current_avg(double &avg, double &std_dev, 
			    double &avg_err) const {
  size_t m_block, m_per_block;
  return current_avg_stats(avg,std_dev,avg_err,m_block,m_per_block);
} 

void scalar_ev::reblock_avg_stats(size_t new_blocks, double &avg, 
				  double &std_dev, double &avg_err,
				  size_t &m_per_block) const {
      
  if (new_blocks==0) {
    O2SCL_ERR2("Requested zero blocks in ",
	       "scalar_ev::reblock_avg_stats().",gsl_einval);
  }
  
  uvector dat(new_blocks);
  dat.set_all(0.0);
  
  // If 'i==0' and 'nperblock==0', then not all of the blocks have
  // data yet, and we use the information from 'last' instead of
  // 'vals'.
  
  // The ratio of the old to new block size
  size_t fact=nblocks/new_blocks;
  if (i==0 && nperblock==0) fact=iblock/new_blocks;
  if (fact==0) {
    O2SCL_ERR2("Not enough data for reblocking ",
	       "in scalar_ev::reblock_avg_stats().",gsl_einval);
  }
  size_t iblock2=0;
  // Compute the sum
  for(size_t j=0;j<fact;j++) {
    for(size_t k=0;k<new_blocks;k++) {
      if (i==0 && nperblock==0) {
	dat[k]+=last[iblock2];
      } else {
	dat[k]+=vals[iblock2];
      }
      iblock2++;
    }
  }
  // Now divide to get averages
  for(size_t k=0;k<new_blocks;k++) {
    dat[k]/=((double)fact);
  }
  avg=vector_mean(new_blocks,dat);
  if (new_blocks>1) {
    std_dev=vector_stddev(new_blocks,dat);
    avg_err=std_dev/sqrt(((double)new_blocks));
  } else {
    std_dev=0.0;
    avg_err=0.0;
  }
  
  if (i==0 && nperblock==0) m_per_block=fact;
  else if (nperblock==0) m_per_block=fact*i;
  else m_per_block=fact*nperblock;
  
  return;
}

void scalar_ev::reblock_avg(size_t new_blocks, double &avg, 
			    double &std_dev, double &avg_err) const {
  size_t m_per_block;
  return reblock_avg_stats(new_blocks,avg,std_dev,avg_err,m_per_block);
}

const uvector &scalar_ev::get_data() const {
  return vals;
}

const double &scalar_ev::operator[](size_t i_block) const {
  return vals[i_block];
}

double &scalar_ev::operator[](size_t i_block) {
  return vals[i_block];
}

vector_ev::vector_ev(size_t n) : expect_val() {
  if (n==0) {
    O2SCL_ERR2("Tried to use zero-length vector in vector_ev::",
	       "vector_ev().",gsl_einval);
  }
  nvec=n;
  current.allocate(nvec);
  current.set_all(0.0);
  vals.allocate(nvec,this->nblocks);
  vals.set_all(0.0);
  last.allocate(nvec,this->nblocks);
  last.set_all(0.0);
}

vector_ev::vector_ev(size_t n, size_t n_blocks, size_t n_per_block) : 
  expect_val(n_blocks,n_per_block) {
  nvec=n;
  if (n_per_block==0) {
    current.allocate(nvec);
    current.set_all(0.0);
    vals.allocate(nvec,this->nblocks);
    last.allocate(nvec,this->nblocks);
    vals.set_all(0.0);
    last.set_all(0.0);
  } else {
    current.allocate(nvec);
    current.set_all(0.0);
    vals.allocate(nvec,this->nblocks);
    vals.set_all(0.0);
  }
}

vector_ev::~vector_ev() {
  free();
}

vector_ev::vector_ev(const vector_ev &ev) : expect_val() {
  nblocks=ev.nblocks;
  nperblock=ev.nperblock;
  iblock=ev.iblock;
  i=ev.i;
  name=ev.name;
  short_name=ev.short_name;
  nvec=ev.nvec;
  current=ev.current;
  vals=ev.vals;
  last=ev.last;
}

vector_ev &vector_ev::operator=(const vector_ev &ev) {
  if (&ev!=this) {
    nblocks=ev.nblocks;
    nperblock=ev.nperblock;
    iblock=ev.iblock;
    i=ev.i;
    name=ev.name;
    short_name=ev.short_name;
    nvec=ev.nvec;
    current=ev.current;
    vals=ev.vals;
    last=ev.last;
  }
  return *this;
}

void vector_ev::set_blocks(size_t n, size_t n_blocks, size_t n_per_block) {

  // The parent calls the error handler if reset() has not been
  // called and then calls free()
  expect_val::set_blocks(n_blocks,n_per_block);

  nvec=n;
  if (n_per_block==0) {
    current.allocate(nvec);
    current.set_all(0.0);
    vals.allocate(nvec,this->nblocks);
    last.allocate(nvec,this->nblocks);
    vals.set_all(0.0);
    last.set_all(0.0);
  } else {
    current.allocate(nvec);
    current.set_all(0.0);
    vals.allocate(nvec,this->nblocks);
    vals.set_all(0.0);
  }
  return;
}

void vector_ev::free() {
  expect_val::free();
  vals.free();
  last.free();
  current.free();
  return;
}

void vector_ev::reset() {
  expect_val::reset();
  current.set_all(0.0);
  vals.set_all(0.0);
  if (last.rows()>0 && last.cols()>0) last.set_all(0.0);
  return;
}
