/*
  -------------------------------------------------------------------
  
  Copyright (C) 2011-2012 Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef EXPECT_VAL_H
#define EXPECT_VAL_H

#include <o2scl/uvector_tlate.h>
#include <o2scl/umatrix_tlate.h>
#include <o2scl/vec_stats.h>

#ifndef DOXYGENP
namespace o2scl {
#endif
  
  /** \brief Expectation value base class
      
      This class is experimental.
      
      This is a base class for a set of classes useful for recording
      the outputs of several iterations of a numerical simulation, and
      summarizing the average, standard deviation, and error in the
      average over all iterations. After each iteration, some
      measurement is performed, and that measurement is added to the
      class with an <tt>add()</tt> functions (e.g. \ref
      scalar_ev::add() ). Autocorrelations are common in numerical
      simulations, and to handle this difficulty the measurements may
      be demarcated into 'blocks'. These blocks may have a variable or
      fixed number of measurements in each block.

      The constructor needs as input the number of blocks to record
      (at least one) and the number of measurements per block. The
      number of measurements per block can be zero to indicate an
      unspecified number of blocks to record.

      If the number of measurements per block is greater than zero,
      then blocks are filled one by one, moving to the next block when
      the requested of measurements (<tt>n_per_block</tt>) in the
      previous block has been provided. If the number of measurements
      per block is zero, then the blocks are filled evenly.

      Current averages are always reported, though not all data is
      always used if there are incomplete blocks or the blocks have
      not yet been filled evenly. Two functions <tt>current_avg()</tt>
      and <tt>current_avg_stats()</tt> are provided in children. The
      latter provides the number of blocks and number of measurements
      per block used in the currently reported statistics. If only one
      block or one measurement is available, the standard deviation is
      reported as zero. If no data is available, then calls to
      <tt>current_avg()</tt> will call the error handler. See, e.g.
      \ref scalar_ev::current_avg() and \ref
      scalar_ev::current_avg_stats().

      If <tt>n_per_block</tt> is nonzero, then children call the error
      handler if they get more than <tt>n_per_block</tt> times
      <tt>n_blocks</tt> measurements.

      \future If \c n_per_block is nonzero and the user adds more
      than \c n_blocks times \c n_per_block data, set \c n_per_block
      to zero, allocate the \c last structure and fall back to the 
      variable block size method. 
  */
  class expect_val {

  protected:

    /// Index denoting the current block number
    size_t iblock;
    
    /// Index for the number of values in the current block (or blocks)
    size_t i;

    /** \brief Total number of blocks (default 1)
	
	This should never be zero.
    */
    size_t nblocks;
    
    /** \brief Number of measurements per block (default 0)

	This is zero to indicate blocks of unspecified size
    */
    size_t nperblock;

  public:

    /// Create with one block of unspecified size
    expect_val();

    /// Copy constructor
    expect_val(const expect_val &ev);

    /// Copy constructor
    expect_val &operator=(const expect_val &ev);

    /** \brief Create with \c n_blocks blocks and \c n_per_block points
	per block

	If this is called with a value of zero for \c n_blocks, then
	the error handler is called.
    */
    expect_val(size_t n_blocks, size_t n_per_block);

    virtual ~expect_val();
    
    /// The name of the expectation value
    std::string name;
    
    /// The shortened name
    std::string short_name;

    /** \brief Reset for \c n_blocks blocks and \c n_per_block points
	per block

	This function resets the currently stored data to zero
	by calling \ref reset(). If this is called with a value
	of zero for \c n_blocks, then the value 1 is assumed.
    */
    virtual void set_blocks(size_t n_blocks, size_t n_per_block);

    /** \brief Get the number of blocks and the number of points per
	block
    */
    virtual void get_blocks(size_t &n_blocks, size_t &n_per_block) const;
    
    /// Free allocated data
    virtual void free();

    /// Clear all the data
    virtual void reset();

    /** \brief Get the block index and the index within the current block
     */
    virtual void get_block_indices(size_t &i_block, 
				   size_t &i_curr_block) const;

    /** \brief Returns true if all blocks have been stored

	If \c n_per_block is greater than zero, then this reports true
	only when all \c n_blocks times \c n_per_block data points
	have been added.

	Otherwise, this reports true whenever the last addition of
	data filled up the last block.
    */
    virtual bool finished() const;

    /** \brief Report progress as a fraction between zero to one 
	(inclusive)
	
	When \c n_per_block is zero, this reports the progress in
	having filled a round of blocks (i.e. whenever the last
	addition of data filled up the last block). This reports zero
	if no data has been added.

	When \c n_per_block is nonzero, this reports the total
	progress on all blocks, reporting \c 1.0 only when all \c
	n_blocks times \c n_per_block data points have been added.
    */
    virtual double progress() const;

  };
  
  /** \brief Scalar expectation value

      See \ref expect_val for some general notes on 
      this and related classes. 

      This class is experimental.

      \hline
      <b>Implementation notes</b>

      In the case where <tt>nperblock</tt> is zero, the vector \ref
      last is used to store the most recent measurements, copying the
      results to \ref vals only when <tt>nblocks</tt> measurements are
      stored in \ref last. The variable <tt>i</tt> is only incremented
      when the vector \ref last is full and the information is copied
      over to \ref vals. This ensures that \ref vals always contains
      averages with an equal number of measurements. The information
      in \ref last is never used by \ref reblock_avg().
  */
  class scalar_ev : public expect_val {
    
  protected:
    
    /** \brief The average for each block
	
	This is a vector of length \c nblocks.
    */
    uvector vals;

    /// The current rolling average
    double current;

    /** \brief The most recent values for each block
	
	This is a vector of length \c nblocks used for when
	\c n_per_block is set to zero.
    */
    uvector last;

  public:
    
    /// Create with one unspecified block
    scalar_ev();
    
    /** \brief Create with \c n_blocks blocks and \c n_per_block points
	block
    */
    scalar_ev(size_t n_blocks, size_t n_per_block);

    virtual ~scalar_ev();
    
    /// Copy constructor
    scalar_ev(const scalar_ev &ev);

    /// Copy constructor
    scalar_ev &operator=(const scalar_ev &ev);

    /** \brief Reset for \c n_blocks blocks and \c n_per_block points
	block
    */
    virtual void set_blocks(size_t n_blocks, size_t n_per_block);

    /// Free allocated data
    virtual void free();

    /// Clear all the data
    virtual void reset();

    /// Add measurement of value \c val
    virtual void add(double val);

    /// \name Report statistics
    //@{
    /** \brief Report current average, standard deviation, and 
	the error in the average and include block information
    */
    virtual void current_avg_stats(double &avg, double &std_dev, 
				   double &avg_err, size_t &m_block,
				   size_t &m_per_block) const;

    /** \brief Report current average, standard deviation, and 
	the error in the average
    */
    virtual void current_avg(double &avg, double &std_dev, 
			     double &avg_err) const;
    
    /** \brief Report average, standard deviation, and 
	the error in the average assuming a new block size

	\future Use recurrence relation for averages here
	rather than dividing at the end.
    */
    virtual void reblock_avg_stats(size_t new_blocks, double &avg, 
				   double &std_dev, double &avg_err,
				   size_t &m_per_block) const;

    /** \brief Report average, standard deviation, and 
	the error in the average assuming a new block size
    */
    virtual void reblock_avg(size_t new_blocks, double &avg, 
			     double &std_dev, double &avg_err) const;
    //@}

    /// \name Direct manipulation of the stored data 
    //@{
    /// Return the current data for all blocks
    const uvector &get_data() const;

    /// Return the current data for block with index \c i_block
    const double &operator[](size_t i_block) const;

    /// Return the current data for block with index \c i_block
    double &operator[](size_t i_block);

    /// Set the data for all blocks
    template<class vec_t> void set_data(vec_t &v) {
      for(size_t ib=0;ib<nblocks;ib++) {
	vals[ib]=v[ib];
      }
      return;
    }
    //@}

  };

  /** \brief Vector expectation value

      See \ref expect_val for some general notes on 
      this and related classes. 

      This class is experimental.

      This is a similar to \ref scalar_ev, except that it allows
      updating and statistics for a set of scalars en masse. The data
      is stored internally in \ref uvector and \ref umatrix objects,
      but the public member functions operate with template types
      which are compatible with any vector class which provides
      <tt>double &operator[]</tt>.

      Internally, data is stored in matrices designating the first
      (row) index as the user-specified vector index and designating
      the second (column) index as the block index.
  */
  class vector_ev : public expect_val {
    
  protected:
    
    /** \brief The most recent values for each block
     */
    umatrix last;

    /** \brief The average for each block
     */
    umatrix vals;
    
    /// The current rolling average
    uvector current;

    /// The size of the vector
    size_t nvec;

  public:
    
    /// Create for a vector of size \c n with no blocking
    vector_ev(size_t n);

    /** \brief Create for a vector of size \c n with \c n_blocks
	blocks and \c n_per_block points block
    */
    vector_ev(size_t n, size_t n_blocks, size_t n_per_block);

    /** \brief Set for a vector of size \c n with \c n_blocks blocks
	and \c n_per_block points block
    */
    virtual void set_blocks(size_t n, size_t n_blocks, size_t n_per_block);

    virtual ~vector_ev();

    /// Copy constructor
    vector_ev(const vector_ev &ev);

    /// Copy constructor
    vector_ev &operator=(const vector_ev &ev);

    /// Free allocated data
    virtual void free();

    /// Clear all the data
    virtual void reset();

    /// Add measurement of value \c val
    template<class vec_t> void add(vec_t &v) {

      // Handle varable block size case separately
      if (nperblock==0) {

	// Store in last vector and increment block index
	for(size_t iv=0;iv<nvec;iv++) {
	  last[iv][iblock]=v[iv];
	}
	iblock++;

	// If last vector is full, add to vals, increment index,
	// and reset block counter
	if (iblock==nblocks) {
	  for(size_t j=0;j<nblocks;j++) {
	    for(size_t k=0;k<nvec;k++) {
	      vals[k][j]+=(last[k][j]-vals[k][j])/(i+1);
	    }
	  }
	  i++;
	  iblock=0;
	}

	return;
      }

      if (iblock==nblocks) {
	O2SCL_ERR("Too much data provided in scalar_ev::add().",
		  gsl_einval);
      }

      // Keep track of the rolling average and increment the index
      for(size_t k=0;k<nvec;k++) {
	current[k]+=(v[k]-current[k])/(i+1);
      }
      i++;

      // If the block is full, store the average and move to the next block
      if (i==nperblock) {
	for(size_t iv=0;iv<nvec;iv++) {
	  vals[iv][iblock]=current[iv];
	}
	iblock++;
	for(size_t k=0;k<nvec;k++) {
	  current[k]=0.0;
	}
	i=0;
      }

      return;
    }

    /// \name Report statistics
    //@{
    /** \brief Report current average, standard deviation, and 
	the error in the average and include block information
    */
    template<class vec_t, class vec2_t, class vec3_t> 
      void current_avg_stats(vec_t &avg, vec2_t &std_dev, 
			     vec3_t &avg_err, size_t &m_block,
			     size_t &m_per_block) const {
      
      if (nperblock==0) {

	// Unspecified block size

	if (i<1) {

	  if (iblock==0) {

	    // No data
	    O2SCL_ERR("No data in scalar_ev::current_avg_stats().",
		      gsl_einval);

	  } else if (iblock==1) {

	    // Only one data point
	    m_block=1;
	    m_per_block=1;
	    for(size_t k=0;k<nvec;k++) {
	      avg[k]=last[k][0];
	      std_dev[k]=0.0;
	      avg_err[k]=0.0;
	    }

	    return;
	  }
          
	  // Only one data point for some of the blocks
	  m_block=iblock;
	  m_per_block=1;
	  for(size_t k=0;k<nvec;k++) {
	    avg[k]=vector_mean(iblock,last[k]);
	    std_dev[k]=vector_stddev(iblock,last[k]);
	    avg_err[k]=std_dev[k]/sqrt(((double)iblock));
	  }

	  return;
	}

	// General case
	m_block=nblocks;
	m_per_block=i;
	for(size_t k=0;k<nvec;k++) {
	  avg[k]=vector_mean(nblocks,vals[k]);
	  std_dev[k]=vector_stddev(nblocks,vals[k]);
	  avg_err[k]=std_dev[k]/sqrt(((double)nblocks));
	}

	return;

      } else if (iblock==0) {

	// We're blocking, but don't have any full blocks
	m_block=1;
	m_per_block=i;
	for(size_t k=0;k<nvec;k++) {
	  avg[k]=current[k];
	  std_dev[k]=0.0;
	  avg_err[k]=0.0;
	}

	return;
      } else if (iblock==1) {

	// We're blocking, but have only one full block so far
	m_block=1;
	m_per_block=nperblock;
	for(size_t k=0;k<nvec;k++) {
	  avg[k]=vals[k][0];
	  std_dev[k]=0.0;
	  avg_err[k]=0.0;
	}

	return;
      }

      // In the blocking case, report the average and std. dev.
      // for all the blocks which have been finished
      m_block=iblock;
      m_per_block=nperblock;
      for(size_t k=0;k<nvec;k++) {
	avg[k]=vector_mean(iblock,vals[k]);
	std_dev[k]=vector_stddev(iblock,vals[k]);
	avg_err[k]=std_dev[k]/sqrt(((double)iblock));
      }
      
      return;
    }

    /** \brief Report current average, standard deviation, and 
	the error in the average
    */
    template<class vec_t, class vec2_t, class vec3_t> 
      void current_avg(vec_t &avg, vec2_t &std_dev, vec3_t &avg_err) const {
      size_t m_per_block, m_block;
      return current_avg_stats(avg,std_dev,avg_err,m_block,m_per_block);
    }

    /** \brief Report average, standard deviation, and 
	the error in the average assuming a new block size
    */
    template<class vec_t, class vec2_t, class vec3_t> 
      void reblock_avg_stats(size_t new_blocks, vec_t &avg, 
			     vec2_t &std_dev, vec3_t &avg_err,
			     size_t &m_per_block) const {

      if (new_blocks==0) {
	O2SCL_ERR2("Requested zero blocks in ",
		   "vector_ev::reblock_avg_stats().",gsl_einval);
      }
      
      umatrix dat(nvec,new_blocks);
      dat.set_all(0.0);
      
      // If 'i==0' and 'nperblock==0', then not all of the blocks have
      // data yet, and we use the information from 'last' instead of
      // 'vals'.
      
      // The ratio of the old to new block size
      size_t fact=nblocks/new_blocks;
      if (i==0 && nperblock==0) fact=iblock/new_blocks;
      if (fact==0) {
	O2SCL_ERR2("Not enough data for reblocking ",
		   "in scalar_ev::reblock_avg_stats().",gsl_einval);
      }
      size_t iblock2=0;
      // Compute the sum
      for(size_t j=0;j<fact;j++) {
	for(size_t k=0;k<new_blocks;k++) {
	  if (i==0 && nperblock==0) {
	    for(size_t iv=0;iv<nvec;iv++) {
	      dat[iv][k]+=last[iv][iblock2];
	    }
	  } else {
	    for(size_t iv=0;iv<nvec;iv++) {
	      dat[iv][k]+=vals[iv][iblock2];
	    }
	  }
	  iblock2++;
	}
      }
      // Now divide to get averages
      for(size_t k=0;k<new_blocks;k++) {
	for(size_t iv=0;iv<nvec;iv++) {
	  dat[iv][k]/=((double)fact);
	}
      }
      for(size_t iv=0;iv<nvec;iv++) {
	avg[iv]=vector_mean(new_blocks,dat[iv]);
	if (new_blocks>1) {
	  std_dev[iv]=vector_stddev(new_blocks,dat[iv]);
	  avg_err[iv]=std_dev[iv]/sqrt(((double)new_blocks));
	} else {
	  std_dev[iv]=0.0;
	  avg_err[iv]=0.0;
	}
      }
      
      if (i==0 && nperblock==0) m_per_block=fact;
      else if (nperblock==0) m_per_block=fact*i;
      else m_per_block=fact*nperblock;
      
      return;
    }
    
    /** \brief Report average, standard deviation, and 
	the error in the average assuming a new block size
    */
    template<class vec_t, class vec2_t, class vec3_t> 
      void reblock_avg(size_t new_blocks, vec_t &avg, 
		       vec2_t &std_dev, vec3_t &avg_err) const {
      size_t m_per_block;
      return reblock_avg_stats(new_blocks,avg,std_dev,avg_err,
			       m_per_block);
    }
    //@}

  };

}

#endif
