/*
  -------------------------------------------------------------------
  
  Copyright (C) 2011-2012 Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#include <o2scl/test_mgr.h>
#include <o2scl/gsl_rnga.h>
#include <o2scl/expect_val.h>

using namespace std;
using namespace o2scl;

int main(int argc, char *argv[]) {
  test_mgr t;
  t.set_output_level(1);

  cout.setf(ios::scientific);
  cout.precision(4);

  gsl_rnga gr;
  int seed=10;
  if (argc>=2) {
    seed=stoi(argv[1]);
  }
  gr.set_seed(seed);

  double avg, sd, avge;

  umatrix rnds(10,10);

  size_t i_block, i_per_block;
  size_t m_block, m_per_block;

  // ------------------------------------------------------------------
  // Class scalar_ev testing
  // ------------------------------------------------------------------

  scalar_ev se;

  se.set_blocks(10,0);
  cout << "n_per_block=0: " << endl;
  cout << "x          i+1 ib ipb mb mpb prog       f "
       << "avg        sd         avge" << endl;

  for(size_t i=0;i<100;i++) {
    double x=gr.random();

    // Store random numbers for use later
    size_t mi=i%10, mj=i/10;
    rnds[mi][mj]=x;

    // Add measurement and get info
    se.add(x);
    se.get_block_indices(i_block,i_per_block);

    // Output
    se.current_avg_stats(avg,sd,avge,m_block,m_per_block);
    if (i<30 || i==99) {
      cout << x << " ";
      cout.width(3);
      cout << i+1 << " ";
      cout.width(2);
      cout << i_block << " ";
      cout.width(2);
      cout << i_per_block << " ";
      cout.width(2);
      cout << m_block << "   ";
      cout.width(2);
      cout << m_per_block << " ";
      cout << se.progress() << " " << se.finished() << " "
	   << avg << " " << sd << " " << avge << endl;
    }

    // Testing
    t.test_gen(i_block==(i+1)%10,"i_block");
    t.test_gen(i_per_block==(i+1)/10,"i_per_block");
    if (i%10==9) {
      t.test_gen(se.finished()==true,"finished 1");
      t.test_rel(se.progress(),1.0,1.0e-10,"progress 1a");
    } else {
      t.test_rel(se.progress(),((double)((i+1)%10))/10.0,1.0e-10,
		 "progress 1b");
    }
  }

  // Test final results
  double avg_last=avg;
  double sd_last=sd;
  double avge_last=avge;
  t.test_rel(0.5,avg,0.2,"avg");
  t.test_rel(0.08,sd,0.4,"sd");
  t.test_rel(0.03,avge,0.7,"avge");
  cout << endl;

  cout << "Reblocking: " << endl;

  se.reblock_avg_stats(1,avg,sd,avge,m_per_block);
  cout.width(2);
  cout << 1 << " " << avg << " " << sd << " " << avge << " " 
       << m_per_block << endl;
  t.test_rel(avg,avg_last,1.0e-12,"reblock 1");
  t.test_gen(m_per_block==100,"reblock 1a");
  double avg1=avg;

  se.reblock_avg_stats(2,avg,sd,avge,m_per_block);
  cout.width(2);
  cout << 2 << " " << avg << " " << sd << " " << avge << " "
       << m_per_block << endl;
  t.test_rel(avg,avg_last,1.0e-12,"reblock 2");
  t.test_gen(m_per_block==50,"reblock 2a");
  double avg2=avg;

  // This average is different because 3 doesn't evenly divide 10
  se.reblock_avg_stats(3,avg,sd,avge,m_per_block);
  cout.width(2);
  cout << 3 << " " << avg << " " << sd << " " << avge << " "
       << m_per_block << endl;
  t.test_rel(avg,avg_last,1.0e-1,"reblock 6");
  t.test_gen(m_per_block==30,"reblock 6a");
  double avg3=avg;

  // This average is different because 4 doesn't evenly divide 10
  se.reblock_avg_stats(4,avg,sd,avge,m_per_block);
  cout.width(2);
  cout << 4 << " " << avg << " " << sd << " " << avge << " "
       << m_per_block << endl;
  t.test_rel(avg,avg_last,1.0e-1,"reblock 3");
  t.test_gen(m_per_block==20,"reblock 3a");
  double avg4=avg;

  se.reblock_avg_stats(5,avg,sd,avge,m_per_block);
  cout.width(2);
  cout << 5 << " " << avg << " " << sd << " " << avge << " "
       << m_per_block << endl;
  t.test_rel(avg,avg_last,1.0e-12,"reblock 4");
  t.test_gen(m_per_block==20,"reblock 4a");
  double avg5=avg;

  se.reblock_avg_stats(10,avg,sd,avge,m_per_block);
  cout.width(2);
  cout << 10 << " " << avg << " " << sd << " " << avge << " "
       << m_per_block << endl;
  t.test_rel(avg,avg_last,1.0e-12,"reblock 5");
  t.test_rel(sd,sd_last,1.0e-12,"reblock 5a");
  t.test_rel(avge,avge_last,1.0e-12,"reblock 5b");
  t.test_gen(m_per_block==10,"reblock 5c");
  double avg10=avg;
  cout << endl;

  // ------------------------------------------------------------------
  // Class scalar_ev testing with n_per_block=10
  // ------------------------------------------------------------------

  cout << "n_per_block=10: " << endl;
  cout << "x          i+1 ib ipb mb mpb prog       f "
       << "avg        sd         avge" << endl;
  se.reset();
  se.set_blocks(10,10);
  for(size_t i=0;i<100;i++) {
    size_t mi=i%10, mj=i/10;
    se.add(rnds[mj][mi]);
    se.get_block_indices(i_block,i_per_block);
    se.current_avg_stats(avg,sd,avge,m_block,m_per_block);
    if (i<30 || i==99) {
      cout << rnds[mj][mi] << " ";
      cout.width(3);
      cout << i+1 << " ";
      cout.width(2);
      cout << i_block << " ";
      cout.width(2);
      cout << i_per_block << " ";
      cout.width(2);
      cout << m_block << "   ";
      cout.width(2);
      cout << m_per_block << " ";
      cout << se.progress() << " " << se.finished() << " "
	   << avg << " " << sd << " " << avge << endl;
    }
    if (i==99) t.test_gen(se.finished()==true,"finished a");
    else t.test_gen(se.finished()==false,"finished b");
    t.test_rel(se.progress(),((double)(i+1))/100.0,1.0e-12,"progress a");
    t.test_gen(i_block==((i+1)/10),"i_block a");
    t.test_gen(i_per_block==((i+1)%10),"i_per_block a");
  }
  t.test_rel(0.5,avg,0.2,"avg");
  t.test_rel(0.08,sd,0.4,"sd");
  t.test_rel(0.03,avge,0.7,"avge");
  t.test_rel(avg_last,avg,1.0e-12,"avg");
  t.test_rel(sd_last,sd,1.0e-12,"sd");
  t.test_rel(avge_last,avge,1.0e-12,"avge");
  cout << endl;

  cout << "Reblocking: " << endl;

  se.reblock_avg_stats(1,avg,sd,avge,m_per_block);
  cout.width(2);
  cout << 1 << " " << avg << " " << sd << " " << avge << " " 
       << m_per_block << endl;
  t.test_rel(avg,avg_last,1.0e-12,"reblock 1");
  t.test_gen(m_per_block==100,"reblock 1a");
  t.test_rel(avg,avg1,1.0e-12,"avg1");

  se.reblock_avg_stats(2,avg,sd,avge,m_per_block);
  cout.width(2);
  cout << 2 << " " << avg << " " << sd << " " << avge << " "
       << m_per_block << endl;
  t.test_rel(avg,avg_last,1.0e-12,"reblock 2");
  t.test_gen(m_per_block==50,"reblock 2a");
  t.test_rel(avg,avg2,1.0e-12,"avg2");

  // This average is different because 3 doesn't evenly divide 10
  se.reblock_avg_stats(3,avg,sd,avge,m_per_block);
  cout.width(2);
  cout << 3 << " " << avg << " " << sd << " " << avge << " "
       << m_per_block << endl;
  t.test_rel(avg,avg_last,1.0e-1,"reblock 6");
  t.test_gen(m_per_block==30,"reblock 6a");
  t.test_rel(avg,avg3,1.0e-12,"avg3");

  // This average is different because 4 doesn't evenly divide 10
  se.reblock_avg_stats(4,avg,sd,avge,m_per_block);
  cout.width(2);
  cout << 4 << " " << avg << " " << sd << " " << avge << " "
       << m_per_block << endl;
  t.test_rel(avg,avg_last,1.0e-1,"reblock 3");
  t.test_gen(m_per_block==20,"reblock 3a");
  t.test_rel(avg,avg4,1.0e-12,"avg4");

  se.reblock_avg_stats(5,avg,sd,avge,m_per_block);
  cout.width(2);
  cout << 5 << " " << avg << " " << sd << " " << avge << " "
       << m_per_block << endl;
  t.test_rel(avg,avg_last,1.0e-12,"reblock 4");
  t.test_gen(m_per_block==20,"reblock 4a");
  t.test_rel(avg,avg5,1.0e-12,"avg5");

  se.reblock_avg_stats(10,avg,sd,avge,m_per_block);
  cout.width(2);
  cout << 10 << " " << avg << " " << sd << " " << avge << " "
       << m_per_block << endl;
  t.test_rel(avg,avg_last,1.0e-12,"reblock 5");
  t.test_gen(m_per_block==10,"reblock 5a");
  t.test_rel(avg,avg10,1.0e-12,"avg10");

  cout << endl;

  // ------------------------------------------------------------------
  // Class vector_ev testing
  // ------------------------------------------------------------------

  vector_ev ve(3);
  uvector arr(3), vavg(3), vsd(3), vavge(3);

  ve.set_blocks(3,10,0);
  cout << "n_per_block=0: " << endl;
  cout << "x          i+1 ib ipb mb mpb prog       f "
       << "avg        sd         avge" << endl;

  for(size_t i=0;i<100;i++) {

    // Use same random numbers
    size_t mi=i%10, mj=i/10;
    arr[0]=rnds[mi][mj];
    arr[1]=0.0;
    arr[2]=rnds[mi][mj];

    // Add measurement and get info
    ve.add(arr);
    ve.get_block_indices(i_block,i_per_block);

    // Output
    ve.current_avg_stats(vavg,vsd,vavge,m_block,m_per_block);
    if (i<30 || i==99) {
      cout << arr[0] << " ";
      cout.width(3);
      cout << i+1 << " ";
      cout.width(2);
      cout << i_block << " ";
      cout.width(2);
      cout << i_per_block << " ";
      cout.width(2);
      cout << m_block << "   ";
      cout.width(2);
      cout << m_per_block << " ";
      cout << ve.progress() << " " << ve.finished() << " "
	   << vavg[0] << " " << vsd[0] << " " << vavge[0] << endl;
    }

    // Testing
    t.test_gen(i_block==(i+1)%10,"i_block");
    t.test_gen(i_per_block==(i+1)/10,"i_per_block");
    if (i%10==9) {
      t.test_gen(ve.finished()==true,"finished 1");
      t.test_rel(ve.progress(),1.0,1.0e-10,"progress 1a");
    } else {
      t.test_rel(ve.progress(),((double)((i+1)%10))/10.0,1.0e-10,
		 "progress 1b");
    }
    t.test_rel(vavg[0],vavg[2],1.0e-12,"avg x");
    t.test_abs(vavg[1],0.0,1.0e-14,"avg y");
    t.test_rel(vsd[0],vsd[2],1.0e-12,"sd x");
    t.test_rel(vavge[0],vavge[2],1.0e-12,"avge x");
  }

  // Test final results
  t.test_rel(avg_last,vavg[0],1.0e-12,"avg");
  t.test_rel(sd_last,vsd[0],1.0e-12,"sd");
  t.test_rel(avge_last,vavge[0],1.0e-12,"avge");
  t.test_rel(avg_last,vavg[2],1.0e-12,"avg");
  t.test_rel(sd_last,vsd[2],1.0e-12,"sd");
  t.test_rel(avge_last,vavge[2],1.0e-12,"avge");
  cout << endl;

  cout << "Reblocking: " << endl;

  ve.reblock_avg_stats(1,vavg,vsd,vavge,m_per_block);
  cout.width(2);
  cout << 1 << " " << vavg[0] << " " << vsd[0] << " " << vavge[0] << " " 
       << m_per_block << endl;
  t.test_rel(vavg[0],avg_last,1.0e-12,"reblock 1");
  t.test_gen(m_per_block==100,"reblock 1a");
  t.test_rel(avg1,vavg[0],1.0e-12,"reblock 1b");

  ve.reblock_avg_stats(2,vavg,vsd,vavge,m_per_block);
  cout.width(2);
  cout << 2 << " " << vavg[0] << " " << vsd[0] << " " << vavge[0] << " "
       << m_per_block << endl;
  t.test_rel(vavg[0],avg_last,1.0e-12,"reblock 2");
  t.test_gen(m_per_block==50,"reblock 2a");
  t.test_rel(avg2,vavg[0],1.0e-12,"reblock 1b");

  // This average is different because 3 doesn't evenly divide 10
  ve.reblock_avg_stats(3,vavg,vsd,vavge,m_per_block);
  cout.width(2);
  cout << 3 << " " << vavg[0] << " " << vsd[0] << " " << vavge[0] << " "
       << m_per_block << endl;
  t.test_rel(vavg[0],avg_last,1.0e-1,"reblock 6");
  t.test_gen(m_per_block==30,"reblock 6a");
  t.test_rel(avg3,vavg[0],1.0e-12,"reblock 1b");

  // This average is different because 4 doesn't evenly divide 10
  ve.reblock_avg_stats(4,vavg,vsd,vavge,m_per_block);
  cout.width(2);
  cout << 4 << " " << vavg[0] << " " << vsd[0] << " " << vavge[0] << " "
       << m_per_block << endl;
  t.test_rel(vavg[0],avg_last,1.0e-1,"reblock 3");
  t.test_gen(m_per_block==20,"reblock 3a");
  t.test_rel(avg4,vavg[0],1.0e-12,"reblock 1b");

  ve.reblock_avg_stats(5,vavg,vsd,vavge,m_per_block);
  cout.width(2);
  cout << 5 << " " << vavg[0] << " " << vsd[0] << " " << vavge[0] << " "
       << m_per_block << endl;
  t.test_rel(vavg[0],avg_last,1.0e-12,"reblock 4");
  t.test_gen(m_per_block==20,"reblock 4a");
  t.test_rel(avg5,vavg[0],1.0e-12,"reblock 1b");

  ve.reblock_avg_stats(10,vavg,vsd,vavge,m_per_block);
  cout.width(2);
  cout << 10 << " " << vavg[0] << " " << vsd[0] << " " << vavge[0] << " "
       << m_per_block << endl;
  t.test_rel(vavg[0],avg_last,1.0e-12,"reblock 5");
  t.test_rel(sd,sd_last,1.0e-12,"reblock 5a");
  t.test_rel(avge,avge_last,1.0e-12,"reblock 5b");
  t.test_gen(m_per_block==10,"reblock 5c");
  t.test_rel(avg10,vavg[0],1.0e-12,"reblock 1b");

  // ------------------------------------------------------------------
  // Class vector_ev testing with n_per_block=10
  // ------------------------------------------------------------------

  cout << "n_per_block=10: " << endl;
  cout << "x          i+1 ib ipb mb mpb prog       f "
       << "avg        sd         avge" << endl;
  ve.reset();
  ve.set_blocks(3,10,10);
  for(size_t i=0;i<100;i++) {

    size_t mi=i%10, mj=i/10;
    arr[0]=rnds[mj][mi];
    arr[1]=0.0;
    arr[2]=rnds[mj][mi];
    ve.add(arr);

    ve.get_block_indices(i_block,i_per_block);
    ve.current_avg_stats(vavg,vsd,vavge,m_block,m_per_block);
    if (i<30 || i==99) {
      cout << rnds[mj][mi] << " ";
      cout.width(3);
      cout << i+1 << " ";
      cout.width(2);
      cout << i_block << " ";
      cout.width(2);
      cout << i_per_block << " ";
      cout.width(2);
      cout << m_block << "   ";
      cout.width(2);
      cout << m_per_block << " ";
      cout << ve.progress() << " " << ve.finished() << " "
	   << vavg[0] << " " << vsd[0] << " " << vavge[0] << endl;
    }
    if (i==99) t.test_gen(ve.finished()==true,"finished a");
    else t.test_gen(ve.finished()==false,"finished b");
    t.test_rel(ve.progress(),((double)(i+1))/100.0,1.0e-12,"progress a");
    t.test_gen(i_block==((i+1)/10),"i_block a");
    t.test_gen(i_per_block==((i+1)%10),"i_per_block a");
    t.test_rel(vavg[0],vavg[2],1.0e-12,"avg x");
    t.test_abs(vavg[1],0.0,1.0e-14,"avg y");
    t.test_rel(vsd[0],vsd[2],1.0e-12,"sd x");
    t.test_rel(vavge[0],vavge[2],1.0e-12,"avge x");
  }

  // Test final results
  t.test_rel(avg_last,vavg[0],1.0e-12,"avg");
  t.test_rel(sd_last,vsd[0],1.0e-12,"sd");
  t.test_rel(avge_last,vavge[0],1.0e-12,"avge");
  t.test_rel(avg_last,vavg[2],1.0e-12,"avg");
  t.test_rel(sd_last,vsd[2],1.0e-12,"sd");
  t.test_rel(avge_last,vavge[2],1.0e-12,"avge");
  cout << endl;

  cout << "Reblocking: " << endl;

  ve.reblock_avg_stats(1,vavg,vsd,vavge,m_per_block);
  cout.width(2);
  cout << 1 << " " << vavg[0] << " " << vsd[0] << " " << vavge[0] << " " 
       << m_per_block << endl;
  t.test_rel(vavg[0],avg_last,1.0e-12,"reblock 1");
  t.test_gen(m_per_block==100,"reblock 1a");
  t.test_rel(vavg[0],avg1,1.0e-12,"avg1");

  ve.reblock_avg_stats(2,vavg,vsd,vavge,m_per_block);
  cout.width(2);
  cout << 2 << " " << vavg[0] << " " << vsd[0] << " " << vavge[0] << " "
       << m_per_block << endl;
  t.test_rel(vavg[0],avg_last,1.0e-12,"reblock 2");
  t.test_gen(m_per_block==50,"reblock 2a");
  t.test_rel(vavg[0],avg2,1.0e-12,"avg2");

  // This average is different because 3 doesn't evenly divide 10
  ve.reblock_avg_stats(3,vavg,vsd,vavge,m_per_block);
  cout.width(2);
  cout << 3 << " " << vavg[0] << " " << vsd[0] << " " << vavge[0] << " "
       << m_per_block << endl;
  t.test_rel(vavg[0],avg_last,1.0e-1,"reblock 6");
  t.test_gen(m_per_block==30,"reblock 6a");
  t.test_rel(vavg[0],avg3,1.0e-12,"avg3");

  // This average is different because 4 doesn't evenly divide 10
  ve.reblock_avg_stats(4,vavg,vsd,vavge,m_per_block);
  cout.width(2);
  cout << 4 << " " << vavg[0] << " " << vsd[0] << " " << vavge[0] << " "
       << m_per_block << endl;
  t.test_rel(vavg[0],avg_last,1.0e-1,"reblock 3");
  t.test_gen(m_per_block==20,"reblock 3a");
  t.test_rel(vavg[0],avg4,1.0e-12,"avg4");

  ve.reblock_avg_stats(5,vavg,vsd,vavge,m_per_block);
  cout.width(2);
  cout << 5 << " " << vavg[0] << " " << vsd[0] << " " << vavge[0] << " "
       << m_per_block << endl;
  t.test_rel(vavg[0],avg_last,1.0e-12,"reblock 4");
  t.test_gen(m_per_block==20,"reblock 4a");
  t.test_rel(vavg[0],avg5,1.0e-12,"avg5");

  ve.reblock_avg_stats(10,vavg,vsd,vavge,m_per_block);
  cout.width(2);
  cout << 10 << " " << vavg[0] << " " << vsd[0] << " " << vavge[0] << " "
       << m_per_block << endl;
  t.test_rel(vavg[0],avg_last,1.0e-12,"reblock 5");
  t.test_gen(m_per_block==10,"reblock 5a");
  t.test_rel(vavg[0],avg10,1.0e-12,"avg10");

  cout << endl;

  t.report();
  return 0;

}
