/*
  -------------------------------------------------------------------
  
  Copyright (C) 2011-2012 Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/hist_ev.h>

using namespace std;
using namespace o2scl;

hist_ev::hist_ev() : expect_val() {
  hsize=0;
}

hist_ev::hist_ev(uniform_grid<double> g, size_t n_blocks) : expect_val(n_blocks,0) {

  h.set_bins(g);
  hsize=h.size();

  last.allocate(hsize,nblocks);
  vals.allocate(hsize,nblocks);
  iblock_bins.allocate(hsize);
  i_bins.allocate(hsize);

  iblock_bins.set_all(0);
  i_bins.set_all(0);
  last.set_all(0.0);
  vals.set_all(0.0);
}

void hist_ev::reset() {
  expect_val::reset();
}

void hist_ev::set_grid_blocks(uniform_grid<double> g, size_t n_blocks) {

  if (i!=0 || iblock!=0) {
    O2SCL_ERR2("Cannot set n_blocks before reset() in ",
	       "expect_val::set_grid_blocks().",gsl_einval);
  }

  expect_val::set_blocks(n_blocks,0);
  
  h.clear();
  h.set_bins(g);

  hsize=h.size();

  last.allocate(hsize,nblocks);
  vals.allocate(hsize,nblocks);
  iblock_bins.allocate(hsize);
  i_bins.allocate(hsize);

  iblock_bins.set_all(0);
  i_bins.set_all(0);
  last.set_all(0.0);
  vals.set_all(0.0);

  return;
}

void hist_ev::add(double x, double val) {

  size_t ibin=h.get_bin_index(x);
    
  // Store in last vector and increment block index
  last[ibin][iblock_bins[ibin]]=val;
  iblock_bins[ibin]++;
  
  // If last vector is full for this bin, add to vals, increment
  // index, and reset block counter
  if (iblock_bins[ibin]==((int)nblocks)) {
    //std::cout << "Reset: " << std::endl;
    for(size_t j=0;j<nblocks;j++) {
      vals[ibin][j]+=(last[ibin][j]-vals[ibin][j])/(i_bins[ibin]+1);
    }
    i_bins[ibin]++;
    iblock_bins[ibin]=0;
  }
  
  if (false) {
    cout << "i_block:     ";
    for(size_t j=0;j<10;j++) {
      cout.width(4);
      cout << iblock_bins[j] << " ";
    }
    cout << endl;
    cout << "i:           ";
    for(size_t j=0;j<10;j++) {
      cout.width(4);
      cout << i_bins[j] << " ";
    }
    cout << endl;
  }
    
  return;
}
    
void hist_ev::current_avg_stats(uvector &reps, uvector &avg, uvector &std_dev, 
				uvector &avg_err, uvector_int &m_block,
				uvector_int &m_per_block) {
      
  // Reallocate all of the vectors
  reps.allocate(hsize);
  avg.allocate(hsize);
  std_dev.allocate(hsize);
  avg_err.allocate(hsize);
  m_block.allocate(hsize);
  m_per_block.allocate(hsize);
      
  for(size_t j=0;j<hsize;j++) {

    reps[j]=h.get_rep_i(j);

    // Unspecified block size
	
    if (i_bins[j]<1) {
	  
      if (iblock_bins[j]==0) {
	
	avg[j]=0.0;
	std_dev[j]=0.0;
	avg_err[j]=0.0;
	m_block[j]=0;
	m_per_block[j]=0;

      } else if (iblock_bins[j]==1) {
	    
	// Only one data point
	avg[j]=last[j][0];
	std_dev[j]=0.0;
	avg_err[j]=0.0;
	m_block[j]=1;
	m_per_block[j]=1;

      } else {
	  
	// Only one data point for some of the blocks
	double *d=last[j];
	avg[j]=vector_mean<double *>(iblock_bins[j],d);
	std_dev[j]=vector_stddev(iblock_bins[j],d);
	avg_err[j]=std_dev[j]/sqrt(((double)(iblock_bins[j])));
	m_block[j]=iblock_bins[j];
	m_per_block[j]=1;

      }

    } else {
	
      // General case
      double *d=vals[j];
      avg[j]=vector_mean(nblocks,d);
      std_dev[j]=vector_stddev(nblocks,d);
      avg_err[j]=std_dev[j]/sqrt(((double)nblocks));
      m_block[j]=nblocks;
      m_per_block[j]=i_bins[j];
	  
    }
	
  }

  return;

}

void hist_ev::current_avg(uvector &reps, uvector &avg, uvector &std_dev, 
			  uvector &avg_err) {
  uvector_int m_block, m_per_block;
  current_avg_stats(reps,avg,std_dev,avg_err,m_block,m_per_block);
  return;
}

#ifdef O2SCL_NEVER_DEFINED

hist_2d_ev::hist_2d_ev(hist_grid hxg, hist_grid hyg, size_t n_blocks) : 
  expect_val(n_blocks,0), hist_2d() {
  
  set_bins(hxg,hyg);
  size_t dim[3]={hsize_x,hsize_y,n_blocks};
  last.tensor_allocate(3,dim);
  vals.tensor_allocate(3,dim);
  iblock_bins.allocate(hsize_x,hsize_y);
  i_bins.allocate(hsize_x,hsize_y);
      
  iblock_bins.set_all(0);
  i_bins.set_all(0);
  // Set all values in last to zero
  size_t tot=last.total_size();
  for(size_t ii=0;ii<tot;ii++) {
    last.unpack_indices(ii,dim);
    last.set(dim,0.0);
  }
}

void hist_2d_ev::add(double x, double y, double val) {

  size_t ibin_x, ibin_y;
  //get_bin(x,y,ibin_x,ibin_y);
      
  // Store in last vector and increment block index
  last.set(ibin_x,ibin_y,iblock_bins[ibin_x][ibin_y],val);
  iblock_bins[ibin_x][ibin_y]++;
      
  // If last vector is full for this bin, add to vals, increment
  // index, and reset block counter
  if (iblock_bins[ibin_x][ibin_y]==((int)nblocks)) {
    for(size_t j=0;j<nblocks;j++) {
      double tmp=(last.get(ibin_x,ibin_y,j)-
		  vals.get(ibin_x,ibin_y,j))/
	(i_bins[ibin_x][ibin_y]+1);
      vals.set(ibin_x,ibin_y,j,tmp);
    }
    i_bins[ibin_x][ibin_y]++;
    iblock_bins[ibin_x][ibin_y]=0;
  }

  return;
}
    
void hist_2d_ev::current_avg_stats(umatrix &avg, umatrix &std_dev, 
				   umatrix &avg_err, umatrix_int &m_block,
				   umatrix_int &m_per_block) {
      
  // Reallocate all of the vectors
  avg.allocate(hsize_x,hsize_y);
  std_dev.allocate(hsize_x,hsize_y);
  avg_err.allocate(hsize_x,hsize_y);
  m_block.allocate(hsize_x,hsize_y);
  m_per_block.allocate(hsize_x,hsize_y);
      
  for(size_t j=0;j<hsize_x;j++) {
    for(size_t k=0;k<hsize_y;k++) {
	  
      // Unspecified block size
	
      if (i_bins[j][k]<1) {
	  
	if (iblock_bins[j][k]==0) {
	
	  avg[j][k]=0.0;
	  std_dev[j][k]=0.0;
	  avg_err[j][k]=0.0;
	  m_block[j][k]=0;
	  m_per_block[j][k]=0;

	} else if (iblock_bins[j][k]==1) {
	    
	  // Only one data point
	  avg[j][k]=last.get(j,k,0);
	  std_dev[j][k]=0.0;
	  avg_err[j][k]=0.0;
	  m_block[j][k]=1;
	  m_per_block[j][k]=1;

	} else {
	  
	  // Only one data point for some of the blocks
	  uvector v(iblock_bins[j][k]);
	  for(int ib=0;ib<iblock_bins[j][k];ib++) {
	    v[ib]=last.get(j,k,ib);
	  }
	  avg[j][k]=vector_mean(iblock_bins[j][k],v);
	  std_dev[j][k]=vector_stddev(iblock_bins[j][k],v);
	  avg_err[j][k]=std_dev[j][k]/sqrt(((double)(iblock_bins[j][k])));
	  m_block[j][k]=iblock_bins[j][k];
	  m_per_block[j][k]=1;
	      
	}
	    
      } else {
	    
	// General case
	uvector v(iblock_bins[j][k]);
	for(size_t ib=0;ib<nblocks;ib++) {
	  v[ib]=vals.get(j,k,ib);
	}
	avg[j][k]=vector_mean(nblocks,v);
	std_dev[j][k]=vector_stddev(nblocks,v);
	avg_err[j][k]=std_dev[j][k]/sqrt(((double)nblocks));
	m_block[j][k]=nblocks;
	m_per_block[j][k]=i_bins[j][k];
	  
      }
	
    }
  }

  return;
}

#endif
