/*
  -------------------------------------------------------------------
  
  Copyright (C) 2011-2012 Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef HIST_EV_H
#define HIST_EV_H

#include <o2scl/tensor.h>
#include <o2scl/expect_val.h>
#include <o2scl/uniform_grid.h>
#include <o2scl/hist.h>
#include <o2scl/hist_2d.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief A set expectation values for histogram bins

      See \ref expect_val for some general notes on this and related
      classes.

      This class is experimental.

      This class computes expectation values of bins in a histogram.
      It is most useful in cases where one does not know a priori how
      many measurements one is going to get for each bin. The class
      automatically arranges each bin into blocks, so that the
      standard deviation and error and the average can be computed
      even when not all bins have been filled with the same number of
      measurements.

      \note This class always assumes that <tt>n_per_block</tt>
      is zero. See \ref expect_val for more details.

      \todo Test set_grid_blocks() function.

      \todo Create copy constructors as in the scalar_ev class.
  */
  class hist_ev : public expect_val {

  protected:

    /// Last measurement for each block and each bin
    umatrix last;
    
    /// Running average for each block and each bin
    umatrix vals;

    /// The value of \c iblock for each bin
    uvector_int iblock_bins;

    /// The value of \c i for each bin
    uvector_int i_bins;

    /** \brief This should always be the same as the size as the
	histogram
    */
    size_t hsize;
    
    /** \brief The associated histogram

	This object is only currently used internally for binning. No
	data is added to it.
    */
    hist h;
    
  public:

    hist_ev();
    
    /// Create a histogram expectation value 
    hist_ev(uniform_grid<double> g, size_t n_blocks);
    
    /// Remove all currently stored data, but keep same block size
    void reset();

    /// Set the histogram grid and the number of blocks
    void set_grid_blocks(uniform_grid<double> g, size_t n_blocks);

    /// Add measurement of value \c val at location \c x
    virtual void add(double x, double val);
    
    /** \brief Report current average, standard deviation, and 
	the error in the average

	This function deallocates any space already allocated for the
	vector parameters and reallocates space as needed. Information
	previously stored in these vectors will be lost.
    */
    virtual void current_avg_stats(uvector &reps, uvector &avg, 
				   uvector &std_dev, uvector &avg_err, 
				   uvector_int &m_block,
				   uvector_int &m_per_block);

    /** \brief Report current average, standard deviation, and 
	the error in the average

	This function deallocates any space already allocated for the
	vector parameters and reallocates space as needed. Information
	previously stored in these vectors will be lost.
    */
    virtual void current_avg(uvector &reps, uvector &avg, uvector &std_dev, 
			     uvector &avg_err);

  };

#ifdef O2SCL_NEVER_DEFINED

  /** \brief A two-dimensional histogram of expectation values

      See \ref expect_val for some general notes on 
      this and related classes. 

      This class is experimental.
   */
  class hist_2d_ev : public expect_val, public hist_2d {

  protected:

    /// Last measurement for each block and each bin
    tensor3 last;
    
    /// Running average for each block and each bin
    tensor3 vals;

    /// The value of \c iblock for each bin
    umatrix_int iblock_bins;

    /// The value of \c i for each bin
    umatrix_int i_bins;

  public:
    
    /// Create a histogram expectation value 
    hist_2d_ev(uniform_grid hxg, uniform_grid hyg, size_t n_blocks);

    /// Add measurement of value \c val at location \c x
    virtual void add(double x, double y, double val);
    
    /** \brief Report current average, standard deviation, and 
	the error in the average
    */
    virtual void current_avg_stats(umatrix &avg, umatrix &std_dev, 
				   umatrix &avg_err, umatrix_int &m_block,
				   umatrix_int &m_per_block);
  };

#endif

}

#endif
