/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef FRDM_MASS_H
#define FRDM_MASS_H

#include <cmath>

#include <o2scl/nucleus.h>
#include <o2scl/nuclear_mass.h>
#include <o2scl/constants.h>

#ifndef DOXYGENP
namespace o2scl {
#endif
  
  /** \brief FRDM semi-empirical mass formula (macroscopic part only
      with no deformation)
      
      The macroscopic part of the finite-range droplet model
      from \ref Moller95 .
  
      Using the relations
      \f[
      \bar{\delta} = (n_n - n_p)/n
      \f]
      and
      \f[
      \bar{\epsilon} = - (n-n_0)/3/n_0
      \f]
      we get
      \f[
      n_n = \frac{1}{2} (1+\bar{\delta}) (1-3 \bar{\epsilon}) n_0
      \f]
      and
      \f[
      n_p = \frac{1}{2} (1-\bar{\delta}) (1-3 \bar{\epsilon}) n_0
      \f]
      Assuming that 
      \f[
      \frac{4 \pi}{3} R_n^3 n_n = N
      \f]
      and 
      \f[
      \frac{4 \pi}{3} R_p^3 n_p = Z
      \f]
      we get
      \f[
      R_n^3 = 3 N / \alpha_n
      \f]
      \f[
      R_p^3 = 3 Z / \alpha_p
      \f]
      where \f$ \alpha \f$'s are
      \f[
      \alpha_n = 2 \pi (1+ \bar{\delta})(1 - 3 \bar{\epsilon}) n_0
      \f]
      \f[
      \alpha_p = 2 \pi (1- \bar{\delta})(1 - 3 \bar{\epsilon}) n_0
      \f]
      Note that the above relations are somehow self-consistent
      because they imply
      \f[
      R^3 n = R_n^3 n_n + R_p^3 n_p
      \f]

      Since we're using (is there a better way?)
      \f[
      R = r_0 A^{1/3}
      \f]
      with \f$ r_0 = 1.16 \f$ fm, then 
      \f$ n_0 = 0.152946 \mathrm{fm}^{-3} \f$.

      \todo Fix pairing energy and double vs. int
      \todo Document drip_binding_energy(), etc.
      \todo Decide on number of fit parameters (10 or 12?) or
      let the user decide
      \todo Document the protected variables
      \todo Set the neutron and proton masses and hbarc to Moller et al.'s 
      values

      \future Add microscopic part.
  */
  class frdm_mass : public nuclear_mass_fit {
  public:

    /// Volume-energy constant in MeV (default 16.247)
    double a1;
    /// Symmetry-energy constant in MeV (default 32.73)
    double J;
    /// Nuclear compressibility constant in MeV (default 240)
    double K;
    /// Surface-energy constant in MeV (default 22.92)
    double a2;
    /// Effective surface-stiffness constant in MeV (default 29.21)
    double Q;
    /// Curvature-energy constant in MeV (default 0)
    double a3;
    /// Charge-asymmetry constant in MeV (default 0.436)
    double ca;
    /// Wigner constant in MeV (default 30)
    double W;
    /** \brief electronic-binding constant in MeV 
	(default \f$ 1.433 \times 10^{-5} \f$ ).
    */
    double ael;
    /// Proton root-mean-square radius in fm (default 0.80)
    double rp;
    /// Nuclear-radius constant in fm (default 1.16)
    double r0;
    /// Hydrogen atom mass excess, 7.289034 MeV
    double MH;
    /// Neutron mass excess, 8.071431 MeV
    double Mn;
    /// Electronic charge squared, 1.4399764 MeV fm
    double e2;
    /// Range of Yukawa-plus-exponential potential, 0.68 fm
    double a;
    /** \brief Range of Yukawa function used to generate nuclear 
	charge distribution, 0.70 fm
    */
    double aden;
    /// Average pairing-gap constant, 4.80 MeV
    double rmac;
    /// Neutron-proton interaction constant, 6.6 MeV
    double h;
    /// Density-symmetry constant, 0 MeV
    double L;
    /// Pre-exponential compressibility-term constant, 60 MeV
    double C;
    /// Exponential compressibility-term range constant, 0.831
    double gamma;
    /// Atomic mass unit, 931.5014 MeV
    double amu;
    
    /// Internal average neutron density
    double nn;
    
    /// Internal average proton density
    double np;

    /// Neutron radius
    double Rn;

    /// Proton radius
    double Rp;

    frdm_mass() {
      MH=7.289034;
      Mn=8.071431;
      e2=1.4399764;
      ael=1.433e-5;
      K=240.0;
      rp=0.8;
      r0=1.16;
      a=0.68;
      aden=0.7;
      rmac=4.8;
      h=6.6;
      W=30.0;
      L=0;
      a3=0;
      a1=16.247;
      a2=22.92;
      J=32.73;
      Q=29.21;
      a0=0.0;
      ca=0.436;
      C=60.0;
      gamma=0.831;
      amu=931.5014;
      //nfit=12;
      nfit=10;
    }

    /// Given \c Z and \c N, return the mass excess in MeV
    virtual double mass_excess_d(double Z, double N) {
      double ret;
      
      double A=Z+N;
      double cN=cbrt(N);
      double cZ=cbrt(Z);
      double cA=cbrt(A);
      double I=(N-Z)/A;
      
      // Ansatz for computing nuclear radius
      // This should imply a saturation density
      double R=r0*cA;

      double y=R/aden;
      double y0=r0*cA/aden;
      double x=R/a;
      double x0=r0*cA/a;

      double x02=x0*x0;
      B1=1.0-3.0/x02+(1.0+x0)*(2.0+3.0/x0+3.0/x02)*exp(-2.0*x0);
      B2=1.0-(1.0+2.0*x0-x02)*exp(-2.0*x0);
      double y02=y0*y0, y03=y02*y0, y04=y03*y0, y05=y04*y0;
      B3=1.0-5.0/y02*(1.0-15.0/8.0/y0+21.0/8.0/y03-0.75*
		      (1.0+4.5/y0+7.0/y02+3.5/y03)*exp(-2.0*y0));
      B4=1.0+5.0*(-3.0/y02+7.5/y03-63.0/4.0/pow(y0,5.0)
		  +0.75*(2.0/y0+12.0/y02+32.0/y03+
			 42.0/y04+21.0/y05)*exp(-2.0*y0));
      
      // Assume spherical shape
      Bs=1.0;
      Bv=1.0;
      Bw=1.0;
      Bk=1.0;
      Br=1.0;
      
      /* Pairing parameters:

	 Deltan goes like N^{-1/3}
	 Deltap goes like Z^{-1/3}
	 deltanp goes like A^{-2/3}
      */
      Deltan=rmac*Bs/cN;
      Deltap=rmac*Bs/cZ;
      deltanp=h/Bs/cA/cA;

      // Coulomb coefficient
      c1=0.6*e2/r0;
      // Volume redistribution coefficient
      c2=1.0/336.0*(1.0/J+18.0/K)*c1*c1;
      // Coulomb exchange coefficient
      c4=1.25*pow(3.0/2.0/o2scl_const::pi,2.0/3.0)*c1;
      // Surface redistribution coefficient
      c5=c1*c1/64.0/Q;
      // Proton form-factor coefficient
      f0=-0.125*145.0/48.0*rp*rp*e2/r0/r0/r0;

      // Average bulk nuclear asymmetry, which
      // goes like \f$ (I,Z A^{-2/3} B1^{-1}) / (1,B1^{-1}) \f$.
      deltabar=(I+3.0/16.0*c1/Q*Z/cA/cA*Bv*Bs/B1)/
	(1.0+2.25*J/Q/cA*Bs*Bs/B1);

      // Average relative deviation of bulk density
      epsbar=(C*exp(-gamma*cA)-2.0*a2*B2/cA+L*deltabar*deltabar+
	      c1*Z*Z/A/cA*B4)/K;
      
      // Pairing contribution
      double tm=0.0, tm2;
      if (((int)Z)%2==1 && ((int)N)%2==1) {
	if (Z==N) tm=1.0/A;
	tm2=Deltap+Deltan-deltanp;
      } else if (((int)Z)%2==1 && ((int)N)%2==0) {
	tm2=Deltap;
      } else if (((int)N)%2==1) {
	tm2=Deltan;
      } else {
	tm2=0.0;
      }
      
      ret=MH*Z+Mn*N+(-a1+J*deltabar*deltabar-0.5*K*epsbar*epsbar)*A
	+(a2*B1+2.25*J*J/Q*deltabar*deltabar*Bs*Bs/B1)*cA*cA
	+a3*cA*Bk+c1*Z*Z/cA*B3-c2*Z*Z*cA*Br-c4*Z*cZ/cA
	-c5*Z*Z*Bw*Bs/B1+f0*Z*Z/A-ca*(N-Z)+W*(fabs(I)+tm)+tm2
	-ael*pow(Z,2.39);
      
      return ret;
    }

    virtual int fit_fun(size_t nv, const ovector_base &x) {
      K=x[0]*200.0;
      r0=x[1];
      W=x[2];
      a1=x[3];
      a2=x[4];
      J=x[5];
      Q=x[6];
      ca=x[7];
      C=x[8]*60.0;
      gamma=x[9];
      //rmac=x[10];
      //h=x[11];
      return 0;

    }

    virtual int guess_fun(size_t nv, ovector_base &x) {
      x[0]=K/200.0;
      x[1]=r0;
      x[2]=W;
      x[3]=a1;
      x[4]=a2;
      x[5]=J;
      x[6]=Q;
      x[7]=ca;
      x[8]=C/60.0;
      x[9]=gamma;
      //x[10]=rmac;
      //x[11]=h;
      return 0;

    }

    /** \brief Return the binding energy in MeV
     */
    virtual double drip_binding_energy_d(double Z, double N,
					 double npout, double nnout,
					 double chi) {
      return (drip_mass_excess_d(Z,N,npout,nnout,chi)+
	      ((Z+N)*o2scl_fm::mass_amu-Z*o2scl_fm::mass_electron-
	       N*o2scl_fm::mass_neutron-Z*o2scl_fm::mass_proton)*
	      o2scl_const::hc_mev_fm);
    }

    /** \brief Given \c Z and \c N, return the mass excess in MeV
     */
    virtual double drip_mass_excess_d(double Z, double N,
				      double np_out, double nn_out,
				      double chi) {
      double ret;
      
      double dN=N;
      double dZ=Z;
      double dA=Z+N;
      double cN=cbrt(dN);
      double cZ=cbrt(dZ);
      double cA=cbrt(dA);
      double I=(dN-dZ)/dA;
      
      // Ansatz for computing nuclear radius
      // This should imply a saturation density
      double R=r0*cA;

      double y=R/aden;
      double y0=r0*cA/aden;
      double x=R/a;
      double x0=r0*cA/a;

      double x02=x0*x0;
      B1=1.0-3.0/x02+(1.0+x0)*(2.0+3.0/x0+3.0/x02)*exp(-2.0*x0);
      B2=1.0-(1.0+2.0*x0-x02)*exp(-2.0*x0);
      double y02=y0*y0, y03=y02*y0, y04=y03*y0, y05=y04*y0;
      B3=1.0-5.0/y02*(1.0-15.0/8.0/y0+21.0/8.0/y03-0.75*
		      (1.0+4.5/y0+7.0/y02+3.5/y03)*exp(-2.0*y0));
      B4=1.0+5.0*(-3.0/y02+7.5/y03-63.0/4.0/pow(y0,5.0)
		  +0.75*(2.0/y0+12.0/y02+32.0/y03+
			 42.0/y04+21.0/y05)*exp(-2.0*y0));
      
      // Assume spherical shape
      Bs=1.0;
      Bv=1.0;
      Bw=1.0;
      Bk=1.0;
      Br=1.0;
      
      /* Pairing parameters:

	 Deltan goes like N^{-1/3}
	 Deltap goes like Z^{-1/3}
	 deltanp goes like A^{-2/3}
      */
      Deltan=rmac*Bs/cN;
      Deltap=rmac*Bs/cZ;
      deltanp=h/Bs/cA/cA;

      // Coulomb coefficient
      c1=0.6*e2/r0;
      // Volume redistribution coefficient
      c2=1.0/336.0*(1.0/J+18.0/K)*c1*c1;
      // Coulomb exchange coefficient
      c4=1.25*pow(3.0/2.0/o2scl_const::pi,2.0/3.0)*c1;
      // Surface redistribution coefficient
      c5=c1*c1/64.0/Q;
      // Proton form-factor coefficient
      f0=-0.125*145.0/48.0*rp*rp*e2/r0/r0/r0;

      // Average bulk nuclear asymmetry, which
      // goes like \f$ (I,Z A^{-2/3} B1^{-1}) / (1,B1^{-1}) \f$.
      deltabar=(I+3.0/16.0*c1/Q*dZ/cA/cA*Bv*Bs/B1)/
	(1.0+2.25*J/Q/cA*Bs*Bs/B1);

      // Average relative deviation of bulk density
      epsbar=(C*exp(-gamma*cA)-2.0*a2*B2/cA+L*deltabar*deltabar+
	      c1*dZ*dZ/dA/cA*B4)/K;
      
      // Saturation density
      double n0=0.152946;

      // Determine densities and radii
      nn=0.5*(1.0+deltabar)*(1.0-3.0*epsbar)*n0;
      np=0.5*(1.0-deltabar)*(1.0-3.0*epsbar)*n0;
      
      Rn=cbrt(3*N/2.0/o2scl_const::pi/(1.0+deltabar)/(1.0-3.0*epsbar)/n0);
      Rp=cbrt(3*Z/2.0/o2scl_const::pi/(1.0-deltabar)/(1.0-3.0*epsbar)/n0);
      
      double tm=0.0, tm2=0.0;
      if (false) {
	// Pairing contribution
	if (((int)Z)%2==1 && ((int)N)%2==1) {
	  if (Z==N) tm=1.0/dA;
	  tm2=Deltap+Deltan-deltanp;
	} else if (((int)Z)%2==1 && ((int)N)%2==0) {
	  tm2=Deltap;
	} else if (((int)N)%2==1) {
	  tm2=Deltan;
	} else {
	  tm2=0.0;
	}
      }
      
      //double xp=np/(nn+np);
      double x3fact=1.0;//+(xp*xp*xp-1.0)/(1.0+exp((xp-0.25)/0.01));
      
      double surf_part=(a2*B1+2.25*J*J/Q*deltabar*deltabar*Bs*Bs/B1)*cA*cA+
	-c5*dZ*dZ*Bw*Bs/B1;
      
      ret=MH*dZ+Mn*dN+(-a1+J*deltabar*deltabar-0.5*K*epsbar*epsbar)*dA
	+a3*cA*Bk+c1*dZ*dZ/cA*B3-c2*dZ*dZ*cA*Br-c4*dZ*cZ/cA
	+f0*dZ*dZ/dA-ca*(dN-dZ)+W*(fabs(I)+tm)+tm2
	-ael*pow(Z,2.39)+surf_part*x3fact;
      
      return ret;
    }

  protected:

    /// Proton pairing coefficient
    double Deltap;
    /// Neutron pairing coefficient
    double Deltan;
    /// Isovector pairing coefficient
    double deltanp;

    /// Average bulk nuclear asymmetry
    double deltabar;
    /// Average relative deviation of bulk density
    double epsbar;

    /// Desc
    double Bs;
    /// Desc
    double Bk;
    /// Desc
    double Br;
    /// Desc
    double Bw;
    /// Desc
    double Bv;

    /// Coulomb energy coefficient
    double c1;
    /// Volume redistribution energy coefficient
    double c2;
    /// Coulomb exchange correction coefficient
    double c4;
    /// Surface redistribution energy coefficient
    double c5;

    /** \brief Coefficient for the proton form-factor correction to the 
	Coulomb energy
    */
    double f0;

    /// Desc
    double a0;

    /// Desc
    double B1;
    /// Desc
    double B2;
    /// Desc
    double B3;
    /// Desc
    double B4;

  };
  
#ifndef DOXYGENP
}
#endif

#endif
