/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#include <o2scl/nuclear_mass.h>
#include <o2scl/hdf_io.h>
#include <o2scl/hdf_nucmass_io.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;
using namespace o2scl_fm;

nuclear_mass_info::nuclear_mass_info() {
  
  std::string tlist[119]=
    {"n","H","He","Li","Be","B","C","N","O","F","Ne","Na","Mg","Al","Si",
     "P","S","Cl","Ar","K","Ca","Sc","Ti","V","Cr","Mn","Fe","Co","Ni",
     "Cu","Zn","Ga","Ge","As","Se","Br","Kr","Rb","Sr","Y","Zr","Nb","Mo",
     "Tc","Ru","Rh","Pd","Ag","Cd","In","Sn","Sb","Te","I","Xe","Cs","Ba",
     "La","Ce","Pr","Nd","Pm","Sm","Eu","Gd","Tb","Dy","Ho","Er","Tm",
     "Yb","Lu","Hf","Ta","W","Re","Os","Ir","Pt","Au","Hg","Tl","Pb","Bi",
     "Po","At","Rn","Fr","Ra","Ac","Th","Pa","U","Np","Pu","Am","Cm","Bk",
     "Cf","Es","Fm","Md","No","Lr","Rf","Db","Sg","Bh","Hs","Mt","Ds","Rg",
     "Cn","Uut","Fl","Uup","Lv","Uus","Uuo"};
  
  for(int i=0;i<nelements;i++) {
    element_list[i]=tlist[i];
    element_table.insert(make_pair(element_list[i],i));
  }

}

int nuclear_mass_info::parse_elstring(std::string ela, int &Z, int &N, 
				      int &A) {

  bool removed=true;
  while (removed) {
    removed=false;
    for(size_t i=0;i<ela.size();i++) {
      if (!isalnum(ela[i])) {
	ela=ela.substr(0,i)+ela.substr(i+1,ela.size()-i-1);
	removed=true;
      }
    }
  }

  // If its just a neutron
  if (ela=="n") {
    Z=0;
    N=1;
    A=1;
    return 0;
  }
  // If its just a proton
  if (ela=="p") {
    Z=1;
    N=0;
    A=1;
    return 0;
  }
  // If its just a deuteron
  if (ela=="d") {
    Z=1;
    N=1;
    A=2;
    return 0;
  }
  // If its just a triton;
  if (ela=="t") {
    Z=1;
    N=2;
    A=3;
    return 0;
  }

  // If the number precedes the element, swap them
  {
    size_t inum=0, ichar=0;
    // Count number of digits
    while (inum<ela.length() && isdigit(ela[inum])) inum++;
    // Count number of letters following the digits
    while (ichar+inum<ela.length() && 
	   isalpha(ela[ichar+inum])) ichar++;
    if (inum>0 && ichar>0 && ichar<4) {
      string stemp;
      // Build a new element string with the characters first...
      for(size_t i=0;i<ichar;i++) {
	stemp+=ela[inum+i];
      }
      // and then the digits
      for(size_t i=0;i<inum;i++) {
	stemp+=ela[i];
      }
      ela=stemp;
    }
  }

  // Change the first character to upper case if necessary
  if (ela[0]>='a' && ela[0]<='z') {
    ela[0]='A'+(ela[0]-'a');
  }
  if (ela.length()<2) {
    O2SCL_ERR2_RET("Element name too short in ",
		   "nuclear_mass::parse_elstring().",gsl_efailed);
  }
  if (ela.length()>3 && isalpha(ela[2])) {
    std::string el=ela.substr(0,3);
    Z=eltoZ(el);
    A=stoi(ela.substr(3,ela.length()-2));
    N=A-Z;
    return 0;
  }
  if (ela.length()>2 && isalpha(ela[1])) {
    std::string el=ela.substr(0,2);
    Z=eltoZ(el);
    A=stoi(ela.substr(2,ela.length()-2));
    N=A-Z;
    return 0;
  }
  std::string el=ela.substr(0,1);
  Z=eltoZ(el);
  A=stoi(ela.substr(1,ela.length()-1));
  N=A-Z;
  return 0;
}

int nuclear_mass_info::eltoZ(std::string el) {
  std::map<std::string,int,string_comp>::iterator 
    eti=element_table.find(el);
  if (eti==element_table.end()) {
    O2SCL_ERR2_RET("Failed to find element in ",
		   "nuclear_mass_info::eltoZ().",-1);
  }
  return eti->second;
}

std::string nuclear_mass_info::Ztoel(size_t Z) {
  if (((int)Z)>=nelements) {
    O2SCL_ERR2("Invalid element in ",
	       "nuclear_mass_info::Ztoel().",gsl_einval);
    return "";
  }
  return element_list[Z];
}

std::string nuclear_mass_info::tostring(size_t Z, size_t N) {
  if (((int)Z)>=nelements) {
    O2SCL_ERR2("Invalid element in ",
	       "nuclear_mass_info::tostring().",gsl_einval);
    return "";
  }
  return element_list[Z]+itos(N+Z);
}

nuclear_mass::nuclear_mass() {
  m_neut=o2scl_fm::mass_neutron*o2scl_const::hc_mev_fm;
  m_prot=o2scl_fm::mass_proton*o2scl_const::hc_mev_fm;
  m_elec=o2scl_fm::mass_electron*o2scl_const::hc_mev_fm;
  m_amu=o2scl_fm::mass_amu*o2scl_const::hc_mev_fm;
}

int nuclear_mass::get_nucleus(int Z, int N, nucleus &n) {
  n.Z=Z;
  n.N=N;
  n.A=Z+N;
  n.mex=mass_excess(Z,N)/o2scl_const::hc_mev_fm;
  n.m=n.mex+((Z+N)*m_amu-Z*m_elec)/o2scl_const::hc_mev_fm;
  n.ms=n.m;
  n.be=n.m-(N*m_neut+Z*m_prot)/o2scl_const::hc_mev_fm;
  if (n.A%2==0) n.g=1.0;
  else n.g=2.0;
  return 0;
}

double nuclear_mass_disc::mass_excess_d(double Z, double N) {
  int Z1=(int)Z;
  int N1=(int)N;
  double mz1n1=mass_excess(Z1,N1);
  double mz1n2=mass_excess(Z1,N1+1);
  double mz2n1=mass_excess(Z1+1,N1);
  double mz2n2=mass_excess(Z1+1,N1+1);
  double mz1=mz1n1+(N-N1)*(mz1n2-mz1n1);
  double mz2=mz2n1+(N-N1)*(mz2n2-mz2n1);
  return mz1+(Z-Z1)*(mz2-mz1);
}

semi_empirical_mass::semi_empirical_mass() {
  B=-16.0;
  Ss=18.0;
  Ec=0.7;
  Sv=23.7;
  Epair=13.0;
  nfit=5;
}

double semi_empirical_mass::mass_excess_d(double Z, double N) {
  double A=Z+N, cA=cbrt(A);
  double EoA=B+Ss/cA+Ec*Z*Z/cA/A+Sv*pow(1.0-2.0*Z/A,2.0);
  
  EoA+=-Epair*(cos(Z*o2scl_const::pi)+cos(N*o2scl_const::pi))/
    2.0/pow(A,1.5);
  
  double ret=EoA*A+Z*(m_prot+m_elec)+N*m_neut-A*m_amu;
  return ret;
}

int semi_empirical_mass::fit_fun(size_t nv, const ovector_base &x) {
  B=-x[0]; Sv=x[1]; Ss=x[2]; Ec=x[3]; Epair=x[4];
  return 0;
}

int semi_empirical_mass::guess_fun(size_t nv, ovector_base &x) {
  x[0]=-B; x[1]=Sv; x[2]=Ss; x[3]=Ec; x[4]=Epair;
  return 0;
}

ame_mass::ame_mass() {
  n=0;
  reference="";
  mass=0;
}

ame_mass::~ame_mass() {
  if (n>0) {
    delete[] mass;
  }
}

double ame_mass::mass_excess(int Z, int N) {
  ame_entry ret;
  ret=get_ZN(Z,N);
  if (ret.Z==0 && ret.N==0) return 0.0;
  return ret.mass/1.0e3;
}

int ame_mass::set_data(int n_mass, ame_entry *m, std::string ref) {
  n=n_mass;
  mass=m;
  reference=ref;
  return 0;
}

bool ame_mass::is_included(int l_Z, int l_N) {
  if (n==0) {
    O2SCL_ERR_RET("No masses loaded in ame_mass::is_included().",
		  gsl_einval);
  }
  for(int i=0;i<n;i++) {
    if (mass[i].Z==l_Z && mass[i].N==l_N) {
      return true;
    }
  }
  return false;
}

bool ame_mass_exp::is_included(int l_Z, int l_N) {
  if (n==0) {
    O2SCL_ERR_RET("No masses loaded in ame_mass_exp::is_included().",
		  gsl_einval);
  }
  for(int i=0;i<n;i++) {
    if (mass[i].Z==l_Z && mass[i].N==l_N && mass[i].mass_acc==0) {
      return true;
    }
  }
  return false;
}

ame_entry ame_mass::get_ZN(int l_Z, int l_N) {
  ame_entry ret;
  ret.Z=0;
  ret.A=0;
  ret.N=0;
  if (n==0) {
    O2SCL_ERR("No masses loaded in ame_mass::get_ZN().",
	      gsl_einval);
    return ret;
  }
  for(int i=0;i<n;i++) {
    if (mass[i].Z==l_Z && mass[i].N==l_N) {
      ret=mass[i];
    }
  }
  return ret;
}

ame_entry ame_mass::get_ZA(int l_Z, int l_A) {
  ame_entry ret;
  ret.Z=0;
  ret.A=0;
  ret.N=0;
  if (n==0) {
    O2SCL_ERR("No masses loaded in ame_mass::get_ZA().",
	      gsl_einval);
    return ret;
  }
  for(int i=0;i<n;i++) {
    if (mass[i].Z==l_Z && mass[i].A==l_A) {
      ret=mass[i];
    }
  }
  return ret;
}

ame_entry ame_mass::get_elA(string l_el, int l_A) {
  ame_entry ret;
  ret.Z=0;
  ret.A=0;
  ret.N=0;
  if (n==0) {
    O2SCL_ERR("No masses loaded in ame_mass::get_elA().",
	      gsl_einval);
    return ret;
  }
  for(int i=0;i<n;i++) {
    if (mass[i].el==l_el && mass[i].A==l_A) {
      ret=mass[i];
    }
  }
  return ret;
}

ame_entry ame_mass::get(string nucleus) {
  if (n==0) {
    O2SCL_ERR("No masses loaded in ame_mass::get().",
	      gsl_einval);
    ame_entry ret;
    ret.Z=0;
    ret.A=0;
    ret.N=0;
    return ret;
  }
  if (isalpha(nucleus[1])) {
    string el=nucleus.substr(0,2);
    int A=stoi(nucleus.substr(2,nucleus.length()-2));
    return get_elA(el,A);
  } 
  string el=nucleus.substr(0,1);
  int A=stoi(nucleus.substr(1,nucleus.length()-1));
  return get_elA(el,A);
}


mnmsk_mass::mnmsk_mass() {
}

mnmsk_mass::~mnmsk_mass() {
  if (n>0) {
    delete[] mass;
  }
}

int mnmsk_mass::set_data(int n_mass, mnmsk_mass_entry *m, std::string ref) {
  n=n_mass;
  mass=m;
  reference=ref;
  last=n/2;
  return 0;
}

double mnmsk_mass::mass_excess(int Z, int N) {
  mnmsk_mass_entry ret;
  ret=get_ZN(Z,N);
  if (ret.Z==0 && ret.N==0) return 0.0;
  return ret.Mth;
}

bool mnmsk_mass::is_included(int l_Z, int l_N) {
  int lo=0, hi=0, mid=last;

  // binary search for the correct Z first
  if (mass[mid].Z!=l_Z) {
    if (mass[mid].Z>l_Z) {
      lo=0;
      hi=mid;
    } else {
      lo=mid;
      hi=n-1;
    }
    while (hi>lo+1) {
      int mp=(lo+hi)/2;
      if (mass[mp].Z<l_Z) {
	lo=mp;
      } else {
	hi=mp;
      }
    }
    mid=lo;
    if (mass[mid].Z!=l_Z) mid=hi;
    if (mass[mid].Z!=l_Z) {
      return false;
    }
  }

  // The cached point was the right one, so we're done
  if (mass[mid].N==l_N) {
    return true;
  }


  // Now look for the right N among all the Z's
  while (mass[mid].Z==l_Z) {
    if (mass[mid].N==l_N) {
      return true;
    } else if (mass[mid].N>l_N) {
      if (mid==0) return false;
      mid--;
    } else {
      if (mid==n-1) return false;
      mid++;
    }
  }
  
  return false;
}

bool mnmsk_mass_exp::is_included(int l_Z, int l_N) {
  int lo=0, hi=0, mid=last;

  // binary search for the correct Z first
  if (mass[mid].Z!=l_Z) {
    if (mass[mid].Z>l_Z) {
      lo=0;
      hi=mid;
    } else {
      lo=mid;
      hi=n-1;
    }
    while (hi>lo+1) {
      int mp=(lo+hi)/2;
      if (mass[mp].Z<l_Z) {
	lo=mp;
      } else {
	hi=mp;
      }
    }
    mid=lo;
    if (mass[mid].Z!=l_Z) mid=hi;
    if (mass[mid].Z!=l_Z) {
      return false;
    }
  }

  // The cached point was the right one, so we're done
  if (mass[mid].N==l_N) {
    if (fabs(mass[mid].Mexp)>1.0e-20 &&
	fabs(mass[mid].Mexp)<1.0e90) {
      return true;
    } else {
      return false;
    }
  }


  // Now look for the right N among all the Z's
  while (mass[mid].Z==l_Z) {
    if (mass[mid].N==l_N) {
      if (fabs(mass[mid].Mexp)>1.0e-20 &&
	  fabs(mass[mid].Mexp)<1.0e90) {
	return true;
      } else {
	return false;
      }
    } else if (mass[mid].N>l_N) {
      if (mid==0) return false;
      mid--;
    } else {
      if (mid==n-1) return false;
      mid++;
    }
  }
  
  return false;
}

double mnmsk_mass_exp::mass_excess(int Z, int N) {
  mnmsk_mass_entry ret;
  ret=get_ZN(Z,N);
  if (ret.Z==0 && ret.N==0) return 0.0;
  return ret.Mexp;
}

mnmsk_mass_entry mnmsk_mass::get_ZN(int l_Z, int l_N) {
  int lo=0, hi=0, mid=last;
  
  mnmsk_mass_entry ret;
  ret.Z=0;
  ret.A=0;
  ret.N=0;
  
  // binary search for the correct Z first
  if (mass[mid].Z!=l_Z) {
    if (mass[mid].Z>l_Z) {
      lo=0;
      hi=mid;
    } else {
      lo=mid;
      hi=n-1;
    }
    while (hi>lo+1) {
      int mp=(lo+hi)/2;
      if (mass[mp].Z<l_Z) {
	lo=mp;
      } else {
	hi=mp;
      }
    }
    mid=lo;
    if (mass[mid].Z!=l_Z) mid=hi;
    if (mass[mid].Z!=l_Z) {
      O2SCL_ERR((((string)"Nuclei with Z=")+itos(l_Z) 
		 +" not found in mnmsk_mass::get_ZN().").c_str(),
		gsl_enotfound);
    }
  }

  // The cached point was the right one, so we're done
  if (mass[mid].N==l_N) {
    ret=mass[mid];
    last=mid;
    return ret;
  }

  // Now look for the right N among all the Z's
  while (mass[mid].Z==l_Z) {
    if (mass[mid].N==l_N) {
      ret=mass[mid];
      last=mid;
      return ret;
    } else if (mass[mid].N>l_N) {
      mid--;
    } else {
      mid++;
    }
  }
  
  O2SCL_ERR((((string)"Nucleus with Z=")+itos(l_Z)+" and N="+itos(l_N)
	     +" not found in mnmsk_mass::get_ZN().").c_str(),gsl_enotfound);
  return ret;
}

hfb_mass::hfb_mass() {
}

hfb_mass::~hfb_mass() {
  if (n>0) {
    delete[] mass;
  }
}

double hfb_mass::mass_excess(int Z, int N) {
  hfb_mass_entry ret;
  ret=get_ZN(Z,N);
  if (ret.Z==0 && ret.N==0) return 0.0;
  return ret.Mcal;
}

int hfb_mass::set_data(int n_mass, hfb_mass_entry *m, std::string ref) {
  n=n_mass;
  mass=m;
  reference=ref;
  last=n/2;
  return 0;
}

bool hfb_mass::is_included(int l_Z, int l_N) {
  int lo=0, hi=0, mid=last;

  // binary search for the correct Z first
  if (mass[mid].Z!=l_Z) {
    if (mass[mid].Z>l_Z) {
      lo=0;
      hi=mid;
    } else {
      lo=mid;
      hi=n-1;
    }
    while (hi>lo+1) {
      int mp=(lo+hi)/2;
      if (mass[mp].Z<l_Z) {
	lo=mp;
      } else {
	hi=mp;
      }
    }
    mid=lo;
    if (mass[mid].Z!=l_Z) mid=hi;
    if (mass[mid].Z!=l_Z) {
      return false;
    }
  }

  // The cached point was the right one, so we're done
  if (mass[mid].N==l_N) {
    return true;
  }

  int it=0;

  // Now look for the right N among all the Z's
  while (mass[mid].Z==l_Z) {

    // This hack is necessary because some nuclei are missing from
    // some of the HFB tables
    if (it>14 && mid!=0 && mass[mid].Z==mass[mid-1].Z && 
	mass[mid].N!=mass[mid-1].N+1) {
      return false;
    }
    if (mass[mid].N==l_N) {
      return true;
    } else if (mass[mid].N>l_N) {
      if (mid==0) return false;
      mid--;
    } else {
      if (mid==n-1) return false;
      mid++;
    }

    it++;
  }
  
  return false;
}

hfb_mass_entry hfb_mass::get_ZN(int l_Z, int l_N) {
  int lo=0, hi=0, mid=last;

  hfb_mass_entry ret;
  ret.Z=0;
  ret.A=0;
  ret.N=0;
  
  // binary search for the correct Z first
  if (mass[mid].Z!=l_Z) {
    if (mass[mid].Z>l_Z) {
      lo=0;
      hi=mid;
    } else {
      lo=mid;
      hi=n-1;
    }
    while (hi>lo+1) {
      int mp=(lo+hi)/2;
      if (mass[mp].Z<l_Z) {
	lo=mp;
      } else {
	hi=mp;
      }
    }
    mid=lo;
    if (mass[mid].Z!=l_Z) mid=hi;
    if (mass[mid].Z!=l_Z) {
      O2SCL_ERR((((string)"Nuclei with Z=")+itos(l_Z) 
		 +" not found in hfb_mass::get_ZN().").c_str(),
		gsl_enotfound);
    }
  }

  // The cached point was the right one, so we're done
  if (mass[mid].N==l_N) {
    ret=mass[mid];
    last=mid;
    return ret;
  }

  // Now look for the right N among all the Z's
  while (mass[mid].Z==l_Z) {
    if (mass[mid].N==l_N) {
      ret=mass[mid];
      last=mid;
      return ret;
    } else if (mass[mid].N>l_N) {
      mid--;
    } else {
      mid++;
    }
  }
  
  O2SCL_ERR((((string)"Nucleus with Z=")+itos(l_Z)+" and N="+itos(l_N)
	     +" not found in hfb_mass::get_ZN().").c_str(),gsl_enotfound);
  return ret;
}

hfb_sp_mass::hfb_sp_mass() {
}

hfb_sp_mass::~hfb_sp_mass() {
  if (n>0) {
    delete[] mass;
  }
}

double hfb_sp_mass::mass_excess(int Z, int N) {
  hfb_sp_mass_entry ret;
  ret=get_ZN(Z,N);
  if (ret.Z==0 && ret.N==0) return 0.0;
  return ret.Mcal;
}

int hfb_sp_mass::set_data(int n_mass, hfb_sp_mass_entry *m, std::string ref) {
  n=n_mass;
  mass=m;
  reference=ref;
  last=n/2;
  return 0;
}

bool hfb_sp_mass::is_included(int l_Z, int l_N) {
  int lo=0, hi=0, mid=last;

  // binary search for the correct Z first
  if (mass[mid].Z!=l_Z) {
    if (mass[mid].Z>l_Z) {
      lo=0;
      hi=mid;
    } else {
      lo=mid;
      hi=n-1;
    }
    while (hi>lo+1) {
      int mp=(lo+hi)/2;
      if (mass[mp].Z<l_Z) {
	lo=mp;
      } else {
	hi=mp;
      }
    }
    mid=lo;
    if (mass[mid].Z!=l_Z) mid=hi;
    if (mass[mid].Z!=l_Z) {
      return false;
    }
  }

  // The cached point was the right one, so we're done
  if (mass[mid].N==l_N) {
    return true;
  }

  int it=0;

  // Now look for the right N among all the Z's
  while (mass[mid].Z==l_Z) {

    // This hack is necessary because some nuclei are missing from
    // some of the HFB tables
    if (it>14 && mid!=0 && mass[mid].Z==mass[mid-1].Z && 
	mass[mid].N!=mass[mid-1].N+1) {
      return false;
    }
    if (mass[mid].N==l_N) {
      return true;
    } else if (mass[mid].N>l_N) {
      if (mid==0) return false;
      mid--;
    } else {
      if (mid==n-1) return false;
      mid++;
    }

    it++;
  }
  
  return false;
}

hfb_sp_mass_entry hfb_sp_mass::get_ZN(int l_Z, int l_N) {
  int lo=0, hi=0, mid=last;

  hfb_sp_mass_entry ret;
  ret.Z=0;
  ret.A=0;
  ret.N=0;
  
  // binary search for the correct Z first
  if (mass[mid].Z!=l_Z) {
    if (mass[mid].Z>l_Z) {
      lo=0;
      hi=mid;
    } else {
      lo=mid;
      hi=n-1;
    }
    while (hi>lo+1) {
      int mp=(lo+hi)/2;
      if (mass[mp].Z<l_Z) {
	lo=mp;
      } else {
	hi=mp;
      }
    }
    mid=lo;
    if (mass[mid].Z!=l_Z) mid=hi;
    if (mass[mid].Z!=l_Z) {
      O2SCL_ERR((((string)"Nuclei with Z=")+itos(l_Z) 
		 +" not found in hfb_mass::get_ZN().").c_str(),
		gsl_enotfound);
    }
  }

  // The cached point was the right one, so we're done
  if (mass[mid].N==l_N) {
    ret=mass[mid];
    last=mid;
    return ret;
  }

  // Now look for the right N among all the Z's
  while (mass[mid].Z==l_Z) {
    if (mass[mid].N==l_N) {
      ret=mass[mid];
      last=mid;
      return ret;
    } else if (mass[mid].N>l_N) {
      mid--;
    } else {
      mid++;
    }
  }
  
  O2SCL_ERR((((string)"Nucleus with Z=")+itos(l_Z)+" and N="+itos(l_N)
	     +" not found in hfb_mass::get_ZN().").c_str(),gsl_enotfound);
  return ret;
}

bool dz_mass::is_included(int l_Z, int l_N) {
  int lo=0, hi=0, mid=last;
  int l_A=l_Z+l_N;

  // binary search for the correct A first
  if (data.get("A",mid)!=l_A) {
    if (data.get("A",mid)>l_A) {
      lo=0;
      hi=mid;
    } else {
      lo=mid;
      hi=n-1;
    }
    while (hi>lo+1) {
      int mp=(lo+hi)/2;
      if (data.get("A",mp)<l_A) {
	lo=mp;
      } else {
	hi=mp;
      }
    }
    mid=lo;
    if (data.get("A",mid)!=l_A) mid=hi;
    if (data.get("A",mid)!=l_A) {
      return false;
    }
  }

  // The cached point was the right one, so we're done
  if (data.get("Z",mid)==l_Z) {
    return true;
  }

  // Now look for the right N among all the Z's
  while (data.get("A",mid)==l_A) {

    if (data.get("Z",mid)==l_Z) {
      return true;
    } else if (data.get("Z",mid)>l_Z) {
      if (mid==0) return false;
      mid--;
    } else {
      if (mid==n-1) return false;
      mid++;
    }
  }
  
  return false;
}

double dz_mass::mass_excess(int l_Z, int l_N) {
  int lo=0, hi=0, mid=last;
  int l_A=l_Z+l_N;
  
  // binary search for the correct A first
  if (data.get("A",mid)!=l_A) {
    if (data.get("A",mid)>l_A) {
      lo=0;
      hi=mid;
    } else {
      lo=mid;
      hi=n-1;
    }
    while (hi>lo+1) {
      int mp=(lo+hi)/2;
      if (data.get("A",mp)<l_A) {
	lo=mp;
      } else {
	hi=mp;
      }
    }
    mid=lo;
    if (data.get("A",mid)!=l_A) mid=hi;
    if (data.get("A",mid)!=l_A) {
      O2SCL_ERR((((string)"Nucleus with Z=")+itos(l_Z)+" and N="+itos(l_N)+
		 " not found in dz_mass::mass_excess().").c_str(),
		gsl_enotfound);
    }
  }
  
  // The cached point was the right one, so we're done
  if (data.get("Z",mid)==l_Z) {
    return data.get("ME",mid);
  }
  
  // Now look for the right N among all the Z's
  while (data.get("A",mid)==l_A) {
    
    if (data.get("Z",mid)==l_Z) {
      return data.get("ME",mid);
    } else if (data.get("Z",mid)>l_Z) {
      if (mid==0) {
	O2SCL_ERR((((string)"Nucleus with Z=")+itos(l_Z)+" and N="+itos(l_N)+
		   " not found in dz_mass::mass_excess().").c_str(),
		  gsl_enotfound);
      }
      mid--;
    } else {
      if (mid==n-1) {
	O2SCL_ERR((((string)"Nucleus with Z=")+itos(l_Z)+" and N="+itos(l_N)+
		   " not found in dz_mass::mass_excess().").c_str(),
		  gsl_enotfound);
      }
      mid++;
    }
  }
  
  O2SCL_ERR((((string)"Nucleus with Z=")+itos(l_Z)+" and N="+itos(l_N)+
	     " not found in dz_mass::mass_excess().").c_str(),
	    gsl_enotfound);
  
  return 0.0;
}

bool ktuy_mass::is_included(int l_Z, int l_N) {
  int lo=0, hi=0, mid=last;

  // binary search for the correct Z first
  if (mass[mid].Z!=l_Z) {
    if (mass[mid].Z>l_Z) {
      lo=0;
      hi=mid;
    } else {
      lo=mid;
      hi=n-1;
    }
    while (hi>lo+1) {
      int mp=(lo+hi)/2;
      if (mass[mp].Z<l_Z) {
	lo=mp;
      } else {
	hi=mp;
      }
    }
    mid=lo;
    if (mass[mid].Z!=l_Z) mid=hi;
    if (mass[mid].Z!=l_Z) {
      return false;
    }
  }

  // The cached point was the right one, so we're done
  if (mass[mid].N==l_N) {
    return true;
  }

  int it=0;

  // Now look for the right N among all the Z's
  while (mass[mid].Z==l_Z) {

    // This hack is necessary because some nuclei are missing from
    // some of the KTUY tables
    if (it>14 && mid!=0 && mass[mid].Z==mass[mid-1].Z && 
	mass[mid].N!=mass[mid-1].N+1) {
      return false;
    }
    if (mass[mid].N==l_N) {
      return true;
    } else if (mass[mid].N>l_N) {
      if (mid==0) return false;
      mid--;
    } else {
      if (mid==n-1) return false;
      mid++;
    }

    it++;
  }
  
  return false;
}

ktuy_mass_entry ktuy_mass::get_ZN(int l_Z, int l_N) {
  int lo=0, hi=0, mid=last;

  ktuy_mass_entry ret;
  ret.Z=0;
  ret.A=0;
  ret.N=0;
  
  // binary search for the correct Z first
  if (mass[mid].Z!=l_Z) {
    if (mass[mid].Z>l_Z) {
      lo=0;
      hi=mid;
    } else {
      lo=mid;
      hi=n-1;
    }
    while (hi>lo+1) {
      int mp=(lo+hi)/2;
      if (mass[mp].Z<l_Z) {
	lo=mp;
      } else {
	hi=mp;
      }
    }
    mid=lo;
    if (mass[mid].Z!=l_Z) mid=hi;
    if (mass[mid].Z!=l_Z) {
      O2SCL_ERR((((string)"Nuclei with Z=")+itos(l_Z) 
		 +" not found in ktuy_mass::get_ZN().").c_str(),
		gsl_enotfound);
    }
  }

  // The cached point was the right one, so we're done
  if (mass[mid].N==l_N) {
    ret=mass[mid];
    last=mid;
    return ret;
  }

  // Now look for the right N among all the Z's
  while (mass[mid].Z==l_Z) {
    if (mass[mid].N==l_N) {
      ret=mass[mid];
      last=mid;
      return ret;
    } else if (mass[mid].N>l_N) {
      mid--;
    } else {
      mid++;
    }
  }
  
  O2SCL_ERR((((string)"Nucleus with Z=")+itos(l_Z)+" and N="+itos(l_N)
	     +" not found in ktuy_mass::get_ZN().").c_str(),gsl_enotfound);
  return ret;
}

dz_mass::dz_mass(std::string model, bool external) {
  n=0;
  
  std::string fname;
  std::string dir=o2scl::lib_settings.get_data_dir();
  if (external) {
    fname=model;
  } else {
    if (model=="96") {
      fname=dir+"/nucmass/du_zu_96.o2";
    } else {
      fname=dir+"/nucmass/du_zu_95.o2";
    }
  }
  
  o2scl_hdf::hdf_file hf;
  hf.open(fname);
  hdf_input(hf,data);
  hf.close();
  
  n=data.get_nlines();
  last=n/2;
}

dz_mass::~dz_mass() {
}
  
ktuy_mass::ktuy_mass(std::string model, bool external) {
  
  n=0;
  
  std::string fname;
  std::string dir=o2scl::lib_settings.get_data_dir();
  if (external) {
    fname=model;
  } else {
    if (model=="04") {
      fname=dir+"/nucmass/ktuy04.o2";
    } else {
      fname=dir+"/nucmass/ktuy05.o2";
    }
  }
  
  table data;
  o2scl_hdf::hdf_file hf;
  hf.open(fname);
  hdf_input(hf,data);
  hf.close();
  
  n=data.get_nlines();

  mass=new ktuy_mass_entry[n];
  for(int i=0;i<n;i++) {
    ktuy_mass_entry kme={((int)(data.get("NN",i)+1.0e-6)),
			 ((int)(data.get("ZZ",i)+1.0e-6)),
			 ((int)(data.get("NN",i)+data.get("ZZ",i)+1.0e-6)),
			 data.get("Mcal",i),data.get("Esh",i),
			 data.get("alpha2",i),data.get("alpha4",i),
			 data.get("alpha6",i)};
    mass[i]=kme;
  }

  last=n/2;
}

ktuy_mass::~ktuy_mass() {
  if (n>0) {
    delete[] mass;
  }
}

double ktuy_mass::mass_excess(int Z, int N) {
  ktuy_mass_entry ret;
  ret=get_ZN(Z,N);
  if (ret.Z==0 && ret.N==0) return 0.0;
  return ret.Mcal;
}
