/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_NUCLEAR_MASS_H
#define O2SCL_NUCLEAR_MASS_H

#include <cmath>
#include <string>
#include <map>
#include <o2scl/nucleus.h>
#include <o2scl/constants.h>
#include <o2scl/table.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /// Nuclear mass info
  class nuclear_mass_info {
  public:
    
    nuclear_mass_info();
    
    /** \brief Parse a string representing an element

	Accepts strings of one of the following forms:
	- <tt>Pb208</tt>
	- <tt>pb208</tt>
	- <tt>Pb 208</tt>
	- <tt>Pb-208</tt>
	- <tt>pb 208</tt>
	- <tt>pb-208</tt>
	or one of the special strings <tt>n</tt>, <tt>p</tt>, <tt>d</tt>
	or <tt>t</tt> for the neutron, proton, deuteron, and triton, 
	respectively. This function also allows the value of A
	to precede the element symbol.

	\note At present, this allows nuclei which don't make sense
	because A<Z, such as Carbon-5.

	\future Warn about malformed combinations like Carbon-5
	\future Right now, <tt>n4</tt> is interpreted incorrectly
	as Nitrogen-4, rather than the tetraneutron.
	
    */
    int parse_elstring(std::string ela, int &Z, int &N, int &A);    

    /** \brief Return Z given the element name
       
        If the string parameter \c el is invalid, the error handler is
        called and the value -1 is returned.
    */
    int eltoZ(std::string el);
    
    /** \brief Return the element name given Z
       
        \note This function returns \c "n" indicating the neutron for
        Z=0, and if the argument \c Z is greater than 118, an empty
        string is returned after calling the error handler.
    */
    std::string Ztoel(size_t Z);

    /** \brief Return a string of the form "Pb208" for a given Z and N

        Note that if \c Z is zero, then and \c 'n' is 
        used to indicate the a nucleus composed entirely of neutrons 
        and if the argument \c Z is greater than 118, an
        empty string is returned (independ.
    */
    std::string tostring(size_t Z, size_t N);

#ifndef DOXYGEN_INTERNAL

  protected:

    /// The number of elements (proton number)
    static const int nelements=119;
    
    /// A map containing the proton numbers organized by element name
    std::map<std::string,int,string_comp> element_table;
    
    /// A convenient typedef for an iterator for element_table
    typedef std::map<std::string,int,string_comp>::iterator table_it;
    
    /// The list of elements organized by proton number
    std::string element_list[nelements];

#endif
    
    
  };
  
  /** \brief Nuclear mass formula base [abstract base]

      This base class provides some default functionality for
      the nuclear mass formulas. For typical usage, use
      \ref ame_mass, \ref mnmsk_mass, \ref mnmsk_mass_exp,
      or \ref semi_empirical_mass.

      Elements 113, 115, 117 and 118 are named "Uut", "Uup", "Uus",
      and "Uuo", respectively.

      Binding energies are determined from mass excesses by
      \f[
      \mathrm{binding~energy} = A u - Z \left(m_p + m_e\right) - N m_n +
      \mathrm{mass~excess}
      \f]
      where \f$ u \f$, \f$ m_n \f$, \f$ m_p \f$ and \f$ m_e \f$
      are all in \ref o2scl_fm . "Total masses" are the mass 
      of the nuclide (without the electron mass contribution)
      \f[
      \mathrm{total~mass} = \mathrm{mass~excess} + A u - Z m_e
      \f]
      Note that \o2 generally ignores contributions from 
      electron binding energies.
      
      Some mass formulas are undefined for sufficiently exotic nuclei.
      You can use the function is_included() to find if a particular
      \nucleus is included or not in a particular mass formula. 

      Generally, descendants of this class only need to provide an
      implementation of \ref mass_excess() and \ref mass_excess_d()
      and possibly a new version of \ref is_included().

      Some common reaction Q-values and separation energies: \n
      \f$\mathrm{Q}(\beta^{-})=\mathrm{M(A,Z)}-\mathrm{M(A,Z+1)}\f$: 
      Beta-decay energy \n
      \f$\mathrm{Q}(2\beta^{-})=\mathrm{M(A,Z)}-\mathrm{M(A,Z+2)}\f$: 
      Double beta-decay energy \n
      \f$\mathrm{Q}(4\beta^{-})=\mathrm{M(A,Z)}-\mathrm{M(A,Z+4)}\f$: 
      Four beta-decay energy \n
      \f$\mathrm{Q}(\alpha)=\mathrm{M(A,Z)}-\mathrm{M(A-4,Z-2)}-
      \mathrm{M(He^{4})}\f$: 
      Alpha-decay energy \n
      \f$\mathrm{Q}(\beta-n)=\mathrm{M(A,Z)}-\mathrm{M(A-1,Z+1)}
      -\mathrm{M(n)}\f$: 
      Beta-delayed neutron emission decay energy \n
      \f$\mathrm{Q}(d,\alpha)=\mathrm{M(A,Z)}-\mathrm{M(A-2,Z-1)}
      -\mathrm{M(He^{4})}-\mathrm{M(H^{2})}\f$: 
      \f$(d,\alpha)\f$ reaction energy \n
      \f$\mathrm{Q}(\mathrm{EC})=\mathrm{M(A,Z)}-\mathrm{M(A,Z-1)}\f$: 
      Electron capture decay energy \n
      \f$\mathrm{Q}(\mathrm{ECp})=\mathrm{M(A,Z)}-\mathrm{M(A-1,Z-2)}\f$: 
      Electron capture with delayed proton emission decay energy \n
      \f$\mathrm{Q}(n,\alpha)=\mathrm{M(A,Z)}-\mathrm{M(A-3,Z-2)}
      -\mathrm{M(He^{4})}+\mathrm{M(n)}\f$: 
      \f$(n,\alpha)\f$ reaction energy \n
      \f$\mathrm{Q}(p,\alpha)=\mathrm{M(A,Z)}-\mathrm{M(A-3,Z-1)}
      -\mathrm{M(He^{4})}+\mathrm{M(p)}\f$: 
      \f$(p,\alpha)\f$ reaction energy \n
      \f$\mathrm{S}(n) = \mathrm{-M(A,Z)}+\mathrm{M(A-1,Z)}+
      \mathrm{M(n)}\f$: Neutron separation energy \n
      \f$\mathrm{S}(p) = \mathrm{-M(A,Z)}+\mathrm{M(A-1,Z-1)}+
      \mathrm{H^{1}}\f$: Proton separation energy \n
      \f$\mathrm{S}(2n) = \mathrm{-M(A,Z)}+\mathrm{M(A-2,Z)}+
      \mathrm{2M(n)}\f$: Two neutron separation energy \n
      \f$\mathrm{S}(2p) = \mathrm{-M(A,Z)}+\mathrm{M(A-2,Z-2)}+
      \mathrm{2 M(H^{1)}}\f$: Two proton separation energy \n

      \future Find a way to include the electron binding energy
      contribution, possibly with the help of a theoretical model.

      \comment
      \future It might be useful to consider a fudge factor 
      to ensure no problems with finite precision arithmetic
      when converting \c double to \c int. 
      11/22/09 - I waffle back and forth on this. I think
      maybe its best to let the user deal with this their own
      way.
      \endcomment

  */
  class nuclear_mass : public nuclear_mass_info {
  public:

    /** \brief Neutron mass in \f$ \mathrm{MeV} \f$ 
	(defaults to o2scl_fm::mass_neutron times o2scl_const::hc_mev_fm)
    */
    double m_neut;
    
    /** \brief Proton mass in \f$ \mathrm{MeV} \f$ 
	(defaults to o2scl_fm::mass_proton times o2scl_const::hc_mev_fm)
    */
    double m_prot;
    
    /** \brief Electron mass in \f$ \mathrm{MeV} \f$ 
	(defaults to o2scl_fm::mass_elec times o2scl_const::hc_mev_fm)
    */
    double m_elec;

    /** \brief Atomic mass unit in \f$ \mathrm{MeV} \f$ 
	(defaults to o2scl_fm::mass_amu times o2scl_const::hc_mev_fm)
    */
    double m_amu;
    
    nuclear_mass();
    
    virtual ~nuclear_mass() {};

    /// Return the type, \c "nuclear_mass".
    virtual const char *type() { return "nuclear_mass"; }

    /** \brief Return false if the mass formula does not include 
	specified nucleus
    */
    virtual bool is_included(int Z, int N) {
      return true;
    }
    
    /** \brief Fill \c n with the information from nucleus with the given
	neutron and proton number
	
	All masses are given in \f$\mathrm{fm}^{-1}\f$. The total mass
	(withouth the electrons) is put in part::m and part::ms, the
	binding energy is placed in nucleus::be, the mass excess in
	nucleus::mex and the degeneracy (part::g) is arbitrarily set
	to 1 for even A nuclei and 2 for odd A nuclei.
    */
    virtual int get_nucleus(int Z, int N, nucleus &n);
    
    /// Given \c Z and \c N, return the mass excess in MeV
    virtual double mass_excess(int Z, int N)=0;

    /// Given \c Z and \c N, return the mass excess in MeV
    virtual double mass_excess_d(double Z, double N)=0;

    /// Return the approximate electron binding energy in MeV
    virtual double electron_binding(double Z) {
      return (14.4381*pow(Z,2.39)+1.55468e-6*pow(Z,5.35))*1.0e-6;
    }
    
    /** \brief Return the binding energy in MeV
	
	The binding energy is defined to be negative for bound 
	nuclei, thus the binding energy per baryon of Pb-208
	is about -8*208 = -1664 MeV.
    */
    virtual double binding_energy(int Z, int N) {
      return (mass_excess(Z,N)+((Z+N)*m_amu-Z*m_elec-N*m_neut-Z*m_prot));
    }

    /** \brief Return the binding energy in MeV
	
	The binding energy is defined to be negative for bound 
	nuclei, thus the binding energy per baryon of Pb-208
	is about -8*208 = -1664 MeV.
    */
    virtual double binding_energy_d(double Z, double N) {
      return (mass_excess_d(Z,N)+((Z+N)*m_amu-Z*m_elec-N*m_neut-Z*m_prot));
    }
    
    /** \brief Return the total mass of the nucleus (without the electrons) 
	in MeV
    */
    virtual double total_mass(int Z, int N) {
      return (mass_excess(Z,N)+((Z+N)*m_amu-Z*m_elec));
    }
    
    /** \brief Return the total mass of the nucleus (without the electrons) 
	in MeV
    */
    virtual double total_mass_d(double Z, double N) {
      return (mass_excess_d(Z,N)+((Z+N)*m_amu-Z*m_elec));
    }

    /** \brief Return the atomic mass of the nucleus in MeV
	(includes electrons and their binding energy)
    */
    virtual double atomic_mass(int Z, int N) {
      return total_mass(Z,N)+Z*m_elec-electron_binding(Z);
    }
    
    /** \brief Return the atomic mass of the nucleus in MeV
	(includes electrons and their binding energy)
    */
    virtual double atomic_mass_d(double Z, double N) {
      return total_mass_d(Z,N)+Z*m_elec-electron_binding(Z);
    }

  };

  /** \brief Discrete nuclear mass formula [abstract base]

      This uses simple linear interpolation to obtain masses of 
      nuclei with non-integer Z and N which may be particularly
      sensitive to the form of the pairing.

      Generally, descendants of this class only need to provide an
      implementation of \ref mass_excess() and possibly a version
      of \ref nuclear_mass::is_included()

   */
  class nuclear_mass_disc : public nuclear_mass {

  public:
    
    /// Given \c Z and \c N, return the mass excess in MeV
    virtual double mass_excess(int Z, int N)=0;

    /// Given \c Z and \c N, return the mass excess in MeV
    virtual double mass_excess_d(double Z, double N);

  };

  /** \brief Continuous nuclear mass formula [abstract base]
      
      Generally, descendants of this class only need to provide an
      implementation of \ref mass_excess_d() and possibly a version
      of \ref nuclear_mass::is_included()
  */
  class nuclear_mass_cont : public nuclear_mass {

  public:
    
    /// Given \c Z and \c N, return the mass excess in MeV
    virtual double mass_excess(int Z, int N) {
      return mass_excess_d(Z,N);
    }
    
    /// Given \c Z and \c N, return the mass excess in MeV
    virtual double mass_excess_d(double Z, double N)=0;
    
  };

  /** \brief Fittable mass formula

      Nuclear mass formulas which are descendants of this class
      can be fit to experiment using \ref mass_fit.

      Within \o2p, this class has only two children,
      \ref frdm_mass and \ref semi_empirical_mass. There
      is also a child \ref ldrop_mass in \o2e.

      \future This shouldn't be a child of nuclear_mass_cont?
  */
  class nuclear_mass_fit : public nuclear_mass_cont {
    
  public:
    
    /// Return the type, \c "nuclear_mass_fit".
    virtual const char *type() { return "nuclear_mass_fit"; }

    /// Number of fitting parameters
    size_t nfit;

    /// Fix parameters from an array for fitting
    virtual int fit_fun(size_t nv, const ovector_base &x)=0;
    
    /// Fill array with guess from present values for fitting
    virtual int guess_fun(size_t nv, ovector_base &x)=0;

  };

  /** \brief Semi-empirical mass formula

      A simple semi-empirical mass formula of the form
      \f[
      E/A = B + S_s \frac{1}{A^{1/3}}+E_c \frac{Z^2}{A^{4/3}}
      + S_v \left(1-\frac{2Z}{A}\right)^2+E_{\mathrm{pair}}(Z,N)
      \f]
      where the pairing energy is given by
      \f[
      E_{\mathrm{pair}}(Z,N) = - \frac{E_{\mathrm{pair}}}{2 A^{3/2}}
      \left[ \cos \left( \pi Z \right)+\cos \left( \pi N \right) \right]
      \f]
      which is equivalent to the traditional prescription
      \f[
      E_{\mathrm{pair}}(Z,N) = \frac{E_{\mathrm{pair}}}{A^{3/2}}
      \times
      \left\{ 
      \begin{array}{rl}
      -1 & \mathrm{N~and~Z~even} \\
      +1 & \mathrm{N~and~Z~odd} \\
      0 & \mathrm{otherwise} 
      \end{array}
      \right.
      \f]
      when \f$ Z \f$ and \f$ N \f$ and integers.

      \note The default parameters are arbitrary, and are not
      determined from a fit.

      There is an example of the usage of this class given in 
      \ref ex_mass_fit_sect.
  */
  class semi_empirical_mass : public nuclear_mass_fit {

  public:

    /// Binding energy (negative and in MeV, default -16)
    double B;

    /// Symmetry energy (in MeV, default 23.7)
    double Sv;

    /// Surface energy (in MeV, default 18)
    double Ss;

    /// Coulomb energy (in MeV, default 0.7)
    double Ec;

    /// Pairing energy (MeV, default 13.0)
    double Epair;
    
    /// Return the type, \c "semi_empirical_mass".
    virtual const char *type() { return "semi_empirical_mass"; }

    semi_empirical_mass();

    /// Given \c Z and \c N, return the mass excess in MeV
    virtual double mass_excess_d(double Z, double N);
    
    /// Fix parameters from an array for fitting
    virtual int fit_fun(size_t nv, const ovector_base &x);

    /// Fill array with guess from present values for fitting
    virtual int guess_fun(size_t nv, ovector_base &x);
    
  };
  
  /** \brief Atomic mass entry structure

      Atomic mass entry data object for \ref ame_mass.

      In cases where the decimal point in the original table was
      replaced with a <tt>#</tt>, the associated accuracy field is set
      to \ref estimated.  In cases where the original table contained
      a asterisk to indicate a value was not calculable, the accuracy
      field is set to \ref not_calculable and the value is zero. If
      \o2 internally computed the value instead of reading it from the
      original table, the accuracy field is set to \ref intl_computed.
      In cases where either \ref orig or \ref bdmode in the original
      table was blank, the string is set to <tt>"blank"</tt>.

      Binding energies are defined with a positive sign, so that 
      lead has a binding energy of +8 MeV. 
  */
  class ame_entry {

    public:

    /// \name Accuracy modes
    //@{
    /// Measured value from source data
    static const int measured=0;
    /// Value estimated in source data
    static const int estimated=1;
    /// Value listed in data as not calculable
    static const int not_calculable=2;
    /// Value computed by \o2
    static const int intl_computed=3;
    //@}
    
    /// N-Z
    int NMZ;
    
    /// Neutron number
    int N;
    
    /// Proton number
    int Z;
    
    /// Atomic number
    int A;
    
    /// Element name
    char el[4];
    
    /// Data origin
    char orig[5];
    
    /// Mass excess (in keV)
    double mass;
    
    /// Mass excess uncertainty (in keV)
    double dmass;

    /// Mass accuracy flag 
    int mass_acc;
    
    /// Binding energy (in keV, given in the '95 data)
    double be;
    
    /// Binding energy uncertainty (in keV, given in the '95 data)
    double dbe;

    /// Binding energy accuracy flag
    int be_acc;
    
    /// Binding energy / A (in keV, given in the '03 data)
    double beoa;
    
    /// Binding energy / A uncertainty (in keV, given in the '03 data)
    double dbeoa;
    
    /// Binding energy / A accuracy flag
    int beoa_acc;

    /// Beta decay mode
    char bdmode[3];
    
    /// Beta-decay energy (in keV)
    double bde;

    /// Beta-decay energy uncertainty (in keV)
    double dbde;
    
    /// Beta-decay energy accuracy flag
    int bde_acc;

    /// ?
    int A2;
    
    /// Atomic mass (in keV)
    double amass;
    
    /// Atomic mass uncertainty (in keV)
    double damass;
    
    /// Atomic mass accuracy flag
    int amass_acc;

  };
  
  /** \brief Mass formula from the Atomic Mass Evaluation (2005 and 1993)
      
      \note This class requires data stored in an HDF file and
      thus requires HDF support for normal usage.

      This class provides an interface to the atomic mass table using
      data from \ref Audi95 and \ref Audi03. 
      
      There are four data sets, selected by the specification of the
      \c version string in the constructor.
      - "95rmd" - "Recommended" data from \ref Audi95 (ame95rmd.o2)
      - "95exp" - "Experimental" data from \ref Audi95 (ame95exp.o2)
      - "03round" - "Rounded" data from \ref Audi03 (ame03round.o2)
      - "03" - Data from \ref Audi03 (default) (ame03.o2)
      
      If any string other than these four is used, the default
      data is loaded. If the constructor cannot find the data file
      (e.g. because of a broken installation), then ame::is_loaded()
      returns false.
      
      The 1995 data provided the binding energy stored in
      ame_entry::be and ame_entry::dbe, while the 2003 data provided
      the binding energy divided by the atomic number stored in
      ame_entry::beoa and ame_entry::dbeoa. When the 1995 data is used
      ame_entry::beoa and ame_entry::dbeoa are calculated automatically,
      and when the 2003 data is used ame_entry::be and ame_entry::dbe
      are calculated automatically. To indicate that \o2 has 
      automatically calculated a value in this way, the associated
      accuracy field is set to \ref ame_entry::intl_computed. 
      
      Note that all uncertainties are 1 sigma uncertainties.

      The functions \ref mass_excess() and \ref
      nuclear_mass::mass_excess_d() directly return the value from the
      Audi et al. data. For consistency, the functions \ref
      nuclear_mass::binding_energy(), \ref
      nuclear_mass::binding_energy_d(), \ref
      nuclear_mass::total_mass(), and \ref
      nuclear_mass::total_mass_d() return values which are
      automatically computed with the \o2 values for the neutron and
      proton mass in \ref o2scl_fm (which are obtained from
      CODATA). In order to obtain the value of the binding energy as
      reported in the original Audi et al. data instead of the value
      computed from the mass excess, you can use the function
      \ref get_ZN(), and access the corresponding entry from
      the Audi et al. data directly. 

      See also the documentation for the class structure for
      each table entry in \ref ame_entry. 

      \future There are definitions of the atomic mass unit
      and other constants that are defined by the 1995 and 2003
      atomic mass evaluations which are not used at present. These
      could be implemented by making a separate copy of the 
      neutron and proton masses inside this class.
      
      \future Create a caching and more intelligent search system for
      the table. The table is sorted by A and then N, so we could
      probably just copy the search routine from mnmsk_mass, which is
      sorted by Z and then N.

      \future Allow the user to select only massses which are not
      estimated.
      
  */
  class ame_mass : public nuclear_mass_disc {
  public:

    /// Create a collection specified by \c version
    ame_mass();

    ~ame_mass();
    
    /// Return the type, \c "ame_mass".
    virtual const char *type() { return "ame_mass"; }

    /** \brief Return false if the mass formula does not include 
	specified nucleus
    */
    virtual bool is_included(int Z, int N);

    /// Given \c Z and \c N, return the mass excess in MeV
    virtual double mass_excess(int Z, int N);
    
    /// Get element with Z=l_Z and N=l_N (e.g. 82,126).
    ame_entry get_ZN(int l_Z, int l_N);
    
    /// Get element with Z=l_Z and A=l_A (e.g. 82,208).
    ame_entry get_ZA(int l_Z, int l_A);
    
    /// Get element with name l_el and A=l_A (e.g. "Pb",208).
    ame_entry get_elA(std::string l_el, int l_A);
    
    /// Get element with string (e.g. "Pb208")
    ame_entry get(std::string nucleus);
    
    /// Returns true if data has been loaded
    bool is_loaded() { return (n>0); }

    /** \brief Set data

	This function is used by the HDF I/O routines.
     */
    int set_data(int n_mass, ame_entry *m, std::string ref);

    /// Return number of entries
    int get_nentries() { return n; }
    
#ifndef DOXYGEN_INTERNAL

  protected:

    /// The number of entries (about 3000).
    int n;
    
    /// The reference for the original data
    std::string reference;
    
    /// The array containing the mass data of length ame::n
    ame_entry *mass;
    
#endif

  };
  
  /** \brief The experimental values from Moller, Nix, Myers and Swiatecki
      
      \note This class requires data stored in an HDF file and
      thus requires HDF support for normal usage.
  */
  class ame_mass_exp : public ame_mass {

  public:

    /** \brief Return false if the mass formula does not include 
	specified nucleus
    */
    virtual bool is_included(int Z, int N);

  };

  /** \brief Mass formula entry structure for Moller, et al.
  */
  typedef struct {
    
    /// Neutron number
    int N;
    
    /// Proton number
    int Z;
    
    /// Atomic number
    int A;
    
    /** \name Ground state deformations (perturbed-spheroid parameterization)
     */
    //@{
    /// Quadrupole
    double eps2;
    /// Octupole
    double eps3;
    /// Hexadecapole
    double eps4;
    /// Hexacontatetrapole
    double eps6;
    /// Hexacontatetrapole without mass asymmetry
    double eps6sym;
    //@}
    
    /** \name Ground state deformations in the spherical-harmonics expansion
     */
    //@{
    /// Quadrupole
    double beta2;
    /// Octupole
    double beta3;
    /// Hexadecapole
    double beta4;
    /// Hexacontatetrapole
    double beta6;
    //@}
    
    /// The ground-state microscopic energy
    double Emic;
    
    /// The theoretical mass excess (in MeV)
    double Mth;
    
    /// The experimental mass excess (in MeV)
    double Mexp;
    
    /// Experimental mass excess error
    double sigmaexp;
    
    /// The ground-state microscopic energy in the FRLDM
    double EmicFL;
    
    /// The theoretical mass excess in the FRLDM
    double MthFL;
    
    /// Spin and pairity of odd proton 
    char spinp[6];
    
    /// Spin and pairity of odd neutron
    char spinn[6];

    /// Lipkin-Nogami proton gap
    double gapp;
    
    /// Lipkin-Nogami neutron gap
    double gapn;

    /// Total binding energy
    double be;

    /// One neutron separation energy
    double S1n;
    
    /// Two neutron separation energy
    double S2n;

    /** \brief Percentage of daughters generated in beta decay after
	beta-delayed neutron emission
    */
    double PA;

    /// Desc
    double PAm1;

    /// Desc
    double PAm2;

    /// Energy released in beta-decay
    double Qbeta;

    /// Half-life w.r.t. GT beta-decay
    double Tbeta;

    /// One proton separation energy
    double S1p;

    /// Two proton separation energy
    double S2p;

    /// Energy released in alpha-decay
    double Qalpha;

    /// Half-life w.r.t. alpha-decay
    double Talpha;
    
  } mnmsk_mass_entry;
  
  /** \brief Mass formula from Moller, Nix, Myers, Swiatecki and Kratz

      This is based on the tables given in \ref Moller95 and 
      \ref Moller97.

      \note This class requires data stored in an HDF file and
      thus requires HDF support for normal usage.
      
      The data containing an object of type moller_mass_entry for 8979
      nuclei is automatically loaded by the constructor. If the file
      (nucmass/mnmsk.o2) is not found, then is_loaded() will return
      <tt>false</tt> and all calls to get_ZN() will return an object
      with \c N=Z=0.

      There are several entries in the original table which are
      blank because they are in some way not known, measured, or
      computable. To distinguish these values from zero, blank entries 
      have been replaced by the number \c 1.0e99. For convenience,
      this value is returned by \ref blank().
  */
  class mnmsk_mass : public nuclear_mass_disc {
  public:
    
    mnmsk_mass();

    virtual ~mnmsk_mass();

    /** \brief Return false if the mass formula does not include 
	specified nucleus
    */
    virtual bool is_included(int Z, int N);
    
    /// Given \c Z and \c N, return the mass excess in MeV
    virtual double mass_excess(int Z, int N);
    
    /** \brief Get the entry for the specified proton and neutron number
	
	This method searches the table using a cached binary search
	algorithm. It is assumed that the table is sorted first by
	proton number and then by neutron number.
    */
    mnmsk_mass_entry get_ZN(int l_Z, int l_N);
    
    /// The value which corresponds to a blank entry
    double blank() { return 1.0e99; };

    /// Neither beta+ or beta- is possible
    double neither() { return 1.0e98; };

    /// The value which corresponds to a blank entry
    double beta_stable() { return 1.0e97; };
    
    /// Both beta+ and beta- are possible
    double beta_plus_and_minus() { return 1.0e96; };
    
    /// The value is greater than 100
    double greater_100() { return 1.0e95; };

    /// The value is greater than \f$ 10^{20} \f$
    double very_large() { return 1.0e94; };

    /// Return the type, \c "mnmsk_mass".
    virtual const char *type() { return "mnmsk_mass"; }
    
    /** \brief Set data

	This function is used by the HDF I/O routines.
     */
    int set_data(int n_mass, mnmsk_mass_entry *m, std::string ref);

#ifndef DOXYGEN_INTERNAL

  protected:
    
    /// The number of entries (about 3000).
    int n;
    
    /// The reference for the original data
    std::string reference;
    
    /// The array containing the mass data of length ame::n
    mnmsk_mass_entry *mass;
    
    /// The last table index for caching
    int last;
    
#endif
    
  };

  /** \brief The experimental values from Moller, Nix, Myers and Swiatecki
      
      \note This class requires data stored in an HDF file and
      thus requires HDF support for normal usage.
  */
  class mnmsk_mass_exp : public mnmsk_mass {

  public:

    /** \brief Return false if the mass formula does not include 
	specified nucleus
    */
    virtual bool is_included(int Z, int N);

    /// Given \c Z and \c N, return the mass excess in MeV
    virtual double mass_excess(int Z, int N);

  };

  /** \brief Mass formula entry structure for HFB mass formula
  */
  typedef struct {
    
    /// Neutron number
    int N;
    
    /// Proton number
    int Z;
    
    /// Atomic number
    int A;
    
    /// Beta 2 deformation
    double bet2;

    /// Beta 4 deformation
    double bet4;

    /// RMS charge radius
    double Rch;

    /// Deformation and Wigner energies
    double def_wig;

    /// Neutron separation energy
    double Sn;

    /// Proton separation energy
    double Sp;

    /// Beta-decay energy
    double Qbet;

    /// Calculated mass excess
    double Mcal;

    /// Error between experimental and calculated mass excess
    double Err;
    
  } hfb_mass_entry;

  /** \brief Version of \ref hfb_mass_entry with spin and parity

      \note This cannot be a child of hfb_mass_entry in order
      for the HDF I/O preprocessor macros, like HOFFSET, to work
   */
  typedef struct {
    
    /// Neutron number
    int N;
    
    /// Proton number
    int Z;
    
    /// Atomic number
    int A;
    
    /// Beta 2 deformation
    double bet2;

    /// Beta 4 deformation
    double bet4;

    /// RMS charge radius
    double Rch;

    /// Deformation and Wigner energies
    double def_wig;

    /// Neutron separation energy
    double Sn;

    /// Proton separation energy
    double Sp;

    /// Beta-decay energy
    double Qbet;

    /// Calculated mass excess
    double Mcal;

    /// Error between experimental and calculated mass excess
    double Err;

    /// Experimental spin
    double Jexp;
    
    /// Theoretical spin
    double Jth;
    
    /// Experimental parity
    int Pexp;

    /// Theoretical parity
    int Pth;

  } hfb_sp_mass_entry;
  
  /** \brief HFB Mass formula 
  */
  class hfb_mass : public nuclear_mass_disc {
    
  public:
    
    /** \brief Create a new mass formula object 
    */
    hfb_mass();

    virtual ~hfb_mass();

    /** \brief Return false if the mass formula does not include 
	specified nucleus
    */
    virtual bool is_included(int Z, int N);
    
    /// Given \c Z and \c N, return the mass excess in MeV
    virtual double mass_excess(int Z, int N);
    
    /** \brief Get the entry for the specified proton and neutron number
	
	This method searches the table using a cached binary search
	algorithm. It is assumed that the table is sorted first by
	proton number and then by neutron number.
    */
    hfb_mass_entry get_ZN(int l_Z, int l_N);
    
    /// Verify that the constructor properly loaded the table
    bool is_loaded() { return (n>0); }
    
    /// The value which corresponds to a blank entry
    double blank() { return 1.0e99; };

    /// Return the type, \c "hfb_mass".
    virtual const char *type() { return "hfb_mass"; }

    /** \brief Set data
	
	This function is used by the HDF I/O routines.
    */
    int set_data(int n_mass, hfb_mass_entry *m, std::string ref);

    /// Return number of entries
    int get_nentries() { return n; }
    
#ifndef DOXYGEN_INTERNAL

  protected:
    
    /// The number of entries (about 3000).
    int n;
    
    /// The reference for the original data
    std::string reference;
    
    /// The array containing the mass data of length ame::n
    hfb_mass_entry *mass;
    
    /// The last table index for caching
    int last;
    
#endif
    
  };

  /** \brief HFB Mass formula with spin and parity information
  */
  class hfb_sp_mass : public nuclear_mass_disc {
    
  public:
    
    /** \brief Create a new mass formula object
     */
    hfb_sp_mass();

    virtual ~hfb_sp_mass();

    /** \brief Return false if the mass formula does not include 
	specified nucleus
    */
    virtual bool is_included(int Z, int N);

    /// Given \c Z and \c N, return the mass excess in MeV
    virtual double mass_excess(int Z, int N);

    /** \brief Get the entry for the specified proton and neutron number
	
	This method searches the table using a cached binary search
	algorithm. It is assumed that the table is sorted first by
	proton number and then by neutron number.
    */
    hfb_sp_mass_entry get_ZN(int l_Z, int l_N);
    
    /// Return the type, \c "hfb_mass".
    virtual const char *type() { return "hfb_sp_mass"; }

    /** \brief Set data
	
	This function is used by the HDF I/O routines.
    */
    int set_data(int n_mass, hfb_sp_mass_entry *m, std::string ref);

#ifndef DOXYGEN_INTERNAL

  protected:
    
    /// The array containing the mass data of length ame::n
    hfb_sp_mass_entry *mass;

    /// The number of entries (about 3000).
    int n;
    
    /// The reference for the original data
    std::string reference;
    
    /// The last table index for caching
    int last;
    
#endif
    
  };
  
  /** \brief Duflo-Zuker mass formula
   */
  class dz_mass : public nuclear_mass_disc {
    
  public:
    
    /** \brief Create a new mass formula object
    */
    dz_mass(std::string model="96", bool external=false);

    virtual ~dz_mass();

    /** \brief Return false if the mass formula does not include 
	specified nucleus
    */
    virtual bool is_included(int Z, int N);
    
    /// Given \c Z and \c N, return the mass excess in MeV
    virtual double mass_excess(int Z, int N);
    
    /// Verify that the constructor properly loaded the table
    bool is_loaded() { return (n>0); }
    
    /// Return the type, \c "dz_mass".
    virtual const char *type() { return "dz_mass"; }

    /// Return number of entries
    int get_nentries() { return n; }
    
#ifndef DOXYGEN_INTERNAL

  protected:
    
    /// The reference for the original data
    std::string reference;
    
    /// Table containing the data
    table data;

    /// The last table index for caching
    int last;

    /// The total number of entries
    int n;
    
#endif
    
  };

  /** \brief Mass formula entry structure for KTUY mass formula
  */
  typedef struct {
    
    /// Neutron number
    int N;
    
    /// Proton number
    int Z;
    
    /// Atomic number
    int A;
    
    /// Calculated mass excess
    double Mcal;

    /// Shell energy
    double Esh;

    /// Alpha 2 deformation
    double alpha2;

    /// Alpha 4 deformation
    double alpha4;

    /// Alpha 6 deformation
    double alpha6;

  } ktuy_mass_entry;

  /** \brief KTUY Mass formula 
  */
  class ktuy_mass : public nuclear_mass_disc {
    
  public:
    
    /** \brief Create a new mass formula object using the specified model
	number
    */
    ktuy_mass(std::string model="05", bool external=false);

    virtual ~ktuy_mass();

    /** \brief Return false if the mass formula does not include 
	specified nucleus
    */
    virtual bool is_included(int Z, int N);
    
    /// Given \c Z and \c N, return the mass excess in MeV
    virtual double mass_excess(int Z, int N);
    
    /** \brief Get the entry for the specified proton and neutron number
	
	This method searches the table using a cached binary search
	algorithm. It is assumed that the table is sorted first by
	proton number and then by neutron number.
    */
    ktuy_mass_entry get_ZN(int l_Z, int l_N);
    
    /// Verify that the constructor properly loaded the table
    bool is_loaded() { return (n>0); }
    
    /// Return the type, \c "ktuy_mass".
    virtual const char *type() { return "ktuy_mass"; }

    /// Return number of entries
    int get_nentries() { return n; }
    
#ifndef DOXYGEN_INTERNAL

  protected:
    
    /// The number of entries (about 3000).
    int n;
    
    /// The reference for the original data
    std::string reference;
    
    /// The array containing the mass data of length ame::n
    ktuy_mass_entry *mass;
    
    /// The last table index for caching
    int last;
    
#endif
    
  };
  
  
#ifndef DOXYGENP
}
#endif

#endif
