/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <gsl/gsl_odeiv.h>

#include <o2scl/test_mgr.h>
#include <o2scl/ovector_tlate.h>
#include <o2scl/funct.h>
#include <o2scl/ode_funct.h>
#include <o2scl/gsl_rkck.h>

int expon(double x, size_t nv, const o2scl::ovector_base &y, 
	   o2scl::ovector_base &dydx) {
  dydx[0]=y[0];
  return 0;
}

int expon_gsl(double x, const double y[], double dydx[], 
	      void *params) {
  dydx[0]=y[0];
  return 0;
}

int expon2(double x, size_t nv, const o2scl::ovector_base &y, 
	    o2scl::ovector_base &dydx) {
  dydx[0]=y[0];
  dydx[1]=y[1];
  return 0;
}

using namespace std;
using namespace o2scl;

int main(void) {
  int i;
  double x, dx=1.0e-1, x_gsl;

  cout.setf(ios::scientific);

  ovector y(2), dydx(2), yout(2), yerr(2), dydx_out(2);
  double y_gsl[1], yerr_gsl[1], dydx_gsl[1], dydx_out_gsl[1];

  gsl_rkck<ode_funct_fptr<>,ovector_base,ovector,
    ovector_alloc> rk;
  gsl_rkck_fast<2,ode_funct_fptr<>,ovector_base,ovector,
    ovector_alloc> rkf;

  test_mgr t;
  t.set_output_level(1);

  // GSL setup
  const gsl_odeiv_step_type *T=gsl_odeiv_step_rkck;
  gsl_odeiv_step *s=gsl_odeiv_step_alloc(T,1);
  gsl_odeiv_system sys={expon_gsl,0,1,0};

  // O2scl function setup
  ode_funct_fptr<ovector_base> od(expon);

  x=1.0;
  y[0]=1.0;

  x_gsl=1.0;
  y_gsl[0]=1.0;

  cout << "x            y(calc)      y(exact)     diff         yerr"
       << "          y'" << endl;
  cout << 1.0 << " " << 1.0 << " " << 1.0 << " " << 0.0 << " " << -0.0 
       << " " << 1.0 << endl;

  // Compute initial derivative
  expon(x,1,y,dydx);
  expon_gsl(x_gsl,y_gsl,dydx_gsl,0);

  for(i=1;i<=40;i++) {

    // Make a step
    rk.step(x,dx,1,y,dydx,y,yerr,dydx,od);

    // Output results
    cout << x+dx << " " << y[0] << " " << exp(x+dx)/exp(1.0) << " "
	 << fabs(y[0]-exp(x+dx)/exp(1.0)) << " " << yerr[0] << " ";
    cout << dydx[0] << endl;

    // Test results
    t.test_rel(y[0],exp(x+dx)/exp(1.0),1.0e-6,"y_calculated-y_exact");
    t.test_abs(yerr[0],0.0,1.0e-6,"y_err");

    // Perform the same step with GSL to compare
    gsl_odeiv_step_apply(s,x_gsl,dx,y_gsl,yerr_gsl,
			 dydx_gsl,dydx_out_gsl,&sys);
    dydx_gsl[0]=dydx_out_gsl[0];

    // Test comparison
    t.test_rel(y_gsl[0],y[0],1.0e-10,"GSL vs. O2scl");
    t.test_rel(yerr_gsl[0],yerr[0],1.0e-9,"GSL vs. O2scl err");
    t.test_rel(dydx_gsl[0],dydx[0],1.0e-9,"GSL vs. O2scl deriv");

    x+=dx;
    x_gsl+=dx;

  }

  // Test gsl_rkck_fast

  x=1.0;
  y[0]=1.0;
  cout << "x            y(calc)      y(exact)     diff         yerr"
       << "          y'" << endl;
  cout << 1.0 << " " << 1.0 << " " << 1.0 << " " << 0.0 << " " << -0.0 
       << " " << 1.0 << endl;
  expon(x,1,y,dydx);
  for(i=1;i<=40;i++) {
    rkf.step(x,dx,1,y,dydx,y,yerr,dydx,od);
    t.test_rel(y[0],exp(x+dx)/exp(1.0),1.0e-6,"y_calculated-y_exact");
    t.test_abs(yerr[0],0.0,1.0e-6,"y_err");
    cout << x+dx << " " << y[0] << " " << exp(x+dx)/exp(1.0) << " "
	 << fabs(y[0]-exp(x+dx)/exp(1.0)) << " " << yerr[0] << " ";
    x+=dx;
    cout << dydx[0] << endl;
  }

  t.report();

  return 0;
}
