/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_ODE_FUNCT_H
#define O2SCL_ODE_FUNCT_H

#include <string>
#include <o2scl/ovector_tlate.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief Ordinary differential equation function [abstract base]

      This base class provides the basic format for specifying
      ordinary differential equations to integrate with the \o2 ODE
      solvers. Select the appropriate child of this class according
      to the kind of functions which are to be given to the solver.
  */
  template <class vec_t=ovector_base> class ode_funct {

  public:

  ode_funct() {}

  virtual ~ode_funct() {}

  /** \brief Compute the \c nv derivatives as a function of the \c nv 
      functions specified in \c y at the point \c x.
  */
  virtual int operator()(double x, size_t nv, const vec_t &y, 
			 vec_t &dydx)=0;
  
#ifndef DOXYGEN_INTERNAL

  private:

  ode_funct(const ode_funct &);
  ode_funct& operator=(const ode_funct&);

#endif

  };

  /** \brief Provide ODE functions in the form of function pointers
  */
  template <class vec_t=ovector_base>
    class ode_funct_fptr : public ode_funct<vec_t> {
  public:

  /** \brief Create an object given a function pointer
   */
  ode_funct_fptr(int (*fp)(double, size_t, const vec_t &, 
			   vec_t &)) {
    fptr=fp;
  }

  virtual ~ode_funct_fptr() {}

  /** \brief Compute the \c nv derivatives as a function of the \c nv 
      functions specified in \c y at the point \c x.
  */
  virtual int operator()(double x, size_t nv, const vec_t &y, 
			 vec_t &dydx) {
    return fptr(x,nv,y,dydx);
  }
  
#ifndef DOXYGEN_INTERNAL

  protected:

  ode_funct_fptr() {};
  
  /// The function pointer
  int (*fptr)(double x, size_t nv, const vec_t &y, 
	      vec_t &dydx);
  
  private:
  
  ode_funct_fptr(const ode_funct_fptr &);
  ode_funct_fptr& operator=(const ode_funct_fptr&);

#endif

  };

  /** \brief Provide ODE functions in the form of function pointers
  */
  template <class param_t, class vec_t=ovector_base>
    class ode_funct_fptr_param : public ode_funct<vec_t> {
  public:

  /** \brief Create an object given a function pointer
   */
  ode_funct_fptr_param(int (*fp)(double, size_t, const vec_t &, 
			   vec_t &, param_t &), param_t &pa) {
    fptr=fp;
    pp=&pa;
  }

  virtual ~ode_funct_fptr_param() {}

  /** \brief Compute the \c nv derivatives as a function of the \c nv 
      functions specified in \c y at the point \c x.
  */
  virtual int operator()(double x, size_t nv, const vec_t &y, 
			 vec_t &dydx) {
    return fptr(x,nv,y,dydx,*pp);
  }
  
#ifndef DOXYGEN_INTERNAL

  protected:

  ode_funct_fptr_param() {};

  /// The function pointer
  int (*fptr)(double x, size_t nv, const vec_t &y, 
	      vec_t &dydx, param_t &);

  /// The parameter
  param_t *pp;
  
  private:
  
  ode_funct_fptr_param(const ode_funct_fptr_param &);
  ode_funct_fptr_param& operator=(const ode_funct_fptr_param&);

#endif

  };

  /** \brief Provide ODE functions in the form of member 
      function pointers
  */
  template <class tclass, class vec_t=ovector_base> 
    class ode_funct_mfptr : public ode_funct<vec_t> {

  public:
    
  /** \brief Create an object given a class and member function pointer
   */
  ode_funct_mfptr
  (tclass *tp, int (tclass::*fp)(double x, size_t nv, 
				 const vec_t &y, vec_t &dydx)) {
    tptr=tp;
    fptr=fp;
  }
  
  virtual ~ode_funct_mfptr() {};
  
  /** \brief Compute the \c nv derivatives as a function of the \c nv 
      functions specified in \c y at the point \c x.
  */
  virtual int operator()(double x, size_t nv, const vec_t &y, 
			 vec_t &dydx) {
    return (*tptr.*fptr)(x,nv,y,dydx);
  }
  
#ifndef DOXYGEN_INTERNAL

  protected:

  /// The pointer to the member function
  int (tclass::*fptr)(double x, size_t nv, const vec_t &y, 
		      vec_t &dydx);

  /// The pointer to the class
  tclass *tptr;

  private:

  ode_funct_mfptr(const ode_funct_mfptr &);
  ode_funct_mfptr& operator=(const ode_funct_mfptr&);

#endif

  };

  /** \brief Provide ODE functions in the form of member 
      function pointers
  */
  template <class tclass, class param_t, class vec_t=ovector_base> 
    class ode_funct_mfptr_param : public ode_funct<vec_t> {

  public:
    
  /** \brief Create an object given a class and member function pointer
   */
  ode_funct_mfptr_param
  (tclass *tp, 
   int (tclass::*fp)(double x, size_t nv, 
		     const vec_t &y, vec_t &dydx, param_t &), param_t &pa) {
    tptr=tp;
    fptr=fp;
    pp=&pa;
  }
  
  virtual ~ode_funct_mfptr_param() {};
  
  /** \brief Compute the \c nv derivatives as a function of the \c nv 
      functions specified in \c y at the point \c x.
  */
  virtual int operator()(double x, size_t nv, const vec_t &y, 
			 vec_t &dydx) {
    return (*tptr.*fptr)(x,nv,y,dydx,*pp);
  }
  
#ifndef DOXYGEN_INTERNAL

  protected:

  /// The pointer to the member function
  int (tclass::*fptr)(double x, size_t nv, const vec_t &y, 
		      vec_t &dydx, param_t &);

  /// The pointer to the class
  tclass *tptr;

  /// The parameter
  param_t *pp;

  private:

  ode_funct_mfptr_param(const ode_funct_mfptr_param &);
  ode_funct_mfptr_param& operator=(const ode_funct_mfptr_param&);

#endif

  };

  /** \brief Provide ODE functions in the form of const member 
      function pointers
  */
  template <class tclass, class vec_t=ovector_base> 
    class ode_funct_cmfptr : public ode_funct<vec_t> {

  public:
    
  /** \brief Create an object given a class and member function pointer
   */
  ode_funct_cmfptr
  (tclass *tp, int (tclass::*fp)(double x, size_t nv, 
				 const vec_t &y, 
				 vec_t &dydx) const) {
    tptr=tp;
    fptr=fp;
  }
  
  virtual ~ode_funct_cmfptr() {};
  
  /** \brief Compute the \c nv derivatives as a function of the \c nv 
      functions specified in \c y at the point \c x.
  */
  virtual int operator()(double x, size_t nv, const vec_t &y, 
			 vec_t &dydx) {
    return (*tptr.*fptr)(x,nv,y,dydx);
  }
  
#ifndef DOXYGEN_INTERNAL

  protected:

  /// The pointer to the member function
  int (tclass::*fptr)(double x, size_t nv, const vec_t &y, 
		      vec_t &dydx) const;

  /// The pointer to the class
  tclass *tptr;

  private:

  ode_funct_cmfptr(const ode_funct_cmfptr &);
  ode_funct_cmfptr& operator=(const ode_funct_cmfptr&);

#endif

  };

#ifndef DOXYGENP
}
#endif

#endif
