/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_GRAPH_H
#define O2SCL_GRAPH_H

#include <string>
#include <Gtypes.h>
#include <TApplication.h>
#include <TArrow.h>
#include <TAxis.h>
#include <TBox.h>
#include <TCanvas.h>
#include <TColor.h>
#include <TF1.h>
#include <TGraph.h>
#include <TGraphErrors.h>
#include <TGaxis.h>
#include <TH1.h>
#include <TH2.h>
#include <TLatex.h>
#include <TLegend.h>
#include <TLine.h>
#include <TList.h>
#include <TMarker.h>
#include <TObjArray.h>
#include <TROOT.h>
#include <TStyle.h>
#include <TPolyLine.h>

#include <o2scl/constants.h>
#include <o2scl/table_units.h>

/** \file graph.h
    \brief Experimental functions for use with Root

    All of the functions in \ref graph.h are documented in
    the \ref o2scl_graph namespace.
*/

/** \brief Namespace for experimental functions for use with Root

    This namespace contains several functions, defined in graph.cpp,
    which simplify plotting with Root. They require that the Root
    libraries have been installed and are not compiled in with the
    normal \o2 libraries. Some of the example plots created for 
    the documentation are based on these functions.
 */
namespace o2scl_graph {
  
  /** \brief Draw a pretty arrow from (x1,y1) to (x2,y2)
      
      The parameter \c size determines the size of the head relative
      to the size of the arrow, \c size2 determines the position of
      the niche (1.0 is no niche), and alpha1 determines the angle of
      the head relative to the line (default is about 20 degrees).
  */
  void arrow(double x1, double y1, double x2, double y2,
	     TLine *&line, TPolyLine *&poly, 
	     double size=0.1, double size2=0.8, 
	     double alpha1=0.35);

  /** \brief Make a canvas and pad
      
      The CanvasName is the default output filename for graphs and
      macros also. The PadName is the name that comes up when you
      right-click on the pad. When making .eps files, the Title field
      is set to "CanvasName.eps: Windowname". This Title field is the
      title that appears, e.g., in a ghostview window. The x-axis
      limits are given by \c lleft and \c lright and the y-axis limits
      are given in \c lbottom and \c ltop. The upper left corner of
      the canvas window is at <tt>(left,top)</tt> and the lower right
      corner is <tt>(right,bottom)</tt>. The parameters \c logx and \c
      logy determine whether or not the x- or y-axes are logarithmic
      instead of linear (the default). The canvas and pad objects
      are created using \c new, and the histogram object \c th1 
      is created with TPad::DrawFrame().
  */
  int new_graph(TCanvas *&c1, TPad *&p1, TH1 *&th1, std::string CanvasName, 
		std::string WindowName, std::string PadName, 
		double lleft, double lbottom, double lright, double ltop,
		int left=0, int top=0, int right=700, int bottom=700, 
		bool logx=false, bool logy=false);
  
  /** \brief Make a canvas and pad with more tick marks

      This is the same as new_graph(), except it adds tick marks
      to the right and top axes, which are created with \c new and
      returned in \c a1 and \c a2 respectively().

      \future Modify this to just directly call new_graph() and add
      the new axes on top.
  */
  int new_graph_ticks(TCanvas *&c1, TPad *&p1, TH1 *&th1, 
		      std::string CanvasName, 
		      std::string WindowName, 
		      std::string PadName, 
		      double lleft, double lbottom, double lright, 
		      double ltop, TGaxis *&a1, TGaxis *&a2,
		      int left=0, int top=0, int right=700, 
		      int bottom=700, bool logx=false, bool logy=false);
  
  /** \brief Graph two columns from a data table

      This function plots the function defined by
      <tt>(scolx,scoly)</tt> given in table \c at with line style \c
      style and line color \c color. A \c TGraph object created with 
      \c new is returned. The \c name of the graph object
      is automatically set to be the same as the \c scoly, but this
      can always be changed afterwards, i.e.
      \code
      g1=table_graph(at,scolx,scoly);
      g1->SetName("Curve name");
      g1->Draw();
      \endcode
  */
  TGraph *table_graph(o2scl::table_units &at, std::string scolx, 
		      std::string scoly, int style=1, int color=1);
  
  /** \brief Plot colums from a data table with error bars
   */
  TGraphErrors *table_graph_errors(o2scl::table_units &at, std::string scolx, 
				   std::string scoly, std::string xerr,
				   std::string yerr, int style=1, int color=1);
  
  /** \brief Make a canvas with a two-up graph, side-by-side

      This function creates a canvas and two pads for a side-by-side
      plot with two different x-axes and the same y-axis. The x-axis
      of the left-hand plot has limits <tt>(lowx1,highx1)</tt> and
      the the x-axis for the right-hand plot has limits
      <tt>(lowx2,highx2)</tt>.
      
   */
  int two_up_graph(TCanvas *&c1, TPad *&p1, TPad *&p2, TH1 *&th1, TH1 *&th2,
		   std::string CanvasName, std::string WindowName, 
		   std::string Pad1Name, std::string Pad2Name,
		   double lowx1, double highx1, double lowx2, double highx2,
		   double lowy, double highy,
		   int left=0, int top=0, int right=1000, int bottom=700,
		   bool logx1=false, bool logx2=false, bool logy=false,
		   double alpha=1.3, double margin=0.1);

  /** \brief Make a canvas with two plots, one on top of the other
      
      The variable \c p1 is on bottom and \c p2 is on top
  */
  int two_up_graphy(TCanvas *&c1, TPad *&p1, TPad *&p2, 
		    TH1 *&th1, TH1 *&th2,
		    std::string CanvasName, std::string WindowName, 
		    std::string Pad1Name, std::string Pad2Name,
		    double lowx, double highx,
		    double lowy1, double highy1, 
		    double lowy2, double highy2,
		    int left=0, int top=0, int right=1000, int bottom=700,
		    bool logx=false, bool logy1=false, bool logy2=false,
		    double alpha=1.3, double margin=0.1);

  /// \name Some useful colors
  //@{
  const int kGray20=kGray;
  const int kGray40=kGray+1;
  const int kGray60=kGray+2;
  const int kGray80=kGray+3;
  //@}

  /// \name Markers
  //@{
  const int m_small_dot=1;
  const int m_plus=2;
  const int m_asterisk=3;
  const int m_circle=4;
  const int m_times=5;
  const int m_med_dot=6;
  const int m_large_dot=7;
  const int m_fill_circle=8;
  const int m_fill_square=21;
  const int m_fill_up_triangle=22;
  const int m_fill_dn_triangle=23;
  const int m_open_circle=24;
  const int m_open_square=25;
  const int m_open_up_triangle=26;
  const int m_open_diamond=27;
  const int m_open_plus=28;
  const int m_fill_star=29;
  const int m_open_star=30;
  const int m_asterisk2=31;
  const int m_open_dn_triangle=32;
  const int m_fill_diamond=33;
  const int m_fill_plus=34;
  //@}

}

#endif

