/*
  -------------------------------------------------------------------
  
  Copyright (C) 2010-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/hist.h>

using namespace std;
using namespace o2scl;

hist::hist() {
  bim1=&dim1;
  bim2=&dim2;
  rmode=rmode_avg;
  extend_rhs=false;
  hsize=0;
}

hist::~hist() {
  ubin.free();
  uwgt.free();
  if (urep.size()>0) urep.free();
  if (user_rep.size()>0) user_rep.free();
}

hist::hist(const hist &h) {
  bim1=&dim1;
  bim2=&dim2;
  rmode=h.rmode;
  extend_rhs=h.extend_rhs;
  hsize=h.hsize;
  ubin=h.ubin;
  urep=h.urep;
  uwgt=h.uwgt;
  user_rep=h.user_rep;
}

hist &hist::operator=(const hist &h) {
  if (&h!=this) {
    bim1=&dim1;
    bim2=&dim2;
    rmode=h.rmode;
    extend_rhs=h.extend_rhs;
    hsize=h.hsize;
    ubin=h.ubin;
    urep=h.urep;
    uwgt=h.uwgt;
    user_rep=h.user_rep;
  }
  return *this;
}

void hist::set_reps_auto() {
  if (urep.size()!=hsize) urep.allocate(hsize);
  for(size_t i=0;i<hsize;i++) {
    if (rmode==rmode_avg) urep[i]=(ubin[i]+ubin[i+1])/2.0;
    else if (rmode==rmode_low) urep[i]=ubin[i];
    else if (rmode==rmode_high) urep[i]=uwgt[i];
    else if (rmode==rmode_gmean) urep[i]=sqrt((ubin[i]*ubin[i+1]));
    else if (rmode==rmode_user) user_rep.allocate(hsize);
    else {
      O2SCL_ERR("Invalid rmode in hist::set_reps().",
		gsl_efailed);
    }
  }
  return;
}

void hist::allocate(size_t n) {
  ubin.allocate(n+1);
  uwgt.allocate(n);
  hsize=n;

  // Set all weights to zero
  uwgt.set_all(0.0);
}

int hist::set_bins(uniform_grid<double> g) {
  if (g.get_nbins()==0) {
    O2SCL_ERR2_RET("Requested zero bins in ",
		   "hist::set_bins().",gsl_einval);
  }
  // Allocate if necessary
  if (g.get_nbins()!=hsize) {
    if (hsize!=0) {
      O2SCL_ERR2_RET("Requested binning change in non-empty ",
		     "histogram in hist::set_bins().",gsl_efailed);
    }
    allocate(g.get_nbins());
  }
  // Set the rep mode from the uniform_grid
  if (g.is_log()) {
    rmode=rmode_gmean;
  } else {
    rmode=rmode_avg;
  }
  // Set the uniform_grid
  g.vector(ubin);
  // Reset reps
  if (urep.size()>0) urep.free();
  return gsl_success;
}

int hist::clear_wgts() {
  uwgt.set_all(0.0);
  return 0;
}

int hist::clear() {
  ubin.free();
  uwgt.free();
  if (urep.size()>0) urep.free();
  if (user_rep.size()>0) user_rep.free();
  hsize=0;
  return 0;
}

size_t hist::get_bin_index(double x) const {
  search_vec<uvector> sv(ubin.size(),ubin);
  // Increasing case
  if (ubin[0]<ubin[hsize]) {
    if (x<ubin[0]) {
      std::string s="Value '"+dtos(x)+"' smaller than smallest "+
	"bin '"+dtos(ubin[0])+"' (inc) in hist::get_bin_index().";
      O2SCL_ERR_RET(s.c_str(),gsl_einval);
    }
    if (!extend_rhs && x>ubin[hsize]) {
      std::string s="Value '"+dtos(x)+"' larger than largest "+
	"bin '"+dtos(ubin[hsize])+"' (inc) in hist::get_bin_index().";
      O2SCL_ERR_RET(s.c_str(),gsl_einval);
    }
    return sv.find_inc(x);
  } else {
    // Decreasing case
    if (x>ubin[0]) {
      std::string s="Value '"+dtos(x)+"' larger than largest "+
	"bin '"+dtos(ubin[0])+"' (dec) in hist::get_bin_index().";
      O2SCL_ERR_RET(s.c_str(),gsl_einval);
    }
    if (!extend_rhs && x<ubin[hsize]) {
      std::string s="Value '"+dtos(x)+"' smaller than smallest "+
	"bin '"+dtos(ubin[hsize])+"' (dec) in hist::get_bin_index().";
      O2SCL_ERR_RET(s.c_str(),gsl_einval);
    }
    return sv.find_dec(x);
  }
}

double &hist::get_bin_low_i(size_t i) {
  if (urep.size()>0) urep.free();
  if (i>=hsize) {
    std::string s="Index "+itos(i)+" must be smaller than "+
      "the histogram size "+itos(hsize)+
      " in hist::get_bin_low().";
    O2SCL_ERR(s.c_str(),gsl_einval);
    return ubin[0];
  }
  return ubin[i];
}

const double &hist::get_bin_low_i(size_t i) const {
  if (i>=hsize) {
    std::string s="Index "+itos(i)+" must be smaller than "+
      "the histogram size "+itos(hsize)+
      " in hist::get_bin_low().";
    O2SCL_ERR(s.c_str(),gsl_einval);
    return ubin[0];
  }
  return ubin[i];
}

double &hist::get_bin_high_i(size_t i) {
  if (urep.size()>0) urep.free();
  if (i>=hsize) {
    std::string s="Index "+itos(i)+" must be smaller than "+
      "the histogram size "+itos(hsize)+
      " in hist::get_bin_high().";
    O2SCL_ERR(s.c_str(),gsl_einval);
    return ubin[0];
  }
  return ubin[i+1];
}

const double &hist::get_bin_high_i(size_t i) const {
  if (i>=hsize) {
    std::string s="Index "+itos(i)+" must be smaller than "+
      "the histogram size "+itos(hsize)+
      " in hist::get_bin_high().";
    O2SCL_ERR(s.c_str(),gsl_einval);
    return ubin[0];
  }
  return ubin[i+1];
}

int hist::set_rep_mode(size_t mode) {
  if (mode>4) {
    O2SCL_ERR2_RET("Valid rep mode values are 0-4 in ",
		   "hist::set_rep_mode().",gsl_einval);
  }
  rmode=mode;
  if (urep.size()>0) urep.free();
  return 0;
}

double &hist::get_rep_i(size_t i) {
  if (urep.size()==0) set_reps_auto();
  if (i>=hsize) {
    std::string s="Index "+itos(i)+" must be smaller than "+
      "the histogram size "+itos(hsize)+
      " in hist::get_bin().";
    O2SCL_ERR(s.c_str(),gsl_einval);
    return urep[0];
  }
  return urep[i];
}

int hist::update(double x, double val) {
  size_t loc=get_bin_index(x);
  uwgt[loc]+=val;
  return 0;
}

const double &hist::get_wgt_i(size_t i) const {
  if (i>=hsize) {
    std::string s="Index "+itos(i)+" must be smaller than "+
      "the histogram size "+itos(hsize)+
      " in hist::get_wgt().";
    O2SCL_ERR(s.c_str(),gsl_einval);
    return uwgt[0];
  }
  return uwgt[i];
}

double &hist::get_wgt_i(size_t i) {
  if (i>=hsize) {
    std::string s="Index "+itos(i)+" must be smaller than "+
      "the histogram size "+itos(hsize)+
      " in hist::get_wgt().";
    O2SCL_ERR(s.c_str(),gsl_einval);
    return uwgt[0];
  }
  return uwgt[i];
}

int hist::set_wgt_i(size_t i, double val) {
  if (i>=hsize) {
    std::string s="Index "+itos(i)+" must be smaller than "+
      "the histogram size "+itos(hsize)+
      " in hist::set().";
    O2SCL_ERR_RET(s.c_str(),gsl_einval);
  }
  uwgt[i]=val;
  return 0;
}

double hist::operator()(double x) {
  if (rmode==rmode_user) {
    interp_t si(*bim1,hsize,user_rep,uwgt);
    return si.interp(x);
  } 
  if (urep.size()==0) set_reps_auto();
  interp_t si(*bim1,hsize,urep,uwgt);
  return si.interp(x);
}

double hist::interp(double x) {
  if (rmode==rmode_user) {
    interp_t si(*bim1,hsize,user_rep,uwgt);
    return si.interp(x);
  } 
  if (urep.size()==0) set_reps_auto();
  interp_t si(*bim1,hsize,urep,uwgt);
  return si.interp(x);
}

double hist::deriv(double x) {
  if (rmode==rmode_user) {
    interp_t si(*bim1,hsize,user_rep,uwgt);
    return si.deriv(x);
  } 
  if (urep.size()==0) set_reps_auto();
  interp_t si(*bim1,hsize,urep,uwgt);
  return si.deriv(x);
}

double hist::deriv2(double x) {
  if (rmode==rmode_user) {
    interp_t si(*bim1,hsize,user_rep,uwgt);
    return si.deriv2(x);
  } 
  if (urep.size()==0) set_reps_auto();
  interp_t si(*bim1,hsize,urep,uwgt);
  return si.deriv2(x);
}

double hist::integ(double x, double y) {
  if (rmode==rmode_user) {
    interp_t si(*bim1,hsize,user_rep,uwgt);
    return si.integ(x,y);
  } 
  if (urep.size()==0) set_reps_auto();
  interp_t si(*bim1,hsize,urep,uwgt);
  return si.integ(x,y);
}


int hist::set_interp(base_interp_mgr<uvector_const_view> &bi1,
			 base_interp_mgr<uvector_const_subvector> &bi2) {
  bim1=&bi1;
  bim2=&bi2;
  return 0;
}

