/*
  -------------------------------------------------------------------
  
  Copyright (C) 2010-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_HIST_H
#define O2SCL_HIST_H

#include <iostream>
#include <o2scl/convert_units.h>
#include <o2scl/smart_interp.h>
#include <o2scl/uvector_tlate.h>
#include <o2scl/uniform_grid.h>

#ifndef DOXYGENP
namespace o2scl {
#endif
  
  /** \brief A one-dimensional histogram class
      
      Experimental.

      One may set the histogram bins using \ref set_bins() or one may
      manually set the limit of one bin using the reference returned
      by get_bin_low(), get_bin_low_i(), get_bin_high(), or
      get_bin_high_i(). Note that if one attempts to set the bins on a
      histogram where the bins have already been set, one must ensure
      that the new and old binnings have the same size. This
      requirement is designed to help prevent accidental data loss.

      The upper edge of bin \c i is always equal to the lower edge
      of bin \c i+1. 

      By convention, all functions which take a bin index as an
      argument have an extra \c "_i" suffix to distinguish them from
      functions which take a floating-point value to be binned as
      their argument.

      Empty histograms have zero size. 

      The \ref set_bins() functions can only be used if (i) the current
      histogram is empty or (ii) the new number of bins is equal to
      the histogram size so that no reallocation is necessary.

      To save space, representative vectors are not allocated until
      they are used.

      \hline
      \b Bin Representatives 

      One way to operate on a histogram as a function is to designate
      a particular value inside each bin which represents the
      coordinate associated with that bin. These "bin representative
      values" are automatically created and can be used by the
      function evaluation and interpolation functions
      operator(double), interp(), deriv(), integ() for some particular
      interpolation type. By default, these representative values are
      taken to be the midpoint of each bin, but this option is
      configurable and the representative values may be set by the
      user for each individual bin.

      \hline

      \todo Check that the actions of set_bins(), clear() and 
      other functions perform the correct actions on the user_rep
      vector. I'm a bit concerned that set_bins_auto() shouldn't
      call user_rep.allocate().

      \todo More documentation and testing.

      \future Would be nice not to have to create a new
      \ref search_vec object in get_bin_index().
      \future Consider adding the analogs of the GSL histogram
      sampling functions 
      \future Add a function which computes the bin sizes?
  */
  class hist {
    
  protected:

    /// Bin locations (N+1)
    uvector ubin;
    
    /// Bin contents (N)
    uvector uwgt;

    /// Bin representative values (N)
    uvector urep;
    
    /// User-defined representative values (N)
    uvector user_rep;
    
    /// Number of bins
    size_t hsize;

    /// Representative mode
    size_t rmode;

    /// A pointer to the interpolation manager
    base_interp_mgr<uvector_const_view> *bim1;
    
    /// A pointer to the subvector interpolation manager
    base_interp_mgr<uvector_const_subvector> *bim2;
    
    /// Default interpolation manager
    def_interp_mgr<uvector_const_view,cspline_interp> dim1;

    /// Default interpolation manager
    def_interp_mgr<uvector_const_subvector,cspline_interp> dim2;

    /// Set the representative array according to current rmode
    void set_reps_auto();

    /// Interpolation typedef
    typedef o2scl_interp_vec<uvector_const_view> interp_t;

    /** \brief Allocate vectors for a histogram of size \c n

	This function also sets all the weights to zero.
     */
    void allocate(size_t n);

  public:

    hist();

    ~hist();

    /// Copy constructor
    hist(const hist &h);

    /// Copy constructor
    hist &operator=(const hist &h);

    /// The histogram size
    size_t size() const {
      return hsize;
    }

    /** \brief If true, allow abcissa larger than largest bin limit
	to correspond to the highest bin (default false)
    */
    bool extend_rhs;

    /// \name Initial bin setup
    //@{
    /** \brief Set bins from a \ref uniform_grid object
	
        \note If the size of the histogram uniform_grid is not equal
	to the current histogram size, this function requires the
	histogram be empty so it can reallocate the vectors for both
	the bins and the weights.
    */
    int set_bins(uniform_grid<double> g);

    /** \brief Set the bins from a vector

        \note If \c n is not equal to the current histogram size,
	this function requires the histogram be empty so it can
	reallocate the vectors for both the bins and the weights.
     */
    template<class vec_t> int set_bins(size_t n, const vec_t &v) {
      if (n!=hsize) {
	if (hsize!=0) {
	  O2SCL_ERR2_RET("Requested binning change in non-empty ",
			 "histogram in hist::set_bins().",gsl_efailed);
	}
	allocate(n-1);
      }
      for(size_t i=0;i<n;i++) ubin[i]=v[i];
      set_reps_auto();
      return gsl_success;
    }
    //@}

    /// \name Weight functions
    //@{
    /// Increment bin for \c x by value \c val
    int update(double x, double val=1.0);

    /// Increment bin with index \c i by value \c val
    int update_i(size_t i, double val=1.0) {
      uwgt[i]+=val;
      return 0;
    }

    /// Return contents of bin with index \c i
    const double &get_wgt_i(size_t i) const;

    /// Return contents of bin with index \c i
    double &get_wgt_i(size_t i);

    /// Return contents of bin for \c x
    const double &get_wgt(double x) const {
      return get_wgt_i(get_bin_index(x));
    }

    /// Return contents of bin for \c x
    double &get_wgt(double x) {
      return get_wgt_i(get_bin_index(x));
    }

    /// Set contents of bin with index \c i to value \c val
    int set_wgt_i(size_t i, double val);

    /// Set contents of bin for \c x to value \c val
    int set_wgt(double x, double val) {
      return set_wgt_i(get_bin_index(x),val);
    }
    
    /// Get a reference to the full y vector
    const uvector &get_wgts() const {
      return uwgt;
    }

    /// Get a reference to the weight for the bin at index \c i
    const double &operator[](size_t i) const {
      return uwgt[i];
    }
    
    /// Get a reference to the weight for the bin at index \c i
    double &operator[](size_t i) {
      return uwgt[i];
    }
    //@}

    /// \name Bin manipulation
    //@{
    /// Get the index of the bin which holds \c x
    size_t get_bin_index(double x) const;

    /// Get the lower edge of bin of index \c i
    double &get_bin_low_i(size_t i);

    /// Get the lower edge of bin of index \c i
    const double &get_bin_low_i(size_t i) const;

    /// Get the upper edge of bin of index \c i
    double &get_bin_high_i(size_t i);
    
    /// Get the upper edge of bin of index \c i
    const double &get_bin_high_i(size_t i) const;

    /// Get the lower edge of bin of index \c i
    double &get_bin_low(double x) {
      return get_bin_low_i(get_bin_index(x));
    }

    /// Get the lower edge of bin of index \c i
    const double &get_bin_low(double x) const {
      return get_bin_low_i(get_bin_index(x));
    }

    /// Get the upper edge of bin of index \c i
    double &get_bin_high(double x) {
      return get_bin_high_i(get_bin_index(x));
    }
    
    /// Get the upper edge of bin of index \c i
    const double &get_bin_high(double x) const {
      return get_bin_high_i(get_bin_index(x));
    }

    /// Get a reference to the full vector of bin specifications
    const uvector &get_bins() const {
      return ubin;
    }
    //@}

    /** \brief Get maximum weight
     */
    double get_max_wgt() {
      double max=uwgt[0];
      for(size_t i=1;i<hsize;i++) {
	if (uwgt[i]>max) max=uwgt[i];
      }
      return max;
    }

    /** \brief Get the index of the maximum weight
     */
    size_t get_max_index() {
      double max=uwgt[0];
      size_t max_ix=0;
      for(size_t i=1;i<hsize;i++) {
	if (uwgt[i]>max) {
	  max=uwgt[i];
	  max_ix=i;
	}
      }
      return max_ix;
    }

    /** \brief Get the representative for the bin with maximum weight
     */
    double get_max_rep() {
      double max=uwgt[0];
      size_t max_ix=0;
      for(size_t i=1;i<hsize;i++) {
	if (uwgt[i]>max) {
	  max=uwgt[i];
	  max_ix=i;
	}
      }
      if (urep.size()==0) set_reps_auto();
      return urep[max_ix];
    }

    /// \name Delete functions
    //@{
    /// Clear the data, but leave the bins as is
    int clear_wgts();

    /// Clear the entire histogram
    int clear();
    //@}

    /// \name Rep modes (default is \c rmode_avg)
    //@{
    static const size_t rmode_avg=0;
    static const size_t rmode_user=1;
    static const size_t rmode_low=2;
    static const size_t rmode_high=3;
    static const size_t rmode_gmean=4;
    //@}
    
    /// \name Representative functions
    //@{
    /// Set the representative x-values for each bin
    template<class vec_t> int set_reps(size_t n, const vec_t &v) {
      rmode=rmode_user;
      if (user_rep.size()>0) user_rep.free();
      user_rep.allocate(n);
      if (user_rep.size()!=n) {
	std::string s="Expected a vector of size "+itos(user_rep.size())+
	  " and got a vector of size "+itos(n)+" in hist::set_reps().";
	O2SCL_ERR_RET(s.c_str(),gsl_einval);
      }
      for(size_t i=0;i<n;i++) user_rep[i]=v[i];
      return gsl_success;
    }

    /// Set mode used to compute bin representatives
    int set_rep_mode(size_t mode);

    /// Get mode used to compute bin representatives
    size_t get_rep_mode() const {
      return rmode;
    }

    /// Return the representative of bin of index \c i
    double &get_rep_i(size_t i);

    /// Return the representative of bin containing \c x
    double &get_rep(double x) {
      return get_rep_i(get_bin_index(x));
    }

    /// Get a reference to the full representative vector
    const uvector &get_reps() {
      if (urep.size()==0) set_reps_auto();
      return urep;
    }

    /// Get a reference to the full data vector
    const uvector &get_user_reps() const {
      return user_rep;
    }
    //@}

    /// \name Evaluation and interpolation functions
    //@{
    /// Return the value of the function at \c x
    double operator()(double x);

    /// Return the value of the function at \c x
    double interp(double x);

    /// Return the derivative of the function at \c x
    double deriv(double x);

    /// Return the second derivative of the function at \c x
    double deriv2(double x);

    /// Return the integral of the function between \c x and \c y
    double integ(double x, double y);

    /// Set the base interpolation objects
    int set_interp(base_interp_mgr<uvector_const_view> &bi1,
		   base_interp_mgr<uvector_const_subvector> &bi2);
    //@}

  };

#ifndef DOXYGENP
}
#endif

#endif
