/*
  -------------------------------------------------------------------
  
  Copyright (C) 2010-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/hist_2d.h>

using namespace std;
using namespace o2scl;

hist_2d::hist_2d() {
  xrmode=rmode_avg;
  yrmode=rmode_avg;
  extend_rhs=false;
}

hist_2d::~hist_2d() {
  xa.free();
  ya.free();
  wgt.free();
  if (xrep.size()>0) xrep.free();
  if (yrep.size()>0) yrep.free();
  if (user_xrep.size()>0) user_xrep.free();
  if (user_yrep.size()>0) user_yrep.free();
}

hist_2d::hist_2d(const hist_2d &h) {
  bim1=&dim1;
  bim2=&dim2;
  xrmode=h.xrmode;
  yrmode=h.yrmode;
  extend_rhs=h.extend_rhs;
  hsize_x=h.hsize_x;
  hsize_y=h.hsize_y;
  xa=h.xa;
  ya=h.ya;
  xrep=h.xrep;
  yrep=h.yrep;
  user_xrep=h.user_xrep;
  user_yrep=h.user_yrep;
  wgt=h.wgt;
}

hist_2d &hist_2d::operator=(const hist_2d &h) {
  if (&h!=this) {
    bim1=&dim1;
    bim2=&dim2;
    xrmode=h.xrmode;
    yrmode=h.yrmode;
    extend_rhs=h.extend_rhs;
    hsize_x=h.hsize_x;
    hsize_y=h.hsize_y;
    xa=h.xa;
    ya=h.ya;
    xrep=h.xrep;
    yrep=h.yrep;
    user_xrep=h.user_xrep;
    user_yrep=h.user_yrep;
    wgt=h.wgt;
  }
  return *this;
}

int hist_2d::set_bins(uniform_grid<double> gx, uniform_grid<double> gy) {
  if (gx.get_nbins()==0 || gy.get_nbins()==0) {
    O2SCL_ERR2_RET("Requested zero bins in ",
		   "hist::set_bins().",gsl_einval);
  }
  allocate(gx.get_nbins(),gy.get_nbins());
  if (gx.is_log()) {
    xrmode=rmode_gmean;
  } else {
    xrmode=rmode_avg;
  }
  gx.vector(xa);
  if (gy.is_log()) {
    yrmode=rmode_gmean;
  } else {
    yrmode=rmode_avg;
  }
  gy.vector(ya);
  wgt.set_all(0.0);
  set_reps_auto();
  return gsl_success;
}

void hist_2d::allocate(size_t nx, size_t ny) {
  xa.allocate(nx+1);
  ya.allocate(ny+1);
  xrep.allocate(nx);
  yrep.allocate(ny);
  wgt.allocate(nx,ny);
  hsize_x=nx;
  hsize_y=ny;
  return;
}

void hist_2d::set_reps_auto() {
  for(size_t i=0;i<hsize_x;i++) {
    if (xrmode==rmode_avg) xrep[i]=(xa[i]+xa[i+1])/2.0;
    else if (xrmode==rmode_low) xrep[i]=xa[i];
    else if (xrmode==rmode_high) xrep[i]=ya[i];
    else if (xrmode==rmode_gmean) xrep[i]=sqrt((xa[i]*xa[i+1]));
    else {
      O2SCL_ERR("Invalid xrmode in hist_2d::set_reps_auto().",
		gsl_efailed);
    }
  }
  for(size_t i=0;i<hsize_y;i++) {
    if (yrmode==rmode_avg) yrep[i]=(ya[i]+ya[i+1])/2.0;
    else if (yrmode==rmode_low) yrep[i]=ya[i];
    else if (yrmode==rmode_high) yrep[i]=ya[i];
    else if (yrmode==rmode_gmean) yrep[i]=sqrt((ya[i]*ya[i+1]));
    else {
      O2SCL_ERR("Invalid yrmode in hist_2d::set_reps_auto().",
		gsl_efailed);
    }
  }
  return;
}

int hist_2d::clear_wgts() {
  wgt.set_all(0.0);
  return 0;
}

int hist_2d::clear() {
  xa.free();
  ya.free();
  user_xrep.free();
  xrep.free();
  user_yrep.free();
  yrep.free();
  hsize_x=0;
  hsize_y=0;
  return 0;
}

void hist_2d::get_bin_indices(double x, double y, 
				  size_t &i, size_t &j) const {
  i=get_x_bin_index(x);
  j=get_y_bin_index(y);
  return;
}

size_t hist_2d::get_x_bin_index(double x) const {
  size_t i;

  // Compute x index
  search_vec<ovector> sv(xa.size(),xa);
  if (xa[0]<xa[hsize_x]) {
    if (x<xa[0]) {
      std::string s="Value '"+dtos(x)+"' smaller than smallest "+
	"bin '"+dtos(xa[0])+"' in hist_2d::get_bin().";
      O2SCL_ERR(s.c_str(),gsl_einval);
    }
    if (!extend_rhs && x>xa[hsize_x]) {
      std::string s="Value '"+dtos(x)+"' larger than largest "+
	"bin '"+dtos(xa[hsize_x])+"' in hist::get_bin().";
      O2SCL_ERR(s.c_str(),gsl_einval);
    }
    i=sv.find_inc(x);
  } else {
    if (x>xa[0]) {
      std::string s="Value '"+dtos(x)+"' larger than largest "+
	"bin '"+dtos(xa[0])+"' in hist_2d::get_bin().";
      O2SCL_ERR(s.c_str(),gsl_einval);
    }
    if (!extend_rhs && x<xa[hsize_x]) {
      std::string s="Value '"+dtos(x)+"' smaller than smallest "+
	"bin '"+dtos(xa[hsize_x])+"' in hist_2d::get_bin().";
      O2SCL_ERR(s.c_str(),gsl_einval);
    }
    i=sv.find_dec(x);
  }

  return i;
}

size_t hist_2d::get_y_bin_index(double y) const {
  size_t j;

  // Compute y index
  search_vec<ovector> sv2(ya.size(),ya);
  if (ya[0]<ya[hsize_y]) {
    if (y<ya[0]) {
      std::string s="Value '"+dtos(y)+"' smaller than smallest "+
	"bin '"+dtos(ya[0])+"' in hist_2d::get_bin().";
      O2SCL_ERR(s.c_str(),gsl_einval);
    }
    if (!extend_rhs && y>ya[hsize_y]) {
      std::string s="Value '"+dtos(y)+"' larger than largest "+
	"bin '"+dtos(ya[hsize_y])+"' in hist::get_bin().";
      O2SCL_ERR(s.c_str(),gsl_einval);
    }
    j=sv2.find_inc(y);
  } else {
    if (y>ya[0]) {
      std::string s="Value '"+dtos(y)+"' larger than largest "+
	"bin '"+dtos(ya[0])+"' in hist_2d::get_bin().";
      O2SCL_ERR(s.c_str(),gsl_einval);
    }
    if (!extend_rhs && y<ya[hsize_y]) {
      std::string s="Value '"+dtos(y)+"' smaller than smallest "+
	"bin '"+dtos(ya[hsize_y])+"' in hist_2d::get_bin().";
      O2SCL_ERR(s.c_str(),gsl_einval);
    }
    j=sv2.find_dec(y);
  }

  return j;
}

double &hist_2d::get_x_low_i(size_t i) {
  if (i>=hsize_x) {
    std::string s="Index "+itos(i)+" must be smaller than "+
      "the histogram size "+itos(hsize_x)+
      " in hist_2d::get_x_low().";
    O2SCL_ERR(s.c_str(),gsl_einval);
    return xa[0];
  }
  return xa[i];
}

const double &hist_2d::get_x_low_i(size_t i) const {
  if (i>=hsize_x) {
    std::string s="Index "+itos(i)+" must be smaller than "+
      "the histogram size "+itos(hsize_x)+
      " in hist_2d::get_x_low().";
    O2SCL_ERR(s.c_str(),gsl_einval);
    return xa[0];
  }
  return xa[i];
}

double &hist_2d::get_x_high_i(size_t i) {
  if (i>=hsize_x) {
    std::string s="Index "+itos(i)+" must be smaller than "+
      "the histogram size "+itos(hsize_x)+
      " in hist_2d::get_x_high().";
    O2SCL_ERR(s.c_str(),gsl_einval);
    return xa[0];
  }
  return xa[i+1];
}

const double &hist_2d::get_x_high_i(size_t i) const {
  if (i>=hsize_x) {
    std::string s="Index "+itos(i)+" must be smaller than "+
      "the histogram size "+itos(hsize_x)+
      " in hist_2d::get_x_high().";
    O2SCL_ERR(s.c_str(),gsl_einval);
    return xa[0];
  }
  return xa[i+1];
}

double &hist_2d::get_y_low_i(size_t j) {
  if (j>=hsize_y) {
    std::string s="Index "+itos(j)+" must be smaller than "+
      "the histogram size "+itos(hsize_y)+
      " in hist_2d::get_y_low().";
    O2SCL_ERR(s.c_str(),gsl_einval);
    return ya[0];
  }
  return ya[j];
}

const double &hist_2d::get_y_low_i(size_t j) const {
  if (j>=hsize_y) {
    std::string s="Index "+itos(j)+" must be smaller than "+
      "the histogram size "+itos(hsize_y)+
      " in hist_2d::get_y_low().";
    O2SCL_ERR(s.c_str(),gsl_einval);
    return ya[0];
  }
  return ya[j];
}

double &hist_2d::get_y_high_i(size_t j) {
  if (j>=hsize_y) {
    std::string s="Index "+itos(j)+" must be smaller than "+
      "the histogram size "+itos(hsize_y)+
      " in hist_2d::get_y_high().";
    O2SCL_ERR(s.c_str(),gsl_einval);
    return ya[0];
  }
  return ya[j+1];
}

const double &hist_2d::get_y_high_i(size_t j) const {
  if (j>=hsize_y) {
    std::string s="Index "+itos(j)+" must be smaller than "+
      "the histogram size "+itos(hsize_y)+
      " in hist_2d::get_y_high().";
    O2SCL_ERR(s.c_str(),gsl_einval);
    return ya[0];
  }
  return ya[j+1];
}

int hist_2d::set_rep_mode(size_t x_mode, size_t y_mode) {
  if (x_mode>4 || y_mode>4) {
    O2SCL_ERR2_RET("Valid rep mode values are 0-4 in ",
		   "hist_2d::set_rep_mode().",gsl_einval);
  }
  xrmode=x_mode;
  yrmode=y_mode;
  set_reps_auto();
  return 0;
}

double &hist_2d::get_x_rep_i(size_t i) {
  if (i>=hsize_x) {
    std::string s="Index "+itos(i)+" must be smaller than "+
      "the histogram size "+itos(hsize_x)+
      " in hist_2d::get_x().";
    O2SCL_ERR(s.c_str(),gsl_einval);
    return xrep[0];
  }
  return xrep[i];
}

const double &hist_2d::get_x_rep_i(size_t i) const {
  if (i>=hsize_x) {
    std::string s="Index "+itos(i)+" must be smaller than "+
      "the histogram size "+itos(hsize_x)+
      " in hist_2d::get_x().";
    O2SCL_ERR(s.c_str(),gsl_einval);
    return xrep[0];
  }
  return xrep[i];
}

double &hist_2d::get_y_rep_i(size_t j) {
  if (j>=hsize_y) {
    std::string s="Index "+itos(j)+" must be smaller than "+
      "the histogram size "+itos(hsize_y)+
      " in hist_2d::get_y().";
    O2SCL_ERR(s.c_str(),gsl_einval);
    return yrep[0];
  }
  return yrep[j];
}

const double &hist_2d::get_y_rep_i(size_t j) const {
  if (j>=hsize_y) {
    std::string s="Index "+itos(j)+" must be smaller than "+
      "the histogram size "+itos(hsize_y)+
      " in hist_2d::get_y().";
    O2SCL_ERR(s.c_str(),gsl_einval);
    return yrep[0];
  }
  return yrep[j];
}

const double &hist_2d::get_wgt_i(size_t i, size_t j) const {
  if (i>=hsize_x || j>=hsize_y) {
    std::string s="Index "+itos(i)+" must be smaller than "+
      "the histogram size "+itos(hsize_x)+
      " in hist_2d::get_wgt().";
    O2SCL_ERR(s.c_str(),gsl_einval);
    return wgt[0][0];
  }
  return wgt[i][j];
}

double &hist_2d::get_wgt_i(size_t i, size_t j) {
  if (i>=hsize_x || j>=hsize_y) {
    std::string s="Index "+itos(i)+" must be smaller than "+
      "the histogram size "+itos(hsize_x)+
      " in hist_2d::get_wgt().";
    O2SCL_ERR(s.c_str(),gsl_einval);
    return wgt[0][0];
  }
  return wgt[i][j];
}

int hist_2d::set_wgt_i(size_t i, size_t j, double val) {
  if (i>=hsize_x || j>=hsize_y) {
    std::string s="Index "+itos(i)+" must be smaller than "+
      "the histogram size "+itos(hsize_x)+
      " in hist_2d::set().";
    O2SCL_ERR_RET(s.c_str(),gsl_einval);
  }
  wgt[i][j]=val;
  return 0;
}

