/*
  -------------------------------------------------------------------
  
  Copyright (C) 2010-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_HIST_2D_H
#define O2SCL_HIST_2D_H

#include <iostream>
#include <o2scl/convert_units.h>
#include <o2scl/smart_interp.h>
#include <o2scl/omatrix_tlate.h>
#include <o2scl/twod_intp.h>
#include <o2scl/uniform_grid.h>

#ifndef DOXYGENP
namespace o2scl {
#endif
  
  /** \brief A two-dimensional histogram class

      Experimental.

      \future Write a function to create a 1-d histogram 
      from a 2-d histogram either by selecting one bin
      in one axis or by marginalizing over one direction.
   */
  class hist_2d {
    
  protected:

    /// Bin locations (Nx+1)
    ovector xa;

    /// Bin locations (Ny+1)
    ovector ya;

    /// Values (Nx,Ny)
    omatrix wgt;

    /// "Central" values for x-axis (N)
    ovector xrep;

    /// "Central" values for y-axis (N)
    ovector yrep;
    
    /// User-defined central values for x-axis (N)
    ovector user_xrep;
    
    /// User-defined central values for y-axis (N)
    ovector user_yrep;
    
    /// Number of x-bins
    size_t hsize_x;

    /// Number of y-bins
    size_t hsize_y;

    /// Rep mode for x
    size_t xrmode;

    /// Rep mode for y
    size_t yrmode;

    /// A pointer to the interpolation manager
    base_interp_mgr<ovector_const_view> *bim1;
    
    /// A pointer to the subvector interpolation manager
    base_interp_mgr<ovector_const_subvector> *bim2;
    
    /// Default interpolation manager
    def_interp_mgr<ovector_const_view,cspline_interp> dim1;

    /// Default interpolation manager
    def_interp_mgr<ovector_const_subvector,cspline_interp> dim2;

    /** \brief Allocate for a histogram of size \c nx, \c ny
	
	This function also sets all the weights to zero.
    */
    void allocate(size_t nx, size_t ny);

    /** \brief An internal function to automatically set
	\ref xrep and \ref yrep
    */
    void set_reps_auto();

  public:

    hist_2d();

    virtual ~hist_2d();

    hist_2d(const hist_2d &h);

    hist_2d &operator=(const hist_2d &h);
    
    /** \brief If true, allow abcissa larger than largest bin limit
	to correspond to the highest bin (default false).
    */
    bool extend_rhs;

    /// \name Initial bin setup
    //@{
    int set_bins(uniform_grid<double> gx, uniform_grid<double> gy);

    /// Set the bins from a vector
    template<class vec_t> int set_bins(size_t nx, vec_t &vx,
				       size_t ny, vec_t &vy) {
      allocate(nx-1,ny-1);
      for(size_t i=0;i<nx;i++) xa[i]=vx[i];
      for(size_t i=0;i<ny;i++) ya[i]=vy[i];
      set_reps_auto();
      return gsl_success;
    }
    //@}

    /// \name Weight functions
    //@{
    /// Increment bin for \c x by value \c val
    int update_i(size_t i, size_t j, double val=1.0) {
      wgt[i][j]+=val;
      return 0;
    }

    /// Increment bin for \c x by value \c val
    int update(double x, double y, double val=1.0) {
      size_t i, j;
      get_bin_indices(x,y,i,j);
      return update_i(i,j,val);
    }

    /// Return contents of bin with index \c i
    const double &get_wgt_i(size_t i, size_t j) const;

    /// Return contents of bin for \c x
    const double &get_wgt(double x, double y) const {
      size_t i, j;
      get_bin_indices(x,y,i,j);
      return get_wgt_i(i,j);
    }

    /// Return contents of bin with index \c i
    double &get_wgt_i(size_t i, size_t j);

    /// Return contents of bin for \c x
    double &get_wgt(double x, double y) {
      size_t i, j;
      get_bin_indices(x,y,i,j);
      return get_wgt_i(i,j);
    }

    /// Set contents of bin with index \c i to value \c val
    int set_wgt_i(size_t i, size_t j, double val);
    
    /// Set contents of bin for \c x to value \c val
    int set_wgt(double x, double y, double val) {
      size_t i, j;
      get_bin_indices(x,y,i,j);
      return set_wgt_i(i,j,val);
    }
    
    /// Get a const reference to the full data vector
    const omatrix &get_wgts() const {
      return wgt;
    }

    /// Get a reference to the full data vector
    omatrix &get_wgts() {
      return wgt;
    }
    //@}
    
    /// \name Delete functions
    //@{
    /// Clear the data, but leave the bins as is
    int clear_wgts();

    /// Clear the entire histogram
    int clear();
    //@}

    /// \name Bin manipulation
    //@{
    /// Get the index of the bin which holds \c x
    void get_bin_indices(double x, double y, size_t &i, size_t &j) const;

    /// Get the index of the bin which holds \c x
    size_t get_x_bin_index(double x) const;
    
    /// Get the indey of the bin which holds \c y
    size_t get_y_bin_index(double y) const;

    /// Get the lower edge of bin of index \c i
    double &get_x_low_i(size_t i);

    /// Get the lower edge of bin of index \c i
    const double &get_x_low_i(size_t i) const;

    /// Get the upper edge of bin of index \c i
    double &get_x_high_i(size_t i);
    
    /// Get the upper edge of bin of index \c i
    const double &get_x_high_i(size_t i) const;

    /// Get the lower edge of bin of index \c j
    double &get_y_low_i(size_t j);

    /// Get the lower edge of bin of index \c j
    const double &get_y_low_i(size_t j) const;

    /// Get the upper edge of bin of index \c j
    double &get_y_high_i(size_t j);
    
    /// Get the upper edge of bin of index \c j
    const double &get_y_high_i(size_t j) const;
    //@}

    /// \name Rep modes (default is \c rmode_avg)
    //@{
    static const size_t rmode_avg=0;
    static const size_t rmode_user=1;
    static const size_t rmode_low=2;
    static const size_t rmode_high=3;
    static const size_t rmode_gmean=4;
    //@}
    
    /// \name Rep functions
    //@{
    /// Set the representative x-values for each bin
    template<class vec_t> int set_reps(size_t nx, vec_t &vx,
				       size_t ny, vec_t &vy) {
      xrmode=rmode_user;
      yrmode=rmode_user;
      if (user_xrep.size()!=nx || user_yrep.size()!=ny) {
	std::string s="Expected vectors of size "+itos(user_xrep.size())+
	  " and got a vector of size "+itos(nx)+" in hist::set_reps().";
	O2SCL_ERR_RET(s.c_str(),gsl_einval);
      }
      for(size_t i=0;i<nx;i++) user_xrep[i]=vx[i];
      for(size_t i=0;i<ny;i++) user_yrep[i]=vy[i];
      return gsl_success;
    }

    /// Set the representative x-values for each bin
    template<class vec_t> int set_x_reps(size_t nx, vec_t &vx) {
      xrmode=rmode_user;
      if (user_xrep.size()!=nx) {
	std::string s="Expected vectors of size "+itos(user_xrep.size())+
	  " and got a vector of size "+itos(nx)+" in hist::set_reps().";
	O2SCL_ERR_RET(s.c_str(),gsl_einval);
      }
      for(size_t i=0;i<nx;i++) user_xrep[i]=vx[i];
      return gsl_success;
    }

    /// Set the representative y-values for each bin
    template<class vec_t> int set_y_reps(size_t ny, vec_t &vy) {
      yrmode=rmode_user;
      if (user_yrep.size()!=ny) {
	std::string s="Eypected vectors of size "+itos(user_yrep.size())+
	  " and got a vector of size "+itos(ny)+" in hist::set_reps().";
	O2SCL_ERR_RET(s.c_str(),gsl_einval);
      }
      for(size_t i=0;i<ny;i++) user_yrep[i]=vy[i];
      return gsl_success;
    }

    /// Set mode used to compute bin reps
    int set_rep_mode(size_t x_mode, size_t y_mode);

    /// Get mode used to compute bin reps
    size_t get_x_rep_mode() const {
      return xrmode;
    }

    /// Get mode used to compute bin reps
    size_t get_y_rep_mode() const {
      return yrmode;
    }

    /// Get a reference to the full vector of bin specifications
    const ovector &get_x_bins() const {
      return xa;
    }

    /// Get a reference to the full vector of bin specifications
    const ovector &get_y_bins() const {
      return ya;
    }

    /// Return the histogram size of the x coordinate
    size_t size_x() const {
      return hsize_x;
    }

    /// Return the histogram size of the y coordinate
    size_t size_y() const {
      return hsize_y;
    }

    /// Get a reference to the user-specified reps for x coordinates
    const ovector &get_user_reps_x() const {
      return user_xrep;
    }

    /// Get a reference to the user-specified reps for y coordinates
    const ovector &get_user_reps_y() const {
      return user_yrep;
    }

    /// Return the rep of bin of index \c i
    double &get_x_rep_i(size_t i);

    /// Return the rep of bin of index \c i
    const double &get_x_rep_i(size_t i) const;

    /// Return the rep of bin of index \c j
    double &get_y_rep_i(size_t j);

    /// Return the rep of bin of index \c j
    const double &get_y_rep_i(size_t j) const;
    //@}

    /// Set up a twod_intp object for interpolation
    int setup_interp(twod_intp &ti, bool x_first=true) {
      ti.set_data(hsize_x,hsize_y,xrep,yrep,wgt,x_first);
      return gsl_success;
    }

  };

#ifndef DOXYGENP
}
#endif

#endif
