/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_EFF_FERMION_H
#define O2SCL_EFF_FERMION_H

#include <string>
#include <iostream>
#include <fstream>
#include <cmath>
#include <o2scl/constants.h>
#include <o2scl/umatrix_tlate.h>
#include <o2scl/funct.h>
#include <o2scl/mm_funct.h>
#include <o2scl/root.h>
#include <o2scl/mroot.h>
#include <o2scl/inte.h>
#include <o2scl/cern_mroot_root.h>
#include <o2scl/misc.h>

#include <o2scl/fermion.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief Fermion class from fitting method
      
      Based on the fitting method of \ref Johns96 which is an update
      of the method from \ref Eggleton73 . This method is approximate,
      but very fast. For a more accurate (but slower) method, use
      rel_fermion.
      
      Given the chemical potential and the temperature the functions
      calc_mu() and pair_mu() work by solving the equation (c.f. Eq. 15
      in \ref Johns96)
      \f[
      \psi= 2 \sqrt{1+f/a}+\log\left(\frac{\sqrt{1+f/a}-1}
      {\sqrt{1+f/a}+1}\right)
      \f]
      for \f$ f \f$ given \f$ \psi=(\mu-m)/T \f$.
      If \f$ f/a<10^{-10} \f$, then the alternative expression
      \f[
      \psi= 2 \left(1+f/(2 a)\right)+\log\left[\frac{f/(2 a)}
      {\left(1+f/(2 a)\right)}\right]
      \f]
      is used. The pressure, energy density, and entropy, are
      determined as polynomials in \f$ f \f$ with a
      set of precomputed coefficients as done in \ref Johns96 .

      If \f$ \psi \f$ is too small (less than about -200), the 
      above procedure fails. To handle this, this class
      uses the \classical result if \f$ \psi \f$ < \ref min_psi,
      where \ref min_psi defaults to -200.

      When the density and temperature is given instead
      (calc_density() and pair_density()), then there are two ways to
      proceed.
      - Use the density to solve for \f$ f \f$ .
      - Use the density to solve for the chemical potential.

      Because the density is a complicated polynomial in \f$ f \f$,
      the former procedure does not work very well even though it
      might be less time consuming. In this class, the density is
      solved for the effective chemical potential instead. The initial
      guess is just taken from the present value of part::nu .

      \note It is important to note that the coefficients are
      static and apply to all objects of type \ref eff_fermion.

      \note The coefficients are stored in static data, and there is
      a small possibility of a memory leak in a multithreaded program
      if more than one instance of \ref eff_fermion tries to initialize
      these coefficients in the constructor. This can easily be 
      avoided by ensuring that the static coefficients are initialized
      by a single thread beforehand. 

      \todo There's still def_err_hnd.set_mode(0) in the testing code,
      probably because the solver has a hard time for extreme values.

      \future Use bracketing to speed up one-dimensional root finding.
  */
  class eff_fermion : public fermion_eval_thermo {

  public:
    
    /// Create a fermion with mass \c mass and degeneracy \c dof 
    eff_fermion();
    
    virtual ~eff_fermion();
    
    /** \name Coefficients for finite-temperature approximation
     */
    //@{
    /** \brief Load coefficients
	
	The argument \c ctype Should be one of the constants below.
    */
    void load_coefficients(int ctype);

    /// A set of coefficients from Jim Lattimer 
    static const int cf_fermilat3=1;
    /// The smaller set of coefficients from \ref Johns96
    static const int cf_fermijel2=2;
    /// The larger set of coefficients from \ref Johns96
    static const int cf_fermijel3=3;
    /** \brief The set of coefficients from \ref Johns96 which retains 
	better thermodynamic consistency
    */
    static const int cf_fermijel3cons=4;
    //@}

    /** \brief Calculate thermodynamic 
	properties as function of chemical potential
	
	If the quantity \f$ (\mu-m)/T \f$ (or \f$ (\nu-m^{*})/T \f$ in
	the case of interacting particles) is less than -200, then
	this quietly sets the density, the scalar density, the energy
	density, the pressure and the entropy to zero and exits.

	\todo Should see if the function actually works if 
	\f$ (\mu-m)/T = -199 \f$ .
    */
    virtual void calc_mu(fermion &f, double temper);

    /** \brief Calculate thermodynamic properties as function of
	density

	\warning This function needs a guess for the chemical
	potential, and will fail if that guess is not sufficiently
	accurate.
    */
    virtual void calc_density(fermion &f, double temper);

    /** \brief Calculate thermodynamic properties with antiparticles
	as function of chemical potential

	\warning This function needs a guess for the chemical
	potential, and will fail if that guess is not sufficiently
	accurate.
    */
    virtual void pair_mu(fermion &f, double temper);

    /** \brief Calculate thermodynamic properties with antiparticles
	as function of density
    */
    virtual void pair_density(fermion &f, double temper);

    /** \brief Set the solver for use in calculating \f$ \psi \f$ 
     */
    int set_psi_root(root<funct> &rp) {
      psi_root=&rp;
      return 0;
    }

    /** \brief Set the solver for use in calculating the chemical
	potential from the density
    */
    int set_density_root(root<funct> &rp) {
      density_root=&rp;
      return 0;
    }
    
    /** \brief If the temperature is less than \c tlimit then the
	zero-temperature functions are used (default 0).
    */
    double tlimit;
    
    /** \brief The default solver for \f$ \psi \f$
     */
    cern_mroot_root<funct> def_psi_root;
    
    /** \brief The default solver for calc_density() and pair_density()
     */
    cern_mroot_root<funct> def_density_root;
    
    /// Return string denoting type ("eff_fermion")
    virtual const char *type() { return "eff_fermion"; }

    /// The minimum value of \f$ \psi \f$ (default -200)
    double min_psi;

#ifndef DOXYGEN_INTERNAL

  protected:

    /// The matrix of coefficients
    umatrix Pmnf;
    /// The parameter \f$ a \f$
    double parma;
    /// The array row size
    int sizem;
    /// The array column size
    int sizen;
    
    /// The solver for \f$ \psi \f$
    root<funct> *psi_root;
    /// The other solver for calc_density()
    root<funct> *density_root;
    
    /// The function which solves for \f$ f \f$ from \f$ \psi \f$.
    double solve_fun(double x, double &psi);
    
    /** \brief Define the function which solves for the chemical 
	potential given the density [protected subclass of \ref 
	eff_fermion]
    */
    class density_fun : public funct {

    protected:

      eff_fermion &ef_;
      fermion &f_;
      double T_;

    public:

      density_fun(eff_fermion &ef, fermion &f, double T);
      
      /// Fix density for \ref eff_fermion::calc_density()
      double operator()(double x);

    };

    /** \brief Define the function which solves for the chemical 
	potential given the density of particles and antiparticles
	[protected subclass of \ref eff_fermion]
    */
    class pair_density_fun : public funct {

    protected:

      eff_fermion &ef_;
      fermion &f_;
      double T_;

    public:

      pair_density_fun(eff_fermion &ef, fermion &f, double T);
      
      /// Fix density for \ref eff_fermion::pair_density()
      double operator()(double x);

    };
    
#endif

  };

#ifndef DOXYGENP
}
#endif

#endif
