/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_REL_FERMION_H
#define O2SCL_REL_FERMION_H

#include <string>
#include <iostream>
#include <fstream>
#include <cmath>
#include <o2scl/constants.h>
#include <o2scl/mroot.h>
#include <o2scl/inte.h>
#include <o2scl/cern_mroot_root.h>
#include <o2scl/gsl_inte_qagiu.h>
#include <o2scl/gsl_inte_qag.h>
#include <o2scl/eff_fermion.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief Equation of state for a relativistic fermion

      This class computes the thermodynamics of a relativistic fermion
      either as a function of the density or the chemical potential.
      It employs direct integration, using two different integrators
      for the degenerate and non-degenerate regimes. The default
      integrators are gsl_inte_qag (for degenerate fermions) and
      gsl_inte_qagiu (for non-degenerate fermions). For the functions
      calc_mu() and calc_density(), if the temperature argument is
      less than or equal to zero, the functions \ref
      fermion_zerot::calc_mu_zerot() and \ref
      fermion_zerot::calc_density_zerot() will be used to compute the
      result.

      \hline 
      <b>Degeneracy parameter:</b>

      Define the degeneracy parameter 
      \f[
      \psi=(\nu-m^{*})/T 
      \f] 
      where \f$ \nu \f$ is the effective chemical potential and \f$
      m^{*} \f$ is the effective mass. For \f$ \psi \f$ greater than
      \ref deg_limit (degenerate regime), a finite interval integrator
      is used and for \f$ \psi \f$ less than \ref deg_limit
      (non-degenerate regime), an integrator over the interval from
      \f$ [0,\infty) \f$ is used. In the case where \ref
      part::inc_rest_mass is false, the degeneracy parameter is
      \f[
      \psi=(\nu+m-m^{*})/T 
      \f] 

      <b>Integration limits:</b>

      The upper limit on the degenerate integration is given by
      \f[
      \mathrm{upper~limit} = \sqrt{{\cal L}^2-m^{*,2}}
      \f]
      where \f$ {\cal L}\equiv u T+\nu \f$ and \f$ u \f$ is \ref
      rel_fermion::upper_limit_fac . In the case where \ref
      part::inc_rest_mass is false, the result is
      \f[
      \mathrm{upper~limit} = \sqrt{(m+{\cal L})^2-m^{*2}}
      \f]
      
      The entropy is only significant at the Fermi surface, thus
      in the degenerate case, the lower limit of the entropy
      integral can be given be determined by the value of \f$ k \f$ 
      which solves
      \f[
      - u = \frac{\sqrt{k^2+m^{* 2}}-\nu}{T}
      \f]
      The solution is 
      \f[
      \mathrm{lower~limit} = \sqrt{(-u T+{\nu})^2-m^{*,2}}
      \f]
      but this solution is only valid if \f$ (m^{*}-\nu)/T < -u \f$.
      In the case where part::inc_rest_mass is false, the result is
      \f[
      \mathrm{lower~limit} = \sqrt{(-u T + m +\nu)^2-m^{*,2}}
      \f]
      which is valid if \f$ (m^{*}-\nu - m)/T < -u \f$.

      <b>Entropy integrand:</b>

      In the degenerate regime, the entropy integrand
      \f[
      - k^2 \left[ f \log f + \left(1-f\right) \log 
      \left(1-f \right) \right]
      \f]
      where \f$ f \f$ is the fermionic distribution function can lose
      precision when \f$ (E^{*} - \nu)/T \f$ is negative and
      sufficiently large in absolute magnitude. Thus when \f$ (E^{*} -
      \nu)/T < S \f$ where \f$ S \f$ is stored in \ref deg_entropy_fac
      (default is -30), the integrand is written as
      \f[
      -k^2 \left( E/T-\nu/T \right) e^{E/T-\nu/T} \, .
      \f]
      If \f$ (E - \nu)/T < S \f$ is less than -1 times \ref exp_limit
      (e.g. less than -200), then the entropy integrand is assumed 
      to be zero.
      
      <b>Non-degenerate integrands:</b>
      
      \comment
      It's not at all clear that this dimensionless form is more
      accurate than other potential alternatives. On the other hand,
      it seems that the uncertainties in the integrations are larger
      than the errors made by the integrand at present.
      \endcomment
      The integrands in the non-degenerate regime are written
      in a dimensionless form, by defining 
      \f$ p = \sqrt{\left(T u + m^{*}\right)^2-m^{* 2}} \f$,
      \f$ \nu \equiv y T \f$, and 
      \f$ m^{*} \equiv \mathrm{mx}~T \f$. 
      The density integrand is 
      \f[
      \left(\mathrm{mx}+u\right) \sqrt{u^2+2 (\mathrm{mx}) u}
      \left(\frac{e^{y}}{e^{\mathrm{mx}+u}+e^{y}}\right) \, , 
      \f]
      the energy integrand is 
      \f[
      \left(\mathrm{mx}+u\right)^2 \sqrt{u^2+2 (\mathrm{mx}) u}
      \left(\frac{e^{y}}{e^{\mathrm{mx}+u}+e^{y}}\right) \, ,
      \f]
      and the entropy integrand is 
      \f[
      \left(\mathrm{mx}+u\right) \sqrt{u^2+2 (\mathrm{mx}) u} 
      \left(t_1+t_2\right) \, ,
      \f]
      where 
      \f{eqnarray*}
      t_1 &=& \log \left(1+e^{y-\mathrm{mx}-u}\right)/
      \left(1+e^{y-\mathrm{mx}-u}\right) \nonumber \\
      t_2 &=& \log \left(1+e^{\mathrm{mx}+u-y}\right)/
      \left(1+e^{\mathrm{mx}+u-y}\right) \, .
      \f}

      \hline 
      <b>Accuracy:</b>

      The default settings for for this class give an accuracy of at
      least 1 part in \f$ 10^6 \f$ (and frequently better than this).

      When the integrators provide numerical uncertainties, these
      uncertainties are stored in \ref unc. In the case of
      calc_density() and pair_density(), the uncertainty from the
      numerical accuracy of the solver is not included. (There is also
      a relatively small inaccuracy due to the mathematical evaluation
      of the integrands which is not included in \ref unc.)
     
      One way to improve the accuracy of the computation is just to 
      decrease the tolerances on the default integration objects. This 
      can be done, using, for example
      \code
      rel_fermion rf(1.0,2.0);
      rf.def_dit.tolx/=1.0e2;
      rf.def_dit.tolf/=1.0e2;
      rf.def_nit.tolx/=1.0e2;
      rf.def_nit.tolf/=1.0e2;
      \endcode
      which decreases the both the relative and absolute tolerances
      for both the degenerate and non-degenerate integrators. If one
      is using either the calc_density() or pair_density() functions,
      one may also have to improve the accuracy of the solver which
      determines the chemical potential from the density. For
      the default solver, this could be done with
      \code
      rf.def_density_root.tolx/=1.0e2;
      rf.def_density_root.tolf/=1.0e2;
      \endcode
      Of course if these tolerances are too small, the calculation
      may fail.

      \hline 
      <b>Todos:</b>

      \future The expressions which appear in in the integrand
      functions density_fun(), etc. could likely be improved,
      especially in the case where inc_rest_mass=false. There should
      not be a need to check if <tt>ret</tt> is finite.

      \future It appears this doesn't compute the uncertainty in the
      chemical potential or density with calc_density(). This could
      be fixed.

      \future I'd like to change the lower limit on the entropy 
      integration, but the value in the code at the moment (stored
      in <tt>ll</tt>) makes bm_part2.cpp worse.

      \future pair_mu() should set the antiparticle integrators
      as done in sn_fermion.
  */
  class rel_fermion : public fermion_eval_thermo {

  public:

    /// \name Numerical parameters
    //@{
    /** \brief The critical degeneracy at which to switch integration 
	techniques (default 2)
    */
    double deg_limit;
    
    /** \brief The limit for exponentials to ensure integrals are finite 
	(default 200)
    */
    double exp_limit;

    /// The factor for the degenerate upper limits (default 20)
    double upper_limit_fac;

    /// A factor for the degenerate entropy integration (default 30)
    double deg_entropy_fac;
    //@}

    /// Storage for the uncertainty
    fermion unc;

    /// Create a fermion with mass \c m and degeneracy \c g
    rel_fermion();

    virtual ~rel_fermion();
    
    /** \brief Calculate properties as function of chemical potential
    */
    virtual void calc_mu(fermion &f, double temper);

    /** \brief Calculate properties as function of density

        This function uses the current value of \c nu (or \c mu if the
	particle is non interacting) for an initial guess to solve for
	the chemical potential. If this guess is too small, then this
	function may fail.
     */
    virtual void calc_density(fermion &f, double temper);

    /** \brief Calculate properties with antiparticles as function of
	chemical potential
    */
    virtual void pair_mu(fermion &f, double temper);

    /** \brief Calculate properties with antiparticles as function of
	density
     */
    virtual void pair_density(fermion &f, double temper);

    /// Calculate effective chemical potential from density
    virtual void nu_from_n(fermion &f, double temper);
    
    /// Set integrators
    int set_inte(inte<funct> &non_it, inte<funct> &deg_it);
    
    /** \brief Set the solver for use in calculating the chemical
	potential from the density */
    int set_density_root(root<funct> &rp) {
      density_root=&rp;
      return 0;
    }
    
    /// The default solver for calc_density().
    cern_mroot_root<funct> def_density_root;

    /// The default integrator for degenerate fermions
    gsl_inte_qag<funct> def_dit;

    /// The default integrator for non-degenerate fermions
    gsl_inte_qagiu<funct> def_nit;

    /// Return string denoting type ("rel_fermion")
    virtual const char *type() { return "rel_fermion"; }

  protected:
    
#ifndef DOXYGEN_INTERNAL
    
    /// The non-degenerate integrator
    inte<funct> *nit;
    /// The degenerate integrator
    inte<funct> *dit;
    /// The solver for calc_density()
    root<funct> *density_root;
    /// Temperature
    double T;
    /// Current fermion pointer
    fermion *fp;

    /// The integrand for the density for non-degenerate fermions
    double density_fun(double u);

    /// The integrand for the energy density for non-degenerate fermions
    double energy_fun(double u);

    /// The integrand for the entropy density for non-degenerate fermions
    double entropy_fun(double u);

    /// The integrand for the density for degenerate fermions
    double deg_density_fun(double u);

    /// The integrand for the energy density for degenerate fermions
    double deg_energy_fun(double u);

    /// The integrand for the entropy density for degenerate fermions
    double deg_entropy_fun(double u);

    /// Solve for the chemical potential given the density
    double solve_fun(double x);

    /// Solve for the chemical potential given the density with antiparticles
    double pair_fun(double x);
    
#endif

  };

#ifndef DOXYGENP
}
#endif

#endif
