/*
  -------------------------------------------------------------------
  
  Copyright (C) 2011-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/

#ifndef GSL_MROOT_BROYDEN_H
#define GSL_MROOT_BROYDEN_H

#include <o2scl/mroot.h>
#include <o2scl/permutation.h>
#include <o2scl/lu.h>

#ifndef DOXYGENP
namespace o2scl {
#endif
  
  /** \brief Multidimensional root-finding using Broyden's method (GSL)

      Experimental.

      See \ref Broyden65.
  */
  template<class func_t=mm_funct<>, class vec_t=ovector_base, 
    class alloc_vec_t=ovector, class alloc_t=ovector_alloc,
    class mat_t=omatrix_base, class alloc_mat_t=omatrix,
    class mat_alloc_t=omatrix_alloc, class jfunc_t=jac_funct<vec_t,mat_t> > 
    class gsl_mroot_broyden : public mroot<func_t,vec_t,jfunc_t> {

    protected:

    /// Desc
    omatrix H;
  
    /// LU decomposition
    omatrix lu;

    /// Permutation object for the LU decomposition
    permutation perm;

    /// Desc
    ovector v;

    /// Desc
    ovector w;

    /// Desc
    ovector y;

    /// Desc
    ovector p;

    /// Desc
    ovector fnew;

    /// Desc
    ovector x_trial;

    /// Desc
    double phi;

    /// Memory allocation object
    alloc_t ao;

    /// Stepsize vector
    alloc_vec_t dx_int;

    /// Function value vector
    alloc_vec_t f_int;

    /// A pointer to the user-specified function
    func_t *user_func;

    /// Function values
    vec_t *user_f;

    /// Initial guess and current solution
    vec_t *user_x;

    /// Initial and current step
    vec_t *user_dx;

    /// Number of variables
    size_t user_nvar;

    /// Size of memory allocated
    size_t mem_size;

    /// Jacobian
    jacobian<func_t,vec_t,mat_t> *ajac;

    /** \brief Clear allocated vectors and matrices

        This function is called by set() before each solve.
     */
    void clear() {
      if (mem_size>0) {
	lu.set_all(0.0);
	for(size_t i=0;i<mem_size;i++) {
	  perm[i]=0;
	}
	H.set_all(0.0);
	v.set_all(0.0);
	w.set_all(0.0);
	y.set_all(0.0);
	fnew.set_all(0.0);
	x_trial.set_all(0.0);
	p.set_all(0.0);
      }
      return;
    }

    public:

    gsl_mroot_broyden() {
      mem_size=0;
      ajac=&def_jac;
      def_jac.epsrel=GSL_SQRT_DBL_EPSILON;
    }
    
    /// Default Jacobian object
    simple_jacobian<func_t,vec_t,mat_t,alloc_vec_t,alloc_t> def_jac;
    
    virtual ~gsl_mroot_broyden() {
      free();
    }

    /// Allocate memory 
    void allocate(size_t n) {

      if (n!=mem_size) {

	free();
      
	lu.allocate(n,n);
	perm.allocate(n);
	H.allocate(n,n);
	v.allocate(n);
	w.allocate(n);
	y.allocate(n);
	fnew.allocate(n);
	x_trial.allocate(n);
	p.allocate(n);
	ao.allocate(dx_int,n);
	ao.allocate(f_int,n);

	mem_size=n;

      }

      return;
    }

    /// Euclidean norm
    double enorm(size_t nvar, const vec_t &ff) {
      double e2=0;
      size_t i;
      for (i=0;i<nvar;i++) {
	double fi=ff[i];
	e2+=fi*fi;
      }
      return sqrt(e2);
    }

    /** \brief Set the function, initial guess, and provide vectors to
	store function values and stepsize

	The initial values of \c f and \c dx are ignored.
    */
    void set(func_t &func, size_t nvar, vec_t &x, vec_t &f, vec_t &dx) {

      if (nvar!=mem_size) allocate(nvar);
      clear();

      user_func=&func;
      user_nvar=nvar;
      user_x=&x;
      user_f=&f;
      user_dx=&dx;

      size_t i, j;
      int signum=0;
      
      func(nvar,x,f);
      
      ajac->set_function(func);
      (*ajac)(nvar,x,f,lu);
      
      o2scl_linalg::LU_decomp<>(nvar,lu,perm,signum);
      o2scl_linalg::LU_invert<omatrix,omatrix,omatrix_col>
      (nvar,lu,perm,H);
      
      for (i=0;i<nvar;i++) {
	for (j=0;j<nvar;j++) {
	  H[i][j]=-H[i][j];
	}
      }
      
      dx.set_all(0.0);
      
      phi=enorm(nvar,f);
      
      return;
    }

    /// Perform an iteration
    int iterate() {

      func_t &func=*user_func;
      size_t nvar=user_nvar;
      vec_t &x=*user_x;
      vec_t &f=*user_f;
      vec_t &dx=*user_dx;
    
      double phi0, phi1, t, lambda;
      
      size_t i, j, iter;
      
      /* p=H f */
      
      for (i=0;i<nvar;i++) {
	double sum=0;
	for (j=0;j<nvar;j++) {
	  sum+=H[i][j]*f[j];
	}
	p[i]=sum;
      }
      
      t=1.0;
      iter=0;
      
      phi0=phi;

      bool new_step=true;
      while (new_step) {
	new_step=false;
      
	for (i=0;i<nvar;i++) {
	  x_trial[i]=x[i]+t*p[i];
	}
      
	{ 
	  int status=func(nvar,x_trial,fnew);
	  if (status != gsl_success) {
	    return gsl_ebadfunc;
	  }
	}
      
	phi1=enorm(nvar,fnew);
      
	iter++;
      
	if (phi1>phi0 && iter<10 && t>0.1) {

	  // [GSL] Full step goes uphill, take a reduced step instead.
	  double theta=phi1/phi0;
	  t*=(sqrt(1.0+6.0*theta)-1.0)/(3.0*theta);
	  new_step=true;
	}
      
	if (new_step==false && phi1>phi0) {

	  // [GSL] Need to recompute Jacobian
	  int signum=0;
	  
	  (*ajac)(nvar,x,f,lu);
	  
	  for (i=0;i<nvar;i++) {
	    for (j=0;j<nvar;j++) {
	      lu[i][j]=-lu[i][j];
	    }
	  }
	  
	  o2scl_linalg::LU_decomp(nvar,lu,perm,signum);
	  o2scl_linalg::LU_invert<omatrix,omatrix,omatrix_col>
	    (nvar,lu,perm,H);
	  o2scl_linalg::LU_solve(nvar,lu,perm,f,p);
	  
	  t=1.0;
	  
	  for (i=0;i<nvar;i++) {
	    x_trial[i]=x[i]+t*p[i];
	  }
	  
	  {
	    int status=func(nvar,x_trial,fnew);
	    if (status != gsl_success) {
	      return gsl_ebadfunc;
	    }
	  }
	  
	  phi1=enorm(nvar,fnew);
	}

      }
      
      /* y=f'-f */
      for (i=0;i<nvar;i++) {
	y[i]=fnew[i]-f[i];
      }
      
      /* v=H y */
      for (i=0;i<nvar;i++) {
	double sum=0.0;
	for (j=0;j<nvar;j++) {
	  sum+=H[i][j]*y[j];
	}
	v[i]=sum;
      }
      
      /* lambda=p dot v */
      lambda=0.0;
      for (i=0;i<nvar;i++) {
	lambda+=p[i]*v[i];
      }
      if (lambda==0) {
	O2SCL_ERR2("Approximation to Jacobian has collapsed in ",
		   "gsl_mroot_broyden::iterate().",gsl_ezerodiv);
      }
      
      /* v'=v+t*p */
      for (i=0;i<nvar;i++) {
	v[i]=v[i]+t*p[i];
      }
      
      /* w^T=p^T H */
      for (i=0;i<nvar;i++) {
	double sum=0;
	for (j=0;j<nvar;j++) {
	  sum+=H[j][i]*p[j];
	}
	w[i]=sum;
      }
      
      /* Hij -> Hij-(vi wj/lambda) */
      for (i=0;i<nvar;i++) {
	double vi=v[i];
	for (j=0;j<nvar;j++) {
	  double wj=w[j];
	  H[i][j]-=vi*wj/lambda;
	}
      }
      
      /* copy fnew into f */
      vector_copy(nvar,fnew,f);
      
      /* copy x_trial into x */
      vector_copy(nvar,x_trial,x);
      for (i=0;i<nvar;i++) {
	dx[i]=t*p[i];
      }
      
      phi=phi1;
      
      return gsl_success;
    }

    /// Desc
    virtual int msolve(size_t n, vec_t &x, func_t &func) {

      int status;

      set(func,n,x,f_int,dx_int);
      
      int iter=0;

      do {
	iter++;
	
	status=iterate();
	if (status) break;

	// ------------------------------------------------------
	// The equivalent of the statement:
	// 
	// status=gsl_multiroot_test_residual(f,this->tol_rel);

	double resid=0.0;
	for(size_t i=0;i<n;i++) {
	  resid+=fabs(f_int[i]);
	}
	if (resid<this->tol_rel) status=gsl_success;
	else status=gsl_continue;
	
	// ------------------------------------------------------
	
	if (this->verbose>0) {
	  this->print_iter(n,x,f_int,iter,resid,this->tol_rel,
			   "gsl_mroot_broyden");
	}
	
      } while (status==gsl_continue && iter<this->ntrial);

      this->last_ntrial=iter;
    
      if (iter>=this->ntrial) {
	if (this->last_conv==0) this->last_conv=gsl_emaxiter;
	O2SCL_CONV2_RET("Function gsl_mroot_broyden::msolve() ",
			"exceeded max. number of iterations.",
			gsl_emaxiter,this->err_nonconv);
      }
      
      if (status!=0) {
	O2SCL_ERR2_RET("Function iterate() failed in ",
		       "gsl_mroot_broyden::solve_set().",gsl_efailed);
      }

      return gsl_success;
    }

    /// Desc
    void free() {
      if (mem_size>0) {
	H.free();
	lu.free();
	perm.free();
	v.free();
	w.free();
	y.free();
	p.free();
	fnew.free();
	x_trial.free();
	ao.free(dx_int);
	ao.free(f_int);
      }
      return;
    }

  };

#ifndef DOXYGENP
}
#endif

#endif
