/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006-2012, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
/* multiroots/hybrid.c
 * multiroots/dogleg.c
 * 
 * Copyright (C) 1996, 1997, 1998, 1999, 2000, 2007 Brian Gough
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 
 * 02110-1301, USA.
 */
#ifndef O2SCL_GSL_MROOT_HYBRIDS_H
#define O2SCL_GSL_MROOT_HYBRIDS_H

#include <string>

#include <gsl/gsl_multiroots.h>
#include <gsl/gsl_linalg.h>

#include <o2scl/ovector_tlate.h>
#include <o2scl/mroot.h>
#include <o2scl/jacobian.h>
#include <o2scl/qr.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief Base functions for \ref gsl_mroot_hybrids
   */
  class hybrids_base {
  public:

    /// Compute the actual reduction
    double compute_actual_reduction(double fnorm0, double fnorm1) {
      double actred;
      if (fnorm1<fnorm0) {
	double u=fnorm1/fnorm0;
	actred=1-u*u;
      } else {
	actred=-1;
      }
      return actred;
    }

    /// Compute the predicted reduction phi1p=|Q^T f + R dx| 
    double compute_predicted_reduction(double fnorm0, double fnorm1) {
      double prered;
      if (fnorm1<fnorm0) {
	double u=fnorm1/fnorm0;
	prered=1-u*u;
      } else {
	prered=0;
      }
      return prered;
    }

    /// Compute \f$ R \cdot g \f$ where \c g is the \gradient
    template<class vec2_t>
      void compute_Rg(size_t N, const omatrix_base &r2, 
		      const ovector_base &gradient2, vec2_t &Rg) {
    
      for (size_t i=0;i<N;i++) {
	double sum=0.0;
	for (size_t j=i;j<N;j++) {
	  sum+=r2[i][j]*gradient2[j];
	}
	Rg[i]=sum;
      }

      return;
    }

    /// Compute \c w and \c v
    template<class vec2_t>
      void compute_wv(size_t n, const ovector_base &qtdf2, 
		      const ovector_base &rdx2, const vec2_t &dx2, 
		      const ovector_base &diag2, double pnorm, 
		      ovector_base &w2, ovector_base &v2) {
    
      for (size_t i=0;i<n;i++) {
	double diagi=diag2[i];
	w2[i]=(qtdf2[i]-rdx2[i])/pnorm;
	v2[i]=diagi*diagi*dx2[i]/pnorm;
      }
    
      return;
    }

    /// Compute \f$ R \cdot \mathrm{dx} \f$
    template<class vec2_t>
      void compute_rdx(size_t N, const omatrix_base &r2, 
		       const vec2_t &dx2, ovector_base &rdx2) {
    
      for (size_t i=0;i<N;i++) {
	double sum=0.0;
	for (size_t j=i;j<N;j++) {
	  sum+=r2[i][j]*dx2[j];
	}
	rdx2[i]=sum;
      }

      return;
    }

    /** \brief Compute the norm of the vector \f$ \vec{v} \f$ defined
	by \f$ v_i = d_i ff_i \f$
    */
    template<class vec2_t, class vec3_t>
      double scaled_enorm(size_t n, const vec2_t &d, const vec3_t &ff) {
      double e2=0.0;
      for (size_t i=0;i<n;i++) {
	double u=d[i]*ff[i];
	e2+=u*u;
      }
      return sqrt(e2);
    }
  
    /// Compute delta
    template<class vec2_t>
      double compute_delta(size_t n, ovector_base &diag2, vec2_t &x2) {
      double Dx=scaled_enorm(n,diag2,x2);
      double factor=100.0;
      return (Dx > 0) ? factor*Dx : factor;
    }

    /** \brief Compute the Euclidean norm of \c ff
      
	\future Replace this with \c dnrm2 from \ref cblas_base.h
    */
    template<class vec2_t> double enorm(size_t N, const vec2_t &ff) {
      double e2=0.0;
      for (size_t i=0;i<N;i++) {
	double fi=ff[i];
	e2+=fi*fi;
      }
      return sqrt(e2);
    }

    /// Compute the Euclidean norm of the sum of \c a and \c b
    double enorm_sum(size_t n, const ovector_base &a, 
			  const ovector_base &b) {
      double e2=0.0;
      for (size_t i=0;i<n;i++) {
	double u=a[i]+b[i];
	e2+=u*u;
      }
      return sqrt(e2);
    }

    /** \brief Compute a trial step and put the result in \c xx_trial
	
	\future Replace with daxpy.
    */
    template<class vec2_t>
      int compute_trial_step(size_t N, vec2_t &xl, vec2_t &dxl, 
			     vec2_t &xx_trial) {
      for (size_t i=0;i<N;i++) {
	xx_trial[i]=xl[i]+dxl[i];
      }
      return 0;
    }

    /** \brief Compute the change in the function value
	
	\future Replace with daxpy?
    */
    template<class vec2_t>
      int compute_df(size_t n, const vec2_t &ff_trial, 
			  const vec2_t &fl, ovector_base &dfl) {
      for (size_t i=0;i<n;i++) {
	dfl[i]=ff_trial[i]-fl[i];
      }
    
      return 0;
    }
   
    /// Compute \c diag, the norm of the columns of the Jacobian
    template<class mat2_t>
      void compute_diag(size_t n, const mat2_t &J2, ovector_base &diag2) {
      for (size_t j=0;j<n;j++) {
	double sum=0;
	for (size_t i=0;i<n;i++) {
	  const double Jij=J2[i][j];
	  sum+=Jij*Jij;
	}
	if (sum == 0) {
	  sum=1.0;
	}
	diag2[j]=sqrt(sum);
      }

      return;
    }
  
    /** \brief Compute \f$ Q^{T} f \f$
      
	\future This is just right-multiplication, so we could
	use the \o2 cblas routines instead.
    */
    template<class vec2_t, class vec3_t, class vec4_t>
      void compute_qtf(size_t N, const vec2_t &q2, 
		       const vec3_t &f2, vec4_t &qtf2) {
      for (size_t j=0;j<N;j++) {
	double sum=0.0;
	for (size_t i=0;i<N;i++) {
	  sum+=q2[i][j]*f2[i];
	}
	qtf2[j]=sum;
      }

      return;
    }
  
    /// Update \c diag 
    template<class mat2_t>
      void update_diag(size_t n, const mat2_t &J2, ovector_base &diag2) {
      for (size_t j=0;j<n;j++) {
	double cnorm, diagj, sum=0.0;
	for (size_t i=0;i<n;i++) {
	  double Jij=J2[i][j];
	  sum+=Jij*Jij;
	}
	if (sum == 0) {
	  sum=1.0;
	}
	cnorm=sqrt(sum);
	diagj=diag2[j];
	if (cnorm > diagj) {
	  diag2[j]=cnorm;
	}
      }

      return;
    }
    
    /** \brief Form appropriate convex combination of the Gauss-Newton
	direction and the scaled gradient direction.

	Using the Gauss-Newton direction given in \c newton (a vector of
	size N), and the gradient direction in \c gradient (also a
	vector of size N), this computes
	\f[
	\mathrm{pp}=\alpha \mathrm{newton}+\beta \mathrm{gradient}
	\f]
    */
    template<class vec2_t>
      void scaled_addition(size_t N, double alpha, ovector_base &newton2, 
			   double beta, ovector_base &gradient2, vec2_t &pp) {
      for (size_t i=0;i<N;i++) {
	pp[i]=alpha*newton2[i]+beta*gradient2[i];
      }

      return;
    }

    /// Compute the Gauss-Newton direction
    int newton_direction(const size_t N, const omatrix_base &r2, 
			 const ovector_base &qtf2, ovector_base &p) {
      size_t i;
      int status=0;

      vector_copy(N,qtf2,p);
      o2scl_cblas::dtrsv(o2scl_cblas::o2cblas_RowMajor,
			 o2scl_cblas::o2cblas_Upper,
			 o2scl_cblas::o2cblas_NoTrans,
			 o2scl_cblas::o2cblas_NonUnit,N,N,r2,p);

      for (i=0;i<N;i++) {
	p[i]*=-1.0;
      }

      return status;
    }

    /// Compute the gradient direction
    void gradient_direction(const size_t M, const size_t N,
			    const omatrix_base &r2, const ovector_base &qtf2,
			    const ovector_base &diag2, ovector_base &g) {
      for (size_t j=0;j<M;j++) {
	double sum=0.0;
	for (size_t i=0;i<N;i++) {
	  sum+=r2[i][j]*qtf2[i];
	}
	g[j]=-sum/diag2[j];
      }

      return;
    }

    /// Compute the point at which the gradient is minimized
    void minimum_step(const size_t N, double gnorm, 
			   const ovector_base &diag2, ovector_base &g) {
      for (size_t i=0;i<N;i++) {
	g[i]=(g[i]/gnorm)/diag2[i];
      }

      return;
    }
  
    /** \brief Take a dogleg step
      
	Given the QR decomposition of an \c n by \c n matrix "A",
	a diagonal matrix \c diag, a vector \c b, and a positive
	number \c delta, this function determines the convex 
	combination \c x of the Gauss-Newton and scaled gradient
	directions which minimizes \f$ A x-b \f$ in the least
	squares sense, subject to the restriction that the 
	Euclidean norm of \f$ d x \f$ is smaller than the 
	value of \c delta.
    */
    template<class vec2_t>
      int dogleg(size_t n, const omatrix_base &r2, const ovector_base &qtf2,
		 const ovector_base &diag2, double delta2,
		 ovector_base &newton2, ovector_base &gradient2, 
		 vec2_t &p) {

      double qnorm, gnorm, sgnorm, bnorm, temp;
      
      // Compute the Gauss-Newton direction
      newton_direction(n,r2,qtf2,newton2);
      qnorm=scaled_enorm(n,diag2,newton2);

      // Test whether the gauss-newton direction is acceptable
      if (qnorm <= delta2) {
	for(size_t i=0;i<n;i++) p[i]=newton2[i];
	return gsl_success;
      }
      
      // Compute the gradient direction
      gradient_direction(n,n,r2,qtf2,diag2,gradient2);
      gnorm=enorm(n,gradient2);
      
      // The special case that the scaled gradient is zero
      if (gnorm == 0) {
	double alpha=delta2/qnorm;
	double beta=0;
	scaled_addition(n,alpha,newton2,beta,gradient2,p);
	return gsl_success;
      }

      // -----------------------------------------------------
      // The next set of statements calculates the point along the
      // scaled gradient at which the quadratic is minimized.

      minimum_step(n,gnorm,diag2,gradient2);
      
      // Use p as temporary space to compute Rg
      compute_Rg(n,r2,gradient2,p); 
      
      temp=enorm(n,p);
      sgnorm=(gnorm/temp)/temp;

      // -----------------------------------------------------
      // Test whether the scaled gradient direction is acceptable

      if (sgnorm>delta2)  {

	double alpha=0;
	double beta=delta2;  

	scaled_addition(n,alpha,newton2,beta,gradient2,p);

	return gsl_success;
      }
      
      // If not, calculate the point along the dogleg step
      // at which the quadratic is minimized
      bnorm=enorm(n,qtf2);
      
      double bg=bnorm/gnorm;
      double bq=bnorm/qnorm;
      double dq=delta2/qnorm;
      double dq2=dq*dq;
      double sd=sgnorm/delta2;
      double sd2=sd*sd;
      
      double t1=bg*bq*sd;
      double u=t1-dq;
      double t2=t1-dq*sd2+sqrt(u*u+(1-dq2)*(1-sd2));
      
      double alpha=dq*(1-sd2)/t2;
      double beta=(1-alpha)*sgnorm;

      // Form the appropriate convex combination of the gauss-newton
      // direction and the scaled gradient direction.

      scaled_addition(n,alpha,newton2,beta,gradient2,p);
      
      return gsl_success;
    }
    
  };

  /** \brief Multidimensional root-finding algorithm using
      Powell's Hybrid method (GSL)
      
      This is a recasted version of the GSL routines which use a
      modified version of Powell's Hybrid method as implemented in the
      HYBRJ algorithm in MINPACK (\ref Garbow80).  Both the scaled and
      unscaled options are available by setting \ref int_scaling (the
      scaled version is the default). If derivatives are not provided,
      they will be computed automatically. This class provides the
      GSL-like interface using allocate(), set() (or set_de() in case
      where derivatives are available), iterate(), and free() and
      higher-level interfaces, msolve() and msolve_de(), which perform
      the solution and the memory allocation automatically. Some
      additional checking is performed in case the user calls the
      functions out of order (i.e. set() without allocate()).
      
      The functions msolve() and msolve_de() use the condition \f$
      \sum_i |f_i|<\f$ \ref mroot::tol_rel to determine if the solver has
      succeeded.
      
      The original GSL algorithm has been modified to shrink the
      stepsize if a proposed step causes the function to return a
      non-zero value. This allows the routine to automatically try to
      avoid regions where the function is not defined. To return to
      the default GSL behavior, set \ref shrink_step to false.

      The default method for numerically computing the Jacobian is
      from \ref simple_jacobian. This default is identical to the GSL
      approach, except that the default value of \ref
      simple_jacobian::epsmin is non-zero. See \ref simple_jacobian
      for more details.

      There is an example for the usage of the multidimensional solver
      classes given in <tt>examples/ex_mroot.cpp</tt>, see the \ref
      ex_mroot_sect .

      \todo Are all the set_all() statements really necessary? Do
      they need to be executed even if memory hasn't been recently
      allocated?

      \future It's kind of strange that set() sets jac_given to false
      and set_de() has to reset it to true. Can this be simplified?

      \future It might be good to make sure that we're directly
      testing to make sure that GSL and O2SCL are more or less
      identical.

      \future Many of these minpack functions could be put in their
      own "minpack_tools" class, or possibly moved to be
      linear algebra routines instead.

      \future There are still some numbers in here which the user
      could have control over, for example, the <tt>nslow2</tt>
      threshold which indicates failure.
  */
  template<class func_t=mm_funct<>, 
    class vec_t=ovector_base, class alloc_vec_t=ovector, 
    class alloc_t=ovector_alloc, class mat_t=omatrix_base, 
    class alloc_mat_t=omatrix, class mat_alloc_t=omatrix_alloc,
    class jfunc_t=jac_funct<vec_t,mat_t> > class gsl_mroot_hybrids : 
  public mroot<func_t,vec_t,jfunc_t>, hybrids_base {
    
#ifndef DOXYGEN_INTERNAL
    
  protected:

  /// Number of iterations
  int iter;
  /// Compute the number of failures
  size_t ncfail;
  /// Compute the number of successes
  size_t ncsuc;
  /// The number of times the actual reduction is less than 0.001
  size_t nslow1;
  /// The number of times the actual reduction is less than 0.1
  size_t nslow2;
  /// The norm of the current function value
  double fnorm;
  /// The limit of the Nuclidean norm
  double delta;
  /// Jacobian
  alloc_mat_t J;
  /// Q matrix from QR decomposition
  omatrix q;
  /// R matrix from QR decomposition
  omatrix r;
  /// The tau vector from QR decomposition
  ovector tau;
  /// The diagonal elements
  ovector diag;
  /// The value of \f$ Q^T f \f$
  ovector qtf;
  /// The Newton direction
  ovector newton;
  /// The gradient direction
  ovector gradient;
  /// The change in the function value
  ovector df;
  /// The value of \f$ Q^T \cdot \mathrm{df} \f$
  ovector qtdf;
  /// The value of \f$ R \cdot \mathrm{dx} \f$
  ovector rdx;
  /// The value of \f$ w=(Q^T df-R dx)/|dx| \f$
  ovector w;
  /// The value of \f$ v=D^2 dx/|dx| \f$
  ovector v;
  
  /// The user-specified Jacobian
  jfunc_t *jac;
    
  /// The automatic Jacobian
  jacobian<func_t,vec_t,mat_t> *ajac;

  /// Memory allocator for objects of type \c alloc_vec_t
  alloc_t ao;

  /// Memory allocator for objects of type \c alloc_mat_t
  mat_alloc_t am;

  /// The value of the derivative
  alloc_vec_t dx;

  /// Trial root
  alloc_vec_t x_trial;
    
  /// Trial function value
  alloc_vec_t f_trial;

  /// The number of equations and unknowns
  size_t dim;

  /// True if the jacobian has been given
  bool jac_given;

  /// The user-specified function
  func_t *fnewp;

  /// True if "set" has been called
  bool set_called;

  /// Finish the solution after set() or set_de() has been called
  int solve_set(size_t nn, vec_t &xx, func_t &ufunc) {

    int status;
    iter=0;

    do {
      iter++;
	
      status=iterate();
      if (status) break;

      // ------------------------------------------------------
      // The equivalent of the statement:
      // 
      // status=gsl_multiroot_test_residual(f,this->tol_rel);

      double resid=0.0;
      for(size_t i=0;i<nn;i++) {
	resid+=fabs(f[i]);
      }
      if (resid<this->tol_rel) status=gsl_success;
      else status=gsl_continue;
	
      // ------------------------------------------------------
	
      if (this->verbose>0) {
	this->print_iter(nn,x,f,iter,resid,this->tol_rel,
			 "gsl_mroot_hybrids");
      }
	
    } while (status==gsl_continue && iter<this->ntrial);
      
    for(size_t i=0;i<nn;i++) {
      xx[i]=x[i];
    }

    this->last_ntrial=iter;
    
    if (((int)iter)>=this->ntrial) {
      if (this->last_conv==0) this->last_conv=gsl_emaxiter;
      O2SCL_CONV2_RET("Function gsl_mroot_hybrids::msolve() ",
		      "exceeded max. number of iterations.",
		      gsl_emaxiter,this->err_nonconv);
    }
      
    if (status!=0) {
      O2SCL_ERR2_RET("Function iterate() failed in ",
		     "gsl_mroot_hybrids::solve_set().",gsl_efailed);
    }

    return gsl_success;
  }

#endif

  public:
      
  gsl_mroot_hybrids() {
    shrink_step=true;
    dim=0;
    ajac=&def_jac;
    def_jac.epsrel=GSL_SQRT_DBL_EPSILON;
    int_scaling=true;
    jac_given=false;
    set_called=false;
  }

  virtual ~gsl_mroot_hybrids() {
    if (dim>0) free();
  }

  /** \brief If true, iterate() will shrink the step-size automatically if
      the function returns a non-zero value (default true)

      The original GSL behavior can be obtained by setting 
      this to \c false.
  */
  bool shrink_step;

  /// If true, use the internal scaling method (default true)
  bool int_scaling;

  /// Default automatic Jacobian object
  simple_jacobian<func_t,vec_t,mat_t,alloc_vec_t,alloc_t> def_jac;

  /// Set the automatic Jacobian object
  virtual int set_jacobian(jacobian<func_t,vec_t,mat_t> &j) {
    ajac=&j;
    return 0;
  }
    
  /** \brief The value of the function at the present iteration
	
      \comment
      We need this to be public so that the user can see if 
      iterate() is working
      \endcomment
  */
  alloc_vec_t f;

  /// The present solution
  alloc_vec_t x;

  /** \brief Perform an iteration
	
      At the end of the iteration, the current value of the solution 
      is stored in \ref x.
  */
  int iterate() {
	
    if (!set_called) {
      O2SCL_ERR2_RET("Function set() not called or most recent call ",
		     "failed in gsl_mroot_hybrids::iterate().",
		     gsl_efailed);
    }

    this->last_conv=0;
 
    double prered,actred;
    double pnorm,fnorm1,fnorm1p;
    double ratio;
    double p1=0.1, p5=0.5, p001=0.001, p0001=0.0001;
    
    /* Compute qtf=Q^T f */
      
    compute_qtf(dim,q,f,qtf);
  
    /* Compute dogleg step */
      
    dogleg(dim,r,qtf,diag,delta,newton,gradient,dx);

    /* Take a trial step */
      
    compute_trial_step(dim,x,dx,x_trial);
    pnorm=scaled_enorm(dim,diag,dx);
    if (iter==1) {
      if (pnorm<delta) {
	delta=pnorm;
      }
    } 

    /* Evaluate function at x+p */
      
    int status;
    
    if (shrink_step==false) {
	  
      // Evaluate the function at the new point, exit if it fails
	
      status=(*fnewp)(dim,x_trial,f_trial);
	  
      if (status != gsl_success) {
	this->last_conv=gsl_ebadfunc;
	std::string str="Function returned non-zero value ("+itos(status)+
	  ") in gsl_mroot_hybrids::iterate().";
	O2SCL_CONV_RET(str.c_str(),o2scl::gsl_ebadfunc,this->err_nonconv);
      }
      
    } else {
	  
      // Evaluate the function at the new point, try to recover
      // if it fails
	  
      status=(*fnewp)(dim,x_trial,f_trial);
	  
      int bit=0;
      while(status!=0 && bit<20) {
	for(size_t ib=0;ib<dim;ib++) {
	  x_trial[ib]=(x_trial[ib]+x[ib])/2.0;
	}
	status=(*fnewp)(dim,x_trial,f_trial);
	bit++;
      }
    
      // Exit if we weren't able to find a new good point

      if (status != gsl_success) {
	this->last_conv=gsl_ebadfunc;
	std::string str="No suitable step found, function returned ("+
	  itos(status)+") in gsl_mroot_hybrids::iterate().";
	O2SCL_CONV_RET(str.c_str(),o2scl::gsl_ebadfunc,this->err_nonconv);
      }
    
    }

    /* Set df=f_trial-f */

    compute_df(dim,f_trial,f,df);
      
    /* Compute the scaled actual reduction */
  
    fnorm1=enorm(dim,f_trial);
    actred=compute_actual_reduction(fnorm,fnorm1);
      
    /* Compute rdx=R dx */
  
    compute_rdx(dim,r,dx,rdx);

    /* Compute the scaled predicted reduction phi1p=|Q^T f+R dx| */
  
    fnorm1p=enorm_sum(dim,qtf,rdx);
    prered=compute_predicted_reduction(fnorm,fnorm1p);

    /* Compute the ratio of the actual to predicted reduction */
  
    if (prered > 0) {
      ratio=actred/prered;
    } else {
      ratio=0;
    }
  
    /* Update the step bound */
  
    if (ratio<p1) {
      ncsuc=0;
      ncfail++;
      delta*=p5;
    } else {
      ncfail=0;
      ncsuc++;
      
      if (ratio >= p5 || ncsuc > 1) {
	delta=GSL_MAX(delta,pnorm/p5);
      }
      if (fabs (ratio-1) <= p1) {
	delta=pnorm/p5;
      }
    }
  
    /* Test for successful iteration */

    if (ratio >= p0001) {
      for(size_t i=0;i<dim;i++) {
	x[i]=x_trial[i];
	f[i]=f_trial[i];
      }
      fnorm=fnorm1;
      iter++;
    }
      
    /* Determine the progress of the iteration */
  
    nslow1++;
    if (actred >= p001) {
      nslow1=0;
    }
  
    if (actred >= p1) {
      nslow2=0;
    }
    
    if (ncfail==2) {

      int jac_ret;
      
      if (jac_given) jac_ret=(*jac)(dim,x,f,J);
      else jac_ret=(*ajac)(dim,x,f,J);
      
      if (jac_ret!=0) {
	this->last_conv=gsl_efailed;
	std::string str="Jacobian failed and returned ("+
	  itos(jac_ret)+") in gsl_mroot_hybrids::iterate().";
	O2SCL_CONV_RET(str.c_str(),gsl_efailed,this->err_nonconv);
      }
      
      nslow2++;
      
      if (iter==1) {
	if (int_scaling) {
	  compute_diag(dim,J,diag);
	}
	delta=compute_delta(dim,diag,x);
      } else {
	if (int_scaling) {
	  update_diag(dim,J,diag);
	}
      }
      
      /* Factorize J into QR decomposition */
      o2scl_linalg::QR_decomp(dim,dim,J,tau);
      o2scl_linalg::QR_unpack(dim,dim,J,tau,q,r);
	
      return gsl_success;
    }
  
    /* Compute qtdf=Q^T df, w=(Q^T df-R dx)/|dx|, v=D^2 dx/|dx| */
  
    compute_qtf(dim,q,df,qtdf);
    compute_wv(dim,qtdf,rdx,dx,diag,pnorm,w,v);

    /* Rank-1 update of the jacobian Q'R'=Q(R+w v^T) */
  
    o2scl_linalg::QR_update(dim,dim,q,r,w,v);

    /* No progress as measured by jacobian evaluations */

    if (nslow2==5) {
      this->last_conv=gsl_enoprogj;
      O2SCL_CONV_RET
	("No progress in Jacobian in gsl_mroot_hybrids::iterate().",
	 gsl_enoprogj,this->err_nonconv);
    }
  
    /* No progress as measured by function evaluations */

    if (nslow1==10) {
      this->last_conv=gsl_enoprog;
      O2SCL_CONV_RET
	("No progress in function in gsl_mroot_hybrids::iterate().",
	 gsl_enoprog,this->err_nonconv);
    }
      
    return gsl_success;
  }
      
  /// Allocate the memory
  int allocate(size_t n) {
    int status;

    if (dim>0) free();

    q.allocate(n,n);
    r.allocate(n,n);
    tau.allocate(n);
    diag.allocate(n);
    qtf.allocate(n);
    newton.allocate(n);
    gradient.allocate(n);
    df.allocate(n);
    qtdf.allocate(n);
    rdx.allocate(n);
    w.allocate(n);
    v.allocate(n);

    q.set_all(0.0);
    r.set_all(0.0);
    tau.set_all(0.0);
    diag.set_all(0.0);
    qtf.set_all(0.0);
    newton.set_all(0.0);
    gradient.set_all(0.0);
    df.set_all(0.0);
    qtdf.set_all(0.0);
    rdx.set_all(0.0);
    w.set_all(0.0);
    v.set_all(0.0);

    am.allocate(J,n,n);

    ao.allocate(x,n);
    ao.allocate(dx,n);
    ao.allocate(f,n);
    ao.allocate(x_trial,n);
    ao.allocate(f_trial,n);

    for(size_t i=0;i<n;i++) {
      x[i]=0.0;
      dx[i]=0.0;
      f[i]=0.0;
    }
      
    dim=n;

    return o2scl::gsl_success;
  }
    
  /// Free the allocated memory
  int free() {

    if (dim>0) {

      ao.free(x_trial);
      ao.free(f_trial);
      ao.free(dx);
      ao.free(x);
      ao.free(f);

      v.free();
      w.free();
      rdx.free();
      qtdf.free();
      df.free();
      gradient.free();
      newton.free();
      qtf.free();
      diag.free();
      tau.free();
      r.free();
      q.free();

      am.free(J,dim);

      dim=0;
    }

    return o2scl::gsl_success;
  }
      
  /// Return the type,\c "gsl_mroot_hybrids".
  virtual const char *type() { return "gsl_mroot_hybrids"; }
    
  /** \brief Solve \c func with derivatives \c dfunc using \c x as 
      an initial guess, returning \c x.

  */
  virtual int msolve_de(size_t nn, vec_t &xx, func_t &ufunc,
			jfunc_t &dfunc) {
    int status;
  
    status=set_de(nn,xx,ufunc,dfunc);
    if (status!=0) {
      O2SCL_ERR2_RET("Function set() failed in ",
		     "gsl_mroot_hybrids::msolve_de().",gsl_efailed);
    }
      
    return solve_set(nn,xx,ufunc);
  }

  /// Solve \c ufunc using \c xx as an initial guess, returning \c xx.
  virtual int msolve(size_t nn, vec_t &xx, func_t &ufunc) {
      
    int status;

    status=set(nn,xx,ufunc);
    if (status!=0) {
      O2SCL_ERR2_RET("Function set() failed in ",
		     "gsl_mroot_hybrids::msolve().",gsl_efailed);
    }
      
    status=solve_set(nn,xx,ufunc);
    for(size_t i=0;i<nn;i++) xx[i]=x[i];

    return status;
  }
      
  /** \brief Set the function, the parameters, and the initial guess 
   */
  int set(size_t nn, vec_t &ax, func_t &ufunc) {

    int status;
  
    if (nn!=dim) { 
      status=allocate(nn);
      if (status!=0) {
	O2SCL_ERR2_RET("Function allocate() failed in ",
		       "gsl_mroot_hybrids::set().",gsl_efailed);
      }
    }
      
    fnewp=&ufunc;

    // Specify function for automatic Jacobian
    ajac->set_function(ufunc);
      
    if (dim==0) {
      O2SCL_ERR2_RET("No memory allocated in ",
		     "gsl_mroot_hybrids::set().",o2scl::gsl_ebadlen);
    }
    
    // Copy the user-specified solution
    for(size_t i=0;i<dim;i++) x[i]=ax[i];
      
    status=ufunc(dim,ax,f);
    if (status!=0) {
      O2SCL_ERR2_RET("Function returned non-zero in ",
		     "gsl_mroot_hybrids::set().",gsl_ebadfunc);
    }
    
    if (jac_given) status=(*jac)(dim,ax,f,J);
    else status=(*ajac)(dim,ax,f,J);

    if (status!=0) {
      O2SCL_ERR2_RET("Jacobian failed in ",
		     "gsl_mroot_hybrids::set().",gsl_efailed);
    }

    iter=1;
    fnorm=enorm(dim,f);
    ncfail=0;
    ncsuc=0;
    nslow1=0;
    nslow2=0;
      
    for(size_t i=0;i<dim;i++) dx[i]=0.0;
      
    /* Store column norms in diag */
  
    if (int_scaling) {
      compute_diag(dim,J,diag);
    } else {
      diag.set_all(1.0);
    }
	
    /* Set delta to factor |D x| or to factor if |D x| is zero */
	
    delta=compute_delta(dim,diag,x);
  
    /* Factorize J into QR decomposition */

    o2scl_linalg::QR_decomp(dim,dim,J,tau);
    o2scl_linalg::QR_unpack(dim,dim,J,tau,q,r);
    set_called=true;
    jac_given=false;

    return gsl_success;
  }

  /** \brief Set the function, the Jacobian, the parameters,
      and the initial guess 
  */
  int set_de(size_t nn, vec_t &ax, func_t &ufunc, jfunc_t &dfunc) {

    fnewp=&ufunc;
    jac=&dfunc;

    // Make sure set() uses the right Jacobian
    jac_given=true;
    
    int ret=set(nn,ax,ufunc);
    
    // Reset jac_given since set() will set it back to false
    jac_given=true;
    set_called=true;
    
    return ret;
  }
  
  };
  
#ifndef DOXYGENP
}
#endif

#endif
